import sys, os

from random import randint
from parsley.config import Config, ConfigParameter
from parsley.loggingex import *
from wooly import Application, Session, Page
from wooly.pages import CssPage, JavascriptPage, ResourcePage
from wooly.server import WebServer
from wooly.devel import DevelPage
from wooly.parameters import IntegerParameter
from mint import *
from time import sleep
from threading import Thread, Event
from urllib import quote
from model import CuminModel, ModelPage, CallPage
from demo import DemoData
from page import MainPage
from stat import StatChartPage
from action import ActionPage
from user import LoginPage, UserSession, UserSessionExpireThread
from datetime import timedelta
from qpid.util import URL

from wooly import Session

log = getLogger("cumin")

class Cumin(Application):
    def __init__(self, config):
        super(Cumin, self).__init__()

        self.config = config

        if self.config.debug:
            self.enable_debug()

        self.home = self.config.home
        self.add_resource_dir(os.path.join(self.home, "resources-wooly"))
        self.add_resource_dir(os.path.join(self.home, "resources"))

        self.model = CuminModel(self, self.config.data)

        self.main_page = MainPage(self, "index.html")
        self.add_page(self.main_page)
        self.set_default_page(self.main_page)

        self.add_page(DevelPage(self, "devel.html"))
        self.add_page(ModelPage(self, "model.xml"))
        self.add_page(CallPage(self, "call.xml"))
        self.add_page(ActionPage(self, "actions.html"))
        self.add_page(StatChartPage(self, "stats.png"))

        unprotected = set()

        self.login_page = LoginPage(self, "login.html")
        self.add_page(self.login_page)
        unprotected.add(self.login_page)

        page = CssPage(self, "cumin.css")
        self.add_page(page)
        unprotected.add(page)

        page = JavascriptPage(self, "cumin.js")
        self.add_page(page)
        unprotected.add(page)

        page = ResourcePage(self, "resource")
        self.add_page(page)
        unprotected.add(page)
        
        self.unprotected_pages = unprotected

        self.user_sessions_by_id = dict()
        self.user_session_expire_thread = UserSessionExpireThread(self)

    def check(self):
        if not os.path.isdir(self.home):
            raise Exception \
                ("Error: cumin home '%s' not found or not a directory" \
                     % self.home)

        self.model.check()

    def init(self):
        self.model.init()

    def start(self):
        self.model.start()
        self.user_session_expire_thread.start()

    def stop(self):
        self.model.stop()

class CuminServer(WebServer):
    def authorized(self, session):
        auth = False
        page = session.page

        if page in self.app.unprotected_pages:
            return True

        id = session.cookies.get("session")

        if id is not None:
            usess = self.app.user_sessions_by_id.get(id)

            if usess is not None:
                timeout = timedelta(seconds=3600)
                now = datetime.now()

                if now > usess.created and now < usess.created + timeout:
                    setattr(session, "user_session", usess)

                    return True

        if self.app.config.user:
            user = Subject.getByName(self.app.config.user)

            assert user

            usess = UserSession(self.app, user)
            session.set_cookie("session", usess.id)

            page.set_redirect_url(session, session.marshal())

            return False

        lpage = self.app.login_page
        lsess = Session(lpage)
        lpage.origin.set(lsess, session.marshal())

        page.set_redirect_url(session, lsess.marshal())

        return False

class CuminConfig(Config):
    def __init__(self):
        super(CuminConfig, self).__init__()

        hdef = os.path.normpath("/var/lib/cumin")
        self.home = os.environ.get("CUMIN_HOME", hdef)

        sdef = os.path.normpath("/usr/share/amqp/amqp.0-10-qpid-errata.xml")
        self.spec = os.environ.get("AMQP_SPEC", sdef)

        if not os.path.isdir(self.home):
            raise Exception("Home path '%s' is not a directory")

        param = ConfigParameter(self, "data", str)
        param.default = "postgresql://cumin@localhost/cumin"

        param = ConfigParameter(self, "log-file", str)
        param.default = os.path.join(self.home, "log", "cumin.log")

        param = ConfigParameter(self, "log-level", str)
        param.default = "warn"

        param = ConfigParameter(self, "debug", bool)
        param.default = False

        param = ConfigParameter(self, "user", str)

        param = ConfigParameter(self, "expire-frequency", int)
        param.default = 600 # 10 minutes

        param = ConfigParameter(self, "expire-threshold", int)
        param.default = 24 * 3600 # 1 day

        param = ConfigParameter(self, "operator-email", str)

    def init(self):
        super(CuminConfig, self).init()

        handler = StreamHandler()
        log.addHandler(handler)
        self.load_file(os.path.join(self.home, "etc", "cumin.conf"))
        self.load_file(os.path.join(os.path.expanduser("~"), ".cumin.conf"))
        log.removeHandler(handler)

        self.init_logging()
        
    def init_logging(self):
        enable_logging("mint", self.log_level, self.log_file)
        enable_logging("cumin", self.log_level, self.log_file)

        if self.debug:
            enable_logging("mint", "debug", sys.stderr)
            enable_logging("cumin", "debug", sys.stderr)
