from cumin.exchange import ExchangeInfo
from cumin.util import sorted_by
from cumin.widgets import SubmitSwitch
from formats import *
from wooly import Template, Writer, Attribute, Parameter, Widget
from wooly.forms import FormInput, FormField, Form
from wooly.parameters import DictParameter
from wooly.resources import StringCatalog

strings = StringCatalog(__file__)

class ExchangeInput(Widget):
    def __init__(self, app, name):
        super(ExchangeInput, self).__init__(app, name)

        self.exchange = None
        self.instance_data = None

        self.name_tmpl = Template(self, "name_html")
        self.key_tmpl = Template(self, "key_html")

        self.form = None

    def init(self):
        super(ExchangeInput, self).init()

        for anc in self.ancestors:
            if isinstance(anc, Form):
                self.form = anc

    def get_args(self, session):
        return (self.exchange,)

    def get_exchange_info(self, session, exchange):
        binding_info = self.form.bindings.dict_param.get(session)
        if str(exchange.id) in binding_info:
            return binding_info[str(exchange.id)]

    def get_exchange_info_for(self, session, exchange, key):
        exchange_info = self.get_exchange_info(session, exchange)
        if exchange_info:
            if key in exchange_info:
                return exchange_info[key]

    def render_exchange_name(self, session, exchange):
        return exchange.name

    def render_exchange_fmt_name(self, session, exchange):
        return fmt_shorten(exchange.name)

    def render_name_path(self, session, *args):
        return DictParameter.sep().join((self.instance_data, "name"))

    def render_exchange_type(self, session, exchange):
        return exchange.type

    def render_exchange_type_path(self, session, exchange):
        return DictParameter.sep().join((self.instance_data, "type"))

    def render_exchange_id(self, session, exchange):
        return exchange.id

    def render_exchange_checked(self, session, exchange):
        exchange_info = self.get_exchange_info(session, exchange)
        if exchange_info:
            if "name" in exchange_info:
                return "checked=\"checked\""

    def render_exchange_name_input(self, session, exchange):
        writer = Writer()
        self.name_tmpl.render(writer, session, exchange)
        return writer.to_string()

    def render_exchange_key_input(self, session, exchange):
        writer = Writer()
        self.key_tmpl.render(writer, session, exchange)
        return writer.to_string()

    def render_onclick(self, session, exchange):
        pass

    def render_list_error(self, session, exchange):
        errors = self.parent.binding_errors.get(session)
        if exchange.name in errors:
            return "<ul class=\"errors\" style=\"margin:0; float:left;\"><li>%s</li></ul>" % \
                "</li><li>".join(errors[exchange.name])

    def render_dict_error(self, session, exchange, key):
        errors = self.parent.binding_errors.get(session)
        if exchange.name in errors:
            exchange_errors = errors[exchange.name]
            if key in exchange_errors:
                return "<ul class=\"errors\" style=\"margin:0; float:left;\"><li>%s</li></ul>" % \
                "</li><li>".join(exchange_errors[key])

    def set_instance_data(self, exchange, dict_key):
        self.exchange = exchange
        self.instance_data = dict_key

class FanoutExchangeInput(ExchangeInput):
    pass

class BindingKeyExchangeInput(ExchangeInput):
    def __init__(self, app, name):
        super(BindingKeyExchangeInput, self).__init__(app, name)

    def render_key_path(self, session, exchange):
        return DictParameter.sep().join((self.instance_data, "key"))

    def render_key_error(self, session, exchange):
        return self.render_list_error(session, exchange)

    def render_key_value(self, session, exchange):
        return self.get_exchange_info_for(session, exchange, "key")

class DirectExchangeInput(BindingKeyExchangeInput):
    pass

class TopicExchangeInput(BindingKeyExchangeInput):
    pass

class XMLExchangeInput(BindingKeyExchangeInput):
    def __init__(self, app, name):
        super(XMLExchangeInput, self).__init__(app, name)

    def render_xquery_path(self, session, exchange):
        return DictParameter.sep().join((self.instance_data, "xquery"))

    def render_headers_class(self, session, exchange):
        exchange_info = self.get_exchange_info(session, exchange)
        if not exchange_info or not "name" in exchange_info:
            return "initial_header_state"

    def render_key_error(self, session, exchange):
        return self.render_dict_error(session, exchange, "key")

    def render_onclick(self, session, exchange):
        return "onclick=\"toggle_row(this, 'xml_extra.%s')\"" % str(exchange.id)

    def render_xml_extra(self, session, exchange):
        return "xml_extra.%s" % str(exchange.id)

    def render_xquery_error(self, session, exchange):
        return self.render_dict_error(session, exchange, "xquery")

    def render_xquery_value(self, session, exchange):
        return self.get_exchange_info_for(session, exchange, "xquery")

    def process_input(self, this_exchange, arguments):
        if "xquery" in this_exchange:
            arguments["xquery"] = this_exchange["xquery"]

class HeadersExchangeInput(BindingKeyExchangeInput):
    def __init__(self, app, name):
        super(HeadersExchangeInput, self).__init__(app, name)

    def render_x_match_path(self, session, exchange):
        return DictParameter.sep().join((self.instance_data, "x-match"))

    def render_mkey_path(self, session, exchange):
        return DictParameter.sep().join((self.instance_data, "mkey"))

    def render_headers_class(self, session, exchange):
        exchange_info = self.get_exchange_info(session, exchange)
        if not exchange_info or not "name" in exchange_info:
            return "initial_header_state"

    def render_all_checked(self, session, exchange):
        checked = self.render_any_checked(session, exchange)
        if not checked:
            return "checked=\"checked\""

    def render_any_checked(self, session, exchange):
        exchange_info = self.get_exchange_info(session, exchange)
        if exchange_info:
            if "x-match" in exchange_info:
                if exchange_info["x-match"] == "any":
                    return "checked=\"checked\""

    def render_mkey1_value(self, session, exchange):
        return self.get_exchange_info_for(session, exchange, "mkey.1")

    def render_mkey2_value(self, session, exchange):
        return self.get_exchange_info_for(session, exchange, "mkey.2")

    def render_mkey3_value(self, session, exchange):
        return self.get_exchange_info_for(session, exchange, "mkey.3")

    def render_mnv1_value(self, session, exchange):
        return self.get_exchange_info_for(session, exchange, "mkey.1.nv")

    def render_mnv2_value(self, session, exchange):
        return self.get_exchange_info_for(session, exchange, "mkey.2.nv")

    def render_mnv3_value(self, session, exchange):
        return self.get_exchange_info_for(session, exchange, "mkey.3.nv")

    def render_key_error(self, session, exchange):
        return self.render_dict_error(session, exchange, "key")

    def render_mkey1_error(self, session, exchange):
        return self.render_dict_error(session, exchange, "mkey.1")

    def render_mkey2_error(self, session, exchange):
        return self.render_dict_error(session, exchange, "mkey.2")

    def render_mkey3_error(self, session, exchange):
        return self.render_dict_error(session, exchange, "mkey.3")

    def render_onclick(self, session, exchange):
        return "onclick=\"toggle_row(this, 'headers_extra.%s')\"" % str(exchange.id)

    def render_headers_extra(self, session, exchange):
        return "headers_extra.%s" % str(exchange.id)

    def process_input(self, this_exchange, arguments):
        # x-match is a radio button, it must have a value
        arguments["x-match"] = this_exchange["x-match"]
        # Fill out the other arguments.
        # The form has input boxes named mkey.* and mkey.*.nv
        # We need to create an arguments dictionary entry
        # of the form {mkey.*.value: mkey.*.nv.value}
        for match_info in this_exchange:
            if this_exchange[match_info]:
                if match_info.startswith("mkey") \
                    and not match_info.endswith("nv"):
                    # find the value in the matching .nv field
                    match_value = self._find_match_value(this_exchange, match_info)
                    # it is valid for the value in the .nv field
                    # to be empty
                    arguments[this_exchange[match_info]] = \
                        match_value or None

    def _find_match_value(self, this_exchange, match_info):
        for m_info in this_exchange:
            if m_info.startswith(match_info):
                if m_info.endswith("nv"):
                    return this_exchange[m_info]

class ExchangeState(SubmitSwitch):
    def __init__(self, app, name):
        super(ExchangeState, self).__init__(app, name)

        self.add_state("c", "Active", bm="phase")
        self.add_state("a", "All", bm="phase")

    def is_all(self, session):
        return self.get(session) == "a"

    def is_active(self, session):
        return self.get(session) == "c"

class ExchangeKeysField(FormField):
    def __init__(self, app, name, title="Initial bindings:"):
        super(ExchangeKeysField, self).__init__(app, name)

        self.dict_param = DictParameter(app, "exchange")
        self.add_parameter(self.dict_param)

        self.direct_input = DirectExchangeInput(app, "direct")
        self.add_child(self.direct_input)

        self.topic_input = TopicExchangeInput(app, "topic")
        self.add_child(self.topic_input)

        self.fanout_input = FanoutExchangeInput(app, "fanout")
        self.add_child(self.fanout_input)

        self.xml_input = XMLExchangeInput(app, "xml")
        self.add_child(self.xml_input)

        self.headers_input = HeadersExchangeInput(app, "headers")
        self.add_child(self.headers_input)

        self.title = title

        self.binding_errors = self.Errors(self, "binding_errors")
        self.add_attribute(self.binding_errors)

        self.state = ExchangeState(app, "phase")
        self.add_child(self.state)

    def get_args(self, session):
        reg = self.frame.get_ancestor("broker").get_object(session)
        return (reg.getDefaultVhost(),)

    class Errors(Attribute):
        def get_default(self, session):
            return dict()

    def render_title(self, session, vhost):
        return self.title

    def render_exchanges(self, session, vhost):
        sortedExchanges = sorted_by(vhost.exchanges)

        # render each exchange we support
        writer = Writer()
        for exchange in sortedExchanges:
            if ExchangeInfo.is_builtin(exchange) or \
                (not exchange._get_qmfDeleteTime() and \
                 not (self.state.is_active(session) and not is_active(exchange))):
                # instance_key gives us a unique path for each exchange
                # we will be rendering
                instance_key = self.dict_param.get_instance_key(str(exchange.id))
                if exchange.type == "direct":
                    if exchange.name:
                        self.direct_input.set_instance_data(exchange, instance_key)
                        writer.write(self.direct_input.render(session))
                elif exchange.type == "topic":
                    if not exchange.name == "qpid.management":
                        self.topic_input.set_instance_data(exchange, instance_key)
                        writer.write(self.topic_input.render(session))
                elif exchange.type == "fanout":
                    self.fanout_input.set_instance_data(exchange, instance_key)
                    writer.write(self.fanout_input.render(session))
                elif exchange.type == "xml":
                    self.xml_input.set_instance_data(exchange, instance_key)
                    writer.write(self.xml_input.render(session))
                elif exchange.type == "headers":
                    self.headers_input.set_instance_data(exchange, instance_key)
                    writer.write(self.headers_input.render(session))

        return writer.to_string()

    def get_binding_errors(self, session, queue_name):

        form_binding_info = self.process_binding_info(session, queue_name)
        binding_info = self.dict_param.get(session)
        berrs = self.binding_errors.get(session)

        for exchange in form_binding_info:
            type = form_binding_info[exchange]["type"]
            if (type == "topic") or (type == "direct"):
                try:
                    val = form_binding_info[exchange]["key"]
                    if not val:
                        raise KeyError
                except KeyError:
                    name = form_binding_info[exchange]["name"]
                    errs = berrs.setdefault(name, list())
                    errs.append("A binding key is required")
            elif type == "headers":
                try:
                    val = form_binding_info[exchange]["key"]
                    if not val:
                        raise KeyError
                except KeyError:
                    name = form_binding_info[exchange]["name"]
                    errs = berrs.setdefault(name, dict())
                    errs["key"] = ["A binding key is required"]
                for key_num in ("1", "2", "3"):
                    mkey = "mkey."+key_num
                    mkeynv = mkey+".nv"
                    if mkeynv in binding_info[exchange] and binding_info[exchange][mkeynv]:
                        try:
                            val = binding_info[exchange][mkey]
                            if not val:
                                raise KeyError
                        except KeyError:
                            name = binding_info[exchange]["name"]
                            if not name in berrs:
                                berrs.setdefault(name, dict())
                            berrs[name][mkey] = ["Missing key"]
            elif type == "xml":
                try:
                    val = form_binding_info[exchange]["key"]
                    if not val:
                        raise KeyError
                except KeyError:
                    name = form_binding_info[exchange]["name"]
                    errs = berrs.setdefault(name, dict())
                    errs["key"] = ["A binding key is required"]
                try:
                    val = binding_info[exchange]["xquery"]
                    if not val:
                        raise KeyError
                except KeyError:
                    name = binding_info[exchange]["name"]
                    if not name in berrs:
                        berrs.setdefault(name, dict())
                    berrs[name]["xquery"] = ["Missing xquery"]

        return (len(berrs), form_binding_info)

    def process_binding_info(self, session, queue_name):
        """ Processes the raw binding_info from the DictParameter into
        a "form_binding_info" dictionary that contains four keys:
        name, key, arguments, and type
        """
        binding_info = self.dict_param.get(session)
        form_binding_info = dict()
        for this_exchange in binding_info:
            # if the exchange checkbox is checked
            if "name" in binding_info[this_exchange]:
                type = binding_info[this_exchange]["type"]
                #if type == "direct":
                #    binding_info[this_exchange]["key"] = queue_name

                form_binding_info[this_exchange] = dict()
                form_binding_info[this_exchange]["name"] = binding_info[this_exchange]["name"]
                if "key" in binding_info[this_exchange] and binding_info[this_exchange]["key"]:
                    form_binding_info[this_exchange]["key"] = binding_info[this_exchange]["key"]
                form_binding_info[this_exchange]["type"] = type

                arguments = dict()
                if type == "headers":
                    self.headers_input.process_input(binding_info[this_exchange], arguments)
                elif type == "xml":
                    self.xml_input.process_input(binding_info[this_exchange], arguments)
                #direct, topic and fanout exchanges don't have aditional arguments
                form_binding_info[this_exchange]["arguments"] = arguments

        return form_binding_info
