from mint import *
from wooly import *
from wooly.widgets import *
from random import random
from psycopg2 import IntegrityError

from configproperty import *
from virtualhost import *
from queue import *
from exchange import *
from brokerlink import *
from client import *
from widgets import *
from parameters import *
from formats import *
from util import *

strings = StringCatalog(__file__)

class BrokerSet(CuminTable, Form):
    def __init__(self, app, name):
        super(BrokerSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.NameColumn(app, "name")
        self.add_column(col)
        self.set_default_column(col)

        col = self.GroupsColumn(app, "groups")
        self.add_column(col)

        col = self.ClusterColumn(app, "cluster")
        self.add_column(col)

        self.groups = self.BrokerSetGroupInput(app, "groups")
        self.add_child(self.groups)

        self.__unregister = self.Unregister(app, "unregister")
        self.add_child(self.__unregister)

        self.groupify = self.Groupify(app, "groupify")
        self.add_child(self.groupify)

    def render_title(self, session, *args):
        count = BrokerRegistration.select().count()
        return "Brokers %s" % fmt_count(count)

    class Unregister(FormButton):
        def render_content(self, session):
            return "Unregister"

        def render_class(self, session):
            return "unregister"

        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            branch = session.branch()
            frame = self.page.main.brokers_remove.show(branch)
            frame.ids.set(branch, ids)
            frame.origin.set(branch, session.marshal())
            self.page.set_redirect_url(session, branch.marshal())

    class BrokerSetGroupInput(BrokerGroupInput):
        def render_submit_id(self, session):
            return self.parent.groupify.path

    class Groupify(FormButton):
        def render_content(self, session):
            return "Add"

        def process_submit(self, session):
            group = self.parent.groups.get(session)

            if group:
                ids = self.parent.ids.get(session)
                self.parent.ids.clear(session)

                for id in ids:
                    try:
                        broker = BrokerRegistration.get(id)
                        broker.addBrokerGroup(group)
                    except IntegrityError:
                        pass

            self.page.set_redirect_url(session, session.marshal())

    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            reg = Identifiable(data["id"])
            href = self.page.main.broker.get_href(session, reg)
            return fmt_link(href, fmt_shorten(data["name"]))

    class GroupsColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Groups"

        def get_orderby_sql(self, session):
            return None

        def render_content(self, session, data):
            broker = BrokerRegistration.get(data["id"])
            count = broker.groups.count()

            if count == 0:
                link = fmt_none()
            elif count < 3:
                links = list()

                for group in broker.groups[:2]:
                    href = self.page.main.broker_group.get_href(session, group)
                    links.append(fmt_link(href, group.name))

                link = ", ".join(links)
            else:
                link = "%i groups" % count

            return link

    class ClusterColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Cluster"

        def render_content(self, session, data):
            return data["cluster"] or fmt_none()

class TopBrokerSet(CuminTable):
    def __init__(self, app, name):
        super(TopBrokerSet, self).__init__(app, name)

        col = self.NameColumn(app, "name")
        self.add_column(col)
        self.set_default_column(col)

    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            reg = Identifiable(data["id"])
            href = self.page.main.broker.get_href(session, reg)
            return fmt_link(href, fmt_shorten(data["name"]))

class BrokerFrame(CuminFrame):
    def __init__(self, app, name):
        super(BrokerFrame, self).__init__(app, name)

        self.object = BrokerRegistrationParameter(app, "id")
        self.add_parameter(self.object)

        self.view = BrokerView(app, "view")
        self.add_mode(self.view)
        self.set_view_mode(self.view)

        self.edit = BrokerEdit(app, "edit")
        self.add_mode(self.edit)
        self.set_edit_mode(self.edit)

        self.queue = QueueFrame(app, "queue")
        self.add_mode(self.queue)

        self.queues_purge = QueueSetPurge(app, "queuespurge")
        self.add_mode(self.queues_purge)

        self.queues_remove = QueueSetRemove(app, "queuesremove")
        self.add_mode(self.queues_remove)

        self.queue_add = QueueAdd(app, "queueadd")
        self.add_mode(self.queue_add)

        self.exchange = ExchangeFrame(app, "exchange")
        self.add_mode(self.exchange)

        self.exchange_add = ExchangeAdd(app, "exchangeadd")
        self.add_mode(self.exchange_add)

        self.exchanges_remove = ExchangeSetRemove(app, "exchangesremove")
        self.add_mode(self.exchanges_remove)

        self.bindings_remove = BindingSetRemove(app, "bindingsremove")
        self.add_mode(self.bindings_remove)

        self.link = PeerFrame(app, "link")
        self.add_mode(self.link)

        self.link_add = BrokerLinkAdd(app, "linkadd")
        self.add_mode(self.link_add)

        self.links_close = BrokerSetClose(app, "linksclose")
        self.add_mode(self.links_close)

        self.connection = ConnectionFrame(app, "conn")
        self.add_mode(self.connection)

        self.connections_close = ConnectionSetClose(app, "connsclose")
        self.add_mode(self.connections_close)

class BrokerStatus(CuminStatus):
    def __init__(self, app, name):
        super(BrokerStatus, self).__init__(app, name)

        self.connected = self.ConnectedAttribute(app, "conn")
        self.add_attribute(self.connected)

    def render_status(self, session, reg):
        if self.connected.get(session):
            return "Connected"
        else:
            return "Disconnected"

    def render_color(self, session, reg):
        if self.connected.get(session):
            return "green"
        else:
            return "yellow"

    class ConnectedAttribute(Attribute):
        def get_default(self, session):
            reg = self.widget.frame.get_object(session)
            connected = False

            try:
                mbroker = self.app.model.data.mintBrokersByUrl[reg.url]

                connected = mbroker.connected
            except KeyError:
                pass

            return connected

class BrokerView(CuminView):
    """
    Despite the name, this is actually a view of two objects, usually
    coupled: a BrokerRegistration and a Broker.
    """

    def __init__(self, app, name):
        super(BrokerView, self).__init__(app, name)

        status = BrokerStatus(app, "status")
        self.add_child(status)

        self.__body = ModeSet(app, "body")
        self.add_child(self.__body)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.__body.add_mode(self.__tabs)

        #self.__tabs.add_tab(BrokerStats(app, "stats"))
        self.__tabs.add_tab(self.BrokerQueueTab(app, "queues"))
        self.__tabs.add_tab(ExchangeSet(app, "exchanges"))
        self.__tabs.add_tab(ConnectionSet(app, "conns"))
        self.__tabs.add_tab(PeerSet(app, "peers"))
        self.__tabs.add_tab(BrokerAccessControl(app, "access"))
        self.__tabs.add_tab(BrokerClustering(app, "cluster"))
        self.__tabs.add_tab(self.BrokerDetailsTab(app, "details"))

        self.__missing = self.BrokerMissing(app, "missing")
        self.__body.add_mode(self.__missing)

        self.__vhost = Attribute(app, "vhost")
        self.add_attribute(self.__vhost)

    class BrokerMissing(Widget):
        pass

    def do_process(self, session, reg):
        reg.sync()

        if reg.broker:
            self.__vhost.set(session, reg.getDefaultVhost())
        else:
            self.__body.show_mode(session, self.__missing)

        super(BrokerView, self).do_process(session, reg)

    def render_script(self, session, reg):
        if reg.broker:
            data = "model.xml?class=broker;id=%i" % reg.broker.id
            return "wooly.setIntervalUpdate('%s', updateBroker, 3000)" % data

    def render_group_links(self, session, reg):
        links = list()

        for group in reg.groups:
            branch = session.branch()
            self.frame.frame.show_broker_group(branch, group).show_view \
                (branch)
            links.append(fmt_olink(branch, group))

        return ", ".join(links)

    class BrokerQueueTab(Widget):
        def __init__(self, app, name):
            super(BrokerView.BrokerQueueTab, self).__init__(app, name)

            self.__queues = QueueSet(app, "items")
            self.add_child(self.__queues)

        def get_args(self, session):
            return self.frame.get_args(session)

        def render_title(self, session, reg):
            vhost = reg.getDefaultVhost()
            return "Queues %s" % fmt_count(vhost.queues.count())

        def render_add_queue_href(self, session, reg):
            branch = session.branch()
            self.frame.queue.add.show(branch, None)
            return branch.marshal()

    class BrokerLogTab(Widget):
        def get_args(self, session):
            return self.frame.get_args(session)

        def render_title(self, session, reg):
            return "Log Messages"

    class BrokerDetailsTab(Widget):
        def __init__(self, app, name):
            super(BrokerView.BrokerDetailsTab, self).__init__(app, name)

            props = self.Properties(app, "properties")
            self.add_child(props)

            actions = self.Actions(app, "actions")
            self.add_child(actions)

        def render_title(self, session):
            return "Details"

        class Properties(CuminProperties):
            def do_get_items(self, session, reg):
                cls = self.app.model.get_class_by_object(reg)
                props = [(x.get_title(session), x.value(session, reg))
                         for x in cls.properties]

                if reg.broker:
                    cls = self.app.model.get_class_by_object(reg.broker)
                    props.extend([(x.get_title(session),
                                   x.value(session, reg.broker))
                                  for x in cls.properties])

                return props

        class Actions(CuminActions):
            def do_get_items(self, session, reg):
                cls = self.app.model.get_class_by_object(reg)
                acts = [(x.get_href(session, reg), x.get_title(session), x.get_enabled(session, reg))
                         for x in cls.actions if x.navigable]

                if reg.broker:
                    cls = self.app.model.get_class_by_object(reg.broker)
                    acts.extend([(x.get_href(session, reg.broker),
                                  x.get_title(session), x.get_enabled(session, reg.broker))
                                 for x in cls.actions if x.navigable])

                return acts

class ModuleNotEnabled(Widget):
    def do_render(self, session):
        return "This module is not enabled"

class BrokerAccessControl(ModeSet):
    def __init__(self, app, name):
        super(BrokerAccessControl, self).__init__(app, name)

        mode = ModuleNotEnabled(app, "notenabled")
        self.add_mode(mode)

        self.__view = BrokerAccessControlView(app, "view")
        self.add_mode(self.__view)

    def get_args(self, session):
        broker = self.frame.get_object(session).broker
        acl = None

        try:
            acl = Acl.selectBy(broker=broker)[0]
        except IndexError:
            pass

        return (acl,)

    def do_process(self, session, acl):
        if acl:
            self.show_mode(session, self.__view)

    def render_title(self, session, acl):
        return "Access Control"

class BrokerAccessControlView(Widget):
    def __init__(self, app, name):
        super(BrokerAccessControlView, self).__init__(app, name)

        props = self.Properties(app, "props")
        self.add_child(props)

        stats = self.Stats(app, "stats", "general")
        self.add_child(stats)

    class Properties(CuminProperties):
        def get_args(self, session):
            return self.parent.parent.get_args(session)

    class Stats(StatSet):
        def get_args(self, session):
            return self.parent.parent.get_args(session)

class BrokerClustering(ModeSet):
    def __init__(self, app, name):
        super(BrokerClustering, self).__init__(app, name)

        mode = ModuleNotEnabled(app, "notenabled")
        self.add_mode(mode)

        self.__view = BrokerClusteringView(app, "view")
        self.add_mode(self.__view)

    def get_args(self, session):
        broker = self.frame.get_object(session).broker
        cluster = None

        try:
            cluster = Cluster.selectBy(broker=broker)[0]
        except IndexError:
            pass

        return (cluster,)

    def do_process(self, session, cluster):
        if cluster:
            self.show_mode(session, self.__view)

    def render_title(self, session, cluster):
        return "Clustering"

class BrokerClusteringView(Widget):
    def __init__(self, app, name):
        super(BrokerClusteringView, self).__init__(app, name)

        props = self.Properties(app, "props")
        self.add_child(props)

        stats = self.Stats(app, "stats", "general")
        self.add_child(stats)

    class Properties(CuminProperties):
        def get_args(self, session):
            return self.parent.parent.get_args(session)

    class Stats(StatSet):
        def get_args(self, session):
            return self.parent.parent.get_args(session)

class BrokerStats(Widget):
    def __init__(self, app, name):
        super(BrokerStats, self).__init__(app, name)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, reg):
        return "Statistics"

class BrokerBrowser(Widget):
    def __init__(self, app, name):
        super(BrokerBrowser, self).__init__(app, name)

        self.group_tmpl = Template(self, "group_html")
        self.profile_tmpl = Template(self, "profile_html")
        self.cluster_tmpl = Template(self, "cluster_html")

        self.group = BrokerGroupParameter(app, "group")
        self.add_parameter(self.group)

        self.profile = BrokerProfileParameter(app, "profile")
        self.add_parameter(self.profile)

        self.cluster = BrokerClusterParameter(app, "cluster")
        self.add_parameter(self.cluster)

        self.brokers = self.BrowserBrokers(app, "brokers")
        self.add_child(self.brokers)

    def render_title(self, session, *args):
        return "Brokers %s" % fmt_count(BrokerRegistration.select().count())

    class BrowserBrokers(BrokerSet):
        def render_sql_where(self, session):
            elems = list()
            group = self.parent.group.get(session)
            #profile = self.parent.profile.get(session)
            #cluster = self.parent.cluster.get(session)

            if group:
                subquery = \
                    "select 1 from broker_group_mapping " + \
                    "where broker_group_id = %i "  % group.id + \
                    "and broker_registration_id = br.id"

                elems.append("exists (%s)" % subquery)

            if elems:
                return "where %s" % " and ".join(elems)

    def render_add_broker_href(self, session):
        branch = session.branch()
        self.frame.brokers_add.show(branch)
        return branch.marshal()

    def render_clear_filters_href(self, session):
        branch = session.branch()
        self.group.set(branch, None)
        self.profile.set(branch, None)
        self.cluster.set(branch, None)
        return branch.marshal()

    def render_group_filters(self, session):
        groups = BrokerGroup.select()
        return self._render_filters(session, groups, self.group_tmpl)

    def render_group_link(self, session, group):
        return self._render_filter_link(session, group, self.group)

    def render_profile_filters(self, session):
        profiles = BrokerProfile.select()
        return self._render_filters(session, profiles, self.profile_tmpl)

    def render_profile_link(self, session, profile):
        return self._render_filter_link(session, profile, self.profile)

    def render_cluster_filters(self, session):
        clusters = BrokerCluster.select()
        return self._render_filters(session, clusters, self.cluster_tmpl)

    def render_cluster_link(self, session, cluster):
        return self._render_filter_link(session, cluster, self.cluster)

    def _render_filters(self, session, collection, template):
        writer = Writer()

        for object in collection:
            template.render(writer, session, object)

        template.render(writer, session, None)

        return writer.to_string()

    def _render_filter_link(self, session, object, param):
        branch = session.branch()
        param.set(branch, object)
        href = branch.marshal()

        name = object and object.name or "Any"

        class_ = param.get(session) is object and "selected"

        return fmt_link(href, name, class_)

class BrokerSetForm(CuminForm, Frame):
    def __init__(self, app, name):
        super(BrokerSetForm, self).__init__(app, name)

        self.name_param = Parameter(app, "name_param");
        self.add_parameter(self.name_param)

        self.names = ListParameter(app, "name", self.name_param)
        self.add_parameter(self.names)

        self.name_errors = self.Errors(self, "name_errors")
        self.add_attribute(self.name_errors)

        self.addr_param = Parameter(app, "addr_param")
        self.add_parameter(self.addr_param)

        self.addrs = ListParameter(app, "address", self.addr_param)
        self.add_parameter(self.addrs)

        self.addr_errors = self.Errors(self, "addr_errors")
        self.add_attribute(self.addr_errors)

        self.group_param = BrokerGroupParameter(app, "group_param")
        self.group_param.default = None
        self.add_parameter(self.group_param)

        self.groups = ListParameter(app, "group", self.group_param)
        self.add_parameter(self.groups)

        self.fields = IntegerParameter(app, "fields")
        self.fields.default = 3
        self.add_parameter(self.fields)

        self.field_tmpl = Template(self, "field_html")
        self.group_tmpl = Template(self, "group_html")

        self.more = self.MoreEntries(app, "more")
        self.add_child(self.more)

    class Errors(Attribute):
        def get_default(self, session):
            return dict()

    def process_display(self, session):
        if self.more.get(session):
            self.fields.set(session, self.fields.get(session) + 3)

    def render_fields(self, session):
        writer = Writer()

        for i in range(self.fields.get(session)):
            self.field_tmpl.render(writer, session, i)

        return writer.to_string()

    def render_field_name_name(self, session, index):
        return self.names.path

    def render_field_name_value(self, session, index):
        names = self.names.get(session)
        if len(names) > index:
            return names[index]

    def render_field_name_errors(self, session, index):
        errors = self.name_errors.get(session)
        if index in errors:
            return "<ul class=\"errors\"><li>%s</li></ul>" % \
                "</li><li>".join(errors[index])

    def render_field_address_name(self, session, index):
        return self.addrs.path

    def render_field_address_value(self, session, index):
        addrs = self.addrs.get(session)
        if len(addrs) > index:
            return addrs[index]

    def render_field_address_errors(self, session, index):
        errors = self.addr_errors.get(session)
        if index in errors:
            return "<ul class=\"errors\"><li>%s</li></ul>" % \
                "</li><li>".join(errors[index])

    def render_field_group_name(self, session, index):
        return self.groups.path

    def render_groups(self, session, index):
        writer = Writer()

        for group in BrokerGroup.select():
            self.group_tmpl.render(writer, session, index, group)

        return writer.to_string()

    def render_group_value(self, session, index, group):
        return group.id

    def render_group_name(self, session, index, group):
        return group.name

    def render_group_selected_attr(self, session, index, group):
        groups = self.groups.get(session)
        if len(groups) > index:
            if groups[index] and group.id == groups[index].id:
                return "selected=\"selected\""

    class MoreEntries(FormButton):
        def render_content(self, session):
            return "More Entries"

        def render_class(self, session):
            return "more"

        def render_type(self, session):
            """ using type="button" so pressing
            Enter on the form will not add more fields """
            return "button"
        
        def render_onclick(self, session):
            """ since this isn't a type="submit" button,
            we need javascript to submit the form when
            the button is clicked """
            return "click_more"

class BrokerSetAdd(BrokerSetForm):
    def process_cancel(self, session):
        branch = session.branch()
        self.page.pop_current_frame(branch)
        self.page.get_current_frame(branch).show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session):
        action = self.app.model.broker_registration.add

        addrs = self.addrs.get(session)
        names = self.names.get(session)
        groups = self.groups.get(session)
        fields = self.fields.get(session)

        if self.validate(session, addrs, names, groups):
            for i in range(0, fields):
                try:
                    addr = addrs[i]
                except:
                    addr = None

                if addr:
                    name = names[i]
                    url = "amqp://%s:%i" % parse_broker_addr(addr)

                    args = {"name": name, "url": url}
                    reg = action.invoke(None, args)

                    if len(groups) > i:
                        group = groups[i]

                        if group:
                            reg.addBrokerGroup(group)

            self.process_cancel(session)

    def validate(self, session, addrs, names, groups):
        nerrs = self.name_errors.get(session)
        aerrs = self.addr_errors.get(session)
        fields = self.fields.get(session)

        for i in range(0, fields):
            try:
                addr, name = addrs[i], names[i]
            except:
                addr = name = None

            if not addr and not name:
                pass # It's just an empty row
            else:
                if not name:
                    errs = nerrs.setdefault(i, list())
                    errs.append("The name field is empty; it is required")
                elif BrokerRegistration.selectBy(name=name).count():
                    errs = nerrs.setdefault(i, list())
                    errs.append("A broker called '%s' already exists" % name)

                if not addr:
                    errs = aerrs.setdefault(i, list())
                    errs.append("The address field is empty; it is required")
                else:
                    #host, port = parse_broker_addr(addr)
                    count = BrokerRegistration.selectBy \
                        (url=addr).count()

                    if count:
                        errs = aerrs.setdefault(i, list())
                        errs.append("The broker at %s " % (addr) +
                                    "is already registered")

        return not len(nerrs) and not len(aerrs)

    def render_title(self, session):
        return "Register New Brokers"

class BrokerEdit(CuminFieldForm):
    def __init__(self, app, name):
        super(BrokerEdit, self).__init__(app, name)

        self.broker_name = UniqueNameField(app, "name", BrokerRegistration)
        self.add_field(self.broker_name)

        self.groups = BrokerGroupCheckboxField(app, "groups")
        self.add_field(self.groups)

    def init(self):
        super(BrokerEdit, self).init()

        self.broker_name.set_object_attr(self.frame.object)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, reg):
        return "Edit Broker '%s'" % reg.name

    def process_cancel(self, session, reg):
        branch = session.branch()
        self.page.pop_current_frame(branch)
        self.page.get_current_frame(branch).show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session, reg):
        errors = self.validate(session)

        if errors:
            pass
        else:
            action = self.app.model.broker_registration.edit
            args = {"name": self.broker_name.get(session)}
            action.invoke(reg, args)

            for group in self.groups.get(session):
                if group not in reg.groups:
                    reg.addBrokerGroup(group)

            for group in reg.groups:
                if group not in self.groups.get(session):
                    reg.removeBrokerGroup(group)

            reg.syncUpdate()

            self.process_cancel(session, reg)

    def process_display(self, session, reg):
        self.broker_name.set(session, reg.name)

        for group in reg.groups:
            self.groups.get(session).append(group)

class BrokerSetRemove(CuminBulkActionForm, Frame):
    def process_item(self, session, id):
        action = self.app.model.broker_registration.remove
        # XXX need to handle object not found case
        action.invoke(BrokerRegistration.get(id))

    def render_title(self, session):
        return "Unregister Brokers"

    def render_item_content(self, session, id):
        return "Unregister Broker '%s'" % BrokerRegistration.get(id).name

from brokergroup import BrokerGroupCheckboxField
