from mint import *
from wooly import *
from wooly.widgets import *
from wooly.forms import *

from broker import BrokerSet
from model import *
from widgets import *
from modelwidgets import *
from parameters import *
from formats import *
from util import *

strings = StringCatalog(__file__)

class BrokerGroupSet(CuminClassTable):
    def __init__(self, app, name):
        cls = app.model.broker_group

        super(BrokerGroupSet, self).__init__(app, name, cls)

        col = CuminClassNameColumn(app, "name", cls)
        self.add_column(col)

class BrokerGroupInputSet(CheckboxInputSet):
    def __init__(self, app, name):
        super(BrokerGroupInputSet, self).__init__(app, name, None)

        item = BrokerGroupParameter(app, "item")
        self.param = ListParameter(app, "param", item)
        self.add_parameter(self.param)

    def do_get_items(self, session, *args):
        return BrokerGroup.select()

    def render_item_value(self, session, group):
        return group.id

    def render_item_content(self, session, group):
        return group.name

    def render_item_checked_attr(self, session, group):
        if group in self.param.get(session):
            return "checked=\"checked\""

class BrokerGroupCheckboxField(FormField):
    def __init__(self, app, name):
        super(BrokerGroupCheckboxField, self).__init__(app, name)

        self.__groups = BrokerGroupInputSet(app, "inputs")
        self.add_child(self.__groups)

    def get(self, session):
        return self.__groups.get(session)

    def render_title(self, session):
        return "Broker Groups"

class BrokerGroupFrame(CuminFrame):
    def __init__(self, app, name):
        super(BrokerGroupFrame, self).__init__(app, name)

        self.object = BrokerGroupParameter(app, "id")
        self.add_parameter(self.object)

        self.view = BrokerGroupView(app, "view")
        self.add_child(self.view)
        self.set_view_mode(self.view)

        self.add = BrokerGroupAdd(app, "add")
        self.add_child(self.add)
        self.set_add_mode(self.add)

        self.edit = BrokerGroupEdit(app, "edit")
        self.add_child(self.edit)
        self.set_edit_mode(self.edit)

        self.remove = BrokerGroupRemove(app, "remove")
        self.add_child(self.remove)
        self.set_remove_mode(self.remove)

    def render_title(self, session, group):
        if group:
            return super(BrokerGroupFrame, self).render_title(session, group)
        else:
            return "Broker Group"

    def render_href(self, session, group):
        if group:
            return super(BrokerGroupFrame, self).render_href(session, group)

class BrokerGroupStatus(CuminStatus):
    pass

class BrokerGroupView(CuminView):
    def __init__(self, app, name):
        super(BrokerGroupView, self).__init__(app, name)

        status = BrokerGroupStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        self.__tabs.add_tab(self.GroupBrokerTab(app, "brokers"))
        self.__tabs.add_tab(CuminDetails(app, "details"))

    class GroupBrokerTab(BrokerSet):
        def get_args(self, session):
            return self.frame.get_args(session)

        def render_title(self, session, group):
            return "Brokers %s" % \
                fmt_count(self.get_item_count(session, group))

        def render_sql_where(self, session, group):
            subquery = \
                "select 1 from broker_group_mapping " + \
                "where broker_group_id = %i " % group.id + \
                "and broker_registration_id = br.id"

            return "where exists (%s)" % subquery

class BrokerGroupForm(CuminFieldForm):
    def __init__(self, app, name):
        super(BrokerGroupForm, self).__init__(app, name)

        self.group_name = UniqueNameField(app, "name", BrokerGroup)
        self.add_field(self.group_name)

    def init(self):
        super(BrokerGroupForm, self).init()

        self.group_name.set_object_attr(self.frame.object)

class BrokerGroupAdd(BrokerGroupForm):
    def process_cancel(self, session):
        branch = session.branch()
        self.page.main.view.show(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session):
        errors = self.validate(session)

        if errors:
            pass
        else:
            args = {"name": self.group_name.get(session)}
            method = self.app.model.broker_group.add
            method.invoke(None, args)

            self.process_cancel(session)

    def render_title(self, session):
        return "Add Group"

class BrokerGroupEdit(BrokerGroupForm):
    def get_args(self, session):
        return self.frame.get_args(session)

    def process_cancel(self, session, group):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session, group):
        errors = self.validate(session)

        if errors:
            pass
        else:
            args = {"name": self.group_name.get(session)}
            method = self.app.model.broker_group.edit
            method.invoke(group, args)

            self.process_cancel(session, group)

    def process_display(self, session, group):
        self.group_name.set(session, group.name)

    def render_title(self, session, group):
        return "Edit Group '%s'" % group.name

class BrokerGroupRemove(CuminConfirmForm, Frame):
    def get_args(self, session):
        return self.frame.get_args(session)

    def process_cancel(self, session, group):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.pop_current_frame(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session, group):
        method = self.app.model.broker_group.remove
        method.invoke(group)

        branch = session.branch()
        self.frame.frame.show_view(branch)
        self.page.pop_current_frame(branch)
        self.page.pop_current_frame(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def render_title(self, session, group):
        return "Remove Broker Group '%s'" % group.name

    def render_submit_content(self, session, group):
        return "Yes, Remove Broker Group '%s'" % group.name

    def render_cancel_content(self, session, group):
        return "No, Cancel"
