import logging

from cairo import *
from random import random
from time import time

from formats import *
from util import *

log = logging.getLogger("cumin.chart")

class TimeSeriesChart(object):
    def __init__(self, width, height):
        self.width = width - 40
        self.height = height - 20
        self.surface = ImageSurface(FORMAT_ARGB32, width, height)
        self.surface.set_device_offset(1.5, 5.5)
        self.x_max = 1
        self.x_min = 0
        self.y_max = 1
        self.y_min = 0

    def plot_values(self, samples, color=(0, 0, 0)):
        cr = Context(self.surface)
        cr.set_line_width(1.5)
        cr.set_source_rgba(color[0], color[1], color[2], 0.66)

        tnow = time()

        for dt, value in samples:
            if value is None:
                continue

            t = secs(dt)

            # prevent line from drawing off right side of grid
            if tnow - t < 0:
                t = tnow

            x = self.width - ((tnow - t) / float(self.x_max)) * self.width
            y = self.height - (value / float(self.y_max)) * self.height
            cr.line_to(x, y)

        cr.stroke()

    def plot_legend(self, titles, colors):
        cr = Context(self.surface)

        for i, item in enumerate(zip(titles, colors)):
            title, color = item
            y = 16 + i * 16

            cr.set_source_rgba(1, 1, 1, 0.75)
            cr.rectangle(4, y - 12, 16 + 6.5 * len(title), 16)
            cr.fill()
            cr.stroke()

            cr.set_source_rgb(*color)
            cr.rectangle(8, y, 8, -8)
            cr.fill()
            cr.stroke()

            cr.move_to(20, y)
            cr.set_source_rgb(0, 0, 0)
            cr.show_text(title)
            cr.stroke()

    def plot_frame(self, color=(0.8, 0.8, 0.8)):
        cr = Context(self.surface)
        cr.set_line_width(1)
        cr.set_source_rgb(*color)

        cr.move_to(self.width, 0)
        cr.line_to(self.width, self.height)
        cr.line_to(0, self.height)

        cr.stroke()

    def plot_x_axis(self, intervals, step):
        cr = Context(self.surface)
        cr.set_line_width(0.2)
        cr.set_source_rgb(0.6, 0.6, 0.6)

        interval = self.width / float(intervals)
        absmax = self.x_max - self.x_min

        for i in range(0, intervals + 1):
            x = self.width - (i * interval)

            cr.move_to(x, 0)
            cr.line_to(x, self.height + 12)
            cr.move_to(x + 2, self.height + 12)

            if i % step == 0:
                value = absmax - (absmax * x / float(self.width))
                text = fmt_duration_brief(value)
                if text:
                    cr.show_text(text)

        cr.stroke()

    def plot_y_axis(self, intervals, step):
        cr = Context(self.surface)
        cr.set_line_width(0.2)
        cr.set_source_rgb(0.6, 0.6, 0.6)

        interval = self.height / float(intervals)

        for i in range(0, intervals + 1):
            y = self.height - (i * interval)

            cr.move_to(0, y)
            cr.line_to(self.width + 3, y)
            cr.move_to(self.width + 4, y + 3)

            if i % step == 0:
                fraction = (self.height - y) / float(self.height)
                value = fraction * self.y_max - self.y_min
                value = int(round(value + self.y_min))

                if value >= 10000000:
                    svalue = "%im" % int(round(value / 1000000.0, -1))
                elif value >= 10000:
                    svalue = "%ik" % int(round(value / 1000.0, -1))
                else:
                    svalue = str(value)
                
                cr.show_text(svalue)

        cr.stroke()

    def write(self, writer):
        self.surface.write_to_png(writer)
