from wooly import *
from wooly.widgets import *
from wooly.tables import *
from datetime import datetime

from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *

strings = StringCatalog(__file__)

class ConnectionSet(CuminTable, Form):
    def __init__(self, app, name):
        super(ConnectionSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.AddressColumn(app, "addr")
        self.add_column(col)

        self.set_default_column(col)

        col = self.SystemConnectionColumn(app, "sysconn")
        self.add_column(col)
        
        col = self.AuthIdentityColumn(app, "authid")
        self.add_column(col)
        
        col = self.FedLinkColumn(app, "fedlink")
        self.add_column(col)
        
        col = self.SentColumn(app, "sent")
        col.alignment = "right"
        self.add_column(col)

        col = self.ReceivedColumn(app, "received")
        col.alignment = "right"
        self.add_column(col)

        self.unit = StateSwitch(app, "unit")
        self.unit.add_state("b", "Bytes")
        self.unit.add_state("f", "Frames")
        self.add_child(self.unit)

        self.__phase = PhaseSwitch(app, "phase")
        self.add_child(self.__phase)

        self.__close = self.Close(app, "close")
        self.add_child(self.__close)

    def get_args(self, session):
        reg = self.frame.get_object(session)
        return (reg.getDefaultVhost(),)

    def get_unit_plural(self, session):
        return self.unit.get(session) == "b" and "Bytes" or "Frames"

    def render_title(self, session, vhost):
        return "Connections %s" % fmt_count(vhost.clientConnections.count())

    def render_sql_where(self, session, vhost):
        elems = list()
        elems.append("l.vhost_id = %(id)r")
        elems.append(self.__phase.get_sql_constraint(session, vhost))
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, vhost):
        return {"id": vhost.id}

    class Close(FormButton):
        def render_content(self, session):
            return "Close"

        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            # XXX Make table a frame to elim this
            self.parent.ids.clear(session)
            href = self.frame.connections_close.get_href(session, ids)
            self.page.set_redirect_url(session, href)

    class AddressColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Address"

        def render_content(self, session, data):
            conn = Identifiable(data["id"])
            href = self.frame.connection.get_href(session, conn)
            return fmt_link(href, fmt_shorten(data["addr"]))

    class SystemConnectionColumn(SqlTableColumn): 
        def render_title(self, session, data):
            return "Connect Type"
    
        def render_content(self, session, data):
            if data['sysconn']:
                return "System"
            else:
                return "Client"
    
    class AuthIdentityColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Auth Id"
    
    class FedLinkColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Fed Link"
    
        def render_content(self, session, data):
            if data['fedlink']:
                return "Yes"
            else:
                return "No"

    class SentColumn(NullSortColumn, FreshDataOnlyColumn):
        def render_title(self, session, data):
            return "%s Sent" % self.parent.get_unit_plural(session)

        def get_column_key(self, session):
            unit = self.parent.unit.get(session)
            return unit == "b" and "bs" or "fs"

        def render_value(self, session, value):
            return fmt_rate(value, "", "sec")

    class ReceivedColumn(NullSortColumn, FreshDataOnlyColumn):
        def render_title(self, session, data):
            return "%s Received" % self.parent.get_unit_plural(session)

        def get_column_key(self, session):
            unit = self.parent.unit.get(session)
            return unit == "b" and "br" or "fr"

        def render_value(self, session, value):
            return fmt_rate(value, "", "sec")

class ConnectionFrame(CuminFrame):
    def __init__(self, app, name):
        super(ConnectionFrame, self).__init__(app, name)

        self.object = ConnectionParameter(app, "id")
        self.add_parameter(self.object)

        view = ConnectionView(app, "view")
        self.add_mode(view)
        self.set_view_mode(view)

        self.__close = ConnectionClose(app, "close")
        self.add_mode(self.__close)

        self.__sessions_detach = SessionSetDetach(app, "sessionsdetach")
        self.add_mode(self.__sessions_detach)

        self.__sessions_close = SessionSetClose(app, "sessionsclose")
        self.add_mode(self.__sessions_close)

    def show_object(self, session, conn):
        reg = conn.vhost.broker.registration
        self.page.main.broker.set_object(session, reg)

        return super(ConnectionFrame, self).show_object(session, conn)

    def show_close(self, session):
        return self.show_mode(session, self.__close)

    def show_sessions_detach(self, session):
        return self.show_mode(session, self.__sessions_detach)

    def show_sessions_close(self, session):
        return self.show_mode(session, self.__sessions_close)

# XXX get rid of this
def doit(error, args):
    pass
    #print error, args
    #print "did it!"

class ConnectionClose(CuminConfirmForm):
    def get_args(self, session):
        return self.frame.get_args(session)

    def process_cancel(self, session, conn):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session, conn):
        action = self.app.model.connection.close
        action.invoke(conn)

        self.process_cancel(session, conn)

    def render_title(self, session, conn):
        return "Close Connection '%s'" % conn.address

    def render_submit_content(self, session, conn):
        return "Yes, Close Connection '%s'" % conn.address

    def render_cancel_content(self, session, conn):
        return "No, Cancel"

class ConnectionSetClose(CuminBulkActionForm):
    def process_return(self, session):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_current_frame(branch, self.frame)
        self.page.set_redirect_url(session, branch.marshal())

    def process_item(self, session, id):
        conn = ClientConnection.get(id)
        action = self.app.model.connection.close
        action.invoke(conn)

    def render_title(self, session):
        return "Close Connections"

    def render_item_content(self, session, id):
        return "Close Connection %s" % ClientConnection.get(id).address

class ConnectionStatus(CuminStatus):
    def render_frames_from(self, session, conn):
        return self.app.model.connection.framesFromClient.rate_html(conn)
        
    def render_frames_to(self, session, conn):
        return self.app.model.connection.framesToClient.rate_html(conn)

    def render_bytes_from(self, session, conn):
        return self.app.model.connection.bytesFromClient.rate_html(conn)
        
    def render_bytes_to(self, session, conn):
        return self.app.model.connection.bytesToClient.rate_html(conn)

class ConnectionView(CuminView):
    def __init__(self, app, name):
        super(ConnectionView, self).__init__(app, name)

        status = ConnectionStatus(app, "status")
        self.add_child(status)
        
        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        self.__tabs.add_tab(ConnectionStats(app, "stats"))

        self.__sessions = SessionSet(app, "sessions")
        self.__tabs.add_tab(self.__sessions)

        self.__tabs.add_tab(CuminDetails(app, "details"))

    def show_sessions(self, session):
        return self.__tabs.show_mode(session, self.__sessions)

    def render_script(self, session, conn):
        data = "model.xml?class=connection;id=%i" % conn.id
        return "wooly.setIntervalUpdate('%s', updateConnection, 3000)" % data

class ConnectionStats(Widget):
    def __init__(self, app, name):
        super(ConnectionStats, self).__init__(app, name)

        self.add_child(StatSet(app, "io", "io"))
        self.add_child(StatSet(app, "general", "general"))

        chart = self.SendReceiveRateChart(app, "sendrecv")
        self.add_child(chart)

    def render_title(self, session):
        return "Statistics"

    class SendReceiveRateChart(StatValueChart):
        def __init__(self, app, name):
            super(ConnectionStats.SendReceiveRateChart, self).__init__(app, name)

            self.stats = ("bytesFromClient", "bytesToClient")
            self.mode = "rate"

        def render_title(self, session, conn):
            return "Bytes Sent and Received"

class SessionSetDetach(CuminBulkActionForm, Frame):
    def process_return(self, session):
        branch = session.branch()
        self.frame.show_view(branch).show_sessions(branch)
        self.page.set_current_frame(branch, self.frame)
        self.page.set_redirect_url(session, branch.marshal())

    def process_item(self, session, id):
        session_ = Session.get(id)
        action = self.app.model.session.detach
        action.invoke(session_)

    def render_title(self, session):
        return "Detach Sessions"
    
    def render_item_content(self, session, id):
        return "Detach Session '%s'" % Session.get(id).name

class SessionSetClose(CuminBulkActionForm, Frame):
    def process_return(self, session):
        branch = session.branch()
        self.frame.show_view(branch).show_sessions(branch)
        self.page.set_current_frame(branch, self.frame)
        self.page.set_redirect_url(session, branch.marshal())

    def process_item(self, session, id):
        session_ = Session.get(id)
        action = self.app.model.session.close
        action.invoke(session_)

    def render_title(self, session):
        return "Close Sessions"
    
    def render_item_content(self, session, id):
        return "Close Session '%s'" % Session.get(id).name

class SessionSet(CuminTable, Form):
    def __init__(self, app, name):
        super(SessionSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.NameColumn(app, "name")
        self.add_column(col)

        self.set_default_column(col)

        col = self.ExpiresColumn(app, "expires")
        self.add_column(col)

        col = self.StatusColumn(app, "status")
        self.add_column(col)

        self.__phase = PhaseSwitch(app, "phase")
        self.add_child(self.__phase)

        self.__detach = self.Detach(app, "detach")
        self.add_child(self.__detach)

        self.__close = self.Close(app, "close")
        self.add_child(self.__close)

    def get_args(self, session):
        return self.frame.get_args(session)
        
    def render_title(self, session, conn):
        return "Sessions %s" % fmt_count(conn.sessions.count())

    def render_sql_where(self, session, conn):
        elems = list()
        elems.append("s.client_connection_id = %(id)r")
        elems.append(self.__phase.get_sql_constraint(session, conn))
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, conn):
        return {"id": conn.id}

    class Detach(FormButton):
        def render_content(self, session):
            return "Detach"

        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            branch = session.branch()
            frame = self.frame.show_sessions_detach(branch)
            frame.ids.set(branch, ids)
            self.page.set_current_frame(branch, frame)
            self.page.set_redirect_url(session, branch.marshal())

    class Close(FormButton):
        def render_content(self, session):
            return "Close"

        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            branch = session.branch()
            frame = self.frame.show_sessions_close(branch)
            frame.ids.set(branch, ids)
            self.page.set_current_frame(branch, frame)
            self.page.set_redirect_url(session, branch.marshal())

    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

    class ExpiresColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Expires"

        def render_value(self, session, value):
            return fmt_datetime(value)

    class StatusColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Attached?"

        def render_content(self, session, data):
            return fmt_predicate(data["attached"])
