import logging

from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *
from wooly.tables import *

from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *
from job import *

strings = StringCatalog(__file__)
log = logging.getLogger("cumin.collector")

class CollectorSet(CuminTable, Form):
    def __init__(self, app, name):
        super(CollectorSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.NameColumn(app, "name")
        self.add_column(col)
        self.set_default_column(col)

        col = self.SystemColumn(app, "system")
        self.add_column(col)

        self.__start = self.StartButton(app, "start")
        self.add_child(self.__start)

        self.__stop = self.StopButton(app, "stop")
        self.add_child(self.__stop)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, collector):
        count = self.get_item_count(session, collector)
        return "Collectors %s" % fmt_count(count)

    def render_sql_where(self, session, collector):
        sql = "qmf_delete_time is null"
        #sql = "((cs.qmf_update_time is null and c.qmf_update_time > now() - interval '10 minutes') or " + \
        #    "(cs.qmf_update_time > now() - interval '10 minutes')" + \
        #    " and qmf_delete_time is null)"
        return "where %s" % sql

    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            coll = Identifiable(data["id"])
            href = self.page.main.pool.collector.get_href(session, coll)
            return fmt_link(href, data["name"])

    class SystemColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "System"

        def render_content(self, session, data):
            id = data["system_id"]

            if id:
                sys = Identifiable(id)
                href = self.page.main.system.get_href(session, sys)
                return fmt_link(href, data["system_name"])

    class StartButton(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)
    
            branch = session.branch()
            frame = self.frame.show_colls_start(branch)
            frame.ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())
            
        def render_content(self, session):
            return "Start"
        
    class StopButton(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)
    
            branch = session.branch()
            frame = self.frame.show_colls_stop(branch)
            frame.ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())
            
        def render_content(self, session):
            return "Stop"
        
class CollectorFrame(CuminFrame):
    def __init__(self, app, name):
        super(CollectorFrame, self).__init__(app, name)

        self.object = CollectorParameter(app, "id")
        self.add_parameter(self.object)

        view = CollectorView(app, "view")
        self.add_mode(view)
        self.set_view_mode(view)

class CollectorView(CuminView):
    def __init__(self, app, name):
        super(CollectorView, self).__init__(app, name)

        status = self.CollectorStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        details = CuminDetails(app, "details")
        self.__tabs.add_tab(details)

    def render_script(self, session, coll):
        data = "model.xml?class=collector;id=%i" % coll.id
        return "wooly.setIntervalUpdate('%s', updateCollector, 3000)" % data

    class CollectorStatus(CuminStatus):
        def render_title(self, session, object):
            return "Collector Status"
            
class CollectorStart(CuminBulkActionForm):
    def render_title(self, session, *args):
        return "Start Collector"
    
    def render_form_heading(self, session, *args):
        return ""

    def render_item_content(self, session, id):
        return "Start Collector '%s'" % Collector.get(id).Name

    def process_item(self, session, item):
        collector = Collector.get(item)
        action = self.app.model.master.start
        action.invoke(collector, {"subsystem": "COLLECTOR"})
        self.process_cancel(session)

class CollectorStop(CuminBulkActionForm):
    def render_title(self, session, *args):
        return "Stop Collector"
    
    def render_form_heading(self, session, *args):
        return ""

    def render_item_content(self, session, id):
        return "Stop Collector '%s'" % Collector.get(id).Name

    def process_item(self, session, item):
        collector = Collector.get(item)
        action = self.app.model.master.stop
        action.invoke(collector, {"subsystem": "COLLECTOR"})
        self.process_cancel(session)
