from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *

from stat import *
from model import *
from widgets import *
from parameters import *
from formats import *
from util import *

strings = StringCatalog(__file__)

class ExchangeInputSet(RadioInputSet):
    def __init__(self, app, name):
        super(ExchangeInputSet, self).__init__(app, name)
        
        param = ExchangeParameter(app, "param")
        self.add_parameter(param)
        self.set_parameter(param)

    def do_get_items(self, session, vhost):
        return sorted_by(vhost.exchanges)

    def render_item_value(self, session, exchange):
        return exchange.id

    def render_item_content(self, session, exchange):
        return exchange.name or "<em>Default</em>"

    def render_item_checked_attr(self, session, exchange):
        return exchange is self.param.get(session) and "checked=\"checked\""

class ExchangeSet(CuminTable, Form):
    def __init__(self, app, name):
        super(ExchangeSet, self).__init__(app, name)

        self.ids = FilteredCheckboxIdColumn(app, "id", self, callback=self.disable_exchange)
        self.add_column(self.ids)

        col = self.NameColumn(app, "name")
        self.add_column(col)
        self.set_default_column(col)

        col = self.ProducersColumn(app, "name")
        col.alignment = "right"
        self.add_column(col)

        col = self.BindingsColumn(app, "bindings")
        col.alignment = "right"
        self.add_column(col)

        col = self.ReceivedColumn(app, "received")
        col.alignment = "right"
        self.add_column(col)

        col = self.RoutedColumn(app, "routed")
        col.alignment = "right"
        self.add_column(col)

        col = self.DroppedColumn(app, "dropped")
        col.alignment = "right"
        self.add_column(col)

        self.unit = UnitSwitch(app, "unit")
        self.add_child(self.unit)

        self.phase = PhaseSwitch(app, "phase")
        self.add_child(self.phase)

        self.__remove = self.Remove(app, "remove")
        self.add_child(self.__remove)

    def get_args(self, session):
        reg = self.frame.get_object(session)
        return (reg.getDefaultVhost(),)
    
    def disable_exchange(self, session, data):
        return data["name"] in ExchangeInfo.get_builtins()
        
    def render_add_exchange_url(self, session, vhost):
        branch = session.branch()
        self.frame.exchange_add.show(branch)
        return branch.marshal()

    def render_title(self, session, vhost):
        return "Exchanges %s" % fmt_count(vhost.exchanges.count())
    
    def render_sql_where(self, session, vhost):
        elems = list()
        elems.append("e.vhost_id = %(id)r")
        elems.append(self.phase.get_sql_constraint(session, vhost))
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, vhost):
        return {"id": vhost.id}

    class Remove(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            branch = session.branch()
            self.frame.exchanges_remove.show(branch).ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())
            
        def render_content(self, session):
            return "Remove"

    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            exchange = Identifiable(data["id"])
            href = self.frame.exchange.get_href(session, exchange)
            name = data["name"] or "<em>Default</em>"
            return fmt_link(href, fmt_shorten(name))

    class ProducersColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Producers"

        def render_content(self, session, data):
            return data["producers"]

    class BindingsColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Bindings"

        def render_content(self, session, data):
            exchange = Identifiable(data["id"])
            href = self.frame.exchange.get_href(session, exchange)
            return fmt_link(href, data["bindings"])

    class ReceivedColumn(NullSortColumn, FreshDataOnlyColumn):
        def render_title(self, session, data):
            return "%s Received" % self.parent.unit.get_brief_plural(session)

        def get_column_key(self, session):
            unit = self.parent.unit.get(session)
            return unit == "b" and "breceived" or "mreceived"

        def render_value(self, session, value):
            return fmt_rate(value, "", "sec")

    class RoutedColumn(NullSortColumn, FreshDataOnlyColumn):
        def render_title(self, session, data):
            return "%s Routed" % self.parent.unit.get_brief_plural(session)

        def get_column_key(self, session):
            unit = self.parent.unit.get(session)
            return unit == "b" and "brouted" or "mrouted"

        def render_value(self, session, value):
            return fmt_rate(value, "", "sec")

    class DroppedColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "%s Dropped" % self.parent.unit.get_brief_plural(session)

        def get_column_key(self, session):
            unit = self.parent.unit.get(session)
            return unit == "b" and "bdropped" or "mdropped"

class ExchangeSetRemove(CuminBulkActionForm):
    def process_return(self, session):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_current_frame(branch, self.frame)
        self.page.set_redirect_url(session, branch.marshal())

    def process_item(self, session, id):
        exchange = Exchange.get(id)
        reg = self.frame.get_object(session)
        action = self.app.model.exchange.remove
        action.invoke(exchange, reg)
 
    def render_title(self, session):
        return "Remove Exchanges"

    def render_item_content(self, session, id):
        return "Remove Exchange '%s'" % Exchange.get(id).name

class ExchangeFrame(CuminFrame):
    def __init__(self, app, name):
        super(ExchangeFrame, self).__init__(app, name)

        self.object = ExchangeParameter(app, "id")
        self.add_parameter(self.object)

        self.view = ExchangeView(app, "view")
        self.add_mode(self.view)
        self.set_view_mode(self.view)
        
        self.remove = ExchangeRemove(app, "remove")
        self.add_mode(self.remove)
        self.set_remove_mode(self.remove)

    def show_object(self, session, exchange):
        reg = exchange.vhost.broker.registration
        self.page.main.broker.set_object(session, reg)

        return super(ExchangeFrame, self).show_object(session, exchange)

    def render_title(self, session, exchange):
        if exchange.name:
            return super(ExchangeFrame, self).render_title(session, exchange)
        else:
            return "Default Exchange"

    def render_href(self, session, exchange):
        if exchange:
            return super(ExchangeFrame, self).render_href(session, exchange)


class ExchangeRemove(CuminConfirmForm):
    def get_args(self, session):
        return self.frame.get_args(session)

    def process_cancel(self, session, exchange):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session, exchange):
        reg = self.frame.frame.get_object(session)
        action = self.app.model.exchange.remove
        action.invoke(exchange, reg)

        branch = session.branch()
        self.frame.frame.show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def render_title(self, session, exchange):
        return "Remove Exchange '%s'" % exchange.name

    def render_submit_content(self, session, exchange):
        return "Yes, Remove Exchange '%s'" % exchange.name

    def render_cancel_content(self, session, exchange):
        return "No, Cancel"

class ExchangeStatus(CuminStatus):
    def render_messages_received(self, session, exchange):
        return self.app.model.exchange.msgReceives.rate_html(exchange)

    def render_messages_routed(self, session, exchange):
        return self.app.model.exchange.msgRoutes.rate_html(exchange)

    def render_messages_dropped(self, session, exchange):
        return self.app.model.exchange.msgDrops.value(exchange)

    def render_bytes_received(self, session, exchange):
        return self.app.model.exchange.byteReceives.rate_html(exchange)

    def render_bytes_routed(self, session, exchange):
        return self.app.model.exchange.byteRoutes.rate_html(exchange)

    def render_bytes_dropped(self, session, exchange):
        return self.app.model.exchange.byteDrops.value(exchange)

class ExchangeView(CuminView):
    def __init__(self, app, name):
        super(ExchangeView, self).__init__(app, name)

        status = ExchangeStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        self.__tabs.add_tab(ExchangeStats(app, "stats"))

        self.__bindings = ExchangeBindingSet(app, "bindings")
        self.__tabs.add_tab(self.__bindings)

        self.__tabs.add_tab(CuminDetails(app, "details"))

    def show_bindings(self, session):
        return self.__tabs.show_mode(session, self.__bindings);

    def render_script(self, session, exchange):
        data = "model.xml?class=exchange;id=%i" % exchange.id
        return "wooly.setIntervalUpdate('%s', updateExchange, 3000)" % data

class ExchangeBindingSet(BindingSet, Form):
    def __init__(self, app, name):
        super(ExchangeBindingSet, self).__init__(app, name)

        self.__remove = self.Remove(app, "remove")
        self.add_child(self.__remove)
        
        self.set_default_column_name("q_id")
    
    def get_visible_columns(self, session):
        return self.get_request_visible_columns(session, ["q_id"])

    def render_title(self, session, exchange):
        return "Queue Bindings %s" % \
            fmt_count(exchange.bindings.count())

    def render_sql_where(self, session, exchange):
        elems = list()
        elems.append("b.exchange_id = %(id)r")
        elems.append(self.phase.get_sql_constraint(session, exchange))
        return "where %s" % " and ".join(elems)

    class Remove(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            href = self.page.main.broker.bindings_remove.get_href(session, ids)
            self.page.set_redirect_url(session, href)
            
        def render_content(self, session):
            return "Remove"

class BindingSetRemove(CuminBulkActionForm):
    def process_item(self, session, id):
        binding = Binding.get(id)
        action = self.app.model.binding.remove
        action.invoke(binding)

    def render_title(self, session):
        return "Remove Binding"
    
    def render_form_heading(self, session, *args):
        return "Remove Binding between:"

    def render_item_content(self, session, id):
        binding = Binding.get(id)
        ename = binding.exchange.name or "Default"
        qname = binding.queue.name
        return "Exchange: %s <i>and</i> Queue: %s" % (ename, qname)

class ExchangeForm(CuminFieldForm):
    def __init__(self, app, name):
        super(ExchangeForm, self).__init__(app, name)

        self.exchange_name = ExchangeNameField(app, "exchange_name")
        self.add_field(self.exchange_name)
 
        self.exchange_type = self.ExchangeTypeField(app, "exchange_type")
        self.add_field(self.exchange_type)
        
        self.more = MoreFieldSet(app, "more")
        self.add_field(self.more)
        
        self.durable = self.ExchangeDurabilityField(app, "durable")
        self.more.add_field(self.durable)
        
        self.sequence = self.SequenceField(app, "sequence")
        self.more.add_field(self.sequence)
        
        self.ive = self.IVEField(app, "ive")
        self.more.add_field(self.ive)
        
    class SequenceField(TwoOptionRadioField):
        def render_title(self, session):
            return "Insert Sequence?"
        
        def render_field_help(self, session):
            return "(Exchange will insert a 'qpid.msg_sequence' field in the message header)"

        def render_title_1(self, session):
            return "Insert Sequence"

        def render_title_2(self, session):
            return "No Sequence"

    class IVEField(TwoOptionRadioField):
        def render_title(self, session):
            return "Initial Value Exchange?"
        
        def render_field_help(self, session):
            return "(Exchange will behave as an 'initial-value-exchange', keeping a reference to the last message forwarded and enqueuing that message to newly bound queues)"

        def render_title_1(self, session):
            return "Initial-Value-Exchange"

        def render_title_2(self, session):
            return "No IVE"

    class ExchangeDurabilityField(TwoOptionRadioField):
        def render_title(self, session):
            return "Durable?"
        
        def render_field_help(self, session):
            return "(Queue is durable)"

        def render_title_1(self, session):
            return "Durable"

        def render_title_2(self, session):
            return "Transient"

    class ExchangeTypeField(RadioField):
        def __init__(self, app, name):
            super(ExchangeForm.ExchangeTypeField, self).__init__(app, name, None)
    
            self.param = Parameter(app, "param")
            self.param.default = "direct"
            self.add_parameter(self.param)
    
            option = self.Direct(app, "direct", self.param)
            self.add_option(option)

            option = self.Topic(app, "topic", self.param)
            self.add_option(option)
    
            option = self.Fanout(app, "fanout", self.param)
            self.add_option(option)
            
            option = self.Headers(app, "headers", self.param)
            self.add_option(option)
            
            option = self.XML(app, "xml", self.param)
            self.add_option(option)
        
        def render_title(self, session):
            return "Exchange Type"
    
        def render_field_help(self, session):
            return "(Type of exchange to add)"
            
        class Direct(RadioFieldOption):
            def render_value(self, session):
                return "direct"
    
            def render_title(self, session):
                return "<em>Direct:</em> Route messages to queues by queue name"
    
        class Topic(RadioFieldOption):
            def render_value(self, session):
                return "topic"
    
            def render_title(self, session):
                return "<em>Topic:</em> Route messages to queues by topic keyword match"
    
        class Fanout(RadioFieldOption):
            def render_value(self, session):
                return "fanout"
    
            def render_title(self, session):
                return "<em>Fan Out:</em> Route message to all queues attached to this exchange"
    
        class Headers(RadioFieldOption):
            def render_value(self, session):
                return "headers"
    
            def render_title(self, session):
                return "<em>Headers:</em> Route message to queues based on content of the message header"
    
        class XML(RadioFieldOption):
            def render_value(self, session):
                return "xml"
    
            def render_title(self, session):
                return "<em>XML:</em> Route message to queues based on XML content of the message"
    
class ExchangeAdd(ExchangeForm):
    def get_args(self, session):
        return (self.frame.get_object(session), )

    def process_cancel(self, session, *args):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session, *args):
        errors = self.validate(session)

        if errors:
            pass
        else:
            reg = self.frame.get_object(session)
            exchange = Exchange()
            exchange.name = self.exchange_name.get(session)
            exchange.type = self.exchange_type.get(session)
            exchange.durable = self.durable.get(session) == "yes" 
            args = {}
            args["reg"] = reg
            args["sequence"] = self.sequence.get(session) == "yes"
            args["ive"] = self.ive.get(session) == "yes"
            
            action = self.app.model.broker.add_exchange
            action.invoke(exchange, args)

            self.process_cancel(session)
            
    def render_title(self, session, reg):
        return "Add Exchange to the Broker '%s'" % reg.name
        
class ExchangeStats(TabbedModeSet):
    def __init__(self, app, name):
        super(ExchangeStats, self).__init__(app, name)

        self.add_child(StatSet(app, "io", "io"))
        self.add_child(StatSet(app, "general", "general"))

        chart = self.ReceiveRouteDropRateChart(app, "recvroutedrop")
        self.add_child(chart)

        chart = StatValueChart(app, "producers")
        chart.stats = ("producerCount",)
        self.add_child(chart)

    def render_title(self, session):
        return "Statistics"

    class ReceiveRouteDropRateChart(StatValueChart):
        def __init__(self, app, name):
            super(ExchangeStats.ReceiveRouteDropRateChart, self).__init__ \
                (app, name)

            self.stats = ("msgReceives", "msgRoutes", "msgDrops")
            self.mode = "rate"

        def render_title(self, session, exchange):
            return "Messages Received, Routed, and Dropped"

class ExchangeProducerSet(ItemSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, exchange):
        return "Producers %s" % \
            fmt_count(self.get_item_count(session, exchange))

    def get_item_count(self, session, exchange):
        return exchange.producers.count()

    def do_get_items(self, session, exchange):
        return exchange.producers

    def render_item_name(self, session, producer):
        return producer.name

class ExchangeInfo(object):
    @classmethod
    def is_builtin(cls, exchange):
        return exchange.name in ExchangeInfo.get_builtins()

    @classmethod
    def get_builtins(cls):
        return ["", "amq.direct", "amq.topic", "amq.match", "amq.fanout", "qpid.management"]
