import logging

from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *
from wooly.tables import *
from time import time

from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *

strings = StringCatalog(__file__)
log = logging.getLogger("cumin.job")

class JobSet(CuminTable, Form):
    def __init__(self, app, name):
        super(JobSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.CustomIdColumn(app, "custom_id")
        self.add_column(col)
        self.set_default_column(col)
        
        col = self.GlobalJobIdColumn(app, "global_job_id")
        col.visible = False
        self.add_column(col)
        
        col = self.CommandColumn(app, "cmd")
        self.add_column(col)

        col = self.AccountingGroupColumn(app, "agroup")
        col.visible = False
        self.add_column(col)

        col = self.SchedulerColumn(app, "scheduler")
        col.visible = False
        self.add_column(col)

        col = self.SubmitterColumn(app, "submitter")
        col.visible = False
        self.add_column(col)

        col = self.ClusterIdColumn(app, "cluster_id")
        col.alignment = "right"
        col.visible = False
        self.add_column(col)
        
        col = self.ConcurrencyLimitsColumn(app, "concurrency_limits")
        #col.visible = False
        self.add_column(col)
        
        col = self.CustomGroupColumn(app, "custom_group")
        col.visible = False
        self.add_column(col)
        
        col = self.CustomPriorityColumn(app, "custom_priority")
        col.visible = False
        self.add_column(col)
        
        col = self.StatusColumn(app, "job_status")
        self.add_column(col)

        col = self.TitleColumn(app, "title")
        col.visible = False
        self.add_column(col)
        
        col = self.ArgsColumn(app, "args")
        col.alignment = "right"
        col.visible = False
        self.add_column(col)
        
        self.phase = JobStatusSwitch(app, "phase")
        self.add_child(self.phase)

    def get_args(self, session):
        return self.frame.get_args(session)

    def get_phase_sql(self, session, phase=None):
        return self.phase.get_sql_constraint(session, phase)
    
    def get_phase_title(self, session):
        state = self.phase.get(session)
        return self.phase.get_title(state)

    def render_sql_where(self, session, *args):
        phase = len(args) > 1 and args[1] or None
        elems = list()
        elems.append(self.get_phase_sql(session, phase))
        elems.append("s.pool = %(pool)s")
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, *args):
        pool = args[0]
        return {"pool": pool.id}

    class ArgsColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Arguments"
        
    class ClusterIdColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Cluster Id"
        
    class ConcurrencyLimitsColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Limits"
        
    class CustomGroupColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Job Group"
        
        def render_content(self, session, data):
            name = data[self.name]

            if name:
                group = Identifiable(name)
                href = self.page.main.pool.job_group.get_href(session, group)
                return fmt_link(href, fmt_shorten(name, 12, 3))
            
    class CustomIdColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "ID"
        
        def render_content(self, session, data):
            id = data[self.name]

            if id:
                job = Identifiable(data["id"])
                href = self.page.main.pool.job.get_href(session, job)
                return fmt_link(href, fmt_shorten(id, 12, 3))

    class CustomPriorityColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Custom Priority"
        
    class GlobalJobIdColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Global Job Id"
        
        def render_content(self, session, data):
            id = data["id"]

            if id:
                job = Identifiable(data["id"])
                href = self.frame.job.get_href(session, job)
                content = fmt_shorten(data[self.name], 12, 3)
                return fmt_link(href, content)
            
    class StatusColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Status"
        
        def render_content(self, session, data):
            stat = data["job_status"]
            return JobStatusInfo.get_status_string(stat)            
            
    class AccountingGroupColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Accounting Group"
        
    class SubmitterColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Submitter"
        
        def render_content(self, session, data):
            name = data["submitter"]

            if name:
                sub = Identifiable(data["submitter_id"])
                href = self.page.main.pool.submitter.get_href(session, sub)
                return fmt_link(href, fmt_shorten(name))
            
    class SchedulerColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Scheduler"
        
        def render_content(self, session, data):
            name = data["scheduler"]

            if name:
                sched = Identifiable(data["scheduler_id"])
                href = self.page.main.pool.scheduler.get_href(session, sched)
                return fmt_link(href, fmt_shorten(name))

    class CommandColumn(ItemTableColumn):
        def render_title(self, session, data):
            return "Command"
        
        def render_content(self, session, data):
            return fmt_shorten(data[self.name], 0, 16)

    class TitleColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Title"

class TopJobSet(TopTable):
    def __init__(self, app, name):
        super(TopJobSet, self).__init__(app, name)

        col = self.NameColumn(app, "name")
        self.add_column(col)

        self.set_default_column(col)

        col = self.DurationColumn(app, "name")
        col.alignment = "right"
        self.add_column(col)

    class NameColumn(TopTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            pool = Identifiable(data["pool"])
            job = Identifiable(data["id"])

            branch = session.branch()
            self.page.main.pool.object.set(session, pool)
            self.page.main.pool.job.object.set(session, job)
            self.page.main.pool.job.show(session)
            return fmt_link(branch.marshal(), fmt_shorten(data["name"]))

    class DurationColumn(TopTableColumn):
        def render_title(self, session, data):
            return "Duration"

        def render_content(self, session, data):
            since = data["q_date"]
            return fmt_duration(time() - secs(since))

class JobTab(JobSet):
    def __init__(self, app, name):
        super(JobTab, self).__init__(app, name)
        
        self.__remove = JobRemoveButton(app, "remove")
        self.add_child(self.__remove)

        self.__hold = JobHoldButton(app, "hold")
        self.add_child(self.__hold)

        self.__release = JobReleaseButton(app, "release")
        self.add_child(self.__release)

        self.job_search = self.JobSearch(app, "job_search")
        self.add_child(self.job_search)

    def find_job(self, session):
        search_term = self.job_search.get(session)

        if search_term:
            object = self.frame.get_object(session)
            rows = self.find_item(session, object)

            if rows:
                try:
                    first = rows[0]
                    job = Identifiable(first["id"])
                    href = self.app.main_page.main.pool.job.get_href(session, job)                    
                    self.page.set_redirect_url(session, href)
                except:
                    self.job_search.set_not_found(session, search_term)
            else:
                self.job_search.set_not_found(session, search_term)
    
            self.job_search.set(session, None)

    def render_find_sql_where(self, session, *args):
        return "j.custom_id = %(custom_id)s"
    
    def get_find_sql_values(self, session, pool):
        return {"custom_id": self.job_search.get(session)}
        #return {"custom_id": self.job_search.get(session), "pool": pool.id}
    
    def get_full_item_count(self, session, *args):
        # request item count will phase=="a"
        new_args = (args[0], "a")
        return self.get_item_count(session, *new_args)

    def get_visible_columns(self, session):
        return self.get_request_visible_columns(session, ["custom_group", "scheduler", "submitter"])

    def render_count(self, session, *args):
        count = self.get_item_count(session, *args)
        phase = self.get_phase_title(session)
        if phase == "All":
            phase = ""
        return "%i %s %s" % (count, phase, count == 1 and "Job" or "Jobs")

    class JobSearch(StringInput):
        """ displays the input box and button used to search for job id """
        
        def __init__(self, app, name):
            super(JobTab.JobSearch, self).__init__(app, name)

            self.__go = self.JobSearchButton(app, "go")
            self.add_child(self.__go)
            
            self.not_found = Attribute(app, "error")
            self.add_attribute(self.not_found)

        def set_not_found(self, session, value):
            self.not_found.set(session, value)
        
        def render_search_prompt(self, session):
            not_found = self.not_found.get(session)
            return not_found and "%s Not Found" % escape_entity(not_found) or self.render_search_default_prompt(session)
        
        def render_search_default_prompt(self, session):
            return "Enter Job ID"
        
        def render_search_class(self, session):
            return self.not_found.get(session) and "search_error" or " "

        def find_job(self, session):
            self.parent.find_job(session)
    
        class JobSearchButton(FormButton):
            def process_submit(self, session):
                self.parent.find_job(session)
    
            def render_content(self, session):
                return "Go"
        
from system import SystemFrame, SystemSet

class JobGroupFrame(CuminFrame):
    def __init__(self, app, name):
        super(JobGroupFrame, self).__init__(app, name)

        self.object = JobGroupParameter(app, "id")
        self.add_parameter(self.object)

        self.view = JobGroupView(app, "view")
        self.add_mode(self.view)
        self.set_view_mode(self.view)

        self.jobs_hold = JobSetHold(app, "jobshold")
        self.add_mode(self.jobs_hold)

        self.jobs_release = JobSetRelease(app, "jobsrelease")
        self.add_mode(self.jobs_release)

        self.jobs_remove = JobSetRemove(app, "jobsremove")
        self.add_mode(self.jobs_remove)

        self.system = SystemFrame(app, "system")
        self.add_mode(self.system)

class JobGroupView(CuminView):
    def __init__(self, app, name):
        super(JobGroupView, self).__init__(app, name)

        summary = CuminSummary(app, "summary")
        self.add_child(summary)

        status = JobGroupStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        self.__tabs.add_tab(JobGroupStats(app, "stats"))
        self.__tabs.add_tab(JobGroupJobSet(app, "jobs"))
        self.__tabs.add_tab(JobGroupSystemSet(app, "systems"))
        #self.__tabs.add_tab(CuminDetails(app, "details"))

class JobGroupStats(Widget):
    def __init__(self, app, name):
        super(JobGroupStats, self).__init__(app, name)

        stats = JobGroupStatSet(app, "general", "general")
        self.add_child(stats)

    def render_title(self, session):
        return "Statistics"
    
class JobGroupStatSet(StatSet):
    def __init__(self, app, name, category):
        super(JobGroupStatSet, self).__init__(app, name, category)
        
        self.jobs = Attribute(app, "jobs")
        self.add_attribute(self.jobs)
        
    def process(self, session):
        group = self.frame.get_args(session)[0]
        if group:
            where_group = "custom_group = '%s'" % group.get_id()
            value = Job.select(where_group).count()
            self.jobs.set(session, value)
        super(JobGroupStatSet, self).process(session)
        
    def render_rate_text(self, session, args):
        return "Percentage"

    def render_item_name(self, session, args):
        stat, object = args
        return stat.name

    def render_item_value(self, session, args):
        stat, group = args
        if stat.name == "Jobs":
            return self.jobs.get(session)
        else:
            state = stat.name
            value = self.get_value(group, state)
        return value

    def get_value(self, group, state):
        where_group = "custom_group = '%s' \
            and job_status = %i" % (group.get_id(), 
                                    JobStatusInfo.get_status_int(state))
        return Job.select(where_group).count()

    def render_item_rate(self, session, args):
        stat, group = args
        jobs = self.jobs.get(session)
        state = stat.name
        if stat.name == "Jobs":
            value = jobs
        else:
            value = self.get_value(group, state)
        percent = (value*1.0) / (jobs*1.0) * 100.0
        return jobs and "%2.1f" % percent or "-"
    
class JobRemoveButton(FormButton):
    def process_submit(self, session):
        ids = self.parent.ids.get(session)
        self.parent.ids.clear(session)

        href = self.frame.jobs_remove.get_href(session, ids)
        self.page.set_redirect_url(session, href)
        
    def render_content(self, session):
        return "Remove"
    
    def render_disabled_attr(self, session, *args):
        phase = self.parent.phase.get(session)
        return phase == "d" and "disabled=\"disabled\"" or None

    def render_class(self, session, *args):
        phase = self.parent.phase.get(session)
        return phase == "d" and "disabled" or "enabled"

class JobHoldButton(FormButton):
    def process_submit(self, session):
        ids = self.parent.ids.get(session)
        self.parent.ids.clear(session)

        href = self.frame.jobs_hold.get_href(session, ids)
        self.page.set_redirect_url(session, href)
        
    def render_content(self, session):
        return "Hold"
    
    def render_disabled_attr(self, session, *args):
        phase = self.parent.phase.get(session)
        return (phase == "h" or phase == "d") and "disabled=\"disabled\"" or None

    def render_class(self, session, *args):
        phase = self.parent.phase.get(session)
        return (phase == "h" or phase == "d") and "disabled" or "enabled"

class JobReleaseButton(FormButton):
    def process_submit(self, session):
        ids = self.parent.ids.get(session)
        self.parent.ids.clear(session)

        href = self.frame.jobs_release.get_href(session, ids)
        self.page.set_redirect_url(session, href)
        
    def render_content(self, session):
        return "Release"
    
    def render_disabled_attr(self, session, *args):
        phase = self.parent.phase.get(session)
        return (phase == "r" or phase == "d") and "disabled=\"disabled\"" or None

    def render_class(self, session, *args):
        phase = self.parent.phase.get(session)
        return (phase == "r" or phase == "d") and "disabled" or "enabled"

class JobGroupJobSet(JobTab):
    def __init__(self, app, name):
        super(JobGroupJobSet, self).__init__(app, name)
        
        self.__remove = JobRemoveButton(app, "remove")
        self.add_child(self.__remove)

        self.__hold = JobHoldButton(app, "hold")
        self.add_child(self.__hold)

        self.__release = JobReleaseButton(app, "release")
        self.add_child(self.__release)

    def get_visible_columns(self, session):
        return self.get_request_visible_columns(session, ["scheduler", "submitter"])
        
    def render_title(self, session, group):
        where_group = "custom_group = '%s'" % group.get_id()
        return "Jobs %s" % fmt_count(Job.select(where_group).count())

    def render_sql_where(self, session, group):
        phase_sql = self.get_phase_sql(session)
        group_sql = "j.custom_group = '%s'" % group.get_id()
        return "where %s" % " and ".join([phase_sql, group_sql])

    def render_count(self, session, *args):
        str = super(JobGroupJobSet, self).render_count(session, *args)
        return "%s in Job Group '%s'" % (str, args[0].id)

class JobGroupStatus(CuminStatus):
    def render_color(self, session, job_group):
        #return JobStatusInfo.get_status_color(job.JobStatus)
        return "green"

    def render_job_status(self, session, job_group):
        #return JobStatusInfo.get_status_string(job.JobStatus)
        return "Status"

class JobGroupSystemSet(SystemSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_sql_where(self, session, group):
        subquery = """
            select 1
            from slot as l
            join job as j on j.custom_id = l.job_id
            where j.custom_group = %(id)s and l.system = s.node_name
        """

        return "where exists (%s)" % subquery

    def get_sql_values(self, session, group):
        return {"id": group.id}

class JobFrame(CuminFrame):
    def __init__(self, app, name):
        super(JobFrame, self).__init__(app, name)

        self.object = JobParameter(app, "id")
        self.add_parameter(self.object)

        view = JobView(app, "view")
        self.add_mode(view)
        self.set_view_mode(view)

        self.__edit_ads = JobAdsEditor(app, "editads")
        self.add_mode(self.__edit_ads)
        
        self.__system = SystemFrame(app, "system")
        self.add_mode(self.__system)

    def show_ads_edit(self, session):
        self.page.set_current_frame(session, self.__edit_ads)
        return self.show_mode(session, self.__edit_ads) 
    
    def show_job_group(self, session, jobgroup):
        self.__job.set_object(session, jobgroup)
        self.page.set_current_frame(session, self.__job)
        self.__job.set_switch(session, "group")
        return self.show_mode(session, self.__job)
    
    def show_system(self, session, system):
        frame = self.show_mode(session, self.__system)
        frame.set_object(session, system)
        return self.page.set_current_frame(session, frame)


class JobView(CuminView):
    def __init__(self, app, name):
        super(JobView, self).__init__(app, name)

        summary = CuminSummary(app, "summary")
        self.add_child(summary)

        status = JobStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        self.__tabs.add_tab(JobAdsViewer(app, "jobads"))
        self.__tabs.add_tab(JobOutput(app, "output"))
        self.__tabs.add_tab(JobSystemSet(app, "systems"))
        self.__tabs.add_tab(CuminDetails(app, "details"))

class JobSystemSet(SystemSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_sql_where(self, session, job):
        return """
            where exists
              (select 1
               from slot
               where system = s.node_name and job_id = %(id)s)
        """

    def get_sql_values(self, session, job):
        return {"id": "%i.%i" % (job.ClusterId, job.ProcId)}

class JobAdsSet(PropertySet):
    def __init__(self, app, name):
        super(JobAdsSet, self).__init__(app, name, )

        self.types = {0: "expression",
                      1: "integer",
                      2: "float",
                      3: "string"}
    def get_args(self, session):
        return self.frame.get_args(session)
           
    def do_get_items(self, session, job):
        items = self.items.get(session)
        if not items:
            items = self.gen_items(session, job)
            # cache the items
            self.items.set(session, items)

        return items

    def get_raw_ads(self, session, job):
        action = self.app.model.job.getad
        return action.do_invoke(job)
    
    def gen_items(self, session, job):
        job_ads = self.get_raw_ads(session, job)
        
        cls = self.app.model.get_class_by_object(job)
        return [self.gen_item(x, job_ads[x]["VALUE"], cls, dtype=self.types[job_ads[x]["TYPE"]]) for x in job_ads]
        # list of dictionaries
        # each disctionary has:
        # name:, value:, type: [, error:] [, property:] [,path:]
        #return [self.gen_item(x, job_ads[x], cls) for x in job_ads]
#       

    def gen_item(self, name, value, cls, path=None, dtype=None, error=None, orig=None):
        """ Generate a dict with name, value, type, error, path, property, orig
        
            This is called with raw GetAd data and with processed data from
            a form submit. With raw data, only the name and value will be present.
            With form data, we might have a path, dtype, or error. dtype is the 
            data type that was remembered from the raw data.
        """
            
        idict = dict()
        idict["name"] = name
        idict["value"] = value
        idict["orig"] = value
        idict["type"] = dtype
        if dtype == "string":
            idict["value"] = self.strip_string_quotes(value)
        if orig:
            idict["orig"] = orig
        if error:
            if "error" in error:
                idict["error"] = error["error"] 

        if name in cls.ad_properties_by_name:
            idict["property"] = cls.ad_properties_by_name[name]
        if path:
            idict["path"] = path
        return idict
    
    def strip_string_quotes(self, value):
        dvalue = value
        if value:
            if value[:1] == "\"" and value[-1:] ==  "\"":
                dvalue = value[1:-1]
        return dvalue
        
class JobPropertyRenderer(TemplateRenderer):
    def render_title(self, session, item):
        title = item["name"]
        if "property" in item:
            property = item["property"]
            if property.title:
                title = property.get_title(session) 
        return escape_amp(title)

    def render_value(self, session, item):
        value = item["value"]
        if "property" in item:
            property = item["property"]
            if property.renderer:
                value = property.renderer(session, value)
        ret = escape_entity(str(value))
        return self.insert_breaks(ret)
                
    def insert_breaks(self, value):
        subwords = list()
        snippets = parse(value, begin_delim="&", end_delim=";")
        for snippet in snippets:
            while len(snippet) > 40:
                subwords.append(snippet[:40])
                snippet = snippet[40:]
            subwords.append(snippet)
        
        return "&#8203;".join(subwords)

    def render_inline_help(self, session, item):
        if "property" in item:
            property = item["property"]
            return property.description

class JobAdsViewer(JobAdsSet):
    def __init__(self, app, name):
        super(JobAdsViewer, self).__init__(app, name)

        self.group_tmpl = Template(self, "group_html")
        self.item_renderer = JobPropertyRenderer(self, "property_html")

    def get_args(self, session):
        return self.frame.get_args(session)
        
    def render_title(self, session, job):
        return "Attributes"

    def do_get_items(self, session, args):
        job = args[0]
        group = args[1]
        group_items = list()
        if job:
            all_items = super(JobAdsViewer, self).do_get_items(session, job)
            for item in all_items:
                if "property" in item:
                    property = item["property"]
                    item_group = property.group
                else:
                    item_group = "Other"
                if item_group == group:
                    group_items.append(item)
                    
        return group_items
    
    def render_properties(self, session, *args):
        items = self.do_get_items(session, *args)
        writer = Writer()

        for item in items:
            self.item_renderer.render(writer, session, item)

        return writer.to_string()

    def render_edit_ads_url(self, session, job):
        branch = session.branch()
        self.frame.show_ads_edit(branch)
        return branch.marshal()

    def render_groups(self, session, job):
        groups = self.app.model.get_ad_groups()
        writer = Writer()
        for group in groups:
            self.group_tmpl.render(writer, session, (job, group,))
        return writer.to_string()

    def render_group_name(self, session, args):
        return args[1]
        
class JobAdsEditor(CuminForm, JobAdsViewer):
    def __init__(self, app, name):
        super(JobAdsEditor, self).__init__(app, name)

        # the parameter that will hold all the field values
        self.ads = DictParameter(app, "params")
        self.add_parameter(self.ads)

        self.item_renderer = EditablePropertyRenderer(self, "property_html")

    def do_get_items(self, session, args):
        job = args[0]
        group = args[1]
        cls = self.app.model.get_class_by_object(job)
        ads = self.ads.get(session)
        if len(ads):
            return [self.gen_item(x, ads[x]["value"], cls, path=self.ads.path, 
                                  dtype=ads[x]["type"], error=ads[x], orig=ads[x]["orig"]) for x in ads 
                                  if self.is_group(x, cls, group)]
        else:
            items = super(JobAdsEditor, self).do_get_items(session, args)
            for item in items:
                item["path"] = self.ads.path
            return items
                
    def is_group(self, name, cls, group):
        if name in cls.ad_properties_by_name:
            property = cls.ad_properties_by_name[name]
            item_group = property.group
        else:
            item_group = "Other"
        return item_group == group

    def process_cancel(self, session, job):
        branch = session.branch()
        self.ads.set(branch, None) # otherwise url is too long
        self.page.pop_current_frame(branch)
        self.page.get_current_frame(branch).show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())
        
    def process_submit(self, session, job):
        ads = self.ads.get(session)
        errors = False
        
        just_ads = dict()
        for field in ads:
            try:
                fval = ads[field]["value"]
            except KeyError:
                ads[field]["value"] = ""
                fval = ""
            ftype = ads[field]["type"]
            if ftype == "integer":
                try:
                    fval = int(fval)
                except:
                    ads[field]["error"] = "Integer value expected"
                    errors = True
            elif ftype == "float":
                try:
                    fval = float(fval)
                except:
                    ads[field]["error"] = "Floating point value expected"
                    errors = True
            elif ftype == "string":
                fval = "\"%s\"" % fval
            else:
                fval = unicode(fval)
            if "orig" in ads[field]:
                orig = ads[field]["orig"]
                if ftype == "integer":
                    orig = int(orig)
                elif ftype == "float":
                    orig = float(orig)
                if fval != orig:
                    just_ads[unicode(field)] = fval
        
        if not errors:
            for field in just_ads:
                action = self.app.model.job.setattribute
                action.invoke(job, [field, just_ads[field]])
            self.process_cancel(session, job)

class JobOutput(Form):
    def __init__(self, app, name):
        super(JobOutput, self).__init__(app, name)

        self.which_file = self.FileSwitch(app, "file")
        self.add_child(self.which_file)
        
        self.first_last = self.FLSwitch(app, "first_last")
        self.add_child(self.first_last)
        
        self.__fetch = self.FetchButton(app, "refresh")
        self.add_child(self.__fetch)

        self.ads = JobAdsSet(app, "ads")
        self.add_child(self.ads)
    
        self.out_file = Parameter(app, "out")
        self.add_parameter(self.out_file)
        
        self.err_file = Parameter(app, "err")
        self.add_parameter(self.err_file)
        
        self.user_file = Parameter(app, "user")
        self.add_parameter(self.user_file)
        
        self.output = self.OutputFile(app, "job_output")
        self.add_child(self.output)

    def get_args(self, session):
        return self.frame.get_args(session)
    
    def render_title(self, session, *args):
        return "Output"
    
    def render(self, session, *args):
        out_file = self.out_file.get(session)
        err_file = self.err_file.get(session)
        user_file = self.user_file.get(session)
        if not out_file:
            job = self.frame.get_args(session)[0]
            items = self.ads.get_raw_ads(session, job)
            out_file = err_file = user_file = None
            if "Out" in items:
                out_file = items["Out"]["VALUE"]
            if "Err" in items:
                err_file = items["Err"]["VALUE"]
            if "UserLog" in items:
                user_file = items["UserLog"]["VALUE"]

            # strip any double quotes surrounding file names
            out_file = self.ads.strip_string_quotes(out_file)
            err_file = self.ads.strip_string_quotes(err_file)
            user_file = self.ads.strip_string_quotes(user_file)
            
            # remember the file names so we don't need to reget them
            self.out_file.set(session, out_file)
            self.err_file.set(session, err_file)
            self.user_file.set(session, user_file)

        # set title for radiotab so mouseover will display file name
        self.which_file.set_file_name(session, "o", out_file)
        self.which_file.set_file_name(session,"e", err_file)
        self.which_file.set_file_name(session,"u", user_file)

        if self.which_file.is_bad(out_file):
            self.which_file.disable(session, "o")
        if self.which_file.is_bad(err_file):
            self.which_file.disable(session, "e")
        if self.which_file.is_bad(user_file):
            self.which_file.disable(session, "u")
            
        return super(JobOutput, self).render(session, *args)
        
    def render_loading(self, session, *args):
        file = self.which_file.get_current_file_name(session)
        return file and "loading..." or "Output, Error, and UserLog file names are invalid."
         
    def get_file_args(self, session):
        first_last = self.first_last.get(session)
        if first_last == "t":
            start = -2048
            end = 0
        else:
            start = 0
            end = 2048
        file = self.which_file.get_current_file_name(session)
        return (file, start, end)
    
    class OutputFile(AjaxField):
        def get_url(self, session):
            job = self.parent.frame.get_args(session)[0]
            if job:
                file, start, end = self.parent.get_file_args(session)
                if file:
                    fl = self.parent.first_last.get(session)
                    args = "%s&%i&%i&%s" % (file, start, end, fl)
                    args = escape_amp(args)
                    return "call.xml?class=job;id=%i;method=fetch;xargs=%s" % (job.id, args)
        
        def do_render(self, session):
            return self.render_script(session)

    class FetchButton(FormButton):
        def render_content(self, session):
            return "Refresh"
        
    class FileSwitch(StateSwitch):
        def __init__(self, app, name):
            super(JobOutput.FileSwitch, self).__init__(app, name)
    
            self.add_state("o", "Output")
            self.add_state("e", "Error")
            self.add_state("u", "UserLog")

            self.disabled = self.DisabledList(app, "disabled")
            self.add_attribute(self.disabled)
            
            self.link_titles = self.Titles(app, "link_titles")
            self.add_attribute(self.link_titles)
            
        class DisabledList(Attribute):
            def get_default(self, session):
                return list()
                        
        class Titles(Attribute):
            def get_default(self, session):
                return dict()
                        
        def disable(self, session, state):
            disabled = self.disabled.get(session)
            disabled.append(state)
            self.disabled.set(session, disabled)
            if state == self.get(session):
                self.select_first_enabled(session)
            
        def set_file_name(self, session, state, link_title):
            link_titles = self.link_titles.get(session)
            link_titles[state] = link_title
            self.link_titles.set(session, link_titles)
            
        def get_file_name(self, session, state):
            link_titles = self.link_titles.get(session)
            return state in link_titles and link_titles[state] or ""
            
        def get_current_file_name(self, session):
            state = self.get(session)
            return self.get_file_name(session, state)
        
        def select_first_enabled(self, session):
            states = self.get_items(session)
            disabled = self.disabled.get(session)                
            for state in states:
                if not state in disabled:
                    self.set(session, state)
                    break

        def render_item_link(self, session, state):
            branch = session.branch()
            self.set(branch, state)
    
            title = self.get_title(state)
            link_titles = self.link_titles.get(session)
            link_title = state in link_titles and link_titles[state] or ""
            disabled = self.disabled.get(session)
            if state in disabled:
                class_ = "disabled"
                href = "javascript:void(0)"
            else:
                class_ = self.get(session) == state and "selected"
                href = branch.marshal()
            return fmt_link(href, title, class_, link_title=link_title)

        def is_bad(self, file):
            bad = False
            if not file:
                bad = True
            elif "/dev/null" in file.lower():
                bad = True
            return bad
            
            
    class FLSwitch(StateSwitch):
        def __init__(self, app, name):
            super(JobOutput.FLSwitch, self).__init__(app, name)
    
            self.add_state("t", "Tail", "Display end of file")
            self.add_state("h", "Head", "Display beginning of file")

class JobStatus(CuminStatus):
    def render_color(self, session, job):
        if job:
            return JobStatusInfo.get_status_color(job.JobStatus)

    def render_job_status(self, session, job):
        if job:
            return JobStatusInfo.get_status_string(job.JobStatus)
        
class JobGroupSet(CuminTable):
    def __init__(self, app, name):
        super(JobGroupSet, self).__init__(app, name)

        self.ids = CheckboxStringIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.GroupColumn(app, "job_group")
        self.add_column(col)

        col = self.JobsCountColumn(app, "jobs")
        col.alignment = "right"
        self.add_column(col)
        
    class GroupColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Job Group"

        def render_content(self, session, data):
            name = data[self.name]

            if name:
                group = Identifiable(name)
                href = self.page.main.pool.job_group.get_href(session, group)
                return fmt_link(href, fmt_shorten(name))
        
    class JobsCountColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Jobs"
        
class JobsAndGroupsTab(Widget):
    def __init__(self, app, name):
        super(JobsAndGroupsTab, self).__init__(app, name)

        self.is_group = GroupSwitch(app, "group")
        self.add_child(self.is_group)

        self.job_tab = JobTab(app, "jobtab")
        self.add_child(self.job_tab)
        
        self.job_group_tab = JobGroupTab(app, "jobgrouptab")
        self.add_child(self.job_group_tab)
    
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, *args):
        return "Jobs %s" % fmt_count(self.job_tab.get_full_item_count(session, *args))

    def show_status_switch(self, session):
        is_group = self.is_group.get(session)
        return is_group == "j"
    
    def set_switch(self, session, switch):
        if switch == "group":
            self.is_group.set(session, "g")
        else:
            self.is_group.set(session, "j")
    
    def render_phase(self, session, *args):
        is_group = self.is_group.get(session)
        if is_group == "j":
            return self.job_tab.phase.render(session)

    def render_jobs(self, session, *args):
        is_group = self.is_group.get(session)
        if is_group == "j":
            return self.job_tab.render(session)
    
    def render_job_groups(self, session, *args):
        is_group = self.is_group.get(session)
        if is_group == "g":
            return self.job_group_tab.render(session)

class JobGroupTab(JobGroupSet, Form):
    def __init__(self, app, name):
        super(JobGroupTab, self).__init__(app, name)
        
        self.__remove = self.Remove(app, "remove")
        self.add_child(self.__remove)

        self.__hold = self.Hold(app, "hold")
        self.add_child(self.__hold)

        self.__release = self.Release(app, "release")
        self.add_child(self.__release)

        self.set_default_column_name("job_group")

    class Remove(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            href = self.frame.job_group_remove.get_href(session, ids)
            self.page.set_redirect_url(session, href)
            
        def render_content(self, session):
            return "Remove"
        
    class Hold(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            href = self.frame.job_group_hold.get_href(session, ids)
            self.page.set_redirect_url(session, href)
            
        def render_content(self, session):
            return "Hold"
        
    class Release(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            href = self.frame.job_group_release.get_href(session, ids)
            self.page.set_redirect_url(session, href)
            
        def render_content(self, session):
            return "Release"

class JobReasonBulkActionForm(CuminBulk):
    def __init__(self, app, name):
        super(JobReasonBulkActionForm, self).__init__(app, name)
        
        self.reason = Parameter(app, "reason")
        self.add_parameter(self.reason)
        
        self.error = Attribute(app, "error")
        self.add_attribute(self.error)

        self.error_tmpl = Template(self, "reason_html")
        
    def render_reason_text(self, session, *args):
        return "Reason"
    
    def render_reason_path(self, session, *args):
        return self.reason.path
    
    def render_reason_value(self, session, *args):
        return self.reason.get(session)
    
    def render_reason_error(self, session, *args):
        if self.error.get(session):
            writer = Writer()
            self.error_tmpl.render(writer, session, *args)
            return writer.to_string()
    
    def render_reason_error_text(self, session, *args):
        return self.error.get(session)
    
    def render_form_heading(self, session, *args):
        pass
    
    def process_submit(self, session, *args):
        reason = self.reason.get(session)
        if not reason:
            self.error.set(session, "Reason is required")
        else:
            super(JobReasonBulkActionForm, self).process_submit(session, *args)

    def process_return(self, session, *args):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_current_frame(branch, self.frame)
        self.page.set_redirect_url(session, branch.marshal())

    def get_reason(self, session, verb):
        """ returns <verb> by username[: <user input reason>] """
        reason = [self.reason.get(session)]
        verb_by = "%s by %s" % (verb, session.user_session.subject.name)
        reason.insert(0, verb_by)
        return ": ".join(reason) 

class JobReasonIntBulkActionForm(JobReasonBulkActionForm, CuminIntIdList):        
    def get_items(self, session, *args):
        return self.ids.get(session)

class JobSetHold(JobReasonIntBulkActionForm):
    def process_item(self, session, id):
        job = Job.get(id)
        action = self.app.model.job.hold
        reason = self.get_reason(session, "held")
        action.invoke(job, reason)
 
    def render_title(self, session):
        ids = self.ids.get(session)
        return len(ids) > 1 and "Hold Jobs" or "Hold Job"

    def render_item_content(self, session, id):
        return "Hold Job '%s'" % Job.get(id).CustomId

    def render_reason_text(self, session, *args):
        return "Hold Reason"
    
class JobSetRelease(JobReasonIntBulkActionForm):
    def process_item(self, session, id):
        job = Job.get(id)
        action = self.app.model.job.release
        reason = self.get_reason(session, "released")
        action.invoke(job, reason)
 
    def render_title(self, session):
        ids = self.ids.get(session)
        return len(ids) > 1 and "Release Jobs" or "Release Job"

    def render_item_content(self, session, id):
        return "Release Job '%s'" % Job.get(id).CustomId

    def render_reason_text(self, session, *args):
        return "Release Reason"
    
class JobSetRemove(JobReasonIntBulkActionForm):
    def process_item(self, session, id):
        job = Job.get(id)
        action = self.app.model.job.remove
        reason = self.get_reason(session, "removed")
        action.invoke(job, reason)
 
    def render_title(self, session):
        ids = self.ids.get(session)
        return len(ids) > 1 and "Remove Jobs" or "Remove Job"

    def render_item_content(self, session, id):
        return "Remove Job '%s'" % Job.get(id).CustomId

    def render_reason_text(self, session, *args):
        return "Remove Reason"
    
class JobReasonStringBulkActionForm(JobReasonBulkActionForm, CuminStringIdList):        
    def get_items(self, session, *args):
        return self.ids.get(session)

class JobGroupHold(JobReasonStringBulkActionForm):
    def process_item(self, session, id):
        sel = "custom_group='%s'" % id
        action = self.app.model.job.hold
        reason = self.get_reason(session, "held")
        for job in Job.select(sel):
            action.invoke(job, reason)
 
    def render_title(self, session):
        ids = self.ids.get(session)
        return len(ids) > 1 and "Hold Job Groups" or "Hold Job Group"

    def render_item_content(self, session, group):
        return group

    def render_reason_text(self, session, *args):
        return "Hold Reason"
    
class JobGroupRelease(JobReasonStringBulkActionForm):
    def process_item(self, session, id):
        sel = "custom_group='%s'" % id
        action = self.app.model.job.release
        reason = self.get_reason(session, "released")
        for job in Job.select(sel):
            action.invoke(job, reason)
 
    def render_title(self, session):
        ids = self.ids.get(session)
        return len(ids) > 1 and "Release Job Groups" or "Release Job Group"

    def render_item_content(self, session, group):
        return group

    def render_reason_text(self, session, *args):
        return "Release Reason"
    
class JobGroupRemove(JobReasonStringBulkActionForm):
    def process_item(self, session, id):
        sel = "custom_group='%s'" % id
        action = self.app.model.job.remove
        reason = self.get_reason(session, "removed")
        for job in Job.select(sel):
            action.invoke(job, reason)
 
    def render_title(self, session):
        ids = self.ids.get(session)
        return len(ids) > 1 and "Remove Job Groups" or "Remove Job Group"

    def render_item_content(self, session, group):
        return group

    def render_reason_text(self, session, *args):
        return "Remove Reason"
    
class JobStatusSwitch(StateSwitch):
    def __init__(self, app, name):
        super(JobStatusSwitch, self).__init__(app, name)

        self.add_state("a", "All")
        self.add_state("r", "Running")
        self.add_state("i", "Idle")
        self.add_state("h", "Held")
        self.add_state("c", "Completed")
        self.add_state("d", "Removed")
        
    def get_sql_constraint(self, session, phase=None):
        if not phase:
            phase = self.get(session)

        alive = "((c.qmf_update_time is null or " + \
            "c.qmf_update_time <= now() - interval '10 minutes')" + \
            " and j.qmf_delete_time is null and j.job_status != %i)" % JobStatusInfo.get_status_int("Removed")

        if phase == "a":
            sql = alive
        elif phase == "r":
            sql = "(j.job_status = %i" % JobStatusInfo.get_status_int("Running") + \
                    " and %s)" % alive
        elif phase == "i":
            sql = "(j.job_status = %i" % JobStatusInfo.get_status_int("Idle") + \
                    " and %s)" % alive
        elif phase == "h":
            sql = "(j.job_status = %i" % JobStatusInfo.get_status_int("Held") + \
                    " and %s)" % alive
        elif phase == "c":
            comotose = "(c.qmf_update_time is null or " + \
                "c.qmf_update_time <= now() - interval '10 minutes')"
            sql = "(j.job_status = %i" % JobStatusInfo.get_status_int("Completed") + \
                    " and %s)" % comotose
        else:
            not_completed = "(j.qmf_delete_time is not null and j.job_status <> %i )" % JobStatusInfo.get_status_int("Completed")
            is_removed = "j.job_status = %i" % JobStatusInfo.get_status_int("Removed")
            sql = " or ".join((not_completed, is_removed)) 

        return sql

class JobStatusInfo(object):
    stat_strings = ["Unexpanded", "Idle", "Running", "Removed", "Completed", "Held", "Submission Error"]
    stat_colors = ["red", "clear", "green", "black", "blue", "yellow", "red"]
    @classmethod
    def get_status_string(cls, stat):
        try:
            return cls.stat_strings[stat]
        except:
            return ""

    @classmethod
    def get_status_color(cls, stat):
        try:
            return cls.stat_colors[stat]
        except:
            return "red"

    @classmethod
    def get_status_int(cls, stat):
        try:
            return cls.stat_strings.index(stat)
        except:
            return -1

    @classmethod
    def get_zipped_colors(cls):
        return zip(cls.stat_strings, cls.stat_colors)
