import logging

from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *
from wooly.tables import *

from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *
from job import *

strings = StringCatalog(__file__)
log = logging.getLogger("cumin.Limits")

class LimitActions(object):
    def fetch_limits(self, session):
        negotiator = self.get_negotiator(session)
        if negotiator:
            limits = self.get_raw_limits(session, negotiator)
            if "timeout" in limits:
                del limits["timeout"]
            return limits
        else:
            return dict()

    def get_negotiator(self, session):
#TODO: find better way to get the negotiator. from pool perhaps?
        pool = self.frame.get_args(session)[0]
        most_recent = None
        negotiators = Negotiator.select("pool='%s'" % pool.id)
        for negotiator in negotiators:
            if negotiator.qmfBrokerId:
                if not most_recent:
                    most_recent = negotiator
                try:
                    if negotiator.statsCurr.qmfUpdateTime > \
                            most_recent.statsCurr.qmfUpdateTime:
                        most_recent = negotiator
                except:
                    pass

        return most_recent

    def set_limit(self, session, limit):
        negotiator = self.get_negotiator(session)
        action = self.app.model.limit.setlimit
        action.invoke(limit, negotiator)
#TODO: this probably shouldn't be called until after the invoke completes
        def completion():
            pass
        negotiator.Reconfig(self.app.model.data, completion)
    
    def get_raw_limits(self, session, negotiator):
        action = self.app.model.negotiator.GetLimits
        return action.do_invoke(negotiator)

class LimitsSet(ItemTable, LimitActions):
    def __init__(self, app, name):
        super(LimitsSet, self).__init__(app, name)

        col = self.NameColumn(app, "name")
        self.add_column(col)
        self.set_default_column(col)

        col = self.CurrentColumn(app, "curr")
        col.alignment = "right"
        self.add_column(col)
        
        col = self.MaxColumn(app, "max")
        col.alignment = "right"
        self.add_column(col)
        
        self.limits = self.Limits(self, "limits")
        self.add_attribute(self.limits)

        self.limit_count = self.LimitCount(app, "limit_count")
        self.add_child(self.limit_count)

    def do_get_items(self, session, *args):
        limits = self.limits.get(session)
        if not len(limits):
            limits = self.fetch_limits(session)
            self.limits.set(session, limits)
        keys = limits.keys()
        keys.sort()
        return [{"name":x, "curr":limits[x]["CURRENT"], "max":limits[x]["MAX"]} for x in keys]

    def render_title(self, session):
        return self.limit_count.get_title(session, "Limits")

    class LimitCount(AjaxField):
        def get_url(self, session):
            negotiator = self.parent.get_negotiator(session)
            return negotiator and \
                "call.xml?class=negotiator;id=%i;method=GetLimitCount" % negotiator.id
        
        def get_title(self, session, title):
            script = self.render_script(session)
            count = script and "?" or "0"
            return "%s <span id=\"%s\"><span class='count'>(%s)</span>%s</span>" % \
                (title, self.name, count, script)

    class NameColumn(ItemTableColumn):
        def render_title(self, session, data):
            return "Name"
            
        def render_content(self, session, data):
            limit = Identifiable(data["name"])
            href = self.frame.limit.get_href(session, limit)
            return fmt_link(href, data["name"])

    class CurrentColumn(ItemTableColumn):
        def render_title(self, session, data):
            return "Current Usage"
            
    class MaxColumn(ItemTableColumn):
        def render_title(self, session, data):
            return "Max Allowance"
            
    class Limits(Attribute):
        def get_default(self, session):
            return dict()

class LimitsFrame(CuminFrame, LimitActions):
    def __init__(self, app, name):
        super(LimitsFrame, self).__init__(app, name)

        self.object = LimitParameter(app, "id")
        self.add_parameter(self.object)

        view = LimitsView(app, "view")
        self.add_mode(view)
        self.set_view_mode(view)

        edit = LimitEdit(app, "edit")
        self.add_mode(edit)
        self.set_edit_mode(edit)
        
    def get_object(self, session):
        obj = self.object.get(session)
        if not getattr(obj, "name", None):
            limits = self.fetch_limits(session)
            obj.name = obj.id
            obj.curr = limits[obj.id]["CURRENT"]
            obj.max = limits[obj.id]["MAX"]
            self.object.set(session, obj)

        return obj
    

class LimitEdit(CuminForm):
    def __init__(self, app, name):
        super(LimitEdit, self).__init__(app, name)

        self.max = Parameter(app, "max")
        self.add_parameter(self.max)

        self.error = self.Errors(self, "error")
        self.add_attribute(self.error)

        self.error_tmpl = Template(self, "error_html")
        
    class Errors(Attribute):
        def get_default(self, session):
            return dict()
    
    def get_args(self, session):
        return self.frame.get_args(session)
    
    def render_title(self, session, *args):
        return "Edit Concurrency Limit '%s'" % args[0].id
    
    def render_input_id(self, session, *args):
        return self.max.path
    
    def render_label(self, session, *args):
        return "Maximum Allowance"
    
    def render_input_name(self, session, *args):
        return self.max.path
    
    def render_input_value(self, session, *args):
        max = self.max.get(session)
        if not max:
            max = args[0].max
        return str(max)
    
    def render_inline_help(self, session, *args):
        return "Set the maximum number of jobs that can run concurrently using this limiter."

    def render_limit_error(self, session, *args):
        error = self.error.get(session)
        if "max" in error:
            writer = Writer()
            self.error_tmpl.render(writer, session, *args)
            return writer.to_string()
        
    def render_error(self, session, *args):
        error = self.error.get(session)
        if "max" in error:
            return error["max"]
        
    def render_original_value(self, session, *args):
        error = self.error.get(session)
        if "max" in error:
            return "Original value was %s" % str(args[0].max)
    
    def process_submit(self, session, *args):
        max = self.max.get(session)
        fmax = 0
        errors = False
        try:
            fmax = float(max)
            if fmax < 1.0 or fmax > 99999.0:
                raise "out of bounds"
        except:
            errors = True
            error = self.error.get(session)
            error["max"] = "Numeric value between 1 and 99999 expected"
            self.error.set(session, error)
            
        if not errors:
            limit = args[0]
            limit.max = fmax
            self.frame.set_limit(session, limit)
            self.process_cancel(session, *args)
    
class LimitsView(CuminView):
    def __init__(self, app, name):
        super(LimitsView, self).__init__(app, name)

        status = LimitsStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        jobs = LimitsJobSet(app, "jobs")
        self.__tabs.add_tab(jobs)
        
        details = CuminDetails(app, "details")
        self.__tabs.add_tab(details)

class LimitsStatus(CuminStatus):
    def render_title(self, session, *args):
        return "Limits Status"

class LimitsJobSet(JobTab):
    def get_visible_columns(self, session):
        return self.get_request_visible_columns(session, 
                                                ["custom_group", 
                                                 "scheduler", 
                                                 "submitter"])

    def get_sql_values(self, session, *args):
        pass

    def render_sql_where(self, session, limit):
        phase_sql = self.get_phase_sql(session)
        limits_sql = self.get_limits_sql(session, limit)
        return "where %s" % " and ".join([phase_sql, limits_sql])

    def render_title(self, session, limit):
        limits_sql = self.get_limits_sql(session, limit)
        return "Jobs %s" % fmt_count(Job.select(limits_sql).count())
    
    def render_count(self, session, *args):
        str = super(LimitsJobSet, self).render_count(session, *args)
        return "%s with Concurrency Limit '%s'" % (str, args[0].id)

    def get_limits_sql(self, session, limit):
        x = limit.id
        return "concurrency_limits similar to '(%s|%s,%%|%%,%s|%%,%s,%%)'" % (x, x, x, x)
