from datetime import datetime, timedelta
from formats import *
from job import *
from parameters import *
from pool import PoolSlotSet, PoolMachineSet, PoolJobStats
from slot import SlotStatSet
from struct import unpack, calcsize
from system import SystemSlotSet
from time import *
from types import *
from util import *
from wooly import *
from wooly.parameters import *
from wooly.widgets import *
from mint.schema import *
import logging


log = logging.getLogger("cumin.model")

class CuminModel(object):
    def __init__(self, app, data_uri):
        self.app = app
        self.data = MintModel(data_uri,
                              dbExpireFrequency=app.config.expire_frequency,
                              dbExpireThreshold=app.config.expire_threshold,
                              debug=False)

        self.data.updateObjects = False
        self.data.dbExpiration = False

        self.classes = list()
        self.invocations = set()

        self.frame = None

        # Messaging
        
        CuminBroker(self)
        CuminQueue(self)
        CuminExchange(self)
        CuminBinding(self)
        CuminConnection(self)
        CuminSession(self)
        CuminLink(self)
        CuminRoute(self)
        CuminBrokerStoreModule(self)
        CuminJournal(self)
        CuminBrokerAclModule(self)
        CuminBrokerClusterModule(self)

        CuminBrokerRegistration(self)
        CuminBrokerGroup(self)

        # Grid

        CuminScheduler(self)
        CuminSubmitter(self)
        CuminJob(self)
        CuminJobGroup(self)
        CuminLimit(self)

        CuminPool(self)
        CuminCollector(self)
        CuminNegotiator(self)
        CuminMaster(self)

        # Systems

        CuminSystem(self)
        CuminSlot(self)

    def check(self):
        self.data.check()

    def init(self):
        self.data.init()

        self.frame = self.app.main_page.main

        for cls in self.classes:
            cls.init()

    def start(self):
        self.data.start()

    def stop(self):
        self.data.stop()
        
    def add_class(self, cls):
        self.classes.append(cls)
        setattr(self, cls.cumin_name, cls)

    def get_ad_groups(self):
        return AdProperty.get_ad_groups()

    def get_class_by_object(self, mint_object):
        for cls in self.classes:
            if cls.mint_class is mint_object.__class__:
                return cls

    def count_invocations(self, status):
        count = 0

        for invoc in self.invocations:
            if invoc.status == status:
                count += 1

        return count

    def write_xml(self, session, writer, objects):
        writer.write("<model>")

        pcount = self.count_invocations("pending")
        ccount = self.count_invocations("OK")
        fcount = len(self.invocations) - pcount - ccount

        writer.write("<invocations pending=\"%i\" " % pcount +
                     "completed=\"%i\" " % ccount +
                     "failed=\"%i\"/>" % fcount)

        for object in objects:
            cls = self.get_class_by_object(object)
            cls.write_xml(session, writer, object)

        writer.write("</model>")

    def show_main(self, session):
        return self.app.main_page.main.show(session)

    def get_main_pool(self):
        for coll in Collector.select():
            return Pool(coll)

class CuminProperty(object):
    def __init__(self, cls, name):
        self.model = cls.model
        self.cumin_class = cls

        self.name = name
        self.title = None
        self.category = "general"
        self.summary = False
        self.escape = True
        self.prefix = None

        self.cumin_class.add_property(self)

    def init(self):
        pass

    def get_title(self, session):
        if self.title:
            return self.title
        else:
            return self.name

    def value(self, session, object):
        value = getattr(object, self.name, None)

        if isinstance(value, datetime):
            value = fmt_datetime(value)

        if isinstance(value, dict):
            value = fmt_dict(value, self.prefix)
            
        return value

class AdProperty(CuminProperty):
    groups = ["Main", "Command Info", "Job Status Info", "Condor Info", "Other"]

    def __init__(self, cls, name):
        # don't call super since we don't want to call add_property
        self.model = cls.model
        self.cumin_class = cls

        self.name = name
        self.title = None
        self.category = "general"
        self.summary = False
        self.escape = True
        self.prefix = None

        self.description = None
        self.example = None
        self.group = "Main"
        self.renderer = None
        self.writable = True

        self.cumin_class.add_ad_property(self)
    
    @classmethod
    def get_ad_groups(cls):
        return cls.groups

class DateAdProperty(AdProperty):
    def __init__(self, cls, name):
        super(DateAdProperty, self).__init__(cls, name)

        self.renderer = self.render_datetime
        self.group = "Other"

    def render_datetime(self, session, value):
        value = datetime.fromtimestamp(int(value))
        return fmt_datetime(value)

class CuminAction(object):
    def __init__(self, cls, name):
        self.model = cls.model
        self.cumin_class = cls

        self.name = name
        self.title = None
        self.summary = False
        self.navigable = True
        self.aggregate = False

        self.cumin_class.add_action(self)

    def init(self):
        pass

    def show(self, session, object):
        raise Exception("Not implemented")

    # XXX we could even cache this if it gets called a lot
    def get_href(self, session, object):
        branch = session.branch()
        self.show(branch, object)
        return branch.marshal()

    def get_title(self, session):
        if self.title:
            return self.title
        else:
            return self.name

    def get_summary(self, session, object):
        pred0 = self.get_title(session)
        pred1 = self.cumin_class.get_object_title(session, object)

        return "%s %s" % (pred0, pred1)

    def get_enabled(self, session, object):
        return True
    
    def get_verb(self, session):
        return None
    
    def invoke(self, object, args={}):
        invoc = CuminActionInvocation(self, object)

        def completion(status, args=None):
            invoc.status = status
            invoc.args = args
            #invoc.prt()

        try:
            try:
                return self.do_invoke(object, args, completion)
            except Exception, e:
                invoc.status = "failed"
                invoc.exception = e

                log.exception(e)
        finally:
            self.model.invocations.add(invoc)

        #invoc.prt()

        return invoc

    def do_invoke(self, object, args, completion):
        pass

    def get_session_by_registration(self, reg):
        assert reg.broker
        assert reg.broker.qmfBrokerId in self.model.data.mintBrokersById

        broker = self.model.data.mintBrokersById[reg.broker.qmfBrokerId]
        return broker.getAmqpSession()

class CuminActionInvocation(object):
    def __init__(self, action, object):
        self.action = action
        self.object = object
        self.when = datetime.now()
        self.status = "pending"
        self.args = None
        self.exception = None

    def get_description(self, session):
        verb = self.action.get_verb(session)

        if not verb:
            verb = self.action.get_title(session)

        if self.object:
            if type(self.object) is list:
                count = len(self.object)
                name = self.action.cumin_class.get_title(session)
                sobject = "%i %s%s" % (count, name, ess(count))
            else:
                cls = self.action.model.get_class_by_object(self.object)
                sobject = cls.get_object_title(session, self.object)
        else:
            sobject = self.action.cumin_class.get_title(session)

        return "%s %s" % (verb, sobject)

    def prt(self):
        print "action", self.action.name, self.object, self.when, \
            self.status, self.args, self.exception

class CuminSetAction(CuminAction):
    def __init__(self, cls, name):
        super(CuminSetAction, self).__init__(cls, name)

        self.aggregate = True

    def get_summary(self, session, object):
        pred0 = self.get_title(session)
        pred1 = self.cumin_class.get_plural_title(session)

        return "%s %s" % (pred0, pred1)

class CuminStat(object):
    def __init__(self, cls, name):
        self.model = cls.model
        self.cumin_class = cls

        self.name = name
        self.unit = None
        self.title = None
        self.category = "general"
        self.summary = False

        self.link_cb = None
        self.highlow = False

        self.cumin_class.add_stat(self)

    def init(self):
        pass

    def samples(self, object, secs):
        stats = object.stats

        now = datetime.now()

        col = self.cumin_class.mint_stats_class.q.qmfUpdateTime
        dt = now - timedelta(seconds=int(secs * 1.5))
        stats = stats.filter(col >= dt)

        stats = stats.orderBy("-qmfUpdateTime")

        samples = list()

        limit = now - timedelta(seconds=secs)
        next_is_last = False

        for stat in stats:
            value = getattr(stat, self.name, 0)
            samples.append((stat.qmfUpdateTime, value))

            if next_is_last:
                break

            if stat.qmfUpdateTime < limit:
                next_is_last = True

        return samples

    def value_html(self, object):
        text = self.value_text(object)

        if text:
            html = text
        else:
            html = fmt_none_brief()

        return html

    def value_text(self, object):
        value = self.value(object)

        if value is None:
            text = ""
        elif type(value) in (int, long):
            if value == 0:
                text = "0"
            else:
                text = "%i" % value
        else:
            text = "%r" % value

        return text

    def value(self, object):
        try:
            return getattr(object.statsCurr, self.name)
        except AttributeError:
            pass

    def rate_html(self, object):
        text = self.rate_text(object)

        if text:
            html = text
        else:
            html = fmt_none_brief()

        return html

    def rate_text(self, object):
        rate = self.rate(object)

        if rate is None:
            text = ""
        elif rate == 0:
            text = "0"
        else:
            text = "%0.2f" % rate

        return text

    def rate(self, object):
        threshold = datetime.now() - timedelta(minutes=10)

        try:
            if object.statsCurr and object.statsCurr.qmfUpdateTime > threshold:
                curr = getattr(object.statsCurr, self.name)
                ctime = object.statsCurr.qmfUpdateTime
                csecs = mktime(ctime.timetuple())

                if object.statsPrev:
                    prev = getattr(object.statsPrev, self.name)
                    ptime = object.statsPrev.qmfUpdateTime
                    psecs = mktime(ptime.timetuple())

                    if curr is not None and prev is not None:
                        return calc_rate(curr, prev, csecs, psecs)
        except AttributeError:
            pass

    def write_xml(self, writer, object):
        value = self.value_text(object)
        rate = self.rate_text(object)

        writer.write("<stat name=\"%s\" value=\"%s\" rate=\"%s\"/>" \
                     % (self.name, value, rate))

class CuminClass(object):
    def __init__(self, model, name, mint_class):
        self.model = model
        self.cumin_name = name
        self.mint_class = mint_class

        self.properties = list()
        self.ad_properties = list()
        self.ad_properties_by_name = dict()
        self.stats = list()
        self.actions = list()

        self.frame = None

        self.model.add_class(self)

    def add_property(self, prop):
        self.properties.append(prop)
        setattr(self, prop.name, prop)

    def add_ad_property(self, prop):
        self.ad_properties.append(prop)
        self.ad_properties_by_name[prop.name] = prop
 
    def add_stat(self, stat):
        self.stats.append(stat)
        setattr(self, stat.name, stat)

    def add_action(self, action):
        self.actions.append(action)
        setattr(self, action.name, action)

    def init(self):
        for prop in self.properties:
            prop.init()

        for ad_prop in self.ad_properties:
            ad_prop.init()

        for stat in self.stats:
            stat.init()

        for action in self.actions:
            action.init()

    def get_title(self, session):
        return "Object"

    def get_plural_title(self, session):
        return self.get_title(session) + "s"

    def get_icon_href(self, session):
        return "resource?name=action-36.png"

    def get_object_href(self, session, object):
        branch = session.branch()
        self.show_object(branch, object).show_view(branch)
        return branch.marshal()

    def get_object_href_by_id(self, session, id):
        return self.get_object_href(session, Identifiable(id))

    def show_object(self, session, object):
        return self.frame.show_object(session, object)

    def get_object_title(self, session, object):
        title = self.get_title(session)
        name = self.get_object_name(object)
        return "%s '%s'" % (title, name)

    def get_object_name(self, object):
        return object.name

    def write_event_xml(self, writer, object):
        writer.write("<events errors=\"%i\" warnings=\"%i\"/>" % (0, 0))

    def write_stat_xml(self, session, writer, object):
        for stat in self.stats:
            stat.write_xml(writer, object)

    def write_xml(self, session, writer, object):
        object.sync()

        writer.write("<%s id=\"%s\" name=\"%s\">" % \
                         (self.cumin_name, str(object.id),
                          self.get_object_name(object)))

        self.write_event_xml(writer, object)
        self.write_stat_xml(session, writer, object)
            
        writer.write("</%s>" % self.cumin_name)

class RemoteClass(CuminClass):
    def __init__(self, model, name, mint_class, mint_stats_class):
        super(RemoteClass, self).__init__(model, name, mint_class)

        self.mint_stats_class = mint_stats_class

        prop = CuminProperty(self, "qmfObjectId")
        prop.title = "Management ID"

        prop = CuminProperty(self, "qmfCreateTime")
        prop.title = "Creation Time"

        prop = CuminProperty(self, "qmfDeleteTime")
        prop.title = "Deletion Time"

        prop = CuminProperty(self, "qmfUpdateTime")
        prop.title = "Last Updated"
        prop.summary = True

class CuminSlot(RemoteClass):
    def __init__(self, model):
        super(CuminSlot, self).__init__(model, "slot", Slot, SlotStats)
        
        prop = CuminProperty(self, "Name")
        prop.title = "Name"
        
        prop = CuminProperty(self, "Machine")
        prop.title = "Machine"
        
        prop = CuminProperty(self, "JobId")
        prop.title = "Current Job Id"
        
        stat = CuminStat(self, "Activity")
        stat.title = "Activity"
        stat.category = "general"

        stat = CuminStat(self, "CondorLoadAvg")
        stat.title = "Condor Load Avg"
        stat.category = "general"

        stat = CuminStat(self, "LoadAvg")
        stat.title = "Load Avg"
        stat.category = "general"

class Visualization(CuminAction):
    def __init__(self, cls, name):
        super(Visualization, self).__init__(cls, name)

        self.itemset = None
        
    def get_xml_response(self, session, object, *args):
        boxes = self.get_boxes(session, object, *args)
        fields = self.get_field_tuples(session)

        writer = Writer()
        writer.write("<%ss>" % self.itemset.name)
        for box in boxes:
            writer.write("<%s id='%s'" % (self.itemset.name, str(box["id"])))
            for field, desc in fields:
                writer.write(" %s='%s'" % (field, box[field]))
            if fields:
                extra = self.get_extra(session, box)
                for name, value in extra:
                    writer.write(" %s='%s'" % (name, value))
                writer.write(" color='%s'/>" % self.get_color(session, box))
        writer.write("</%ss>" %self.itemset.name)
        return writer.to_string()

    def get_boxes(self, session, object, *args):
        if args:
            self.itemset.set_args(session, *args)
        box_list = self.itemset.get_items(session, object)
        return box_list

    def get_color(self, session, box):
        return ""  

    def get_extra(self, session, box):
        return list()
        
    def get_colors(self):
        return list()
    
class SlotVisualization(Visualization):
    # list of status/colors in the order we want them displayed
    # in the legend
    load_colors = [("Idle", "clear"), 
                   ("Busy", "green"),
                   ("Suspended", "red"),
                   ("Vacating", "orange"),
                   ("Killing", "blue"),
                   ("Benchmarking", "yellow"),
                   ("Unknown", "grey")]

    load_states = [("Unclaimed", "Unclaimed"), 
                   ("Claimed", "Claimed"),
                   ("Owner", "Owner"),
                   ("Matched", "Matched"),
                   ("Preempting", "Preempting")]

    def get_field_tuples(self, session):
        return [("name", "Name"), ("machine", "Machine"), ("job_id", "Job"), ("state", "State"), ("activity", "Activity")]

    def get_extra(self, session, slot):
        return [("jid", slot["jid"])]

    def get_color(self, session, slot):
        activity = slot["activity"]
        for status, color in self.load_colors:
            if status == activity:
                return color
        return "grey"  

    def get_colors(self):
        return self.load_colors
    
    def get_states(self):
        return self.load_states
    
class CuminSystem(RemoteClass):
    def __init__(self, model):
        super(CuminSystem, self).__init__(model, "system", Sysimage, SysimageStats)

        prop = CuminProperty(self, "uuid")
        prop.title = "System ID"
        
        prop = CuminProperty(self, "nodeName")
        prop.title = "Address"
        prop.summary = True

        prop = CuminProperty(self, "osName")
        prop.title = "Kernel"

        prop = CuminProperty(self, "release")
        prop.title = "Kernel Release"

        prop = CuminProperty(self, "version")
        prop.title = "Kernel Version"

        prop = CuminProperty(self, "machine")
        prop.title = "Architecture"

        prop = CuminProperty(self, "distro")
        prop.title = "Distribution"

        stat = CuminStat(self, "memFree")
        stat.title = "Memory Free"
        stat.highlow = True

        stat = CuminStat(self, "swapFree")
        stat.title = "Swap Free"

        stat = self.LoadStat(self, "loadAverage1Min")
        stat.title = "1 Minute Load Average"
        stat.highlow = True

        stat = self.LoadStat(self, "loadAverage5Min")
        stat.title = "5 Minute Load Average"

        stat = self.LoadStat(self, "loadAverage10Min")
        stat.title = "10 Minute Load Average"

        stat = CuminStat(self, "procTotal")
        stat.title = "Total processes"

        stat = CuminStat(self, "procRunning")
        stat.title = "Running processes"

        action = self.SystemSlotVisualization(self, "slots")
        action.navigable = False

        #action = CuminAction(self, "ping")
        #action.title = "Send Ping"
        #action.summary = True

    def init(self):
        self.frame = self.model.frame.system

    def get_title(self, session):
        return "System"

    def get_icon_href(self, session):
        return "resource?name=system-36.png"

    def get_object_name(self, object):
        return object.nodeName

    class LoadStat(CuminStat):
        def value_text(self, object):
            value = self.value(object)
            text = value and "%0.2f" % value or ""
            return text

    class SystemSlotVisualization(SlotVisualization):
        def __init__(self, cls, name):
            super(CuminSystem.SystemSlotVisualization, self).__init__(cls, name)
            
            self.itemset = self.ModelSystemSlotSet(cls.model.app, "slot")
            self.itemset.items.path = "CuminSystem.SystemSlotVisualization.slot"
                        
        def get_slot_count(self, session, system):
            items = self.itemset.get_items(session, system)
            return len(items)
            
        class ModelSystemSlotSet(SystemSlotSet):
            def render_sql_limit(self, session, *args):
                pass
            
            def render_sql_orderby(self, session, *args):
                return "order by machine, name asc"
        
class CuminMaster(RemoteClass):
    def __init__(self, model):
        super(CuminMaster, self).__init__(model, "master", Master, MasterStats)

        self.Start(self, "start")
        self.Stop(self, "stop")

    def get_title(self, session):
        return "System"

    def get_icon_href(self, session):
        return "resource?name=system-36.png"

    def show_object(self, session, system):
        return self.model.frame.system.show_object(session, system)

    def get_object_name(self, object):
        return object.nodeName

    class Start(CuminAction):
        def do_invoke(self, object, args, completion):
            system_name = object.System
            try:
                master = Master.select("System = '%s'" % system_name)[0]
            except IndexError:
                raise Exception("Master daemon not running")
            master.Start(self.model.data, completion, args["subsystem"])
        
    class Stop(CuminAction):
        def do_invoke(self, object, args, completion):
            system_name = object.System
            try:
                master = Master.select("System = '%s'" % system_name)[0]
            except IndexError:
                raise Exception("Master daemon not running")
            #master.Stop(self.model.data, completion, args["subsystem"])
        
class CuminBroker(RemoteClass):
    def __init__(self, model):
        super(CuminBroker, self).__init__(model, "broker", Broker, BrokerStats)

        prop = CuminProperty(self, "port")
        prop.title = "Port"

        prop = CuminProperty(self, "workerThreads")
        prop.title = "Worker Threads"

        prop = CuminProperty(self, "maxConns")
        prop.title = "Max Connections"

        prop = CuminProperty(self, "connBacklog")
        prop.title = "Connection Backlog"

        prop = CuminProperty(self, "stagingThreshold")
        prop.title = "Staging Threshold"

        prop = CuminProperty(self, "mgmtPubInterval")
        prop.title = "Mgmt. Publish Interval"

        prop = CuminProperty(self, "clusterName")
        prop.title = "Cluster Name"

        prop = CuminProperty(self, "version")
        prop.title = "Version"

        prop = CuminProperty(self, "dataDir")
        prop.title = "Data Directory"
        
        stat = self.StatusStat(self, "connection")
        stat.category = "status"

        action = self.AddExchange(self, "add_exchange")
        action.summary = True

        action = self.AddLink(self, "add_link")
        action.summary = True

        action = self.AddQueue(self, "add_queue")
        action.summary = True

    def init(self):
        self.frame = self.model.frame.broker

    def get_icon_href(self, session):
        return "resource?name=broker-36.png"

    def show_object(self, session, broker):
        reg = broker.registration
        return super(CuminBroker, self).show_object(session, reg)

    def get_object_name(self, broker):
        return broker.id

    class StatusStat(CuminStat):
        def value_text(self, broker):
            connected = False
            if broker:
                try:
                    mbroker = self.model.data.mintBrokersById[broker.qmfBrokerId]
                    connected = mbroker.connected
                except KeyError:
                    pass
                    
            if connected:
                return "Connected"
            else:
                return "Disconnected"

        def rate_text(self, record):
            return ""

    class AddExchange(CuminAction):
        MSG_SEQUENCE = "qpid.msg_sequence"
        IVE = "qpid.ive"

        def get_title(self, session):
            return "Add Exchange"

        def get_verb(self, session):
            return "Add"
        
        def show(self, session, reg):
            frame = self.cumin_class.show_object(session, reg)
            frame.set_object(session, reg)
            return frame.exchange_add.show(session)
        
        def do_invoke(self, exchange, args, completion):

            reg = args["reg"]
            declArgs = {}
            if args["sequence"]:
                declArgs[self.MSG_SEQUENCE] = 1
            if args["ive"]:
                declArgs[self.IVE] = 1
            
            session = self.get_session_by_registration(reg)
            session.exchange_declare(exchange=exchange.name, 
                                     type=exchange.type, 
                                     durable=exchange.durable, 
                                     arguments=declArgs)
            # if the above call fails, an exception is
            # raised and we won't get here
            completion("OK")

    class AddLink(CuminAction):
        def get_title(self, session):
            return "Add Broker Link"

        def get_verb(self, session):
            return "Add"
        
        def show(self, session, reg):
            frame = self.cumin_class.show_object(session, reg)
            frame.set_object(session, reg)
            return frame.link_add.show(session)
        
        def do_invoke(self, link, args, completion):
            reg = args["reg"]
            username = args["username"]
            password = args["password"]
            transport = args["transport"]
            if username == "anonymous":
                authMechanism = "ANONYMOUS"
            else:
                authMechanism = "PLAIN"

            broker = reg._get_broker()
            broker.connect(self.model.data, completion, 
                        link.host, link.port, link.durable,
                        authMechanism, username, password, transport)
            
    class AddQueue(CuminAction):
        FILECOUNT = "qpid.file_count"
        FILESIZE  = "qpid.file_size"
        MAX_QUEUE_SIZE  = "qpid.max_size"
        MAX_QUEUE_COUNT  = "qpid.max_count"
        POLICY_TYPE  = "qpid.policy_type"
        CLUSTER_DURABLE = "qpid.persist_last_node"
        LVQ = "qpid.last_value_queue"
        OPTIMISTIC_CONSUME = "qpid.optimistic_consume"

        def get_title(self, session):
            return "Add Queue"
        
        def get_verb(self, session):
            return "Add"
        
        def show(self, session, reg):
            frame = self.cumin_class.show_object(session, reg)
            frame.set_object(session, reg)
            return frame.queue_add.show(session)
        
        def do_invoke(self, queue, args, completion):

            reg = args["reg"]
            declArgs = {}
            if queue.durable:
                if args["file_count"]:
                    declArgs[self.FILECOUNT] = args["file_count"]
                if args["file_size"]:
                    declArgs[self.FILESIZE] = args["file_size"]

            if args["policy"]:
                declArgs[self.POLICY_TYPE] = args["policy"]
                if args["q_size"]:
                    declArgs[self.MAX_QUEUE_SIZE] = args["q_size"]
                if args["q_count"]:
                    declArgs[self.MAX_QUEUE_COUNT] = args["q_count"]

            if args["cluster_durable"]:
                declArgs[self.CLUSTER_DURABLE] = 1
            if args["lvq"]:
                declArgs[self.LVQ] = 1
            if args["optimistic"]:
                declArgs[self.OPTIMISTIC_CONSUME] = 1
                
            session = self.get_session_by_registration(reg)
            session.queue_declare(queue=queue.name, 
                        durable=queue.durable, 
                        arguments=declArgs)

            # optionally bind to exchanges
            binding_info = args['exchange_keys']
            do_bind(session, queue, binding_info)
            completion("OK")

def do_bind(session, queue, binding_info):
        for exchange in binding_info:
            if "key" in binding_info[exchange]:
                binding_key = binding_info[exchange]["key"] 
            else:
                binding_key = None

            session.exchange_bind(queue=queue.name,
                exchange=binding_info[exchange]["name"],
                binding_key=binding_key,
                arguments=binding_info[exchange]["arguments"])

class CuminQueue(RemoteClass):
    def __init__(self, model):
        super(CuminQueue, self).__init__(model, "queue", Queue, QueueStats)
        
        prop = CuminProperty(self, "name")
        prop.title = "Name"

        prop = CuminProperty(self, "durable")
        prop.title = "Durable?"
        prop.summary = True

        prop = CuminProperty(self, "autoDelete")
        prop.title = "Auto Delete?"
        prop.summary = True

        prop = CuminProperty(self, "exclusive")
        prop.title = "Exclusive?"
        prop.summary = True

        prop = CuminProperty(self, "arguments")
        prop.title = "Arguments"
        prop.prefix = "qpid."
        prop.escape = False

        stat = CuminStat(self, "consumerCount")
        stat.title = "Consumers"
        stat.unit = "consumer"
        stat.highlow = True

        stat = CuminStat(self, "bindingCount")
        stat.title = "Bindings"
        stat.unit = "binding"
        stat.highlow = True

        stat = CuminStat(self, "msgDepth")
        stat.title = "Message Depth"
        stat.highlow = True

        stat = CuminStat(self, "msgTotalEnqueues")
        stat.title = "Msgs. Enqueued"
        stat.unit = "message"
        stat.category = "io"

        stat = CuminStat(self, "msgTotalDequeues")
        stat.title = "Msgs. Dequeued"
        stat.unit = "message"
        stat.category = "io"

        stat = CuminStat(self, "byteDepth")
        stat.title = "Byte Depth"
        stat.unit = "byte"
        stat.highlow = True

        stat = CuminStat(self, "byteTotalEnqueues")
        stat.title = "Bytes Enqueued"
        stat.unit = "byte"
        stat.category = "io"

        stat = CuminStat(self, "byteTotalDequeues")
        stat.title = "Bytes Dequeued"
        stat.unit = "byte"
        stat.category = "io"

        stat = CuminStat(self, "unackedMessages")
        stat.title = "Msgs. Unacked"
        stat.unit = "message"
        stat.category = "io"
        stat.highlow = True

        stat = CuminStat(self, "messageLatencyMin")
        stat.title = "Min. Msg. Latency"
        stat.unit = "nanosecond"
        stat.category = "io"

        stat = CuminStat(self, "messageLatencyMax")
        stat.title = "Max. Msg. Latency"
        stat.unit = "nanosecond"
        stat.category = "io"

        stat = CuminStat(self, "messageLatencyAverage")
        stat.title = "Avg. Msg. Latency"
        stat.unit = "nanosecond"
        stat.category = "io"

        # Transactional

        stat = CuminStat(self, "msgTxnEnqueues")
        stat.title = "Msgs. Enqueued"
        stat.unit = "message"
        stat.category = "io.txn"

        stat = CuminStat(self, "msgTxnDequeues")
        stat.title = "Msgs. Dequeued"
        stat.unit = "message"
        stat.category = "io.txn"

        stat = CuminStat(self, "byteTxnEnqueues")
        stat.title = "Bytes Enqueued"
        stat.unit = "byte"
        stat.category = "io.txn"

        stat = CuminStat(self, "byteTxnDequeues")
        stat.title = "Bytes Dequeued"
        stat.unit = "byte"
        stat.category = "io.txn"

        stat = CuminStat(self, "enqueueTxnStarts")
        stat.title = "Enq. Trans. Started"
        stat.unit = "transaction"
        stat.category = "txn"

        stat = CuminStat(self, "enqueueTxnCommits")
        stat.title = "Enq. Trans. Committed"
        stat.unit = "transaction"
        stat.category = "txn"

        stat = CuminStat(self, "enqueueTxnRejects")
        stat.title = "Enq. Trans. Rejected"
        stat.unit = "transaction"
        stat.category = "txn"

        stat = CuminStat(self, "enqueueTxnCount")
        stat.title = "Enq. Trans. Pending"
        stat.unit = "transaction"
        stat.category = "txn"
        stat.highlow = True

        stat = CuminStat(self, "dequeueTxnStarts")
        stat.title = "Deq. Trans. Started"
        stat.unit = "transaction"
        stat.category = "txn"

        stat = CuminStat(self, "dequeueTxnCommits")
        stat.title = "Deq. Trans. Committed"
        stat.unit = "transaction"
        stat.category = "txn"

        stat = CuminStat(self, "dequeueTxnRejects")
        stat.title = "Deq. Trans. Rejected"
        stat.unit = "transaction"
        stat.category = "txn"

        stat = CuminStat(self, "dequeueTxnCount")
        stat.title = "Deq. Trans. Pending"
        stat.unit = "transaction"
        stat.category = "txn"
        stat.highlow = True

        # Durable

        stat = CuminStat(self, "msgPersistEnqueues")
        stat.title = "Msgs. Enqueued"
        stat.unit = "message"
        stat.category = "io.durable"

        stat = CuminStat(self, "msgPersistDequeues")
        stat.title = "Msgs. Dequeued"
        stat.unit = "message"
        stat.category = "io.durable"

        stat = CuminStat(self, "bytePersistEnqueues")
        stat.title = "Bytes Enqueued"
        stat.unit = "byte"
        stat.category = "io.durable"

        stat = CuminStat(self, "bytePersistDequeues")
        stat.title = "Bytes Dequeued"
        stat.unit = "byte"
        stat.category = "io.durable"

        action = self.Purge(self, "purge")
        action.summary = True
        
        action = self.Remove(self, "remove")
        action.summary = True
        
        action = self.Bind(self, "bind")
        action.summary = True

    def init(self):
        self.frame = self.model.frame.broker.queue

        action = self.MoveMessages(self, "move_messages")
        action.summary = True

    def get_title(self, session):
        return "Queue"

    def get_icon_href(self, session):
        return "resource?name=queue-36.png"

    class Purge(CuminAction):
        def show(self, session, queue):
            frame = self.model.frame.broker
            frame.queue.set_object(session, queue)
            return frame.queue.purge.show(session)

        def get_title(self, session):
            return "Purge"

        def do_invoke(self, queue, args, completion):
            queue.purge(self.model.data, completion, args["request"])

    class Remove(CuminAction):
        def show(self, session, queue):
            frame = self.model.frame.broker
            frame.queue.set_object(session, queue)
            return frame.queue.remove.show(session)

        def get_title(self, session):
            return "Remove"

        def do_invoke(self, queue, args, completion):
            reg = queue.vhost.broker.registration
            session = self.get_session_by_registration(reg)
            session.queue_delete(queue=queue.name)
                
            completion("OK")

    class Bind(CuminAction):
        def show(self, session, queue):
            frame = self.model.frame.broker
            frame.queue.set_object(session, queue)
            return frame.queue.binding_add.show(session)

        def get_title(self, session):
            return "Bind"

        def do_invoke(self, queue, args, completion):
            reg = queue.vhost.broker.registration
            session = self.get_session_by_registration(reg)
            binding_info = args['exchange_keys']
            do_bind(session, queue, binding_info)

            completion("OK")

    class MoveMessages(CuminAction):
        def show(self, session, queue):
            frame = self.cumin_class.show_object(session, queue)
            return frame.move_messages.show(session)

        def get_title(self, session):
            return "Move Messages"

        def do_invoke(self, queue, args, completion):
            broker = queue.vhost.broker
            broker.queueMoveMessages(self.model.data, completion, args["srcQueue"], args["destQueue"], args["qty"])

class CuminExchange(RemoteClass):
    def __init__(self, model):
        super(CuminExchange, self).__init__(model, "exchange",
                                            Exchange, ExchangeStats)
        
        prop = CuminProperty(self, "name")
        prop.title = "Name"

        prop = CuminProperty(self, "type")
        prop.title = "Type"
        prop.summary = True

        prop = CuminProperty(self, "durable")
        prop.title = "Durable?"
        prop.summary = True

        prop = CuminProperty(self, "arguments")
        prop.title = "Arguments"
        prop.prefix = "qpid."
        prop.escape = False

        stat = CuminStat(self, "producerCount")
        stat.title = "Producers"
        stat.unit = "producer"
        stat.highlow = True

        stat = CuminStat(self, "bindingCount")
        stat.title = "Bindings"
        stat.unit = "binding"
        stat.highlow = True

        stat = CuminStat(self, "msgReceives")
        stat.title = "Msgs. Received"
        stat.unit = "message"
        stat.category = "io"

        stat = CuminStat(self, "msgRoutes")
        stat.title = "Msgs. Routed"
        stat.unit = "message"
        stat.category = "io"

        stat = CuminStat(self, "msgDrops")
        stat.title = "Msgs. Dropped"
        stat.unit = "message"
        stat.category = "io"

        stat = CuminStat(self, "byteReceives")
        stat.title = "Bytes Received"
        stat.unit = "message"
        stat.category = "io"

        stat = CuminStat(self, "byteRoutes")
        stat.title = "Bytes Routed"
        stat.unit = "message"
        stat.category = "io"

        stat = CuminStat(self, "byteDrops")
        stat.title = "Bytes Dropped"
        stat.unit = "message"
        stat.category = "io"

        action = self.Remove(self, "remove")
        action.summary = True

    def init(self):
        self.frame = self.model.frame.broker.exchange

    def get_title(self, session):
        return "Exchange"
    
    def get_icon_href(self, session):
        return "resource?name=exchange-36.png"

    def get_object_title(self, session, object):
        name = self.get_object_name(object)

        if name:
            return super(CuminExchange, self).get_object_title(session, object)
        else:
            return "Default Exchange"

    class Remove(CuminAction):
        def get_title(self, session):
            return "Remove"
        
        def show(self, session, exchange):
            frame = self.cumin_class.show_object(session, exchange)
            return frame.remove.show(session)
        
        def do_invoke(self, exchange, reg, completion):
            session = self.get_session_by_registration(reg)
            session.exchange_delete(exchange=exchange.name)
            
            completion("OK")

class CuminBinding(RemoteClass):
    def __init__(self, model):
        super(CuminBinding, self).__init__(model, "binding",
                                           Binding, BindingStats)

        prop = CuminProperty(self, "bindingKey")
        prop.title = "Binding Key"

        stat = CuminStat(self, "msgMatched")
        stat.title = "Msgs. Matched"
        stat.unit = "message"

        action = self.Remove(self, "remove")
        action.summary = True

    def get_title(self, session):
        return "Binding"
    
    def get_object_name(self, binding):
        return "between %s and %s" % (binding.exchange.name, binding.queue.name)

    class Remove(CuminAction):
        def get_title(self, session):
            return "Remove"
        
        def show(self, session, binding):
            raise Exception("XXX")

            frame = self.cumin_class.show_object(session, binding)
            frame = frame.exchange.show_object(session, binding)
            return frame.remove.show(session)
        
        def do_invoke(self, binding, args, completion):
            reg = binding.exchange.vhost.broker.registration
            session = self.get_session_by_registration(reg)
            session.exchange_unbind(queue=binding.queue.name, 
                                    exchange=binding.exchange.name, 
                                    binding_key=binding.bindingKey)
 
            completion("OK")


class CuminRoute(RemoteClass):
    def __init__(self, model):
        super(CuminRoute, self).__init__(model, "route",
                                           Bridge, BridgeStats)

        prop = CuminProperty(self, "key")
        prop.title = "Route Key"
        
        prop = CuminProperty(self, "tag")
        prop.title = "Tag"
        
        prop = CuminProperty(self, "excludes")
        prop.title = "Excludes"
         
        prop = CuminProperty(self, "durable")
        prop.title = "Durable"
        
        prop = CuminProperty(self, "dest")
        prop.title = "Exchange"
         
        action = self.Remove(self, "remove")
        #action.summary = True

    def get_title(self, session):
        return "Route"
    
    def get_object_name(self, route):
        return route.src
    
    class Remove(CuminAction):
        def get_title(self, session):
            return "Remove"
        
        def show(self, session, peer):
            frame = self.cumin_class.show_object(session, peer)
            frame = frame.show_peer(session, peer)
            return frame.show_remove(session)
        
        def do_invoke(self, bridge, args, completion):
            bridge.close(self.model.data, completion)
    
class CuminConnection(RemoteClass):
    def __init__(self, model):
        super(CuminConnection, self).__init__(model, "connection",
                                              ClientConnection,
                                              ClientConnectionStats)

        prop = CuminProperty(self, "address")
        prop.title = "Address"

        prop = CuminProperty(self, "SystemConnection")        
        prop.title = "System Connection?"                     

        prop = CuminProperty(self, "authIdentity")
        prop.title = "Auth Identity?"

        prop = CuminProperty(self, "federationLink")
        prop.title = "Federation Link"
                  
        stat = CuminStat(self, "closing")
        stat.title = "Closing Down"
        stat.category = "general"

        stat = CuminStat(self, "bytesFromClient")
        stat.title = "Bytes Sent"
        stat.unit = "byte"
        stat.category = "io"

        stat = CuminStat(self, "bytesToClient")
        stat.title = "Bytes Received"
        stat.unit = "byte"
        stat.category = "io"

        stat = CuminStat(self, "framesFromClient")
        stat.title = "Frames Sent"
        stat.unit = "frame"
        stat.category = "io"

        stat = CuminStat(self, "framesToClient")
        stat.title = "Frames Received"
        stat.unit = "frame"
        stat.category = "io"

        action = self.Close(self, "close")
        action.summary = True

    def init(self):
        self.frame = self.model.frame.broker.connection

    def get_title(self, session):
        return "Connection"
    
    def get_icon_href(self, session):
        return "resource?name=client-36.png"

    def get_object_name(self, conn):
        return conn.address

    class Close(CuminAction):
        def show(self, session, conn):
            frame = self.cumin_class.show_object(session, conn)
            return frame.show_close(session)

        def get_title(self, session):
            return "Close"

        def do_invoke(self, conn, args, completion):
            session_ids = set()

            for broker in self.model.data.mintBrokersByQmfBroker:
                session_ids.add(broker.getSessionId())

            for sess in conn.sessions:
                if sess.name in session_ids:
                    raise Exception \
                        ("Cannot close management connection %s" % \
                             conn.address)

            conn.close(self.model.data, completion)

class CuminSession(RemoteClass):
    def __init__(self, model):
        super(CuminSession, self).__init__(model, "session",
                                           Session, SessionStats)

        prop = CuminProperty(self, "name")
        prop.title = "Name"

        prop = CuminProperty(self, "channelId")
        prop.title = "Channel ID"

        prop = CuminProperty(self, "detachedLifespan")
        prop.title = "Detached Lifespan"

        stat = CuminStat(self, "expireTime")
        stat.title = "Expiration"
        stat.category = "general"

        stat = CuminStat(self, "framesOutstanding")
        stat.title = "Frames Outstanding"
        stat.unit = "frame"
        stat.category = "general"

        stat = CuminStat(self, "attached")
        stat.title = "Attached"
        stat.category = "general"

        self.Close(self, "close")
        self.Detach(self, "detach")
        self.ResetLifespan(self, "resetLifespan")
        self.SolicitAck(self, "solicitAck")

    def get_title(self, session):
        return "Session"
    
    class Close(CuminAction):
        def get_title(self, session):
            return "Close"

        def do_invoke(self, sess, args, completion):
            for broker in self.model.data.mintBrokersByQmfBroker:
                if sess.name == broker.getSessionId():
                    raise Exception \
                        ("Cannot close management session %s" % sess.name)

            sess.close(self.model.data, completion)

    class Detach(CuminAction):
        def get_title(self, session):
            return "Detach"

        def do_invoke(self, sess, args, completion):
            for broker in self.model.data.mintBrokersByQmfBroker:
                if sess.name == broker.getSessionId():
                    raise Exception \
                        ("Cannot detach management session %s" % sess.name)

            sess.detach(self.model.data, completion)

    class ResetLifespan(CuminAction):
        def get_title(self, session):
            return "Reset Lifespan"

        def do_invoke(self, object, args, completion):
            object.resetLifespan(self.model.data, completion)

    class SolicitAck(CuminAction):
        def get_title(self, session):
            return "Solicit Acknowledgment"

        def do_invoke(self, object, args, completion):
            object.solicitAck(self.model.data, completion)

class CuminLink(RemoteClass):
    def __init__(self, model):
        super(CuminLink, self).__init__(model, "link", Link, LinkStats)

        prop = CuminProperty(self, "useSsl")
        prop.title = "Use SSL"
        prop.summary = True

        prop = CuminProperty(self, "durable")
        prop.title = "Durable"
        prop.summary = True

        stat = CuminStat(self, "closing")
        stat.title = "Closing Down"

        stat = CuminStat(self, "framesFromPeer")
        stat.title = "Frames from Peer"
        stat.unit = "frame"
        stat.category = "io"

        stat = CuminStat(self, "framesToPeer")
        stat.title = "Frames to Peer"
        stat.unit = "frame"
        stat.category = "io"

        stat = CuminStat(self, "bytesFromPeer")
        stat.title = "Bytes from Peer"
        stat.unit = "byte"
        stat.category = "io"

        stat = CuminStat(self, "bytesToPeer")
        stat.title = "Bytes to Peer"
        stat.unit = "byte"
        stat.category = "io"

        action = self.Close(self, "close")
        action.summary = True
        
        action = self.Bridge(self, "bridge")
        action.summary = True

    def init(self):
        self.frame = self.model.frame.broker.link
        
    def get_title(self, session):
        return "Broker Link"

    def get_object_name(self, link):
        return "%s:%d" % (link.host, link.port)

    class Bridge(CuminAction):
        def show(self, session, link):
            frame = self.cumin_class.show_object(session, link)
            frame.set_object(session, link)
            return frame.show_bridge_add(session)
        
        def get_title(self, session):
            return "Add Route"
        
        def get_verb(self, session):
            return "Add Route"
        
        def do_invoke(self, link, args, completion):
            durable = args["durable"]
            src = args["exchange"]
            dest = args["exchange"] # see qpid-route
            key = args["key"]
            tag = args["tag"]
            dynamic = args["dynamic"]
            excludes = args["excludes"]
            sync = args["sync"]
            
            link.bridge(self.model.data, completion, 
                        durable, src, dest, key, 
                        tag, excludes, False, False, dynamic, sync)

    class Close(CuminAction):
        def show(self, session, link):
            frame = self.cumin_class.show_object(session, link)
            return frame.show_remove(session)
        
        def get_title(self, session):
            return "Close"

        def do_invoke(self, link, args, completion):
            link.close(self.model.data, completion)

class CuminBrokerStoreModule(RemoteClass):
    def __init__(self, model):
        super(CuminBrokerStoreModule, self).__init__ \
            (model, "store", Store, StoreStats)

        prop = CuminProperty(self, "location")
        prop.title = "Location"

class CuminJournal(RemoteClass):
    def __init__(self, model):
        super(CuminJournal, self).__init__(model, "journal",
                                           Journal, JournalStats)

        prop = CuminProperty(self, "name")
        prop.title = "Name"
        prop.summary = True

        prop = CuminProperty(self, "directory")
        prop.title = "Directory"

        stat = CuminStat(self, "initialFileCount")
        stat.title = "Initial File Count"
        stat.unit = "file"
        stat.category = "io.journal"

        stat = CuminStat(self, "dataFileSize")
        stat.title = "Data File Size"
        stat.unit = "byte"
        stat.category = "io.journal"

        stat = CuminStat(self, "recordDepth")
        stat.title = "Record Depth"
        stat.unit = "record"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "recordEnqueues")
        stat.title = "Record Enqueues"
        stat.unit = "record"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "outstandingAIOs")
        stat.title = "Outstanding AIOs"
        stat.unit = "aio"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "freeFileCount")
        stat.title = "Free Files"
        stat.unit = "file"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "availableFileCount")
        stat.title = "Avail. Files"
        stat.unit = "file"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "writeWaitFailures")
        stat.title = "Write Wait Failures"
        stat.unit = "failure"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "writeBusyFailures")
        stat.title = "Write Busy Failures"
        stat.unit = "failure"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "readRecordCount")
        stat.title = "Read Records"
        stat.unit = "record"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "readBusyFailures")
        stat.title = "Read Busy Failures"
        stat.unit = "failure"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "writePageCacheDepth")
        stat.title = "Write Page Cache Depth"
        stat.unit = "page"
        stat.category = "io.journal"
        
        stat = CuminStat(self, "readPageCacheDepth")
        stat.title = "Read Page Cache Depth"
        stat.unit = "page"
        stat.category = "io.journal"
        
class CuminBrokerAclModule(RemoteClass):
    def __init__(self, model):
        super(CuminBrokerAclModule, self).__init__ \
            (model, "store", Acl, AclStats)

        prop = CuminProperty(self, "policyFile")
        prop.title = "Policy File"

        prop = CuminProperty(self, "enforcingAcl")
        prop.title = "Enforcing?"

        prop = CuminProperty(self, "transferAcl")
        prop.title = "Transfer?"

        prop = CuminProperty(self, "lastAclLoad")
        prop.title = "Last Loaded"

        stat = CuminStat(self, "aclDenyCount")
        stat.title = "Denials"
        stat.unit = "denial"

class CuminBrokerClusterModule(RemoteClass):
    def __init__(self, model):
        super(CuminBrokerClusterModule, self).__init__ \
            (model, "broker_cluster_module", Cluster, ClusterStats)

        prop = CuminProperty(self, "clusterName")
        prop.title = "Name"

        prop = CuminProperty(self, "clusterID")
        prop.title = "ID"

        prop = CuminProperty(self, "publishedURL")
        prop.title = "URL"

        prop = CuminProperty(self, "clusterSize")
        prop.title = "Size"

        prop = CuminProperty(self, "status")
        prop.title = "Status"

        prop = CuminProperty(self, "members")
        prop.title = "Members"

class CuminBrokerRegistration(CuminClass):
    def __init__(self, model):
        super(CuminBrokerRegistration, self).__init__ \
            (model, "broker_registration", BrokerRegistration)

        prop = CuminProperty(self, "url")
        prop.title = "URL"
        prop.summary = True

        action = self.Add(self, "add")
        action.title = "Add"
        action.navigable = False

        action = self.Edit(self, "edit")
        action.title = "Edit"
        action.summary = True

        action = self.Remove(self, "remove")
        action.title = "Unregister"
        action.summary = True

    def init(self):
        self.frame = self.model.frame.broker

    def get_title(self, session):
        return "Broker"

    def get_icon_href(self, session):
        return "resource?name=broker-36.png"

    class Add(CuminAction):
        def do_invoke(self, object, args, completion):
            assert object is None

            try:
                object = self.cumin_class.mint_class(**args)

                completion("OK")

                return object
            except Exception, e:
                log.exception("Action failed")
                completion(e.message or "failed")

    class Edit(CuminAction):
        def show(self, session, object):
            frame = self.cumin_class.show_object(session, object)
            return frame.show_edit(session)

        def do_invoke(self, object, args, completion):
            try:
                object.set(**args)
                object.syncUpdate()
                completion("OK")
            except Exception, e:
                completion(e.message or "failed")

    class Remove(CuminAction):
        def show(self, session, object):
            frame = self.model.frame.brokers_remove
            frame.ids.set(session, [object.id])

            branch = session.branch()
            self.model.frame.view.show(branch)
            frame.origin.set(session, branch.marshal())

            return frame.show(session)

        def do_invoke(self, object, args, completion):
            try:
                object.destroySelf();
                object.syncUpdate()
                completion("OK")
            except Exception, e:
                completion(e.message or "failed")

class CuminBrokerGroup(CuminClass):
    def __init__(self, model):
        super(CuminBrokerGroup, self).__init__ \
            (model, "broker_group", BrokerGroup)

        action = self.Add(self, "add")
        action.title = "Add"
        action.navigable = False

        action = self.Edit(self, "edit")
        action.title = "Edit"
        action.summary = True

        action = self.Remove(self, "remove")
        action.title = "Remove"
        action.summary = True

        action = self.RemoveSet(self, "remove_set")
        action.title = "Remove"

    def init(self):
        self.frame = self.model.frame.broker_group

    def get_title(self, session):
        return "Broker Group"

    def get_icon_href(self, session):
        return "resource?name=group-36.png"

    class Add(CuminAction):
        def show(self, session, object):
            frame = self.cumin_class.show_object(session, object)
            return frame.show_add(session)

        def do_invoke(self, object, args, completion):
            assert object is None

            try:
                object = self.cumin_class.mint_class(**args)

                completion("OK")

                return object
            except Exception, e:
                log.exception("Action failed")
                completion(e.message or "failed")

    class Edit(CuminAction):
        def show(self, session, object):
            frame = self.cumin_class.show_object(session, object)
            return frame.show_edit(session)

        def do_invoke(self, object, args, completion):
            try:
                object.set(**args)
                object.syncUpdate()
                completion("OK")
            except Exception, e:
                completion(e.message or "failed")

    class Remove(CuminAction):
        def show(self, session, object):
            frame = self.model.frame.broker_group
            frame.object.set(session, object)
            return frame.remove.show(session)

        def do_invoke(self, object, args, completion):
            try:
                object.destroySelf();
                object.syncUpdate()
                completion("OK")
            except Exception, e:
                completion(e.message or "failed")

    class RemoveSet(CuminSetAction):
        def show(self, session, groups):
            frame = self.model.frame.broker_groups_remove
            frame.objects.set(session, groups)
            return frame.show(session)

        def do_invoke(self, groups, args, completion):
            for group in groups:
                group.destroySelf()
                group.syncUpdate()

            completion("OK")

class Pool(object):
    def __init__(self, collector):
        assert collector

        self.collector = collector
        self.id = self.collector.Pool
        self.name = self.collector.Name
        
    def get(cls, id):
        for coll in Collector.select("pool='%s'" % id):
            return Pool(coll)
    get = classmethod(get)

    def sync(self):
        pass

class CuminPool(CuminClass):
    def __init__(self, model):
        super(CuminPool, self).__init__(model, "pool", Pool)

        prop = CuminProperty(self, "id")
        prop.title = "Collector ID"
        prop.summary = True
        
        stat = self.PercentStat(self, "running")
        stat.title = "Running Jobs"

        stat = self.PercentStat(self, "completed")
        stat.title = "Completed Jobs"

        stat = self.PercentStat(self, "idle")
        stat.title = "Idle Jobs"

        stat = self.PercentStat(self, "held")
        stat.title = "Held Jobs"

        stat = self.PercentStat(self, "removed")
        stat.title = "Removed Jobs"

        stat = self.PercentStat(self, "total")
        stat.title = "Total Jobs"

        stat = self.StatusStat(self, "idl")
        stat.category = "status"

        stat = self.StatusStat(self, "all")
        stat.category = "status"

        action = self.PoolSlotVisualization(self, "slots")
        action.navigable = False

        action = self.VisMachine(self, "machines")
        action.navigable = False

        action = self.SeeAllPools(self, "allpools")
        action.summary = True

        self.fake_stats = self.FakeJobStats(self, "fakestats")
        self.fake_stats.navigable = False

        self.status = self.PoolStatus(self, "poolstatus")
        self.status.navigable = False

    def init(self):
        self.frame = self.model.frame.pool

    def get_title(self, session):
        return "Pool"

    def get_object_title(self, session, pool):
        if pool.name == "main":
            title = "Main Pool"
        else:
            title = super(CuminPool, self).get_object_title(session, pool)

        return title

    def get_icon_href(self, session):
        return "resource?name=cluster-36.png"

    class SeeAllPools(CuminAction):
        def get_title(self, session):
            return "See All Pools"

        def get_href(self, session, object):
            branch = session.branch()
            self.model.frame.view.show(branch)
            return branch.marshal()

    def write_stat_xml(self, session, writer, object):
        record = self.fake_stats.get_stat_record(session, object)
        for stat in self.stats:
            if stat.category == "general":
                stat.write_xml(writer, record)

        record = self.status.get_stat_record(session, object)
        for stat in self.stats:
            if stat.category == "status":
                stat.write_xml(writer, record)

    class StatusStat(CuminStat):
        def value_text(self, record):
            if record:
                state = self.name
                value = record[state]
                return str(value)
            return ""

        def rate_text(self, record):
            return ""

    class PercentStat(CuminStat):
        def value_text(self, record):
            if record:
                state = self.name
                value = record[state]
                return str(value)
            return ""

        def rate_text(self, record):
            if record:
                state = self.name
                value = record[state]
                total = record["total"]
                if total:
                    percent = float(value) / float(total) * 100.0
                return total and "%2.2f" % percent or "-"
            return ""

    class PoolSlotVisualization(SlotVisualization):
        def __init__(self, cls, name):
            super(CuminPool.PoolSlotVisualization, self).__init__(cls, name)
            
            self.itemset = self.ModelPoolSlotSet(cls.model.app, "slot")
            self.itemset.items.path = "CuminPool.PoolSlotVisualization.slot"

        def set_machine(self, session, machine):
            self.itemset.set_machine(session, machine)
        
        def get_machine(self, session):
            return self.itemset.get_machine(session)
        
        class ModelPoolSlotSet(PoolSlotSet):
            def __init__(self, app, name):
                super(CuminPool.PoolSlotVisualization.ModelPoolSlotSet, self).__init__(app, name)
    
                self.__machine = None
                
            def set_args(self, session, machine):
                self.set_machine(session, machine)
                
            def set_machine(self, session, machine):
                self.__machine= machine
                
            def get_machine(self, session):
                return self.__machine
            
            def render_sql_where(self, session, pool):
                elems = list()
                elems.append("s.pool = %(pool)s")
                recent = self.get_recent_sql_where(session)
                if recent:
                    elems.append(recent)
                  
                if self.__machine:
                    elems.append("machine = %(machine)s")

                return "where %s" % " and ".join(elems)

            def get_sql_values(self, session, pool):
                values = {"pool": pool.id}
                machine = self.__machine
                if machine:
                    values["machine"] = machine
                return values

            def render_sql_limit(self, session, *args):
                pass
            
            def render_sql_orderby(self, session, *args):
                return "order by machine, name asc"

    class FakeJobStats(CuminAction):
        def __init__(self, cls, name):
            super(CuminPool.FakeJobStats, self).__init__(cls, name)

            self.itemset = PoolJobStats(cls.model.app, "stats")
            self.itemset.items.path = "CuminPool.FakeJobStats.stats"
            
        def get_stat_record(self, session, pool):
            cursor = self.itemset.get_items(session, pool)
            rows = self.itemset.cursor_to_rows(cursor)
            if len(rows):
                return rows[0]

    class PoolStatus(CuminAction):
        def __init__(self, cls, name):
            super(CuminPool.PoolStatus, self).__init__(cls, name)

            self.itemset = self.StatusPoolSlotSet(cls.model.app, "stats")
            self.itemset.items.path = "CuminPool.PoolStatus.stats"

        def get_stat_record(self, session, pool):
            rows = self.itemset.get_items(session, pool)
            if len(rows):
                return rows[0]

        class StatusPoolSlotSet(SlotStatSet):
            def render_sql_where(self, session, pool):
                elems = list()
                elems.append("s.pool = %(pool)s")
                recent = self.get_recent_sql_where(session)
                if recent:
                    elems.append(recent)

                return "where %s" % " and ".join(elems)

            def get_sql_values(self, session, pool):
                values = {"pool": pool.id}
                return values

    class VisMachine(Visualization):
        load_colors = [("Idle", "clear"), 
                       ("> 0%", "green"),
                       ("> 25%", "yellow"),
                       ("> 50%", "blue"),
                       ("> 75%", "green3")]

        def __init__(self, cls, name):
            super(CuminPool.VisMachine, self).__init__(cls, name)

            self.itemset = PoolMachineSet(cls.model.app, "machine")
            self.itemset.items.path = "CuminPool.VisMachine.machine"

        def get_field_tuples(self, session):
            return [("machine", "Machine"), ("busy", "Busy Slots"), ("idle", "Idle Slots")]
        
        def get_color(self, session, machine):
            total = float(machine["total"])
            busy = float(machine["busy"])
            work = total > 0.0 and busy / total or 0.0
            
            percents = [0.0, 0.25, 0.5, 0.75, 1.0]
            for i in range(len(percents)):
                if work <= percents[i]:
                    return self.load_colors[i][1]
            return "grey"  

        def get_colors(self):
            return self.load_colors
        
class CuminLimit(CuminClass):
    def __init__(self, model):
        super(CuminLimit, self).__init__ \
            (model, "limit", Limit)
            
        prop = CuminProperty(self, "name")
        prop.title = "Name"

        prop = CuminProperty(self, "curr")
        prop.summary = True
        prop.title = "Current"

        prop = CuminProperty(self, "max")
        prop.summary = True
        prop.title = "Maximum"
        
        action = self.Edit(self, "edit")
        action.title = "Edit"
        action.summary = True

        action = self.SetLimit(self, "setlimit")
        action.navigable = False
        action.title = "Set Limit"

    def init(self):
        self.frame = self.model.frame.pool.limit

    def get_title(self, session):
        return "Concurrency Limit"
    
    def get_object_title(self, session, limit):
        title = self.get_title(session)
        name = limit.id
        return "%s '%s'" % (title, name)

    class Edit(CuminAction):
        def show(self, session, limit):
            frame = self.cumin_class.show_object(session, limit)
            frame = frame.show_edit(session)
            return frame
    
    class SetLimit(CuminAction):
        def show(self, session, job):
            pass
        
        def get_title(self, session):
            return "Set Limit for"
        
        def do_invoke(self, limit, negotiator, completion):
            Name = limit.id
            Max = limit.max
            negotiator.SetLimit(self.model.data, completion, Name, Max)
            #negotiator.SetLimit(self.model.data, completion, Name, str(Max))
    
class CuminJobGroup(CuminClass):
    def __init__(self, model):
        super(CuminJobGroup, self).__init__ \
            (model, "job_group", JobGroup)

        stat = CuminStat(self, "Running")
        stat.title = "Running Jobs"

        stat = CuminStat(self, "Completed")
        stat.title = "Completed Jobs"

        stat = CuminStat(self, "Idle")
        stat.title = "Idle Jobs"

        stat = CuminStat(self, "Held")
        stat.title = "Held Jobs"

        stat = CuminStat(self, "Jobs")
        stat.title = "Total Jobs"

        action = self.Hold(self, "hold")
        action.title = "Hold"
        action.summary = True

        action = self.Release(self, "release")
        action.title = "Release"
        action.summary = True

        action = self.Remove(self, "remove")
        action.title = "Remove"
        action.summary = True

    def init(self):
        self.frame = self.model.frame.pool.job_group

    def get_object_title(self, session, group):
        title = self.get_title(session)
        name = group.get_id()
        return "%s '%s'" % (title, name)

    def get_title(self, session):
        return "Job Group"

    def get_icon_href(self, session):
        return "resource?name=group-36.png"
    
    class Hold(CuminAction):
        def show(self, session, group):
            frame = self.model.frame.pool.job_group_hold.show(session)
            frame.ids.set(session, [group.id])
            return frame

    class Release(CuminAction):
        def show(self, session, group):
            frame = self.model.frame.pool.job_group_release.show(session)
            frame.ids.set(session, [group.id])
            return frame

    class Remove(CuminAction):
        def show(self, session, group):
            frame = self.model.frame.pool.job_group_remove.show(session)
            frame.ids.set(session, [group.id])
            return frame

class CuminJob(RemoteClass):
    def __init__(self, model):
        super(CuminJob, self).__init__(model, "job", Job, JobStats)

        ### Main Group
        prop = AdProperty(self, "Owner")
        prop.description = "The submitter of the job"
        prop.writable = False

        prop = AdProperty(self, "BufferBlockSize")
        prop.example = "32768"

        ### Condor Info Group
        prop = AdProperty(self, "CondorVersion")
        prop.group = "Condor Info"
        prop.writable = False

        prop = AdProperty(self, "CondorPlatform")
        prop.group = "Condor Info"
        prop.writable = False

        ### Command Info Group
        prop = AdProperty(self, "Args")
        prop.description = "Arguments passed to job Cmd"
        prop.group = "Command Info"

        prop = AdProperty(self, "Cmd")
        prop.description = "Command that will run the job"
        prop.group = "Command Info"
        prop.title = "Command"

        prop = AdProperty(self, "In")
        prop.description = "The file where the job's standard input is read"
        prop.group = "Command Info"

        prop = AdProperty(self, "Iwd")
        prop.description = "Command Input Working Directory"
        prop.group = "Command Info"
        prop.title = "Working Directory"

        prop = AdProperty(self, "Out")
        prop.description = "The file where the job's standard output is written"
        prop.example = "'/dev/null' or '~/logs/'"
        prop.group = "Command Info"

        ######## Job Status Info
        prop = self.JobStatusProperty(self, "JobStatus")
        prop.description = "The current job status"
        prop.group = "Job Status Info"
        prop.renderer = prop.render_status
        prop.title = "Job Status"
        prop.writable = False

        prop = AdProperty(self, "HoldReasonCode")
        prop.group = "Job Status Info"
        prop.title = "Hold Reason Code"
        prop.writable = False
        
        prop = AdProperty(self, "ExitStatus")
        prop.description = "Status when job completes"
        prop.group = "Job Status Info"
        prop.title = "Exit Status"
        prop.writable = False
        
        prop = AdProperty(self, "ProcId")
        prop.description = "The id of the job within its cluster. Proc Ids are unique within a cluster."
        prop.group = "Job Status Info"
        prop.title = "Proc Id"
        prop.writable = False

        ######## Other
        prop = AdProperty(self, "ClusterId")
        prop.description = "The id of the cluster the job belongs to"
        prop.group = "Other"
        prop.title = "Cluster ID"
        prop.writable = False
        
        prop = DateAdProperty(self, "QDate")
        prop.description = "When the job was submitted"
        prop.writable = False

        prop = DateAdProperty(self, "JobStartDate")
        prop.group = "Main"
        prop.writable = False
        
        prop = DateAdProperty(self, "JobCurrentStartDate")
        prop.writable = False
        
        prop = DateAdProperty(self, "LastSuspensionTime")
        prop.writable = False
        
        prop = DateAdProperty(self, "ShadowBday")
        prop.writable = False
        
        prop = DateAdProperty(self, "LastJobLeaseRenewal")
        prop.writable = False
        
        prop = DateAdProperty(self, "EnteredCurrentStatus")
        prop.writable = False
        
        prop = DateAdProperty(self, "CommittedTime")
        prop.writable = False
        
        prop = DateAdProperty(self, "LastMatchTime")
        prop.writable = False
        
        ######## Properties
        prop = self.GroupProperty(self, "CustomGroup")
        prop.title = "Job Group"
        prop.summary = True
        prop.escape = False

        prop = self.SchedulerProperty(self, "scheduler")
        prop.title = "Scheduler"
        prop.summary = True
        prop.escape = False

        prop = self.SubmitterProperty(self, "submitter")
        prop.title = "Submitter"
        prop.summary = True
        prop.escape = False

        prop = CuminProperty(self, "AccountingGroup")
        prop.title = "Accounting Group"

        prop = CuminProperty(self, "Args")
        prop.title = "Args"

        prop = CuminProperty(self, "ClusterId")
        prop.title = "Cluster ID"

        prop = CuminProperty(self, "Cmd")
        prop.title = "Command"

        prop = CuminProperty(self, "ConcurrencyLimits")
        prop.title = "Limits"

        prop = CuminProperty(self, "CustomId")
        prop.title = "Custom ID"

        prop = CuminProperty(self, "CustomPriority")
        prop.title = "Custom Priority"

        prop = CuminProperty(self, "GlobalJobId")
        prop.title = "Global Job ID"

        prop = CuminProperty(self, "JobStatus")
        prop.title = "Job Status"

        prop = CuminProperty(self, "Note")
        prop.title = "Note"

        prop = CuminProperty(self, "ProcId")
        prop.title = "Proc ID"

        prop = CuminProperty(self, "QDate")
        prop.title = "QDate"

        prop = CuminProperty(self, "Requirements")
        prop.title = "Requirements"

        prop = CuminProperty(self, "Title")
        prop.title = "Title"

        prop = CuminProperty(self, "UserLog")
        prop.title = "User Log"

        prop = CuminProperty(self, "HoldReason")
        prop.title = "Hold Reason"

        ######## Actions
        action = self.Hold(self, "hold")
        action.summary = True
        
        action = self.Release(self, "release")
        action.summary = True
        
        action = self.Remove(self, "remove")
        action.summary = True

        action = self.SetAttribute(self, "setattribute")
        action.navigable = False
        
        action = self.GetAd(self, "getad")
        action.navigable = False
        
        action = self.Fetch(self, "fetch")
        action.navigable = False

    def init(self):
        self.frame = self.model.frame.pool.job

    def get_title(self, session):
        return "Job"
    
    def get_object_name(self, job):
        return job.CustomId

    def get_icon_href(self, session):
        return "resource?name=job-36.png"

    class JobStatusProperty(AdProperty):
        def render_status(self, session, status):
            return JobStatusInfo.get_status_string(status)            
            
    class GroupProperty(CuminProperty):
        def value(self, session, job):
            group = JobGroup(job.CustomGroup)
            href = self.model.frame.pool.job_group.get_href(session, group)
            content = fmt_shorten(group.id, 12, 3)
            return fmt_link(href, group.get_id())
                
    class SubmitterProperty(CuminProperty):
        def value(self, session, job):
            sub = job.submitter
            href = self.model.frame.pool.submitter.get_href(session, sub)
            return fmt_link(href, sub.Name)
        
    class SchedulerProperty(CuminProperty):
        def value(self, session, job):
            sched = job.scheduler
            href = self.model.frame.pool.scheduler.get_href(session, sched)
            return fmt_link(href, sched.Name)

    class GetAd(CuminAction):
        def do_invoke(self, job):
            self.job_ads = {"":{"VALUE": "", "TYPE": 0}}
            self.got_data = False
            
            def completion(status, job_ads):
                try:
                    self.job_ads = job_ads["JobAd"]
                    self.got_data = True
                except:
                    pass
                
            def predicate():
                return self.got_data
            
            try:
                job.GetAd(self.model.data, completion, None)
            except Exception, e:
                return self.job_ads
            
            # wait for up to 20 seconds for completion to be called
            wait(predicate, timeout=20)
            return self.job_ads
        
    class Fetch(CuminAction):
        def get_xml_response(self, session, job, *args):
            file, start, end, tail = args
            self.job_output = None
            self.got_data = False
            self.reason = None
            
            def completion(status, job_output):
                if "Data" in job_output:
                    raw = job_output["Data"]
                    #remove the last partial line from the buffer 
                    lindex = -1
                    ord_nl = ord('\n')
                    while ord(raw[-lindex:][0]) != ord_nl:
                        lindex = lindex + 1
                    self.job_output = raw[:-lindex]
                else:
                    self.reason = status

                self.got_data = True
                
            def predicate():
                return self.got_data
            
            try:
                data = None
                job.Fetch(self.model.data, completion, file, start, end, data)
                # wait for up to 20 seconds for completion to be called
                wait(predicate, timeout=20)
                if not self.got_data:
                    self.reason = "time out"
            except Exception, e:
                self.reason = e 

            if self.reason:
                self.job_output = "Unable to get file at %s. Reason: %s" % (file, self.reason)
                    
            return self.job_output and "<fetch><output>%s</output><tail>%s</tail></fetch>" % (escape_entity(self.job_output), tail)

    class Hold(CuminAction):
        def show(self, session, job):
            frame = self.model.frame.pool.jobs_hold.show(session)
            frame.ids.set(session, [job.id])
            return frame
        
        def get_title(self, session):
            return "Hold"

        def do_invoke(self, job, reason, completion):
            job.Hold(self.model.data, completion, reason)
        
        def get_enabled(self, session, job):
            is_held = JobStatusInfo.get_status_int("Held") == job.JobStatus
            is_deleted = job.qmfDeleteTime is not None
            return not (is_held or is_deleted)
    
    class Release(CuminAction):
        def show(self, session, job):
            frame = self.model.frame.pool.jobs_release.show(session)
            frame.ids.set(session, [job.id])
            return frame
        
        def get_title(self, session):
            return "Release"

        def do_invoke(self, job, reason, completion):
            job.Release(self.model.data, completion, reason)

        def get_enabled(self, session, job):
            is_held = JobStatusInfo.get_status_int("Held") == job.JobStatus
            is_deleted = job.qmfDeleteTime is not None
            return is_held and not is_deleted

    class Remove(CuminAction):
        def show(self, session, job):
            frame = self.model.frame.pool.jobs_remove.show(session)
            frame.ids.set(session, [job.id])
            return frame
        
        def get_title(self, session):
            return "Remove"

        def do_invoke(self, job, reason, completion):
            job.Remove(self.model.data, completion, reason)

        def get_enabled(self, session, job):
            is_deleted = job.qmfDeleteTime is not None
            return not is_deleted

    class SetAttribute(CuminAction):
        def show(self, session, job):
            pass
        
        def get_title(self, session):
            return "Save Ad for"
        
        def do_invoke(self, job, args, completion):
            Name = args[0]
            Value = args[1]
            job.SetAttribute(self.model.data, completion, Name, str(Value))

class GetStartedAction(CuminAction):
    def get_xml_response(self, session, object, *args):
        updateTime = object.statsCurr and object.statsCurr.qmfUpdateTime \
            or object.qmfUpdateTime
        delta = secs(updateTime) - secs(datetime.now()) 
        conf = "<age>%i</age>" % -delta
        rect = "<updatetime>%s</updatetime>" % fmt_duration(delta)

        return "%s%s" % (conf, rect)
    
class CuminScheduler(RemoteClass):
    def __init__(self, model):
        super(CuminScheduler, self).__init__(model, "scheduler",
                                             Scheduler, SchedulerStats)

        prop = CuminProperty(self, "Name")
        prop.title = "Name"

        prop = CuminProperty(self, "JobQueueBirthdate")
        prop.title = "Job Queue Birthdate"

        prop = CuminProperty(self, "MaxJobsRunning")
        prop.title = "Max Jobs Running"

        prop = CuminProperty(self, "MyAddress")
        prop.title = "Contact Address"

        prop = CuminProperty(self, "PublicNetworkIpAddr")
        prop.title = "Public IP Address"

        prop = CuminProperty(self, "Machine")
        prop.title = "Machine"
        prop.summary = True

        prop = CuminProperty(self, "System")
        prop.title = "System"

        stat = CuminStat(self, "NumUsers")
        stat.title = "Users"

        stat = CuminStat(self, "TotalRunningJobs")
        stat.title = "Running Jobs"

        stat = CuminStat(self, "TotalIdleJobs")
        stat.title = "Idle Jobs"

        stat = CuminStat(self, "TotalHeldJobs")
        stat.title = "Held Jobs"

        stat = CuminStat(self, "TotalJobAds")
        stat.title = "Job Ads"

        stat = CuminStat(self, "TotalRemovedJobs")
        stat.title = "Removed Jobs"

        action = self.Start(self, "start")
        action.summary = True
        
        action = self.Stop(self, "stop")
        action.summary = True

        action = GetStartedAction(self, "GetStarted")
        action.navigable = False

    def init(self):
        self.frame = self.model.frame.pool.scheduler
        
    def get_title(self, session):
        return "Scheduler"

    def get_object_name(self, sched):
        return sched.Name

    class Start(CuminAction):
        def show(self, session, sched):
            frame = self.model.frame.pool.show_scheds_start(session)
            frame.ids.set(session, [sched.id])
            return frame
        
        def get_title(self, session):
            return "Start"

    class Stop(CuminAction):
        def show(self, session, sched):
            frame = self.model.frame.pool.show_scheds_stop(session)
            frame.ids.set(session, [sched.id])
            return frame
        
        def get_title(self, session):
            return "Stop"

class CuminSubmitter(RemoteClass):
    def __init__(self, model):
        super(CuminSubmitter, self).__init__(model, "submitter",
                                             Submitter, SubmitterStats)

        prop = CuminProperty(self, "Name")
        prop.title = "Name"

        stat = CuminStat(self, "RunningJobs")
        stat.title = "Running Jobs"

        stat = CuminStat(self, "IdleJobs")
        stat.title = "Idle Jobs"

        stat = CuminStat(self, "HeldJobs")
        stat.title = "Held Jobs"

    def init(self):
        self.frame = self.model.frame.pool.submitter
        
    def get_title(self, session):
        return "Submitter"

    def get_object_name(self, sub):
        return sub.Name

    def show_object(self, session, sub):
        pool = self.model.get_main_pool()

        assert pool

        self.model.frame.pool.set_object(session, pool)

        return self.frame.show_object(session, sub)

class CuminCollector(RemoteClass):
    def __init__(self, model):
        super(CuminCollector, self).__init__(model, "collector",
                                             Collector, CollectorStats)

        prop = CuminProperty(self, "Name")
        prop.title = "Name"

        prop = CuminProperty(self, "PublicNetworkIpAddr")
        prop.title = "IP Address"

        action = self.Start(self, "start")
        action.summary = True

        action = self.Stop(self, "stop")
        action.summary = True

        action = GetStartedAction(self, "GetStarted")
        action.navigable = False

    def init(self):
        self.frame = self.model.frame.pool.collector
        
    def get_title(self, session):
        return "Collector"

    def get_object_name(self, coll):
        return coll.Name

    class Start(CuminAction):
        def show(self, session, collector):
            frame = self.model.frame.pool.show_colls_start(session)
            ids = [collector.id]
            frame.ids.set(session, ids)
            return frame
        
        def get_title(self, session):
            return "Start"

    class Stop(CuminAction):
        def show(self, session, collector):
            frame = self.model.frame.pool.show_colls_stop(session)
            ids = [collector.id]
            frame.ids.set(session, ids)
            return frame
        
        def get_title(self, session):
            return "Stop"

class CuminNegotiator(RemoteClass):
    def __init__(self, model):
        super(CuminNegotiator, self).__init__(model, "negotiator",
                                              Negotiator, NegotiatorStats)

        prop = CuminProperty(self, "Name")
        prop.title = "Name"

        prop = CuminProperty(self, "Machine")
        prop.title = "Machine"

        prop = CuminProperty(self, "MyAddress")
        prop.title = "Address"
        
        prop = CuminProperty(self, "DaemonStartTime")
        prop.title = "Start Time"

        stat = CuminStat(self, "MonitorSelfAge")
        stat.title = "Age"
        stat.category = "general"

        stat = CuminStat(self, "MonitorSelfCPUUsage")
        stat.title = "CPU Usage"
        stat.category = "general"

        stat = CuminStat(self, "MonitorSelfImageSize")
        stat.title = "Image Size"
        stat.category = "general"

        stat = CuminStat(self, "MonitorSelfRegisteredSocketCount")
        stat.title = "Registered Socket Count"
        stat.category = "general"

        stat = CuminStat(self, "MonitorSelfResidentSetSize")
        stat.title = "Resident Set Size"
        stat.category = "general"

        action = self.Start(self, "start")
        action.summary = True
        
        action = self.Stop(self, "stop")
        action.summary = True

        action = self.GetLimits(self, "GetLimits")
        action.navigable = False

        action = self.GetLimitCount(self, "GetLimitCount")
        action.navigable = False

        action = GetStartedAction(self, "GetStarted")
        action.navigable = False

    def init(self):
        self.frame = self.model.frame.pool.negotiator
        
    def get_title(self, session):
        return "Negotiator"

    def get_object_name(self, neg):
        return neg.Name

    class Start(CuminAction):
        def show(self, session, neg):
            frame = self.model.frame.pool.show_negs_start(session)
            frame.ids.set(session, [neg.id])
            return frame
        
        def get_title(self, session):
            return "Start"

    class Stop(CuminAction):
        def show(self, session, neg):
            frame = self.model.frame.pool.show_negs_stop(session)
            frame.ids.set(session, [neg.id])
            return frame
        
        def get_title(self, session):
            return "Stop"

    class GetLimitCount(CuminAction):
        def get_xml_response(self, session, negotiator, *args):
            action = self.model.negotiator.GetLimits
            limits = action.do_invoke(negotiator)
            error = "False"
            if "timeout" in limits:
                val = 0
                error = "True"
            else:
                val = len(limits)
            err = "<error value=\"%s\" />" % error
            count = "<count value=\"%i\" />" % val
            return "%s%s" % (count, err)

    class GetLimits(CuminAction):
        def do_invoke(self, negotiator):
            self.lim = dict()
            self.got_data = False
            
            def completion(status, data):
                self.lim = data["Limits"]
                self.got_data = True
                
            def predicate():
                return self.got_data
            
            try:
                negotiator.GetLimits(self.model.data, completion, None)
            except Exception, e:
                return self.lim
            
            # wait for up to 5 seconds for completion to be called
            wait(predicate, timeout=5)
            if not self.got_data:
                self.lim["timeout"] = True
            return self.lim

class ModelPage(Page):
    def __init__(self, app, name):
        super(ModelPage, self).__init__(app, name)

        self.__class = CuminClassParameter(app, "class")
        self.add_parameter(self.__class)

        param = IntegerParameter(app, "param")
        self.add_parameter(param)

        self.__ids = ListParameter(app, "id", param)
        self.add_parameter(self.__ids)

    def get_content_type(self, session):
        return Page.xml_content_type

    def do_render(self, session):
        writer = Writer()
        writer.write(Page.xml_1_0_declaration)

        cls = self.__class.get(session)
        objects = list()

        if cls:
            for id in self.__ids.get(session):
                objects.append(cls.mint_class.get(id))
        
        self.app.model.write_xml(session, writer, objects)
        
        return writer.to_string()

class CallPage(Page):
    def __init__(self, app, name):
        super(CallPage, self).__init__(app, name)

        self.__class = CuminClassParameter(app, "class")
        self.add_parameter(self.__class)

        self.__id = IntegerParameter(app, "id")
        self.add_parameter(self.__id)
        
        self.__method = Parameter(app, "method")
        self.add_parameter(self.__method)

        self.__args = Parameter(app, "xargs")
        self.add_parameter(self.__args)

    def get_content_type(self, session):
        return Page.xml_content_type

    def do_render(self, session):
        writer = Writer()
        writer.write(Page.xml_1_0_declaration)

        cls = self.__class.get(session)
        id = self.__id.get(session)
        xargs = self.__args.get(session)

        if cls:
            object = cls.mint_class.get(id)
            method = self.__method.get(session)
            for action in cls.actions:
                if action.name == method:
                    args = xargs and xargs.split("&") or list()
                    data = action.get_xml_response(session, object, *args)
                    self.write_xml(writer, data)
        
        return writer.to_string()

    def write_xml(self, writer, data):
        writer.write("<data>")
        writer.write(data)
        writer.write("</data>")
