from wooly import *
from wooly.forms import *

from model import *
from widgets import *
from util import *

from wooly.widgets import Link

strings = StringCatalog(__file__)

class CuminActionButton(FormButton):
    def __init__(self, app, name, action, object=None):
        super(CuminActionButton, self).__init__(app, name)

        self.action = action
        self.object = object

    def process_submit(self, session, *args):
        branch = session.branch()
        object = None

        if self.object:
            object = self.object.get(session)

        self.action.show(branch, object)

        self.page.set_redirect_url(session, branch.marshal())

    def render_content(self, session, *args):
        return self.action.get_title(session)

class CuminActionLink(Link):
    def __init__(self, app, name, action, object=None):
        super(CuminActionLink, self).__init__(app, name)

        self.action = action
        self.object = object

    def edit_session(self, session, *args):
        object = None

        if self.object:
            object = self.object.get(session)

        self.action.show(session, object)

    def render_content(self, session, *args):
        return self.action.get_title(session)

class CuminActionForm(CuminFieldForm):
    def __init__(self, app, name, action, object=None):
        super(CuminActionForm, self).__init__(app, name)

        self.action = action
        self.object = object

    def render_title(self, session, *args):
        return self.action.get_summary(session, self.object.get(session))

    def process_submit(self, session, *args):
        object = None

        if self.object:
            object = self.object.get(session)

        self.action.invoke(object)

        self.process_cancel(session, *args)

class CuminSetActionForm(CuminForm):
    def __init__(self, app, name, action, object):
        super(CuminSetActionForm, self).__init__(app, name)

        assert isinstance(action, CuminAction)
        assert action.aggregate
        assert isinstance(object, Parameter)

        self.action = action

        self.objects = ListParameter(app, "param", object)
        self.add_parameter(self.objects)

        body = self.SetActionBody(app, "body")
        self.add_child(body)

    def render_title(self, session, *args):
        return self.action.get_summary(session, self.objects.get(session))

    def process_submit(self, session, *args):
        self.action.invoke(self.objects.get(session))

        self.objects.set(session, None)

        self.process_cancel(session, *args)

    class SetActionBody(ItemSet):
        def render_message(self, session, *args):
            verb = self.parent.action.get_title(session)
            noun = self.parent.action.cumin_class.get_plural_title(session)

            return "%s the following %s" % (verb, noun)

        def do_get_items(self, session, *args):
            return self.parent.objects.get(session)

        def render_item_content(self, session, item):
            cls = self.parent.action.cumin_class
            return cls.get_object_title(session, item)

class CuminClassTable(CuminTableWithControls, Form):
    def __init__(self, app, name, cumin_class):
        super(CuminClassTable, self).__init__(app, name)

        self.cumin_class = cumin_class

        self.selection = CuminClassCheckboxColumn(app, "id", self.cumin_class)
        self.add_column(self.selection)

        self.actions = WidgetSet(app, "actions")
        self.actions.html_class = "actions"
        self.add_child(self.actions)

        actions = [x for x in self.cumin_class.actions if x.aggregate]

        for action in actions:
            button = CuminActionButton \
                (app, action.name, action, self.selection.param)
            self.actions.add_child(button)

        self.add = None

        if hasattr(self.cumin_class, "add"):
            self.add = CuminActionLink(app, "add", self.cumin_class.add)
            self.add.html_class = "action"
            self.add_child(self.add)

    def render_title(self, session, *args):
        title = self.cumin_class.get_plural_title(session)
        count = fmt_count(self.cumin_class.mint_class.select().count())

        return "%s %s" % (title, count)

    def render_plural_title(self, session, *args):
        return self.cumin_class.get_plural_title(session)

    def render_add_link(self, session, *args):
        if self.add:
            return self.add.render(session, *args)

    def render_actions(self, session, *args):
        if self.actions.children:
            return self.actions.render(session, *args)

class CuminClassCheckboxColumn(CheckboxInputColumn):
    def __init__(self, app, name, cumin_class):
        item = CuminObjectParameter(app, "item", cumin_class)

        super(CuminClassCheckboxColumn, self).__init__(app, name, item)

        self.cumin_class = cumin_class

class CuminClassNameColumn(SqlTableColumn):
    def __init__(self, app, name, cumin_class):
        super(CuminClassNameColumn, self).__init__(app, name)

        self.cumin_class = cumin_class

        self.id_column = "id"
        self.name_column = "name"

    def render_title(self, session, data):
        return "Name"

    def render_content(self, session, data):
        id = data[self.id_column]
        name = data[self.name_column]

        href = self.cumin_class.get_object_href_by_id(session, id)

        return fmt_link(href, name)
