import logging

from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *
from wooly.tables import *

from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *
from job import *

strings = StringCatalog(__file__)
log = logging.getLogger("cumin.negotiator")

class NegotiatorSet(CuminTable, Form):
    def __init__(self, app, name):
        super(NegotiatorSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.NameColumn(app, "name")
        self.add_column(col)
        self.set_default_column(col)
        
        col = self.SystemColumn(app, "system")
        self.add_column(col)

        self.__start = self.StartButton(app, "start")
        self.add_child(self.__start)

        self.__stop = self.StopButton(app, "stop")
        self.add_child(self.__stop)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, negotiator):
        count = self.get_item_count(session, negotiator)
        return "Negotiators %s" % fmt_count(count)

    def render_sql_where(self, session, negotiator):
        sql = "((c.qmf_update_time is not null and " + \
            "c.qmf_update_time > now() - interval '10 minutes')" + \
            " and qmf_delete_time is null)"
        return "where %s" % sql

    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            branch = session.branch()

            neg = Identifiable(data["id"])
            self.app.model.negotiator.show_object(branch, neg)

            return fmt_olink(branch, neg, name=data["name"])

    class SystemColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "System"

        def render_content(self, session, data):
            id = data["system_id"]

            if id:
                sys = Identifiable(id)
                href = self.page.main.system.get_href(session, sys)
                return fmt_link(href, data["system_name"])

    class StartButton(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)
    
            branch = session.branch()
            frame = self.frame.show_negs_start(branch)
            frame.ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())
            
        def render_content(self, session):
            return "Start"
        
    class StopButton(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)
    
            branch = session.branch()
            frame = self.frame.show_negs_stop(branch)
            frame.ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())
            
        def render_content(self, session):
            return "Stop"
        
class NegotiatorFrame(CuminFrame):
    def __init__(self, app, name):
        super(NegotiatorFrame, self).__init__(app, name)

        self.object = NegotiatorParameter(app, "id")
        self.add_parameter(self.object)

        self.view = NegotiatorView(app, "view")
        self.add_mode(self.view)
        self.set_view_mode(self.view)

class NegotiatorView(CuminView):
    def __init__(self, app, name):
        super(NegotiatorView, self).__init__(app, name)

        status = self.NegotiatorStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        details = CuminDetails(app, "details")
        self.__tabs.add_tab(details)

    class NegotiatorStatus(StartStopStatus):
        def get_url(self, session):
            negotiator = self.get_args(session)[0]
            if negotiator:
                return "call.xml?class=negotiator;id=%i;method=GetStarted" % negotiator.id
    
        def render_title(self, session, *args):
            return "Negotiator Status"
    
class NegStart(CuminBulkActionForm):
    def render_title(self, session, *args):
        return "Start Negotiator"
    
    def render_form_heading(self, session, *args):
        return ""

    def render_item_content(self, session, id):
        return "Start Negotiator '%s'" % Negotiator.get(id).Name

    def process_item(self, session, item):
        negotiator = Negotiator.get(item)
        action = self.app.model.master.start
        action.invoke(negotiator, {"subsystem": "NEGOTIATOR"})
        self.process_cancel(session)

class NegStop(CuminBulkActionForm):
    def render_title(self, session, *args):
        return "Stop Negotiator"
    
    def render_form_heading(self, session, *args):
        return ""

    def render_item_content(self, session, id):
        return "Stop Negotiator '%s'" % Negotiator.get(id).Name

    def process_item(self, session, item):
        negotiator = Negotiator.get(item)
        action = self.app.model.master.stop
        action.invoke(negotiator, {"subsystem": "NEGOTIATOR"})
        self.process_cancel(session)

