import logging

from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *
from wooly.tables import *

from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *
from job import *
from scheduler import SchedulerSet, SchedulerFrame, SchedulerStart, SchedulerStop
from submitter import SubmitterSet, SubmitterFrame
from collector import CollectorSet, CollectorFrame, CollectorStart, CollectorStop
from negotiator import NegotiatorSet, NegotiatorFrame, NegStart, NegStop
from limits import LimitsSet, LimitsFrame
from slot import SlotSet, MachineSet, SlotStatSet

strings = StringCatalog(__file__)
log = logging.getLogger("cumin.pool")

class PoolSet(CuminTable):
    def __init__(self, app, name):
        super(PoolSet, self).__init__(app, name)

        col = self.NameColumn(app, "name")
        self.add_column(col)

        self.set_default_column(col)

        col = self.JobsColumn(app, "jobs")
        self.add_column(col)

        col = self.StatusColumn(app, "status")
        self.add_column(col)

    def render_title(self, session):
        count = self.get_item_count(session)
        return "Pools %s" % fmt_count(count)

    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            pool = Identifiable(data["id"])
            self.page.main.set_last_pool(session, data["id"])
            href = self.frame.pool.get_href(session, pool)
            
            return fmt_link(href, data["name"])

    class JobsColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Jobs"

        def render_content(self, session, data):
            return data["jobs"]

    class StatusColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Status"

        def render_content(self, session, data):
            return "Active"

class PoolFrame(CuminFrame):
    def __init__(self, app, name):
        super(PoolFrame, self).__init__(app, name)

        self.object = PoolParameter(app, "id")
        self.add_parameter(self.object)

        self.view = PoolView(app, "view")
        self.add_mode(self.view)
        self.set_view_mode(self.view)

        self.job = JobFrame(app, "job")
        self.add_mode(self.job)
        
        self.job_group = JobGroupFrame(app, "jobgroup")
        self.add_mode(self.job_group)
        
        self.scheduler = SchedulerFrame(app, "sched")
        self.add_mode(self.scheduler)

        self.submitter = SubmitterFrame(app, "sub")
        self.add_mode(self.submitter)

        self.collector = CollectorFrame(app, "coll")
        self.add_mode(self.collector)

        self.limit = LimitsFrame(app, "limit")
        self.add_mode(self.limit)

        self.negotiator = NegotiatorFrame(app, "neg")
        self.add_mode(self.negotiator)

        self.__startneg = NegStart(app, "startneg")
        self.add_mode(self.__startneg)

        self.__stopneg = NegStop(app, "stopneg")
        self.add_mode(self.__stopneg)

        self.__startcoll = CollectorStart(app, "startcoll")
        self.add_mode(self.__startcoll)
        
        self.__stopcoll = CollectorStop(app, "stopcoll")
        self.add_mode(self.__stopcoll)
        
        self.__startsched = SchedulerStart(app, "startsched")
        self.add_mode(self.__startsched)
        
        self.__stopsched = SchedulerStop(app, "stopsched")
        self.add_mode(self.__stopsched)
        
        self.jobs_hold = JobSetHold(app, "jobshold")
        self.add_mode(self.jobs_hold)

        self.jobs_release = JobSetRelease(app, "jobsrelease")
        self.add_mode(self.jobs_release)

        self.jobs_remove = JobSetRemove(app, "jobsremove")
        self.add_mode(self.jobs_remove)

        self.job_group_hold = JobGroupHold(app, "jobgrouphold")
        self.add_mode(self.job_group_hold)
        
        self.job_group_release = JobGroupRelease(app, "jobgrouprelease")
        self.add_mode(self.job_group_release)
        
        self.job_group_remove = JobGroupRemove(app, "jobgroupremove")
        self.add_mode(self.job_group_remove)
        
    def show_negs_start(self, session):
        self.page.set_current_frame(session, self.__startneg)
        return self.show_mode(session, self.__startneg)
    
    def show_negs_stop(self, session):
        self.page.set_current_frame(session, self.__stopneg)
        return self.show_mode(session, self.__stopneg)
    
    def show_scheds_start(self, session):
        self.page.set_current_frame(session, self.__startsched)
        return self.show_mode(session, self.__startsched)
    
    def show_scheds_stop(self, session):
        self.page.set_current_frame(session, self.__stopsched)
        return self.show_mode(session, self.__stopsched)
    
    def show_colls_start(self, session):
        self.page.set_current_frame(session, self.__startcoll)
        return self.show_mode(session, self.__startcoll)
    
    def show_colls_stop(self, session):
        self.page.set_current_frame(session, self.__stopcoll)
        return self.show_mode(session, self.__stopcoll)
    
class PoolView(CuminView):
    def __init__(self, app, name):
        super(PoolView, self).__init__(app, name)

        status = PoolStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        stats = PoolStats(app, "stats")
        self.__tabs.add_tab(stats)

        jobs = JobsAndGroupsTab(app, "jobs")
        self.__tabs.add_tab(jobs)

        self.scheds = PoolSchedulerSet(app, "scheds")
        self.__tabs.add_tab(self.scheds)

        subs = PoolSubmitterSet(app, "subs")
        self.__tabs.add_tab(subs)

        self.colls = PoolCollectorSet(app, "colls")
        self.__tabs.add_tab(self.colls)

        self.negs = PoolNegotiatorSet(app, "negs")
        self.__tabs.add_tab(self.negs)

        limits = self.LimitsTab(app, "limits")
        self.__tabs.add_tab(limits)

    class LimitsTab(LimitsSet):
        pass

    def render_script(self, session, pool):
        data = "model.xml?class=pool;id=%s" % pool.id
        return "wooly.setIntervalUpdate('%s', updatePool, 4000)" % data

    def set_collector_tab(self, session):
        self.__tabs.set_selected_mode(session, self.colls)
        
    def set_negotiator_tab(self, session):
        self.__tabs.set_selected_mode(session, self.negs)
        
    def set_scheduler_tab(self, session):
        self.__tabs.set_selected_mode(session, self.scheds)
        
class PoolSchedulerSet(SchedulerSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def get_sql_values(self, session, pool):
        return {"pool": pool.id}

    def render_title(self, session, pool):
        count = self.get_item_count(session, pool)
        return "Schedulers %s" % fmt_count(count)

    def render_sql_where(self, session, pool):
        return "where s.pool = %(pool)s"

class PoolSubmitterSet(SubmitterSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def get_sql_values(self, session, pool):
        return {"pool": pool.id}

    def render_title(self, session, pool):
        count = self.get_item_count(session, pool)
        return "Submitters %s" % fmt_count(count)

    def render_sql_where(self, session, pool):
        return "where d.pool = %(pool)s"
    
class PoolCollectorSet(CollectorSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def get_sql_values(self, session, pool):
        return {"pool": pool.id}

    def render_title(self, session, pool):
        count = self.get_item_count(session, pool)
        return "Collectors %s" % fmt_count(count)

    def render_sql_where(self, session, pool):
        return "where c.pool = %(pool)s"
    
class PoolNegotiatorSet(NegotiatorSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def get_sql_values(self, session, pool):
        return {"pool": pool.id}

    def render_title(self, session, pool):
        count = self.get_item_count(session, pool)
        return "Negotiators %s" % fmt_count(count)

    def render_sql_where(self, session, pool):
        return "where n.pool = %(pool)s"

class PoolSlotSet(SlotSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_sql_where(self, session, pool):
        elems = list()
        elems.append("s.pool = %(pool)s")
        recent = self.get_recent_sql_where(session)
        if recent:
            elems.append(recent)
          
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, pool):
        return {"pool": pool.id}

    def render_title(self, session, pool):
        count = self.get_item_count(session, pool)
        return "Slots %s" % fmt_count(count)
    
    def filter(self, session, system, slots):
        return slots

class PoolMachineSet(MachineSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_sql_where(self, session, pool):
        elems = list()
        elems.append("s.pool = %(pool)s")
        recent = self.get_recent_sql_where(session)
        if recent:
            elems.append(recent)
          
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, pool):
        return {"pool": pool.id}
    
    def render_sql_orderby(self, session, pool):
        return "order by id asc"

class PoolStats(Widget):
    def __init__(self, app, name):
        super(PoolStats, self).__init__(app, name)

        stats = PoolStatSet(app, "general", "general")
        self.add_child(stats)

        self.machine_grid = self.MachineVisualization(app, "machine_grid")
        self.add_child(self.machine_grid)

        self.slot_grid = self.SlotVisualization(app, "slot_grid")
        self.add_child(self.slot_grid)

        self.all_fits = Attribute(app, "can_show_all")
        self.add_attribute(self.all_fits)
        
        self.machine_param = Parameter(app, "machine_param")
        self.add_parameter(self.machine_param)
    
    def process(self, session):
        super(PoolStats, self).process(session)
        all_slots = self.slot_grid.show_all_slots.get(session)
        machine = self.machine_param.get(session)
        all_fits = True
        
        if all_slots == "a":
            machine = None
            
        if not machine:
            self.slot_grid.set_machine(session, None)
            all_fits = self.slot_grid.will_it_fit(session)
            if not all_fits:
                machine = self.machine_grid.get_1st_machine(session)
            else:
                machine = ""

        self.all_fits.set(session, all_fits)
        self.slot_grid.set_machine(session, machine)
        
    def render_title(self, session):
        return "Statistics"
    
    def render_machine_title(self, session):
        slots_fit = self.all_fits.get(session)
        if not slots_fit:
            pool = self.frame.get_args(session)[0]
            return "Machines on %s" % pool.name
    
    def render_machine_help(self, session):
        slots_fit = self.all_fits.get(session)
        if not slots_fit:
            return "Select a Machine to view its slots"    
    
    def render_slot_job_url(self, session):
        job = Identifiable("XXX")
        return self.page.main.pool.job.get_href(session, job)
    
    class SlotVisualization(StatUtilizationGrid):
        def __init__(self, app, name):
            super(PoolStats.SlotVisualization, self).__init__(app, name)

            self.show_all_slots = self.ShowAllSlots(app, "show_all_slots")
            self.add_child(self.show_all_slots)
            
        def render(self, session, *args):
            slots_fit = self.parent.all_fits.get(session)
            if slots_fit:
                return super(PoolStats.SlotVisualization, self).render(session, *args)
            
        def get_cells(self, session):
            pool = self.frame.get_args(session)[0]
            action = self.app.model.pool.slots
            return action.get_boxes(session, pool)
        
        def render_title(self, session):
            pool = self.frame.get_args(session)[0]
            machine = self.get_machine(session)
            return machine and "Slots on (%s)" % machine or "Slots on %s" % pool.name

        def set_machine(self, session, machine):
            action = self.app.model.pool.slots
            action.set_machine(session, machine)
            
        def get_machine(self, session):
            action = self.app.model.pool.slots
            return action.get_machine(session)
            
        def get_colors(self, session):
            action = self.app.model.pool.slots
            return action.get_colors()
            
        def get_states(self, session):
            action = self.app.model.pool.slots
            return action.get_states()
            
        def get_color(self, session, job):
            action = self.app.model.pool.slots
            return action.get_color(session, job)
    
        def get_contents(self, session, slot):
            return ""

        def get_href(self, session, slot):
            return "#"

        def get_url(self, session):
            pool = self.parent.frame.get_args(session)[0]
            machine = self.get_machine(session)
            return "call.xml?class=pool;id=%s;method=slots;xargs=%s" % (pool.id, machine)

        def get_sticky_info(self, session):
            action = self.app.model.pool.slots
            return action.get_field_tuples(session)

        class ShowAllSlots(StateSwitch):
            def __init__(self, app, name):
                super(PoolStats.SlotVisualization.ShowAllSlots, self).__init__(app, name)
                
                self.add_state("a", "All")
                self.add_state("m", "Machine")
        
            def render(self, session):
                state = self.get(session)
                if state == "m":
                    return super(PoolStats.SlotVisualization.ShowAllSlots, self).render(session)
                
            def render_href(self, session):
                state = self.get(session)
                if state == "m":
                    branch = session.branch()
                    self.set(branch, "a")
                    return branch.marshal()
                else:
                    return "#"
            
            def render_class(self, session):
                state = self.get(session)
                return state == "a" and "disabled" or "enabled"
                
    class MachineVisualization(StatUtilizationGrid):
        def render(self, session, *args):
            slots_fit = self.parent.all_fits.get(session)
            if not slots_fit:
                return super(PoolStats.MachineVisualization, self).render(session, *args)
            
        def get_cells(self, session):
            pool = self.frame.get_args(session)[0]
            action = self.app.model.pool.machines
            return action.get_boxes(session, pool)
        
        def render_title(self, session):
            return "Machines"

        def render_vis_help(self, session):
            return "Click on a machine to view its slots"

        def get_1st_machine(self, session):
            pool = self.frame.get_args(session)[0]
            action = self.app.model.pool.machines
            boxes = action.get_boxes(session, pool)
            return boxes[0]["machine"]
        
        def get_colors(self, session):
            action = self.app.model.pool.machines
            return action.get_colors()
            
        def get_color(self, session, machine):
            action = self.app.model.pool.machines
            return action.get_color(session, machine)
    
        def get_contents(self, session, machine):
            return ""

        def get_href(self, session, data):
            machine = data["machine"]
            branch = session.branch()
            self.parent.machine_param.set(branch, machine)
            self.parent.slot_grid.show_all_slots.set(branch, "m")
            return branch.marshal()

        def get_url(self, session):
            pool = self.parent.frame.get_args(session)[0]
            return "call.xml?class=pool;id=%s;method=machines" % pool.id
        
        def get_sticky_info(self, session):
            action = self.app.model.pool.machines
            return action.get_field_tuples(session)

class PoolStatSet(StatSet):
    def render_rate_text(self, session, args):
        return "Percentage"
    
    def do_get_items(self, session, pool):
        stats = super(PoolStatSet, self).do_get_items(session, pool)

        action = self.app.model.pool.fakestats
        record = action.get_stat_record(session, pool)

        fake_stats = list()
        for stat in stats:
            fake_stats.append((stat[0], record))
        
        return fake_stats

class PoolJobStats(CuminTable):
    def render_sql_where(self, session, pool):
        elems = list()
        elems.append("s.pool = %(pool)s")
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, pool):
        values = {"pool": pool.id}
        return values

class PoolStatus(CuminStatus):
    def render_title(self, session, pool):
        return "Pool Status"

    def render_status(self, session, pool):
        action = self.app.model.pool.poolstatus
        record = action.get_stat_record(session, pool)

        return record and "<div><span>%i</span> of <span>%i</span> slots idle</div>" % \
            (record["idl"], record["all"]) or ""
