import logging

from mint.schema import Queue, Journal, Exchange
from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *
from wooly.tables import *
from datetime import datetime
from sqlobject.sqlbuilder import LEFTJOINOn

from binding import *
from exchange import ExchangeInputSet
from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *

strings = StringCatalog(__file__)
log = logging.getLogger("cumin.queue")

class QueueSet(CuminTable, Form):
    def __init__(self, app, name):
        super(QueueSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.NameColumn(app, "name")
        self.add_column(col)

        self.set_default_column(col)

        col = self.ConsumersColumn(app, "consumers")
        col.alignment = "right"
        self.add_column(col)

        col = self.BindingsColumn(app, "bindings")
        col.alignment = "right"
        self.add_column(col)

        col = self.EnqueuedColumn(app, "enqueued")
        col.alignment = "right"
        self.add_column(col)

        col = self.DequeuedColumn(app, "dequeued")
        col.alignment = "right"
        self.add_column(col)

        col = self.DepthColumn(app, "depth")
        col.alignment = "right"
        self.add_column(col)

        self.unit = UnitSwitch(app, "unit")
        self.add_child(self.unit)

        self.phase = PhaseSwitch(app, "phase")
        self.add_child(self.phase)

        self.__purge = self.Purge(app, "purge")
        self.add_child(self.__purge)

        self.__remove = self.Remove(app, "remove")
        self.add_child(self.__remove)

    def render_add_queue_url(self, session, vhost):
        branch = session.branch()
        self.frame.queue_add.show(branch)
        return branch.marshal()
    
    def get_args(self, session):
        reg = self.frame.get_object(session)
        return (reg.getDefaultVhost(),)

    def render_title(self, session, vhost):
        return "Queues %s" % fmt_count(Queue.select().count())

    def render_sql_where(self, session, vhost):
        elems = list()
        elems.append("q.vhost_id = %(id)r")
        elems.append(self.phase.get_sql_constraint(session, vhost))
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, vhost):
        return {"id": vhost.id}

    class Purge(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            branch = session.branch()
            self.frame.queues_purge.show(branch).ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())

        def render_content(self, session):
            return "Purge"
        
    class Remove(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            branch = session.branch()
            self.frame.queues_remove.show(branch).ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())
            
        def render_content(self, session):
            return "Remove"
        
    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            queue = Identifiable(data["id"])
            href = self.frame.queue.get_href(session, queue)
            return fmt_link(href, fmt_shorten(data["name"]))

    class ConsumersColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Consumers"

        def render_value(self, session, value):
            if value is None:
                return fmt_none_brief()
            else:
                return value

    class BindingsColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Bindings"

        def render_content(self, session, data):
            queue = Identifiable(data["id"])
            branch = session.branch()
            self.frame.queue.object.set(branch, queue)
            self.frame.queue.view.bindings.show(branch)
            return fmt_link(branch.marshal(), data["bindings"] or "0")

    class EnqueuedColumn(NullSortColumn, FreshDataOnlyColumn):
        def render_title(self, session, data):
            return "%s Enqueued" % self.parent.unit.get_brief_plural(session)

        def get_column_key(self, session):
            unit = self.parent.unit.get(session)
            return unit == "b" and "benqueued" or "menqueued"

        def render_value(self, session, value):
            return fmt_rate(value, "", "sec")

    class DequeuedColumn(NullSortColumn, FreshDataOnlyColumn):
        def render_title(self, session, data):
            return "%s Dequeued" % self.parent.unit.get_brief_plural(session)

        def get_column_key(self, session):
            unit = self.parent.unit.get(session)
            return unit == "b" and "bdequeued" or "mdequeued"

        def render_value(self, session, value):
            return fmt_rate(value, "", "sec")

    class DepthColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "%s Depth" % self.parent.unit.get_brief_singular(session)

        def get_column_key(self, session):
            unit = self.parent.unit.get(session)
            return unit == "b" and "bdepth" or "mdepth"

        def render_value(self, session, value):
            if value is None:
                return fmt_none_brief()
            else:
                return value

class TopQueueSet(TopTable):
    def __init__(self, app, name):
        super(TopQueueSet, self).__init__(app, name)

        col = self.NameColumn(app, "name")
        self.add_column(col)

        self.set_default_column(col)

        col = self.EnqueuesColumn(app, "enqueues")
        col.alignment = "right"
        self.add_column(col)

    class NameColumn(TopTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            reg = Identifiable(data["registration_id"])
            queue = Identifiable(data["id"])

            branch = session.branch()
            self.page.main.broker.object.set(session, reg)
            self.page.main.broker.queue.object.set(session, queue)
            self.page.main.broker.queue.show(session)
            return fmt_link(branch.marshal(), fmt_shorten(data["name"]))

    class EnqueuesColumn(TopTableColumn):
        def render_title(self, session, data):
            return "Recent Enqueues"

class QueueFrame(CuminFrame):
    def __init__(self, app, name):
        super(QueueFrame, self).__init__(app, name)

        self.object = QueueParameter(app, "id")
        self.add_parameter(self.object)

        self.view = QueueView(app, "view")
        self.add_mode(self.view)
        self.set_view_mode(self.view)

        self.purge = QueuePurge(app, "purge")
        self.add_mode(self.purge)
        
        self.binding_add = QueueBindingAdd(app, "bindingadd")
        self.add_mode(self.binding_add)

        self.remove = QueueRemove(app, "remove")
        self.add_mode(self.remove)
        self.set_remove_mode(self.remove)

        self.move_messages = QueueMoveMessages(app, "move_messages")
        self.add_mode(self.move_messages)

    def show_object(self, session, queue):
        reg = queue.vhost.broker.registration
        self.page.main.broker.set_object(session, reg)

        return super(QueueFrame, self).show_object(session, queue)

    def render_href(self, session, queue):
        if queue:
            return super(QueueFrame, self).render_href(session, queue)

    def render_title(self, session, queue):
        if queue:
            return super(QueueFrame, self).render_title(session, queue)
        else:
            return "Queue"

    def show_move(self, session):
        return self.show_mode(session, self.move_messages)

class QueueStatus(CuminStatus):
    def render_consumers(self, session, queue):
        return self.app.model.queue.consumers.value(queue)

    def render_messages_enqueued(self, session, queue):
        return self.app.model.queue.msgTotalEnqueues.rate_html(queue)

    def render_messages_dequeued(self, session, queue):
        return self.app.model.queue.msgTotalDequeues.rate_html(queue)

    def render_message_depth(self, session, queue):
        return self.app.model.queue.msgDepth.value(queue)

    def render_message_depth_accel(self, session, queue):
        return self.app.model.queue.msgDepth.rate_html(queue)

    def render_bytes_enqueued(self, session, queue):
        return self.app.model.queue.byteTotalEnqueues.rate_html(queue)

    def render_bytes_dequeued(self, session, queue):
        return self.app.model.queue.byteTotalDequeues.rate_html(queue)

    def render_byte_depth(self, session, queue):
        return self.app.model.queue.byteDepth.value(queue)

    def render_byte_depth_accel(self, session, queue):
        return self.app.model.queue.byteDepth.rate_html(queue)

class QueueView(CuminView):
    def __init__(self, app, name):
        super(QueueView, self).__init__(app, name)

        status = QueueStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        self.__tabs.add_tab(QueueStats(app, "stats"))

        self.bindings = QueueBindingSet(app, "bindings")
        self.__tabs.add_tab(self.bindings)

        self.details = CuminDetails(app, "details")
        self.__tabs.add_tab(self.details)

    def show_bindings(self, session):
        self.__tabs.show_mode(session, self.bindings);

    def render_script(self, session, queue):
        data = "model.xml?class=queue;id=%i" % queue.id
        return "wooly.setIntervalUpdate('%s', updateQueue, 3000)" % data

class QueueBindingSet(BindingSet, Form):
    def __init__(self, app, name):
        super(QueueBindingSet, self).__init__(app, name)

        self.__remove = self.Remove(app, "remove")
        self.add_child(self.__remove)
        
        self.set_default_column_name("e_id")
        
    def get_visible_columns(self, session):
        return self.get_request_visible_columns(session, ["e_id"])

    def render_add_queue_binding_url(self, session, *args):
        branch = session.branch()
        self.frame.binding_add.show(branch)
        return branch.marshal()
    
    def render_sql_where(self, session, queue):
        elems = list()
        elems.append("b.queue_id = %(id)r")
        elems.append(self.phase.get_sql_constraint(session, queue))
        return "where %s" % " and ".join(elems)

    def render_title(self, session, queue):
        return "Exchange Bindings %s" % \
            fmt_count(queue.bindings.count())

    class Remove(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)

            href = self.frame.frame.bindings_remove.get_href \
                (session, ids)
            self.page.set_redirect_url(session, href)
            
        def render_content(self, session):
            return "Remove"

class QueueForm(CuminFieldForm):
    def __init__(self, app, name):
        super(QueueForm, self).__init__(app, name)

        self.namef = NameField(app, "name")
        self.add_field(self.namef)

        self.more = MoreFieldSet(app, "more")
        self.add_field(self.more)
        
        self.durable = self.QueueDurabilityField(app, "durable")
        self.more.add_field(self.durable)
        
        self.cluster_durable = self.ClusterDurabilityField(app, "cluster_durable")
        self.more.add_field(self.cluster_durable)
        
        self.lvq = self.LVQField(app, "lvq")
        self.more.add_field(self.lvq)
        
        self.optimistic = self.OptimisticField(app, "optimistic")
        self.more.add_field(self.optimistic)

        self.file_count = self.FileCountField(app, "file_count")
        self.file_count.input.param.default = 8
        self.more.add_field(self.file_count)
        
        self.file_size = self.FileSizeField(app, "file_size")
        self.file_size.input.param.default = 24
        self.more.add_field(self.file_size)
        
        self.policy = self.PolicyField(app, "policy")
        self.more.add_field(self.policy)
        
        self.q_size = self.QSizeField(app, "q_size")
        self.more.add_field(self.q_size)
        
        self.q_count = self.QCountField(app, "q_count")
        self.more.add_field(self.q_count)
        
        self.bindings = ExchangeKeysField(app, "bindings")
        self.add_field(self.bindings)
        
    def validate(self, session, queue_name):
        super_error = super(QueueForm, self).validate(session)
        (errors, form_binding_info) = self.bindings.get_binding_errors(session, queue_name)
        return (errors or super_error, form_binding_info)
        
    class QCountField(IntegerField):
        def render_title(self, session):
            return "<div style=\"padding-left: 1em;\">Max Queue Count</div>"
        
        def render_field_help(self, session):
            return "(Maximum in-memory queue size as a number of messages. Applies if Policy is set.)"
        
    class QSizeField(IntegerField):
        def render_title(self, session):
            return "<div style=\"padding-left: 1em;\">Max Queue Size</div>"
        
        def render_field_help(self, session):
            return "(Maximum in-memory queue size as bytes. Applies if Policy is set.)"
        
    class FileCountField(IntegerField):
        def render_title(self, session):
            return "File Count"
        
        def render_field_help(self, session):
            return "(Number of files in queue's persistence journal)"
        
    class FileSizeField(IntegerField):
        def render_title(self, session):
            return "File Size"
        
        def render_field_help(self, session):
            return "(File size in pages - 64Kb/page)"
    
    class QueueDurabilityField(TwoOptionRadioField):
        def render_title(self, session):
            return "Durable?"
        
        def render_field_help(self, session):
            return "(Queue is durable)"

        def render_title_1(self, session):
            return "Durable"

        def render_title_2(self, session):
            return "Transient"

    class ClusterDurabilityField(TwoOptionRadioField):
        def render_title(self, session):
            return "Cluster Durable?"
        
        def render_field_help(self, session):
            return "(Queue becomes durable if there is only one functioning cluster node)"
            
        def render_title_1(self, session):
            return "Cluster Durable"

        def render_title_2(self, session):
            return "Not Cluster Durable"

    class LVQField(TwoOptionRadioField):
        def render_title(self, session):
            return "Enable Last Value Queue?"
        
        def render_field_help(self, session):
            return "(Enable LVQ behavior on the queue)"
            
        def render_title_1(self, session):
            return "Enabled"

        def render_title_2(self, session):
            return "Not Enabled"

    class OptimisticField(TwoOptionRadioField):
        def render_title(self, session):
            return "Enable Optimistic Consume?"
        
        def render_field_help(self, session):
            return "(Enable optimistic consume on the queue)"
            
        def render_title_1(self, session):
            return "Enabled"

        def render_title_2(self, session):
            return "Not Enabled"

    class PolicyField(RadioField):
        def __init__(self, app, name):
            super(QueueForm.PolicyField, self).__init__(app, name, None)
    
            self.param = Parameter(app, "param")
            self.param.default = "none"
            self.add_parameter(self.param)
    
            option = self.NoneField(app, "none", self.param)
            self.add_option(option)
            
            option = self.Reject(app, "reject", self.param)
            self.add_option(option)
            
            option = self.Flow(app, "flow", self.param)
            self.add_option(option)
    
            option = self.Ring(app, "ring", self.param)
            self.add_option(option)
    
            option = self.RingStrict(app, "ring_strict", self.param)
            self.add_option(option)
    
        def render_title(self, session):
            return "Policy-type"
    
        def render_field_help(self, session):
            return "(Action taken when queue limit is reached)"
            
        class NoneField(RadioFieldOption):
            def render_value(self, session):
                return "none"
    
            def render_title(self, session):
                return "None"
    
        class Reject(RadioFieldOption):
            def render_value(self, session):
                return "reject"
    
            def render_title(self, session):
                return "Reject"
    
        class Flow(RadioFieldOption):
            def render_value(self, session):
                return "flow"
    
            def render_title(self, session):
                return "Flow to disc"
    
        class Ring(RadioFieldOption):
            def render_value(self, session):
                return "ring"
    
            def render_title(self, session):
                return "Ring"
    
        class RingStrict(RadioFieldOption):
            def render_value(self, session):
                return "ring_strict"
    
            def render_title(self, session):
                return "Ring Strict"

class QueueAdd(QueueForm):
    def get_args(self, session):
        return (self.frame.get_object(session), )

    def process_cancel(self, session, *args):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session, reg):
        queue_name = self.namef.get(session)
        durable = self.durable.get(session)

        (errors, form_binding_info) = self.validate(session, queue_name)
        if errors:
            pass
        else:
            # since we are adding a queue, we don't have
            # an existing queue to pass, so create a
            # blank one
            queue = Queue()
            queue.name = queue_name
            queue.durable = (durable == "yes")

            args = {}
            args["reg"] = reg
            args["exchange_keys"] = form_binding_info
            args["cluster_durable"] = self.cluster_durable.get(session) == "yes"
            args["lvq"] = self.lvq.get(session) == "enable"
            args["optimistic"] = self.optimistic.get(session) == "yes"
            policy = self.policy.get(session)
            if policy == "none":
                policy = None
            args["policy"] = policy
            
            file_count = self.file_count.get(session)
            file_size = self.file_size.get(session)
            if file_count == self.file_count.input.param.default:
                file_count = None
            if file_size == self.file_size.input.param.default:
                file_size = None
            args["file_count"] = file_count
            args["file_size"] = file_size

            args["q_count"] = self.q_count.get(session)
            args["q_size"] = self.q_size.get(session)
            
            action = self.app.model.broker.add_queue
            action.invoke(queue, args)
            
            # navigate back to main queue frame
            self.process_cancel(session)

    def render_title(self, session, reg):
        return "Add Queue to the Broker '%s'" % reg.name

class QueueRemove(CuminConfirmForm):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, queue):
        return "Remove Queue '%s'" % queue.name

    def process_submit(self, session, queue):
        action = self.app.model.queue.remove
        action.invoke(queue)

        self.process_cancel(session, queue)

    def render_submit_content(self, session, queue):
        return "Yes, Remove Queue '%s'" % queue.name

    def render_cancel_content(self, session, queue):
        return "No, Cancel"

class QueuePurge(CuminFieldForm):
    def __init__(self, app, name):
        super(QueuePurge, self).__init__(app, name)

        self.purge_request = MultiplicityField(app, "purge_request")
        self.add_field(self.purge_request)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, queue):
        return "Purge Messages from  Queue '%s'" % queue.name

    def process_submit(self, session, queue):
        request_amt = self.purge_request.get(session)
        args = dict()

        if request_amt == "all":
            args["request"] = 0
        elif request_amt == "top":
            args["request"] = 1
        elif request_amt == "N":
            args["request"] = self.purge_request.top_n.get_n_value(session)
        else:
            raise Exception("Wrong Value")

        action = self.app.model.queue.purge
        action.invoke(queue, args)

        self.process_cancel(session, queue)

    def render_submit_content(self, session, queue):
        return "Yes, Purge Messages from Queue '%s'" % queue.name

    def render_cancel_content(self, session, queue):
        return "No, Cancel"

class QueueSetPurge(CuminBulkActionForm):
    def process_return(self, session):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_current_frame(branch, self.frame)
        self.page.set_redirect_url(session, branch.marshal())

    def process_item(self, session, id):
        queue = Queue.get(id)

        args = dict()
        args["request"] = 0

        action = self.app.model.queue.purge
        action.invoke(queue, args)

    def render_title(self, session):
        return "Purge Queues"

    def render_item_content(self, session, id):
        return "Purge Queue '%s'" % Queue.get(id).name


class QueueSetRemove(CuminBulkActionForm):
    def process_return(self, session):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_current_frame(branch, self.frame)
        self.page.set_redirect_url(session, branch.marshal())

    def process_item(self, session, id):
        queue = Queue.get(id)
        action = self.app.model.queue.remove
        action.invoke(queue)
 
    def render_title(self, session):
        return "Remove Queues"

    def render_item_content(self, session, id):
        return "Remove Queue '%s'" % Queue.get(id).name

class BindSummaryPropertiesField(FormField):
    def __init__(self, app, name):
        super(BindSummaryPropertiesField, self).__init__(app, name)

        self.sum_props = self.SummaryProperties(app, "properties")
        self.add_child(self.sum_props)

        self.prop_tmpl = Template(self, "properties_html")
        
    class SummaryProperties(CuminProperties):
        def do_get_items(self, session, queue):
            return [("Name", queue.name),
                    ("Durable", queue.durable),
                    ("Exclusive", queue.exclusive),
                    ("Auto-Delete", queue.autoDelete)]

    def render_inputs(self, session, *args):
        writer = Writer()
        self.prop_tmpl.render(writer, session, args)
        return writer.to_string()

    def render_prop_items(self, session, args):
        return self.sum_props.render_items(session, *args)

    def get_args(self, session):
        return (self.frame.frame.get_object(session),)

class QueueBindingAdd(CuminFieldForm):
    def __init__(self, app, name):
        super(QueueBindingAdd, self).__init__(app, name)

        self.props = BindSummaryPropertiesField(app, "props")
        self.add_field(self.props)

        self.bindings = ExchangeKeysField(app, "bindings", title="Bind to Exchange:")
        self.add_field(self.bindings)
        
        self.errors = self.Errors(self, "errors")
        self.add_attribute(self.errors)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_form_error(self, session, *args):
        errors = self.errors.get(session)
        if "no_exchanges" in errors:
            return "<ul class=\"errors\" style=\"margin:0; float:left;\"><li>%s</li></ul>" % \
                "</li><li>".join(errors["no_exchanges"])
        
    def process_cancel(self, session, *args):
        branch = session.branch()
        self.frame.show_view(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session, queue):
        (errors, form_binding_info) = self.bindings.get_binding_errors(session, queue.name)

        if not len(form_binding_info):
            # no exhchanges were selected is not an
            # error that ExchangeKeysField looks for
            errors = self.errors.get(session)
            errs = errors.setdefault("no_exchanges", list())
            errs.append("At least one exchange must be selected")
            errors = True
            
        if errors:
            pass
        else:
            reg = self.frame.frame.get_object(session)
            args = {
                    "exchange_keys": form_binding_info,
                    "reg": reg}
            
            action = self.app.model.queue.bind
            action.invoke(queue, args)
            
            # navigate back to main queue frame
            self.process_cancel(session)

    def render_title(self, session, *args):
        queue = self.frame.get_object(session)
        return "Add Binding to the Queue '%s'" % queue.name

    class Errors(Attribute):
        def get_default(self, session):
            return dict()
        
class QueueBindingRemove(CuminConfirmForm):
    def get_args(self, session):
        return self.frame.get_args(session)

    def process_submit(self, session, binding):
        log.info("Removing binding %s" % binding)
        
        self.process_cancel(session, binding)

    def render_title(self, session, binding):
        return "Remove Binding"

    def render_submit_content(self, session, binding):
        return "Yes, Remove Binding"

    def render_cancel_content(self, session, binding):
        return "No, Cancel"

class QueueStats(TabbedModeSet):
    def __init__(self, app, name):
        super(QueueStats, self).__init__(app, name)

        self.add_tab(QueueStatsGeneral(app, "gen"))
        self.add_tab(QueueStatsDurability(app, "dur"))
        self.add_tab(QueueStatsTransactions(app, "txn"))

    def render_title(self, session):
        return "Statistics"

class QueueStatsGeneral(Widget):
    def __init__(self, app, name):
        super(QueueStatsGeneral, self).__init__(app, name)

        self.add_child(StatSet(app, "general", "general"))
        self.add_child(StatSet(app, "io", "io"))

        chart = self.EnqueueDequeueRateChart(app, "enqdeq")
        self.add_child(chart)

        chart = self.DepthChart(app, "depth")
        self.add_child(chart)

        chart = StatValueChart(app, "consumers")
        chart.stats = ("consumerCount",)
        self.add_child(chart)

    def render_title(self, session):
        return "General"

    class EnqueueDequeueRateChart(StatValueChart):
        def __init__(self, app, name):
            super(QueueStatsGeneral.EnqueueDequeueRateChart,
                  self).__init__(app, name)

            self.stats = ("msgTotalEnqueues", "msgTotalDequeues")
            self.mode = "rate"

        def render_title(self, session, queue):
            return "Messages Enqueued and Dequeued"

    class DepthChart(StatValueChart):
        def __init__(self, app, name):
            super(QueueStatsGeneral.DepthChart, self).__init__(app, name)

            self.stats = ("msgDepth",)

        def render_title(self, session, queue):
            return "Queue Depth"

class QueueStatsDurability(Widget):
    def __init__(self, app, name):
        super(QueueStatsDurability, self).__init__(app, name)

        self.add_child(StatSet(app, "io", "io.durable"))
        self.add_child(JournalStats(app, "jrnl", "io.journal"))

        chart = self.EnqueueDequeueRateChart(app, "enqdeq")
        self.add_child(chart)

    def render_title(self, session):
        return "Durability"

    class EnqueueDequeueRateChart(StatValueChart):
        def __init__(self, app, name):
            super(QueueStatsDurability.EnqueueDequeueRateChart,
                  self).__init__(app, name)

            self.stats = ("msgPersistEnqueues", "msgPersistDequeues")
            self.mode = "rate"

        def render_title(self, session, queue):
            return "Durable Messages Enqueued and Dequeued"

class JournalStats(StatSet):
    def get_args(self, session):
        queue = self.frame.get_args(session)[0]

        try:
            jrnl = Journal.selectBy(queue=queue)[0]
        except IndexError:
            jrnl = None

        return (jrnl,)

    def render_title(self, session, jrnl):
        return "Journal"

    def do_render(self, session, jrnl):
        if jrnl:
            return super(JournalStats, self).do_render(session, jrnl)
        else:
            return "<div class=\"iblock\">%s</div>" % fmt_none()

class QueueStatsTransactions(Widget):
    def __init__(self, app, name):
        super(QueueStatsTransactions, self).__init__(app, name)

        self.add_child(StatSet(app, "transactions", "txn"))
        self.add_child(StatSet(app, "io", "io.txn"))

        chart = self.EnqueueTransactionRateChart(app, "enqtxn")
        self.add_child(chart)
        
        chart = self.DequeueTransactionRateChart(app, "deqtxn")
        self.add_child(chart)
        
        chart = self.EnqueueDequeueRateChart(app, "enqdeq")
        self.add_child(chart)
        
    def render_title(self, session):
        return "Transactions"

    class EnqueueTransactionRateChart(StatValueChart):
        def __init__(self, app, name):
            super(QueueStatsTransactions.EnqueueTransactionRateChart,
                  self).__init__(app, name)

            self.stats = ("enqueueTxnStarts", "enqueueTxnCommits",
                          "enqueueTxnRejects")
            self.mode = "rate"

        def render_title(self, session, queue):
            return "Enqueue Transaction Operations per Second"

    class DequeueTransactionRateChart(StatValueChart):
        def __init__(self, app, name):
            super(QueueStatsTransactions.DequeueTransactionRateChart,
                  self).__init__(app, name)

            self.stats = ("dequeueTxnStarts", "dequeueTxnCommits",
                          "dequeueTxnRejects")
            self.mode = "rate"

        def render_title(self, session, queue):
            return "Dequeue Transaction Operations per Second"

    class EnqueueDequeueRateChart(StatValueChart):
        def __init__(self, app, name):
            super(QueueStatsTransactions.EnqueueDequeueRateChart,
                  self).__init__(app, name)

            self.stats = ("msgTxnEnqueues", "msgTxnDequeues")
            self.mode = "rate"

        def render_title(self, session, queue):
            return "Transactional Messages Enqueued and Dequeued"

class QueueConsumerSet(PaginatedItemSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, queue):
        count = fmt_count(self.get_item_count(session, queue))
        return "Consumers %s" % count

    def get_item_count(self, session, queue):
        return queue.consumers.count()

    def do_get_items(self, session, queue):
        start, end = self.get_bounds(session)
        return queue.consumers[start:end]

    def render_item_name(self, session, consumer):
        return consumer.name

    def render_item_messages_consumed(self, session, consumer):
        return self.app.model.consumer.msgsConsumed.value(consumer)

    def render_item_messages_consumed_rate(self, session, consumer):
        return self.app.model.consumer.msgsConsumed.rate_html(consumer)

    def render_item_bytes_consumed(self, session, consumer):
        return self.app.model.consumer.bytesConsumed.value(consumer)

    def render_item_bytes_consumed_rate(self, session, consumer):
        return self.app.model.consumer.bytesConsumed.rate_html(consumer)

    def render_item_unacked_messages(self, session, consumer):
        return self.app.model.consumer.unackedMessages.value(consumer)


class QueueSelectField(FormField):
    def __init__(self, app, name, form):
        super(QueueSelectField, self).__init__(app, name)

        param = QueueParameter(app, "param")
        self.param = param
        self.add_parameter(param)
        
        self.queue_set = self.QueueInputSet(app, "queue_set", param)
        self.add_child(self.queue_set)
        
    def get(self, session):
        return self.param.get(session)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, queue):
        return "Select the destination queue"

    def render_inputs(self, session, queue):
        return self.queue_set.render(session)

    class QueueInputSet(RadioInputSet):
        def get_args(self, session):
            return self.frame.get_args(session)

        def do_get_items(self, session, queue):
            queue_list_full = sorted_by(list(queue.vhost.queues))
            delta = timedelta(minutes=10)
            queue_list = []
            for _queue in queue_list_full:
                if (_queue.qmfUpdateTime > (datetime.now() - delta)) and (_queue.name != queue.name):
                    queue_list.append(_queue)
            return queue_list
        
        def render_item_value(self, session, queue):
            return queue.id
        
        def render_item_content(self, session, queue):
            return queue.name or "<em>Default</em>"
        
        def render_item_checked_attr(self, session, queue):
            return queue is self.param.get(session) and "checked=\"checked\"" or None

class QueueMoveMessages(CuminFieldForm):
    def __init__(self, app, name):
        super(QueueMoveMessages, self).__init__(app, name)

        self.move_to = QueueSelectField(app, "move_to", self)
        self.add_field(self.move_to)

        self.move_qty = MultiplicityField(app, "move_qty")
        self.add_field(self.move_qty)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, queue):
        return "Move Messages from  Queue '%s'" % queue.name

    def process_submit(self, session, queue):
        request_qty = self.move_qty.get(session)
        args = dict()
        args["destQueue"] = self.move_to.get(session).name
        args["srcQueue"] = queue.name

        if request_qty == "all":
            args["qty"] = 0
        elif request_qty == "top":
            args["qty"] = 1
        elif request_qty == "N":
            args["qty"] = self.move_qty.top_n.get_n_value(session)
        else:
            raise Exception("Wrong Value")

        action = self.app.model.queue.move_messages
        action.invoke(queue, args)
        self.process_cancel(session, queue)

    def render_submit_content(self, session, queue):
        return "Yes, Move Messages from Queue '%s'" % queue.name

    def render_cancel_content(self, session, queue):
        return "No, Cancel"

