import logging

from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *
from wooly.tables import *

from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *
from job import *

strings = StringCatalog(__file__)
log = logging.getLogger("cumin.scheduler")

class SchedulerSet(CuminTable, Form):
    def __init__(self, app, name):
        super(SchedulerSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.NameColumn(app, "name")
        self.add_column(col)
        self.set_default_column(col)

        col = self.SystemColumn(app, "system")
        self.add_column(col)

        col = self.UsersColumn(app, "users")
        col.alignment = "right"
        self.add_column(col)

        col = self.RunningJobsColumn(app, "running")
        col.alignment = "right"
        self.add_column(col)

        col = self.HeldJobsColumn(app, "held")
        col.alignment = "right"
        self.add_column(col)

        self.__start = self.StartButton(app, "start")
        self.add_child(self.__start)

        self.__stop = self.StopButton(app, "stop")
        self.add_child(self.__stop)

    def render_title(self, session, *args):
        return "Schedulers %s" % fmt_count(Scheduler.select().count())

    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            sched = Identifiable(data["id"])
            href = self.page.main.pool.scheduler.get_href(session, sched)
            return fmt_link(href, data["name"])

    class UsersColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Users"

    class RunningJobsColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Running Jobs"

    class HeldJobsColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Held Jobs"

    class SystemColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "System"

        def render_content(self, session, data):
            id = data["system_id"]

            if id:
                sys = Identifiable(id)
                href = self.page.main.system.get_href(session, sys)
                return fmt_link(href, data["system_name"])

    class StartButton(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)
    
            branch = session.branch()
            frame = self.frame.show_scheds_start(branch)
            frame.ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())
            
        def render_content(self, session):
            return "Start"
        
    class StopButton(FormButton):
        def process_submit(self, session):
            ids = self.parent.ids.get(session)
            self.parent.ids.clear(session)
    
            branch = session.branch()
            frame = self.frame.show_scheds_stop(branch)
            frame.ids.set(branch, ids)
            self.page.set_redirect_url(session, branch.marshal())
            
        def render_content(self, session):
            return "Stop"
        
class SchedulerFrame(CuminFrame):
    def __init__(self, app, name):
        super(SchedulerFrame, self).__init__(app, name)

        self.object = SchedulerParameter(app, "id")
        self.add_parameter(self.object)

        view = SchedulerView(app, "view")
        self.add_mode(view)
        self.set_view_mode(view)

class SchedulerView(CuminView):
    def __init__(self, app, name):
        super(SchedulerView, self).__init__(app, name)

        status = self.SchedulerStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        stats = SchedulerStats(app, "stats")
        self.__tabs.add_tab(stats)

        jobs = SchedulerJobSet(app, "jobs")
        self.__tabs.add_tab(jobs)
        
        details = CuminDetails(app, "details")
        self.__tabs.add_tab(details)

    def render_script(self, session, sched):
        data = "model.xml?class=scheduler;id=%i" % sched.id
        return "wooly.setIntervalUpdate('%s', updateScheduler, 3000)" % data

    class SchedulerStatus(StartStopStatus):
        def get_url(self, session):
            scheduler = self.get_args(session)[0]
            if scheduler:
                return "call.xml?class=scheduler;id=%i;method=GetStarted" % scheduler.id
    
        def render_title(self, session, *args):
            return "Scheduler Status"
        
class SchedulerStats(Widget):
    def __init__(self, app, name):
        super(SchedulerStats, self).__init__(app, name)

        stats = StatSet(app, "general", "general")
        self.add_child(stats)

        chart = self.UsersChart(app, "users")
        chart.duration.param.default = "h"
        self.add_child(chart)

        chart = self.JobsChart(app, "jobs")
        chart.duration.param.default = "h"
        self.add_child(chart)

    def render_title(self, session):
        return "Statistics"

    class UsersChart(StatValueChart):
        def __init__(self, app, name):
            super(SchedulerStats.UsersChart, self).__init__(app, name)

            self.stats = ("NumUsers",)

        def render_title(self, session, sched):
            return "Users"

    class JobsChart(StatValueChart):
        def __init__(self, app, name):
            super(SchedulerStats.JobsChart, self).__init__ \
                (app, name)

            self.stats = ("TotalRunningJobs", "TotalIdleJobs", "TotalHeldJobs")

        def render_title(self, session, sched):
            return "Jobs"

class SchedulerJobSet(JobTab):
    def __init__(self, app, name):
        super(SchedulerJobSet, self).__init__(app, name)
        
    def get_visible_columns(self, session):
        return self.get_request_visible_columns(session, ["custom_group", "submitter"])

    def render_sql_where(self, session, scheduler):
        phase_sql = self.get_phase_sql(session)
        scheduler_sql = "j.scheduler_id = %i" % scheduler.id
        return "where %s" % " and ".join([phase_sql, scheduler_sql])

    def render_title(self, session, scheduler):
        where_scheduler = "scheduler_id = %i" % scheduler.id
        return "Jobs %s" % fmt_count(Job.select(where_scheduler).count())

class SchedulerStart(CuminBulkActionForm):
    def render_title(self, session, *args):
        return "Start Scheduler"
    
    def render_form_heading(self, session, *args):
        return ""

    def render_item_content(self, session, id):
        return "Start Scheduler '%s'" % Scheduler.get(id).Name

    def process_item(self, session, item):
        scheduler = Scheduler.get(item)
        action = self.app.model.master.start
        action.invoke(scheduler, {"subsystem": "SCHEDD"})
        self.process_cancel(session)

class SchedulerStop(CuminBulkActionForm):
    def render_title(self, session, *args):
        return "Stop Scheduler"
    
    def render_form_heading(self, session, *args):
        return ""

    def render_item_content(self, session, id):
        return "Stop Scheduler '%s'" % Scheduler.get(id).Name

    def process_item(self, session, item):
        scheduler = Scheduler.get(item)
        action = self.app.model.master.stop
        action.invoke(scheduler, {"subsystem": "SCHEDD"})
        self.process_cancel(session)

