from wooly import *
from wooly.widgets import *
from mint import *
from math import sqrt

from widgets import *
from parameters import *
from util import *
from formats import *

strings = StringCatalog(__file__)

class StatSet(ItemSet):
    unit_abbrevs = dict()
    unit_abbrevs["message"] = "msg"
    unit_abbrevs["transaction"] = "trans"
    
    def __init__(self, app, name, category):
        super(StatSet, self).__init__(app, name)

        self.category = category

        self.object = Attribute(app, "object");
        self.add_attribute(self.object);

    def get_args(self, session):
        return self.frame.get_args(session)

    def do_get_items(self, session, object):
        stats = list()
        cls = self.app.model.get_class_by_object(object)

        for stat in cls.stats:
            if self.category == stat.category:
                stats.append((stat, object))
        
        return stats

    def render_item_title(self, session, args):
        stat, object = args
        return stat.title

    def render_rate_text(self, session, args):
        return "Per Second"
    
    def render_item_name(self, session, args):
        stat, object = args
        return stat.name

    def render_item_value(self, session, args):
        stat, object = args
        
        #if stat.link_cb:
        #    branch = session.branch()
        #    stat.link_cb(self.page, branch, object)
        #    return fmt_link(branch.marshal(), stat.value(object))
        #else:

        return stat.value_html(object)

    def render_item_rate(self, session, args):
        stat, object = args
        
        #if False and stat.highlow: #XXX
        #    return "<small>high</small>&nbsp;<span>%i</span> " + \
        #           "<small>low</small>&nbsp;<span>%i</span>" \
        #           % (stat.high(object) or 0, stat.low(object) or 0)
        #else:

        return stat.rate_html(object)

class StatValueChart(Widget):
    def __init__(self, app, name):
        super(StatValueChart, self).__init__(app, name)

        self.mode = None

        self.stats = ()

        self.stats_tmpl = Template(self, "stat_html")

        self.duration = DurationSwitch(app, "duration")
        self.add_child(self.duration)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_href(self, session, object):
        params = list()

        cls = self.app.model.get_class_by_object(object).cumin_name

        params.append("class=%s" % cls)

        params.append("id=%i" % object.id)

        for stat in self.stats:
            params.append("stat=%s" % stat)

        duration = self.duration.get(session)

        if duration == "h":
            params.append("duration=%i" % 3600)
        elif duration == "d":
            params.append("duration=%i" % 86400)

        if self.mode:
            params.append("mode=%s" % self.mode)

        return "stats.png?" + ";".join(params)

    def render_title(self, session, object):
        cls = self.app.model.get_class_by_object(object)
        return getattr(cls, self.stats[0]).title

    def render_stats(self, session, object):
        writer = Writer()

        for name in self.stats:
            cls = self.app.model.get_class_by_object(object)
            stat = getattr(cls, name)
            self.stats_tmpl.render(writer, session, stat, object)

        return writer.to_string()

    def render_stat_mode(self, session, stat, object):
        return self.mode

    def render_stat_color(self, session, stat, object):
        return "rgb(255,0,0)"

    def render_stat_value(self, session, stat, object):
        if self.mode == "rate":
            html = stat.rate_html(object)
        else:
            html = stat.value_html(object)

        return html

    def render_stat_name(self, session, stat, object):
        return stat.name

class StatUtilizationGrid(Widget):
    def __init__(self, app, name):
        super(StatUtilizationGrid, self).__init__(app, name)

        self.cells = self.GridCells(app, "grid")
        self.add_child(self.cells)
        
        legend = self.Legend(app, "grid_legend")
        legend.get_method = self.get_colors
        self.add_child(legend)
        
        legend1 = self.Legend(app, "grid_legend1")
        legend1.get_method = self.get_states
        self.add_child(legend1)
        
        ajax = self.Updater(app, "grid_updater")
        self.add_child(ajax)
        
        self.max_px = 480.0
        self.max_item_px = 28.0
        self.max_columns = 20.0
        self.max_background_cols = 15 # max columns with background image

    def render_title(self, session):
        return "Utilization"
    
    def render_name(self, session):
        return self.name
    
    def render_vis_help(self, session):
        pass
    
    def will_it_fit(self, session):
        count = len(self.cells.get_items(session))
        columns = self.cells.calculate_columns(count)
        return columns <= self.max_columns
    
    def get_cell_class(self, session):
        count = len(self.cells.get_items(session))
        columns = self.cells.calculate_columns(count)
        return columns > self.max_background_cols and "small" or "big"
    
    def get_grid_width(self, session):
        count = len(self.cells.get_items(session))
        columns = self.cells.calculate_columns(count)
        # we have approx 480px to work with
        width = float(columns) * self.max_item_px
        if width > self.max_px:
            width = self.max_px
        extra = 3.0 * columns # left right border plus 1 right margin
        if columns > self.max_background_cols:
            extra = 0
        return int(width + extra)
    
    def render_grid_width(self, session):
        width = self.get_grid_width(session) or self.max_px
        return "%ipx" % int(width)

    class GridCells(ItemSet):
        def __init__(self, app, name):
            super(StatUtilizationGrid.GridCells, self).__init__(app, name)
            
            self.width = Attribute(app, "width")
            self.add_attribute(self.width)

            self.sticky = self.Sticky(app, "sticky_info")
            self.add_child(self.sticky)

        def render(self, session, *args):
            count = len(self.get_items(session))
            if count == 0:
                return "0 slots found"
            else:
                return super(StatUtilizationGrid.GridCells, self).render(session, *args)

        def do_get_items(self, session):
            cells = self.parent.get_cells(session)
            self.width.set(session, 
                           self.calculate_cell_width(len(cells)))
            return cells
        
        def render_cell_id(self, session, cell):
            return cell["id"]

        def render_cell_width(self, session, cell):
            return self.width.get(session)
        
        def render_cell_class(self, session, cell):
            return self.parent.get_cell_class(session)
        
        def calculate_columns(self, count):
            sq = sqrt(count)
            isq = int(sq)
            if sq > isq:
                isq = isq + 1
            return isq
            
        def calculate_cell_width(self, count):
            columns = self.calculate_columns(count)
            width = self.parent.max_item_px
            if columns:
                min = self.parent.max_px / float(columns)
                width = min > self.parent.max_item_px and self.parent.max_item_px or min
            return "%ipx" % int(width)

        def render_href(self, session, cell):
            return self.parent.get_href(session, cell)
        
        def render_color(self, session, cell):
            return self.parent.get_color(session, cell)
        
        def render_contents(self, session, cell):
            return self.parent.get_contents(session, cell)
    
        def get_sticky_info(self, session):
            return self.parent.get_sticky_info(session)

        def render_sticky_rows(self, session, cell):
            return self.sticky.render_items(session, cell)
    
        class Sticky(ItemSet):
            def do_get_items(self, session, cell):
                info = self.parent.get_sticky_info(session)
                return [(inf, cell["id"]) for inf in info]
            
            def render_sticky_name(self, session, item):
                return item[0][0]
            
            def render_sticky_title(self, session, item):
                return item[0][1]
            
            def render_sticky_object_id(self, session, item):
                return item[1]
        
    def get_cells(self, session):
        """ should return a list of dictionaries """
        return list()
    
    def get_href(self, session, cell):
        return "#"

    def get_color(self, session, cell):
        return "clear"
    
    def get_contents(self, session, cell):
        return ""

    def get_colors(self, session):
        return ["clear"]
        
    def get_states(self, session):
        return ["Unclaimed"]
        
    def get_url(self, session):
        """ returns something like
            "call.xml?class=system;id=%i;method=slots" % system.id
        """
        pass

    def get_sticky_info(self, session):
        return [("Name", "Display Title")]
    
    def get_fn(self, session):
        return self.name
    
    def got_fn(self, session):
        return self.name
    
    def elem_id(self, session):
        return self.name

    class Legend(ItemSet):
        def render(self, session, *args):
            if len(self.parent.cells.get_items(session)) > 0:
                return super(StatUtilizationGrid.Legend, self).render(session, *args)
            
        def do_get_items(self, session, *args):
            return self.get_method(session)

        def render_legend_text(self, session, color):
            return color[0]
        
        def render_legend_color(self, session, color):
            return color[1]
    
    class Updater(AjaxField):
        def do_render(self, session):
            return self.render_script(session)
        
        def get_url(self, session):
            return self.parent.get_url(session)

        def get_fn(self, session):
            return self.parent.get_fn(session)
        
        def got_fn(self, session):
            return self.parent.got_fn(session)
        
        def elem_id(self, session):
            return self.parent.elem_id(session)


class StatChartPage(Page):
    def __init__(self, app, name):
        super(StatChartPage, self).__init__(app, name)

        self.class_ = CuminClassParameter(app, "class")
        self.add_parameter(self.class_)

        self.id = IntegerParameter(app, "id")
        self.add_parameter(self.id)

        param = Parameter(app, "param")
        self.add_parameter(param)

        self.stats = ListParameter(app, "stat", param)
        self.add_parameter(self.stats)

        self.mode = Parameter(app, "mode")
        self.add_parameter(self.mode)

        # In seconds
        self.duration = IntegerParameter(app, "duration")
        self.duration.default = 600
        self.add_parameter(self.duration)

    def get_content_type(self, session):
        return "image/png"

    def get_cache_control(self, session):
        return "no-cache"

    def get_args(self, session):
        cls = self.class_.get(session).mint_class
        id = self.id.get(session)
        return (cls.get(id),)

    def do_render(self, session, object):
        chart = TimeSeriesChart(360, 100)

        cls = self.class_.get(session)
        stats = [getattr(cls, x) for x in self.stats.get(session)]

        samples = dict()
        values = dict()

        duration = self.duration.get(session)

        if self.mode.get(session) == "rate":
            for stat in stats:
                os = stat.samples(object, duration)
                ns = list()
                prev = None

                for sample in reversed(os):
                    if prev is not None:
                        rate = calc_rate(sample[1], prev[1],
                                         secs(sample[0]), secs(prev[0]))

                        ns.insert(0, (sample[0], rate))

                    prev = sample

                samples[stat] = ns
        else:
            for stat in stats:
                samples[stat] = stat.samples(object, duration)

        for stat in stats:
            values[stat] = [x[1] for x in samples[stat]]

        max_value = 1
        min_value = 0

        for stat in stats:
            vals = values[stat]
            if vals:
                max_value = max(nvl(max(vals), 0), max_value)
                min_value = min(nvl(min(vals), 0), min_value)

        max_value = round(max_value * 1.1 + 1)

        if min_value < 0:
            min_value = round(min_value * 1.1 - 1)

        chart.x_max = duration
        chart.x_min = 0
        chart.y_max = max_value
        chart.y_min = min_value

        x_intervals = 8
        x_step = 2

        if chart.x_max == 600:
            x_intervals = 10

        y_intervals = 6
        y_step = 2

        if chart.y_max < 3:
            y_step = 3

        chart.plot_x_axis(x_intervals, x_step)
        chart.plot_y_axis(y_intervals, y_step)

        colors = ((1,0,0), (0,0,1), (0,1,0))

        for stat, color in zip(stats, colors):
            chart.plot_values(samples[stat], color=color)
        
        chart.plot_frame()

        if self.mode.get(session) == "rate":
            titles = ["%s / sec" % x.title for x in stats]
        else:
            titles = [x.title for x in stats]

        chart.plot_legend(titles, colors)

        writer = Writer()
        chart.write(writer)
        return writer.to_string()

class DurationSwitch(StateSwitch):
    def __init__(self, app, name):
        super(DurationSwitch, self).__init__(app, name)

        self.add_state("m", "10 minutes")
        self.add_state("h", "1 hour")
        self.add_state("d", "1 day")
