from mint import *
from wooly import *
from wooly.widgets import *

from stat import *
from widgets import *
from parameters import *
from formats import *
from util import *

strings = StringCatalog(__file__)

class SystemSet(CuminTable, Form):
    def __init__(self, app, name):
        super(SystemSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)

        col = self.NameColumn(app, "name")
        self.add_column(col)
        self.set_default_column(col)

        col = self.KernelColumn(app, "kernel")
        self.add_column(col)

        col = self.ArchColumn(app, "arch")
        self.add_column(col)

        col = self.FreeMemoryColumn(app, "mem_free")
        col.alignment = "right"
        self.add_column(col)

        col = self.LoadColumn(app, "load")
        col.alignment = "right"
        self.add_column(col)
        
    def render_title(self, session, *args):
        return "Systems %s" % fmt_count(self.get_item_count(session, *args))
        
    class NameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            system = Identifiable(data["id"])
            href = self.page.main.system.get_href(session, system)
            return fmt_link(href, fmt_shorten(data["name"]))

    class KernelColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Kernel"

    class ArchColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Arch"

    class FreeMemoryColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Free Memory"

        def render_content(self, session, data):
            mem_free = data["mem_free"]

            if mem_free:
                return "%i KB" % mem_free
            else:
                return fmt_none_brief()

    class LoadColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Load Average"

        def render_content(self, session, data):
            load = data["load"]

            if load:
                return "%0.3f" % load
            else:
                return fmt_none_brief()

class TopSystemSet(TopTable):
    def __init__(self, app, name):
        super(TopSystemSet, self).__init__(app, name)

        col = self.NameColumn(app, "name")
        self.add_column(col)

        self.set_default_column(col)

        col = self.LoadColumn(app, "load")
        col.alignment = "right"
        self.add_column(col)

    class NameColumn(TopTableColumn):
        def render_title(self, session, data):
            return "Name"

        def render_content(self, session, data):
            system = Identifiable(data["id"])
            href = self.page.main.system.get_href(session, system)
            return fmt_link(href, fmt_shorten(data["name"]))

    class LoadColumn(TopTableColumn):
        def render_title(self, session, data):
            return "Load Average"

        def render_content(self, session, data):
            return "%0.3f" % data["load_avg"]

class SystemFrame(CuminFrame):
    def __init__(self, app, name):
        super(SystemFrame, self).__init__(app, name)

        self.object = SystemParameter(app, "id")
        self.add_parameter(self.object)

        self.view = SystemView(app, "view")
        self.add_mode(self.view)
        self.set_view_mode(self.view)

class SystemStatus(CuminStatus):
    def render_mem_free(self, session, system):
        return self.app.model.system.memFree.value_html(system)

    def render_load_average(self, session, system):
        return self.app.model.system.loadAverage1Min.value_html(system)

class SystemStats(Widget):
    def __init__(self, app, name):
        super(SystemStats, self).__init__(app, name)

        self.grid = self.SlotUtilizationGrid(app, "slot_grid")
        self.add_child(self.grid)
        
        self.add_child(StatSet(app, "stats", "general"))

        chart = StatValueChart(app, "freemem")
        chart.stats = ("memFree",)
        self.add_child(chart)

        chart = StatValueChart(app, "loadavg")
        chart.stats = ("loadAverage1Min",)
        self.add_child(chart)

    def render_title(self, session):
        return "Statistics"

    def render_slot_job_url(self, session):
        job = Identifiable("XXX")
        return self.page.main.pool.job.get_href(session, job)

    class SlotUtilizationGrid(StatUtilizationGrid):
        def render(self, session):
            cells = self.get_cells(session)
            if len(cells) > 0:
                return super(SystemStats.SlotUtilizationGrid, self).render(session)

        def get_cells(self, session):
            system = self.frame.get_args(session)[0]
            action = self.app.model.system.slots
            return action.get_boxes(session, system)

        def render_title(self, session):
            return "Slot Utilization"

        def get_colors(self, session):
            action = self.app.model.system.slots
            return action.get_colors()
            
        def get_color(self, session, slot):
            action = self.app.model.system.slots
            return action.get_color(session, slot)
    
        def get_contents(self, session, slot):
            return ""
        
        def get_href(self, session, slot):
            branch = session.branch()
            try:
                job = Job.select("custom_id = '%s'" % slot.JobId)[0]
            except Exception, e:
                return "#"
            return self.page.main.pool.job.get_href(branch, job)

        def get_url(self, session):
            system = self.parent.frame.get_args(session)[0]
            return "call.xml?class=system;id=%i;method=slots" % system.id

        def get_sticky_info(self, session):
            action = self.app.model.system.slots
            return action.get_field_tuples(session)

class SystemView(CuminView):
    def __init__(self, app, name):
        super(SystemView, self).__init__(app, name)

        summary = CuminSummary(app, "summary")
        self.add_child(summary)

        status = SystemStatus(app, "status")
        self.add_child(status)

        self.__tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.__tabs)

        self.__tabs.add_tab(SystemStats(app, "stats"))
        self.__tabs.add_tab(SystemJobSet(app, "jobs"))
        self.__tabs.add_tab(SystemSlotSet(app, "slots"))
        self.__tabs.add_tab(SystemServices(app, "services"))
        self.__tabs.add_tab(CuminDetails(app, "details"))

    def render_script(self, session, system):
        data = "model.xml?class=system;id=%i" % system.id
        return "wooly.setIntervalUpdate('%s', updateSystem, 3000)" % data

from job import JobTab

class SystemJobSet(JobTab):
    def render_title(self, session, system):
        return "Grid Jobs %s" % fmt_count(self.get_item_count(session, system))

    def render_sql_where(self, session, system):
        elems = list()
        elems.append("s.system = %(nodeName)s")
        elems.append(self.get_phase_sql(session))
        return "where %s" % " and ".join(elems)

    def get_sql_values(self, session, system):
        return {"nodeName": system.nodeName}
        
from slot import SlotSet

class SystemSlotSet(SlotSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, system):
        # dont' use self.get_item_count because it will cause a costly sql query
        # instead, use the systemslotset in model so it will be cached
        count = self.app.model.system.slots.get_slot_count(session, system)
        return "Grid Slots %s" % fmt_count(count)

    def render_sql_where(self, session, system):
        elems = list()
        elems.append("system = %(nodeName)s")
        recent = self.get_recent_sql_where(session)
        if recent:
            elems.append(recent)
          
        return "where %s" % " and ".join(elems)
        
    def get_sql_values(self, session, system):
        return {"nodeName": system.nodeName}

class SystemServices(ItemSet):
    def render_title(self, session, *args):
        return "Services"
    
    def get_args(self, session):
        return (self.frame.get_object(session),)

    def do_get_items(self, session, system):
        daemons = list()
        daemon_types = [Scheduler, Collector, Negotiator]
        sql = "system = '%s'" % system.nodeName
        
        for daemon in daemon_types:
            system_daemons = daemon.select(sql)
            for devil in system_daemons: 
                try:
                    daemons.append(devil)
                except Exception, e:
                    pass

        # using a loop instead of an sql select with an outer join
        brokers = Broker.select()
        for broker in brokers:
            if broker.system.nodeName == system.nodeName:
                daemons.append(BrokerRegistration.get(broker.registrationID))

        return daemons
    
    def render_item_content(self, session, item):
        if isinstance(item, BrokerRegistration):
            reg = Identifiable(item.id)
            self.page.main.set_messaging_tab(session)
            href = self.page.main.broker.get_href(session, reg)
            return fmt_link(href, item.url)
        else:
            pool = model.Pool.get(item.Pool)
            self.page.main.pool.set_object(session, pool)
            self.page.main.show_grid_tab(session)
            daemon = Identifiable(item.id)
            if isinstance(item, Collector):
                self.page.main.pool.view.set_collector_tab(session)
                href = self.page.main.pool.collector.get_href(session, daemon)
            elif isinstance(item, Scheduler):
                self.page.main.pool.view.set_scheduler_tab(session)
                href = self.page.main.pool.scheduler.get_href(session, daemon)
            elif isinstance(item, Negotiator):
                self.page.main.pool.view.set_negotiator_tab(session)
                href = self.page.main.pool.negotiator.get_href(session, daemon)
            return fmt_link(href, item.Name)

    def render_item_type(self, session, item):
        if isinstance(item, Collector):
            return "Collector"
        elif isinstance(item, Scheduler):
            return "Scheduler"
        elif isinstance(item, Negotiator):
            return "Negotiator"
        elif isinstance(item, BrokerRegistration):
            return "Broker"
        else:
            return "Daemon"
