import sys, os, signal
from mint import *
from mint.schema import *
from traceback import print_exc, extract_tb
from datetime import datetime
from threading import Thread
from popen2 import Popen4
from shutil import copyfileobj
from tempfile import mkdtemp
import qpid, quirk, wooly

from cumin import Cumin
from util import *
import time

class TestBroker(Thread):
    def __init__(self, path, port):
        super(TestBroker, self).__init__()

        self.path = path
        self.port = port

        self.temp_dir = mkdtemp()
        self.data_dir = os.path.join(self.temp_dir, "data")

        self.command = (path,
                        "--port", str(port),
                        "--auth", "no",
                        "--trace",
                        "--data-dir", self.data_dir)

        self.output_path = os.path.join(self.temp_dir, "output")

        self.setDaemon(True)

    def run(self):
        print "Saving qpidd output to %s" % self.temp_dir

        out = open(self.output_path, "w")

        try:
            pop = Popen4(self.command)
            copyfileobj(pop.fromchild, out)
            exit_code = pop.wait()
            print exit_code
        finally:
            if pop:
                os.kill(pop.pid, signal.SIGKILL)

            kpid, stat = os.waitpid(pop.pid, os.WNOHANG)

            if kpid == 0:
                print "Test qpidd wouldn't die"

class TestEnvironment(object):
    def __init__(self, app, broker_host, broker_port, spec_path):
        self.app = app

        self.broker_conn = quirk.Connection(broker_host, broker_port,
                                            spec_path)
        self.broker_queue = quirk.Queue("cumin.queue")
        self.broker_exchange = quirk.Exchange("cumin.exchange")

        self.broker_registration = None
        self.broker = None
        self.vhost = None
        self.queue = None
        self.exchange = None
        self.broker_group = None

    def init(self):
        self.broker_conn.open()
        session = quirk.Session(self.broker_conn, "test")
        session.open()

        try:
            self.broker_queue.declare(session)
            self.broker_exchange.declare(session)
        finally:
            session.close()

    def run_test(self, test):
        session = TestSession(self)
        test.run(session)
        return session

    def page_and_session(self, page=None):
        if page is None:
            page = self.app.main_page

        s = wooly.Session(page)

        return page, s

    def check_redirect(self, p, s):
        redirect = p.get_redirect_url(s)

        if redirect is None:
            raise Exception("Expected redirect")

        s = wooly.Session.unmarshal(p.app, redirect)
        p = s.page

        p.process(s)
        p.render(s)

class TestSession(object):
    def __init__(self, env):
        self.env = env
        self.id = short_id()

        self.stack = list()
        self.passed = list()
        self.failed = list()

    def report(self, out):
        out.write("Tests:")
        out.write(os.linesep)
        out.write(os.linesep)

        self.stack[0].report(sys.stdout, 1)
        out.write(os.linesep)

        out.write("Failures:")
        out.write(os.linesep)
        out.write(os.linesep)

        if self.failed:
            for call in self.failed:
                out.write("  %s: %s" % \
                              (call.test.path(), call.get_exception_text()))
                out.write(os.linesep)

                for spec in extract_tb(call.traceback):
                    file, line, func, text = spec
                    out.write("    File \"%s\", line %i, in %s" % spec[:3])
                    out.write(os.linesep)
                    out.write("      %s" % spec[3])
                    out.write(os.linesep)

                out.write(os.linesep)
        else:
            out.write("  None")
            out.write(os.linesep)
            out.write(os.linesep)

        out.write("Summary:")
        out.write(os.linesep)
        out.write(os.linesep)

        passed, failed = len(self.passed), len(self.failed)
        out.write("  %i passed, %i failed" % (passed, failed))
        out.write(os.linesep)

class Test(object):
    def __init__(self, env, parent):
        self.env = env
        self.name = self.__class__.__name__
        self.parent = parent
        self.children = list()
        
        if parent:
            self.parent.children.append(self)

    def path(self):
        if self.parent is None:
            return self.name
        else:
            return self.parent.path() + "/" + self.name

    def run(self, session):
        call = TestCall(session, self)
        call.open()
        try:
            try:
                self.do_run(session)
                session.passed.append(call)
            except Exception, e:
                call.exception = e
                call.traceback = sys.exc_info()[2]
                session.failed.append(call)
        finally:
            call.close()

    def do_run(self, session):
        self.run_children(session)

    def run_children(self, session):
        for child in self.children:
            child.run(session)

    def __repr__(self):
        return self.__class__.__name__

class TestCall(object):
    def __init__(self, session, test):
        self.session = session
        self.test = test
        self.exception = None
        self.traceback = None

        self.caller = None
        self.callees = list()

        self.start = None
        self.end = None

    def open(self):
        if self.session.stack:
            self.caller = self.session.stack[-1]
            self.caller.callees.append(self)

        self.session.stack.append(self)

        self.start = time.clock()

    def close(self):
        self.end = time.clock()

        if len(self.session.stack) > 1:
            self.session.stack.pop()

    def report(self, writer, depth):
        entry = ("  " * depth) + str(self.test)
        status = "-"

        if self.exception:
            status = self.get_exception_text()

        writer.write("%-40s %s" % (entry, status))
        writer.write(os.linesep)

        for call in self.callees:
            call.report(writer, depth + 1)

    def get_exception_text(self):
        if self.exception:
            message = self.exception.__class__.__name__

            if hasattr(self.exception, "message"):
                message = message + ": " + self.exception.message

            return message

class MainTest(Test):
    def __init__(self, env):
        super(MainTest, self).__init__(env, None)

        BrokerGroupTest(env, self)
        BrokerRegistrationTest(env, self)

class BrokerLinkTest(Test):
    def __init__(self, env, parent):
        super(BrokerLinkTest, self).__init__(env, parent)
        
        RouteTest(env, self)
        self.LinkRemove(env, self)

    def do_run(self, session):
        p, s = self.env.page_and_session()

        p.main.broker.set_object(s, self.env.broker_registration)

        form = p.main.broker.link_add
        form.show(s)

        form.host.set(s, "localhost")
        form.port.set(s, "9991")
        form.submit(s)

        p.process(s)

        self.env.check_redirect(p, s)

        vhost = self.env.vhost

        def predicate():
            for item in Link.selectBy(vhost=vhost, host="localhost", port=9991):
                self.env.link = item
                return True

        wait(predicate)

        self.run_children(session)

    class LinkRemove(Test):
        def do_run(self, session):
            p, s = self.env.page_and_session()

            p.main.broker.set_object(s, self.env.broker_registration)

            form = p.main.broker.links_close
            form.show(s)
            form.ids.set(s, [self.env.link.id])
            form.submit(s)

            p.process(s)

            self.env.check_redirect(p, s)
            
            def predicate():
                return self.env.link.qmfDeleteTime

            wait(predicate)

class RouteTest(Test):
    def __init__(self, env, parent):
        super(RouteTest, self).__init__(env, parent)
        
        self.Add(env, self)
        self.Remove(env, self)

    def do_run(self, session):
        self.run_children(session)

    class Add(Test):
        def do_run(self, session):
            p, s = self.env.page_and_session()
    
            p.main.broker.set_object(s, self.env.broker_registration)
            p.main.broker.link.set_object(s, self.env.link)

            form = p.main.broker.link.bridge_add
            form.show(s)

            vhost = self.env.vhost
            exchange = Exchange.selectBy(vhost=vhost, name=self.env.broker_exchange.name)[0]
            form.exchange.set(s, str(exchange.id))

            form.key.set(s, "cumin.key")
            form.tag.set(s, "cumin.tag")
            form.excludes.set(s, "cumin.tag")
    
            form.submit(s)

            p.process(s)
    
            self.env.check_redirect(p, s)
    
            def predicate():
                for item in Bridge.selectBy(link=self.env.link, dest=self.env.broker_exchange.name, key="cumin.key"):
                    self.env.bridge = item
                    return True
    
            wait(predicate)
    
    class Remove(Test):
        def do_run(self, session):
            p, s = self.env.page_and_session()

            p.main.broker.set_object(s, self.env.broker_registration)
            p.main.broker.link.set_object(s, self.evn.link)

            form = p.main.broker.link.routes_close
            form.show(s)
            form.ids.set(s, [self.env.bridge.id])
            form.submit(s)

            p.process(s)

            self.env.check_redirect(p, s)
            
            def predicate():
                return self.env.bridge.qmfDeleteTime

            wait(predicate)

class BrokerRegistrationTest(Test):
    def __init__(self, env, parent):
        super(BrokerRegistrationTest, self).__init__(env, parent)

        BrokerTest(env, self)
        self.AddToGroup(env, self)
        self.Remove(env, self)

    def do_run(self, session):
        p, s = self.env.page_and_session()

        form = p.main.brokers_add
        form.show(s)

        form.names.get(s).append(session.id)
        addr = "%s:%s" % \
            (self.env.broker_conn.host, self.env.broker_conn.port)
        form.addrs.get(s).append(addr)
        form.groups.get(s).append(None)
        form.submit(s)
 
        p.process(s)

        self.env.check_redirect(p, s)

        try:
            self.env.broker_registration = BrokerRegistration.selectBy \
                (name=session.id)[0]
        except IndexError:
            raise Exception("Registration not created")
            
        self.run_children(session)

    class AddToGroup(Test):
        def do_run(self, session):
            raise Exception("Not implemented")

    class Remove(Test):
        def do_run(self, session):
            try:
                reg = BrokerRegistration.selectBy(name=session.id)[0]
            except IndexError:
                raise Exception("Broker missing")

            p, s = self.env.page_and_session()

            form = p.main.brokers_remove
            form.show(s)
            form.ids.set(s, [reg.id])
            form.submit(s)

            p.process(s)

            self.env.check_redirect(p, s)

            try:
                reg = BrokerRegistration.selectBy(name=session.id)[0]
                raise Exception("Registration not deleted")
            except IndexError:
                pass

class BrokerGroupTest(Test):
    def __init__(self, env, parent):
        super(BrokerGroupTest, self).__init__(env, parent)

        self.Edit(env, self)
        self.Remove(env, self)

    def do_run(self, session):
        p, s = self.env.page_and_session()

        form = p.main.broker_group.add
        form.show(s)
        form.group_name.set(s, session.id)
        form.submit(s)

        p.process(s)

        self.env.check_redirect(p, s)

        try:
            self.env.broker_group = BrokerGroup.selectBy(name=session.id)[0]
        except IndexError:
            raise Exception("Broker group not found")

        self.run_children(session)

    class Edit(Test):
        def do_run(self, session):
            p, s = self.env.page_and_session()

            p.main.broker_group.set_object(s, self.env.broker_group)

            form = p.main.broker_group.edit
            form.show(s)
            form.group_name.set(s, session.id + "%;&#\\")
            form.submit(s)

            p.process(s)

            self.env.check_redirect(p, s)

    class Remove(Test):
        def do_run(self, session):
            p, s = self.env.page_and_session()

            p.main.broker_group.set_object(s, self.env.broker_group)

            form = p.main.broker_group.remove
            form.show(s)
            form.submit(s)

            p.process(s)

            self.env.check_redirect(p, s)

class BrokerTest(Test):
    def __init__(self, env, parent):
        super(BrokerTest, self).__init__(env, parent)

        VhostTest(env, self)

    def do_run(self, session):
        reg = self.env.broker_registration

        wait(lambda: reg.broker)

        self.env.broker = reg.broker

        self.run_children(session)

class VhostTest(Test):
    def __init__(self, env, parent):
        super(VhostTest, self).__init__(env, parent)

        QueueTest(env, self)
        ExchangeTest(env, self)
        ConnectionTest(env, self)
        BrokerLinkTest(env, self)

    def do_run(self, session):
        def predicate():
            for item in Vhost.selectBy(broker=self.env.broker, name="/"):
                return True

        wait(predicate)

        try:
            self.env.vhost = Vhost.selectBy \
                (broker=self.env.broker, name="/")[0]
        except IndexError:
            raise Exception("Vhost not found")

        self.run_children(session)

class BindQueueTest(Test):
    def __init__(self, env, parent):
        super(BindQueueTest, self).__init__(env, parent)

        self.BindDirect(env, self)
        self.BindTopic(env, self)
        self.BindFanout(env, self)
        self.BindHeaders(env, self)

    def do_run(self, session):
        self.run_children(session)
        
    class BindDirect(Test):
        def __init__(self, env, parent):
            super(BindQueueTest.BindDirect, self).__init__(env, parent)

            self.RemoveBindDirect(env, self)

        def do_run(self, session):
            p, s = self.env.page_and_session()

            p.main.broker.set_object(s, self.env.broker.registration)
            p.main.broker.queue.set_object(s, self.env.queue)

            form = p.main.broker.queue.binding_add
            form.show(s)

            binding = dict()
            binding["test"] = {}
            direct = binding["test"]
            direct["name"] = "amq.direct"
            direct["type"] = "direct"
            direct["key"] = "amq.direct.key"

            form.bindings.dict_param.set(s, binding)
            
            form.submit(s)

            p.process(s)

            self.env.check_redirect(p, s)

            def predicate():
                for item in Binding.selectBy(queue=self.env.queue, bindingKey=self.env.queue.name):
                    self.env.binding_direct = item
                    return True

            wait(predicate)

        class RemoveBindDirect(Test):
            def do_run(self, session):
                p, s = self.env.page_and_session()

                p.main.broker.set_object(s, self.env.broker_registration)

                form = p.main.broker.bindings_remove.show(s)
                form.ids.set([self.env.binding_direct.id])
                form.submit(s)

                p.process(s)

                self.env.check_redirect(p, s)

                def predicate():
                    return self.env.binding_direct.qmfDeleteTime

                wait(predicate)

    class BindTopic(Test):
        def __init__(self, env, parent):
            super(BindQueueTest.BindTopic, self).__init__(env, parent)

            self.RemoveBindTopic(env, self)

        def do_run(self, session):
            p, s = self.env.page_and_session()

            p.main.broker.set_object(s, self.env.broker.registration)
            p.main.broker.queue.set_object(s, self.env.queue)

            form = p.main.broker.queue.binding_add
            form.show(s)

            binding = dict()
            binding["test"] = {}
            direct = binding["test"]
            direct["name"] = "amq.topic"
            direct["type"] = "topic"
            direct["key"] = "topic.key"
            form.bindings.dict_param.set(s, binding)
            
            form.submit(s)
            p.process(s)

            self.env.check_redirect(p, s)

            def predicate():
                for item in Binding.selectBy(queue=self.env.queue, bindingKey="topic.key"):
                    self.env.binding_topic = item
                    return True

            wait(predicate)

        class RemoveBindTopic(Test):
            def do_run(self, session):
                p, s = self.env.page_and_session()

                p.main.broker.set_object(s, self.env.broker_registration)

                form = p.main.broker.bindings_remove.show(s)
                form.ids.set([self.env.binding_topic.id])
                form.submit(s)

                p.process(s)

                self.env.check_redirect(p, s)

                def predicate():
                    return self.env.binding_topic.qmfDeleteTime

                wait(predicate)

    class BindFanout(Test):
        def __init__(self, env, parent):
            super(BindQueueTest.BindFanout, self).__init__(env, parent)

            self.RemoveBindFanout(env, self)

        def do_run(self, session):
            p, s = self.env.page_and_session()

            p.main.broker.set_object(s, self.env.broker.registration)
            p.main.broker.queue.set_object(s, self.env.queue)

            form = p.main.broker.queue.binding_add
            form.show(s)

            binding = dict()
            binding["test"] = {}
            direct = binding["test"]
            direct["name"] = "amq.fanout"
            direct["type"] = "fanout"
            form.bindings.dict_param.set(s, binding)
            
            form.submit(s)
            p.process(s)

            self.env.check_redirect(p, s)

            def predicate():
                for item in Binding.selectBy(queue=self.env.queue):
                    if item.exchange.name == "amq.fanout":
                        self.env.binding_fanout = item
                        return True

            wait(predicate)

        class RemoveBindFanout(Test):
            def do_run(self, session):
                p, s = self.env.page_and_session()

                p.main.broker.set_object(s, self.env.broker_registration)

                form = p.main.broker.bindings_remove.show(s)
                form.ids.set([self.env.binding_fanout.id])
                form.submit(s)

                p.process(s)

                self.env.check_redirect(p, s)

                def predicate():
                    return self.env.binding_fanout.qmfDeleteTime

                wait(predicate)

    class BindHeaders(Test):
        def do_run(self, session):
            p, s = self.env.page_and_session()

            p.main.broker.set_object(s, self.env.broker.registration)
            p.main.broker.queue.set_object(s, self.env.queue)

            form = p.main.broker.queue.binding_add
            form.show(s)

            binding = dict()
            binding["test"] = {}
            direct = binding["test"]
            direct["name"] = "amq.match"
            direct["type"] = "headers"
            direct["key"] = "headers.key"
            direct["x-match"] = "all"
            direct["mkey.1"] = "key1"
            direct["mkey.1.nv"] = "name.value.1"
            form.bindings.dict_param.set(s, binding)
            
            form.submit(s)
            p.process(s)

            self.env.check_redirect(p, s)

            def predicate():
                for item in Binding.selectBy(queue=self.env.queue, bindingKey="headers.key"):
                    return True

            wait(predicate)

            # if it timed out, raise an exception
            try:
                Binding.selectBy(queue=self.env.queue, bindingKey="headers.key")[0]
            except IndexError:
                raise Exception("Headers Binding not added")

class AddQueueTest(Test):
    def __init__(self, env, parent):
        super(AddQueueTest, self).__init__(env, parent)

        BindQueueTest(env, self)

    def do_run(self, session):
        name = "cumin.queue.%s" % session.id
        p, s = self.env.page_and_session()

        reg = self.env.broker_registration
        p.main.broker.set_object(s, reg)

        form = p.main.broker.queue_add
        form.show(s)
        form.namef.set(s, name)
        form.submit(s)

        p.process(s)

        self.env.check_redirect(p, s)

        vhost = self.env.vhost

        # wait for newly created queue to show up
        def predicate():
            for item in Queue.selectBy(vhost=vhost, name=name):
                return True

        wait(predicate)

        # if it timed out, raise an exception
        try:
            self.env.queue = Queue.selectBy(vhost=vhost, name=name)[0]
        except IndexError:
            raise Exception("Queue %s not added" % name)

        self.run_children(session)


class QueueTest(Test):
    def __init__(self, env, parent):
        super(QueueTest, self).__init__(env, parent)

        AddQueueTest(env, self)
        self.Remove(env, self)
        
    def do_run(self, session):
        vhost = self.env.vhost
        name = self.env.broker_queue.name

        def predicate():
            for item in Queue.selectBy(vhost=vhost, name=name):
                self.env.broker_queue = item
                return True

        wait(predicate)

        self.run_children(session)

    class Remove(Test):
        def do_run(self, session):
            p, s = self.env.page_and_session()

            reg = self.env.broker_registration
            p.main.broker.set_object(s, reg)

            form = p.main.broker.queues_remove
            form.show(s)
            form.ids.set(s, [self.env.queue.id])
            form.submit(s)

            p.process(s)

            self.env.check_redirect(p, s)

            # wait for newly created queue to get marked as deleted
            def predicate():
                return self.env.queue.qmfDeleteTime

            wait(predicate)

class AddExchangeTest(Test):
    def do_run(self, session):
        name = "cumin.exchange.%s" % session.id
        p, s = self.env.page_and_session()

        reg = self.env.broker_registration
        p.main.broker.set_object(s, reg)

        form = p.main.broker.exchange_add
        form.show(s)

        form.exchange_name.set(s, name)
        form.exchange_type.set(s, "topic")
        form.submit(s)

        p.process(s)

        self.env.check_redirect(p, s)

        vhost = self.env.vhost
        # wait for newly created exchange to show up
        def predicate():
            for item in Exchange.selectBy(vhost=vhost, name=name):
                self.env.added_exchange = item
                return True

        wait(predicate)
            
class ExchangeTest(Test):
    def __init__(self, env, parent):
        super(ExchangeTest, self).__init__(env, parent)

        AddExchangeTest(env, self)
        self.Remove(env, self)

    def do_run(self, session):
        vhost = self.env.vhost
        name = self.env.broker_exchange.name

        def predicate():
            for item in Exchange.selectBy(vhost=vhost, name=name):
                self.env.broker_exchange = item
                return True

        wait(predicate)

        self.run_children(session)

    class Remove(Test):
        def do_run(self, session):
            # try to delete it
            p, s = self.env.page_and_session()

            reg = self.env.broker_registration
            p.main.broker.set_object(s, reg)

            form = p.main.broker.exchanges_remove
            form.show(s)
            form.ids.set(s, [self.env.added_exchange.id])
            form.submit(s)

            p.process(s)

            self.env.check_redirect(p, s)

            # wait for newly created exchange to get marked as deleted
            def predicate():
                return self.env.added_exchange.qmfDeleteTime
        
            wait(predicate)
        
class ConnectionTest(Test):
    def do_run(self, session):
        raise Exception("Not implemented")

        vhost = self.env.vhost
        address = self.env.broker_conn.host + ":" + \
            str(self.env.broker_conn.port)

        try:
            self.env.conn = ClientConnection.selectBy \
                (vhost=vhost, address=address)[0]
        except IndexError:
            raise Exception("ClientConnection not found")

        self.run_children(session)

def wait(predicate, timeout=30):
    start = time.time()

    while True:
        if predicate():
            return

        if time.time() - start > timeout:
            raise Exception("Operation timed out")

        time.sleep(1)
