import sys, os, re, signal

from parsley.config import *
from parsley.command import *
from wooly.devel import *
from mint import *
from getpass import getpass
from psycopg2 import IntegrityError
from subprocess import Popen
from time import sleep

from util import *
from cumin import Cumin, CuminServer, CuminConfig
from test import *

def prompt_password():
    password = None

    while password is None:
        once = getpass("Enter new password: ")
        twice = getpass("Confirm new password: ")
        
        if once == twice:
            password = once
        else:
            print "Passwords don't match; try again"

    return password

class BaseCuminTool(Command):
    def __init__(self, name):
        super(BaseCuminTool, self).__init__(None, name)

        self.config = CuminConfig()

        opt = CommandOption(self, "data")
        opt.argument = "URI"
        opt.description = "Connect to database at URI"

        opt = CommandOption(self, "log-file")
        opt.argument = "PATH"
        opt.description = "Log to file at PATH"

        opt = CommandOption(self, "log-level")
        opt.argument = "LEVEL"
        opt.description = "Log messages at or above LEVEL " + \
            "('debug', 'info', 'warn', 'error')"

        opt = CommandOption(self, "debug")
        opt.description = "Enable debugging; print logging to console"

    def check(self):
        if os.getuid() not in (os.stat(__file__).st_uid, 0):
            print "Error: You have insufficient privileges"
            sys.exit(1)

    def init(self):
        super(BaseCuminTool, self).init()

        if os.getuid() == 0:
            # Drop privileges
            os.setuid(os.stat(__file__).st_uid)

        self.config.init()

        try:
            import psyco
            psyco.full()
        except ImportError:
            pass

    def run(self):
        try:
            opts, args = self.parse(sys.argv)
        except CommandException, e:
            print "Error: %s" % e.message
            e.command.print_help()
            sys.exit(1)

        if "help" in opts:
            self.print_help()
            sys.exit(0)

        self.do_run(opts, args)

    def do_run(self, opts, args):
        raise Exception("Not implemented")

    def main(self):
        self.check()
        self.init()
        self.run()

class CuminAdminTool(BaseCuminTool):
    def __init__(self, name):
        super(CuminAdminTool, self).__init__(name)

        self.description = "Cumin administration commands"
        self.database = None # Set in init

        command = self.CreateSchema(self, "create-schema")
        command.description = "Create the database schema"

        command = self.DropSchema(self, "drop-schema")
        command.description = "Drop the database schema; requires --force"

        opt = CommandOption(command, "force")
        opt.description = "Don't complain and just do it"

        command = self.AddQmfServer(self, "add-qmf-server")
        command.arguments = ("NAME", "URL")
        command.description = "Add a new QMF server called NAME at URL"

        command = self.RemoveQmfServer(self, "remove-qmf-server")
        command.arguments = ("URL",)
        command.description = "Remove QMF server called NAME; requires --force"

        opt = CommandOption(command, "force")
        opt.description = "Don't complain and just do it"

        command = self.ListQmfServers(self, "list-qmf-servers")
        command.description = "List existing QMF servers"

        command = self.AddUser(self, "add-user")
        command.arguments = ("NAME", "[PASSWORD]")
        command.description = "Add a new user called NAME"

        command = self.RemoveUser(self, "remove-user")
        command.arguments = ("NAME",)
        command.description = "Remove user called NAME; requires --force"

        opt = CommandOption(command, "force")
        opt.description = "Don't complain and just do it"

        command = self.ListUsers(self, "list-users")
        command.description = "List existing users"

        command = self.ListRoles(self, "list-roles")
        command.description = "List existing roles"

        command = self.Assign(self, "assign")
        command.description = "Add USER to ROLE"
        command.arguments = ("USER", "ROLE")

        command = self.Unassign(self, "unassign")
        command.description = "Remove USER from ROLE"
        command.arguments = ("USER", "ROLE")

        command = self.ChangePassword(self, "change-password")
        command.description = "Change password of USER"
        command.arguments = ("USER",)

    def init(self):
        super(CuminAdminTool, self).init()

        self.database = MintDatabase(self.config.data)
        self.database.check()
        self.database.init()

    def run(self):
        if self.config.debug:
            self.config.prt()

        args = sys.argv[1:]

        try:
            opts, remaining = self.parse_options(args)
        except CommandException, e:
            print "Error: %s" % e.message
            e.command.print_help()
            sys.exit(1)

        try:
            scommand = remaining[0]
        except IndexError:
            self.print_help()
            sys.exit(1)

        try:
            command = self.commands_by_name[scommand]
        except KeyError:
            print "Error: Command '%s' is unrecognized" %  scommand
            self.print_help()
            sys.exit(1)

        try:
            opts, args = command.parse(remaining)

            if "help" in opts:
                command.print_help()
                return

            command.run(opts, args)
        except CommandException, e:
            print "Error: %s" % e.message
            e.command.print_help()
            sys.exit(1)

    class CreateSchema(Command):
        def run(self, opts, args):
            main = os.path.join(self.parent.config.home, "sql", "schema.sql")
            indexes = os.path.join(self.parent.config.home, "sql", "indexes.sql")
            triggers = os.path.join(self.parent.config.home, "sql", "triggers.sql")

            self.parent.database.createSchema((main, indexes, triggers))

    class DropSchema(Command):
        def run(self, opts, args):
            if "force" in opts:
                self.parent.database.dropSchema()
                print "The schema is dropped"
            else:
                raise CommandException \
                    (self, "Command create-schema requires --force")

    class CheckSchema(Command):
        def run(self, opts, args):
            self.parent.database.checkSchema()

    class AddQmfServer(Command):
        def run(self, opts, args):
            try:
                name, url = args[1:]
            except IndexError:
                raise CommandException(self, "NAME and URL are required")

            for reg in BrokerRegistration.selectBy(name=name):
                print "Error: a broker called '%s' already exists" % name
                sys.exit(1)

            for reg in BrokerRegistration.selectBy(url=url):
                print "Error: a broker at %s already exists" % url
                sys.exit(1)

            url = url.strip()

            expr = re.compile("^amqps?://")

            if not expr.match(url):
                url = "amqp://%s" % url

            expr = re.compile(":[0-9]+$")

            if not expr.match(url):
                url = "%s:5672" % url

            reg = BrokerRegistration(name=name, url=url)
            reg.syncUpdate()

    class RemoveQmfServer(Command):
        def run(self, opts, args):
            try:
                name = args[1]
            except IndexError:
                raise CommandException(self, "NAME is required")

            for reg in BrokerRegistration.selectBy(name=name):
                break

            if reg:
                reg.destroySelf()
                reg.syncUpdate()
            else:
                raise CommandException \
                    (self, "Broker '%s' is unknown", reg.name)

    class ListQmfServers(Command):
        def run(self, opts, args):
            regs = BrokerRegistration.select()

            print "  ID Name                 URL"
            print "---- -------------------- ---------------------------------"

            for reg in regs:
                print "%4i %-20s %-20s" % (reg.id, reg.name, reg.url)

            count = regs.count()

            print "(%i server%s found)" % (count, ess(count))

    class AddUser(Command):
        def run(self, opts, args):
            try:
                name = args[1]
            except IndexError:
                raise CommandException(self, "NAME is required")

            if Subject.selectBy(name=name).count():
                print "Error: a user called '%s' already exists" % name
                sys.exit(1)

            try:
                password = args[2]
            except IndexError:
                password = prompt_password()
            
            crypted = crypt_password(password)

            try:
                subject = Subject(name=name, password=crypted)

                for role in Role.selectBy(name="user"):
                    subject.addRole(role)
                    break

                assert role

                subject.syncUpdate()
            except IntegrityError:
                print "Error: a user called '%s' already exists" % name
                sys.exit(1)

            print "User '%s' is added" % name

    class RemoveUser(Command):
        def run(self, opts, args):
            if "force" in opts:
                if len(args) != 2:
                    print "Error: no user name given"
                    sys.exit(1)

                name = args[1]
                subjects = Subject.selectBy(name=name)
                
                if subjects.count():
                    for subject in subjects:
                        subject.destroySelf()
                        break
                else:
                    raise CommandException(self, "User '%s' is unknown" % name)

                print "User '%s' is removed" % name
            else:
                raise CommandException \
                    (self, "Command remove-user requires --force yes")

    class ListUsers(Command):
        def run(self, opts, args):
            subjects = Subject.select(orderBy='name')

            print "  ID Name                 Roles"
            print "---- -------------------- --------------------"

            for subject in subjects:
                roles = ", ".join([x.name for x in list(subject.roles)])

                print "%4i %-20s %-20s" % (subject.id, subject.name, roles)

            count = subjects.count()
            print "(%i user%s found)" % (count, ess(count))

    class ListRoles(Command):
        def run(self, opts, args):
            roles = Role.select(orderBy='name')

            print "  ID Name"
            print "---- --------------------"

            for role in roles:
                print "%4i %s" % (role.id, role.name)

            count = roles.count()
            print "(%i role%s found)" % (count, ess(count))

    class Assign(Command):
        def run(self, opts, args):
            if len(args) != 3:
                raise CommandException(self, "USER and ROLE are required")

            subject = Subject.getByName(args[1])

            if not subject:
                raise CommandException \
                    (self, "User '%s' is unknown" % subject.name)

            role = Role.getByName(args[2])

            if not role:
                raise CommandException \
                    (self, "Role '%s' is unknown" % role.name)

            try:
                subject.addRole(role)
                subject.syncUpdate()

                print "User '%s' is added to role '%s'" % \
                    (subject.name, role.name)
            except IntegrityError:
                msg = "User '%s' is already assigned to role '%s'" % \
                    (subject.name, role.name)
                raise CommandException(self, msg)

    class Unassign(Command):
        def run(self, opts, args):
            if len(args) != 3:
                raise CommandException(self, "USER and ROLE are required")

            subject = Subject.getByName(args[1])

            if not subject:
                raise CommandException \
                    (self, "User '%s' is unknown" % subject.name)

            role = Role.getByName(args[2])

            if not role:
                raise CommandException \
                    (self, "Role '%s' is unknown" % role.name)

            subject.removeRole(role)
            subject.syncUpdate()

            print "User '%s' is removed from role '%s'" % \
                (subject.name, role.name)

    class ChangePassword(Command):
        def run(self, opts, args):
            try:
                ssubject = args[1]
            except IndexError:
                raise CommandException(self, "USER is required")

            subject = Subject.getByName(ssubject)
            
            if not subject:
                raise CommandException\
                    (self, "User '%s' is unknown" % subject.name)

            crypted = crypt_password(prompt_password())

            subject.password = crypted
            subject.syncUpdate()

            print "Password of user '%s' is changed" % subject.name

class CuminServerTool(BaseCuminTool):
    def __init__(self, name):
        super(CuminServerTool, self).__init__(name)

        signal.signal(signal.SIGTERM, self.sigTermHandler)

        self.description = "Cumin web server"

        param = ConfigParameter(self.config, "addr", str)
        param.default = "localhost"

        param = ConfigParameter(self.config, "port", int)
        param.default = 45672

        param = ConfigParameter(self.config, "ssl", bool)
        param.default = False

        opt = CommandOption(self, "addr")
        opt.argument = "ADDR"
        opt.description = "Run web server at address ADDR"

        opt = CommandOption(self, "port")
        opt.argument = "PORT"
        opt.description = "Run web server on port PORT"

        opt = CommandOption(self, "ssl")
        opt.description = "Serve web pages using SSL"

    def sigTermHandler(self, signum, frame):
        sys.exit(1)

    class MintProcess(object):
        def __init__(self):
            self.proc = None

        def start(self, config, opts):
            args = ["mint-server",
                    "--data", config.data,
                    "--log-file", config.log_file,
                    "--log-level", config.log_level,
                    "--expire-frequency", str(config.expire_frequency),
                    "--expire-threshold", str(config.expire_threshold)]

            if "debug" in opts or config.debug:
                args.append("--debug")

            self.proc = Popen(args)

        def stop(self):
            if self.proc.poll is not None:
                os.kill(self.proc.pid, signal.SIGTERM)

            for i in range(30):
                code = self.proc.poll()

                if code is not None:
                    log.debug("Mint subprocess %i terminated", self.proc.pid)
                    return

                sleep(1)

            os.kill(self.proc.pid, signal.SIGKILL)

            log.warn("Mint subprocess %i wouldn't go gracefully, killed",
                     self.proc.pid)
                    
    def do_run(self, opts, args):
        self.config.load_dict(opts)

        self.config.ssl = "ssl" in opts

        if self.config.debug:
            self.config.prt()

        app = Cumin(self.config)

        try:
            app.check()
        except Exception, e:
            log.exception(e)
            sys.exit(1)

        app.init()

        server = CuminServer(app, self.config.addr, self.config.port)

        if self.config.ssl:
            cpath = os.path.join(self.config.home, "etc", "cumin.crt")
            kpath = os.path.join(self.config.home, "etc", "cumin.key")

            if os.path.isfile(cpath):
                server.set_ssl_cert_path(cpath)
            else:
                log.error("SSL certificate file '%s' not found" % cpath)
                sys.exit(1)

            if os.path.isfile(kpath):
                server.set_ssl_key_path(kpath)
            else:
                log.error("SSL key file '%s' not found" % kpath)
                sys.exit(1)

        mint_proc = self.MintProcess()

        try:
            try:
                mint_proc.start(self.config, opts)
            except:
                log.exception("Failed starting mint process")

            app.start()

            try:
                server.start()
            finally:
                server.stop()
        finally:
            app.stop()
            mint_proc.stop()

class CuminTestTool(BaseCuminTool):
    def __init__(self, name):
        super(CuminTestTool, self).__init__(name)

        self.description = "Cumin test tool"

        opt = CommandOption(self, "broker")
        opt.argument = "ADDR"
        opt.description = "Use existing broker at ADDR"

    def do_run(self, opts, args):
        self.config.load_dict(opts)

        if self.config.debug:
            self.config.prt()

        app = Cumin(self.config)

        try:
            app.check()
        except Exception, e:
            if hasattr(e, "message"):
                print e.message

            sys.exit(1)

        usr_sbin = "/usr/sbin"
        if usr_sbin not in sys.path:
            sys.path.append(usr_sbin)

        app.init()

        if "broker" in opts:
            host, port = parse_broker_addr(opts["broker"])
        else:
            host, port = "localhost", randint(49152, 65535)

            broker = TestBroker("qpidd", port)
            broker.start()

        env = TestEnvironment(app, host, port, self.config.spec)
        env.init();

        app.start()

        try:
            session = env.run_test(MainTest(env))
            session.report(sys.stdout)
        finally:
            app.stop()

class CuminBenchTool(BaseCuminTool):
    def __init__(self, name):
        super(CuminBenchTool, self).__init__(name)

        self.description = "Cumin benchmarking tool"

        opt = CommandOption(self, "hits")
        opt.argument = "COUNT"
        opt.description = "Stop after COUNT page hits"

        opt = CommandOption(self, "profile")
        opt.description = "Enable profiling"

        opt = CommandOption(self, "check-xml")
        opt.description = "Check that page output is well-formed xml"

    def do_run(self, opts, args):
        self.config.load_dict(opts)

        if self.config.debug:
            self.config.prt()

        app = Cumin(self.config)

        try:
            app.check()
        except Exception, e:
            if hasattr(e, "message"):
                print e.message

            sys.exit(1)

        app.init()

        harness = BenchmarkHarness(app, "check-xml" in opts)

        app.start()

        try:
            harness.run(int(opts.get("hits", "1000")))
        finally:
            app.stop()
