from wooly import *
from wooly.widgets import *
from wooly.resources import *
from time import sleep
from threading import Thread, Event
from random import randint
from datetime import datetime, timedelta
from crypt import crypt
from qpid.datatypes import uuid4

from widgets import *
from util import *

from wooly import Session

strings = StringCatalog(__file__)

log = logging.getLogger("cumin.user")

class UserSession(object):
    def __init__(self, app, subject):
        self.app = app
        self.subject = subject
        self.id = str(uuid4())
        self.created = datetime.now()

        self.app.user_sessions_by_id[self.id] = self

    def delete(self):
        del self.app.user_sessions_by_id[self.id]

class UserSessionExpireThread(Thread):
    def __init__(self, app):
        super(UserSessionExpireThread, self).__init__()

        self.app = app
        self.setDaemon(True)

    def run(self):
        while True:
            self.expire_sessions()
            sleep(60)

    def expire_sessions(self):
        when = datetime.now() - timedelta(hours=2)
        count = 0

        for session in self.app.user_sessions_by_id.values():
            if session.created < when:
                session.delete()
                count += 1

        log.debug("Expired %i user sessions", count)

class LoginPage(Page):
    def __init__(self, app, name):
        super(LoginPage, self).__init__(app, name)

        self.logout = BooleanParameter(app, "logout")
        self.add_parameter(self.logout)

        self.origin = self.Origin(app, "origin")
        self.add_parameter(self.origin)

        form = LoginForm(app, "form")
        self.add_child(form)

    def do_process(self, session):
        if self.logout.get(session):
            #id = session.get_cookie("session")
            session.expire_cookie("session")

        super(LoginPage, self).do_process(session)

    def render_title(self, session):
        return "Log In"

    class Origin(Parameter):
        def get_default(self, session):
            return Session(self.app.main_page).marshal()

class LoginForm(FieldForm):
    def __init__(self, app, name):
        super(LoginForm, self).__init__(app, name)

        self.__login_invalid = Attribute(app, "login_invalid")
        self.add_attribute(self.__login_invalid)

        self.__name = self.Name(app, "name")
        self.add_field(self.__name)
        self.__name.input.size = 20

        self.__password = self.Password(app, "password")
        self.add_field(self.__password)
        self.__password.input.size = 20

        self.__submit = self.Submit(app, "submit")
        self.add_child(self.__submit)

    def do_process(self, session):
        if self.__submit.get(session):
            name = self.__name.get(session)
            password = self.__password.get(session)

            errors = self.validate(session)

            if not errors:
                try:
                    user = Subject.selectBy(name=name)[0]
                except IndexError:
                    self.__login_invalid.set(session, True)
                    return

                crypted = user.password

                if crypted and crypt(password, crypted) == crypted:
                    # You're in!

                    usess = UserSession(self.app, user)
                    session.set_cookie("session", usess.id)

                    url = self.page.origin.get(session)

                    self.page.set_redirect_url(session, url)
                else:
                    self.__login_invalid.set(session, True)

    def render_operator_link(self, session):
        email = self.app.config.operator_email

        if email:
            return "<a href=\"mailto:%s\">site operator</a>" % email
        else:
            return "site operator"

    def render_login_invalid(self, session):
        if self.__login_invalid.get(session):
            return self.get_string("login_invalid")

    class Name(StringField):
        def render_title(self, session):
            return "User Name"

    class Password(PasswordField):
        def render_title(self, session):
            return "Password"

    class Submit(FormButton):
        def render_content(self, session):
            return "Submit"

class ChangePasswordForm(CuminFieldForm):
    def __init__(self, app, name):
        super(ChangePasswordForm, self).__init__(app, name)

        self.__errors = self.Errors(app, "errors")
        self.add_attribute(self.__errors)

        self.__current = self.Current(app, "current")
        self.add_field(self.__current)
        self.__current.input.size = 20

        self.__new0 = self.New0(app, "new0")
        self.add_field(self.__new0)
        self.__new0.input.size = 20

        self.__new1 = self.New1(app, "new1")
        self.add_field(self.__new1)
        self.__new1.input.size = 20

    def process_cancel(self, session):
        branch = session.branch()
        self.page.main.view.show(branch)
        self.page.set_redirect_url(session, branch.marshal())

    def process_submit(self, session):
        curr = self.__current.get(session)
        new0 = self.__new0.get(session)
        new1 = self.__new1.get(session)

        errors = self.validate(session)

        subject = session.user_session.subject
        crypted = subject.password

        if crypt_password(curr, crypted) != crypted:
            msg = "The current password you entered is incorrect"
            self.__errors.get(session).append(msg)
            errors.append(msg)

        if new0 != new1:
            msg = "The new passwords do not match"
            self.__errors.get(session).append(msg)
            errors.append(msg)

        if not errors:
            subject.password = crypt_password(new0)
            subject.syncUpdate()

            self.process_cancel(session)

    def render_title(self, session):
        return "Change Password"

    def render_form_error(self, session):
        errors = self.__errors.get(session)

        if errors:
            return "<ul class=\"errors\" style=\"margin:0; float:left;\"><li>" + \
                ("</li><li>".join(errors)) + \
                "</li></ul>"

    class Current(PasswordField):
        def render_title(self, session):
            return "Current Password"

    class New0(PasswordField):
        def render_title(self, session):
            return "New Password"

    class New1(PasswordField):
        def render_title(self, session):
            return "Repeat New Password"

    class Errors(Attribute):
        def get_default(self, session):
            return list()
