from datetime import datetime, timedelta
from sqlobject import sqlhub
from sqlobject.sresults import SelectResults
from wooly import *
from wooly.widgets import *
from wooly.forms import *
from wooly.tables import *
from mint.schema import *

from parameters import *
from charts import *
from formats import *
from util import *

import time

strings = StringCatalog(__file__)

class CuminPage(Page):
    def __init__(self, app, name):
        super(CuminPage, self).__init__(app, name)

        self.__frames = self.FramesAttribute(app, "frames")
        self.add_attribute(self.__frames)

        self.__modal = Attribute(app, "modal")
        self.add_attribute(self.__modal)

    def save_session(self, session):
        if self.app.debug:
            self.app.debug.sessions.append(session)

    def set_modal(self, session, modal):
        self.__modal.set(session, modal)

    def get_frames(self, session):
        return self.__frames.get(session)

    def render_class(self, session):
        return self.__modal.get(session) and "modal" or "modeless"

    class FramesAttribute(Attribute):
        def get_default(self, session):
            return list()

class CuminFrame(Frame, ModeSet):
    def __init__(self, app, name):
        super(CuminFrame, self).__init__(app, name)

        self.object = None
        self.__view = None
        self.__add = None
        self.__edit = None
        self.__remove = None

    def show_object(self, session, object):
        self.object.set(session, object)
        self.__view.show(session)
        return self

    def get_args(self, session):
        if self.object:
            return (self.get_object(session),)
        else:
            return ()

    def get_object(self, session):
        if self.object:
            return self.object.get(session)

    def set_object(self, session, object):
        return self.object.set(session, object)

    def get_href(self, session, object):
        branch = session.branch()
        self.object.set(branch, object)
        self.__view.show(branch)
        return branch.marshal()

    def set_view_mode(self, mode):
        self.__view = mode

    def show_view(self, session):
        return self.show_mode(session, self.__view)

    def set_add_mode(self, mode):
        self.__add = mode

    def show_add(self, session):
        return self.show_mode(session, self.__add)

    def set_edit_mode(self, mode):
        self.__edit = mode

    def show_edit(self, session):
        return self.show_mode(session, self.__edit)

    def set_remove_mode(self, mode):
        self.__remove = mode

    def show_remove(self, session):
        return self.show_mode(session, self.__remove)

    def do_process(self, session, *args):
        self.page.get_frames(session).append(self)

        super(CuminFrame, self).do_process(session, *args)

    def render_href(self, session, *args):
        branch = session.branch()
        self.__view.show(branch)
        return branch.marshal()

    def render_title(self, session, *args):
        obj = self.get_object(session)

        cls = self.app.model.get_class_by_object(obj)
        if cls:
            return fmt_shorten(cls.get_object_title(session, obj), 16, 4)

class CuminView(Widget):
    def __init__(self, app, name):
        super(CuminView, self).__init__(app, name)

        summary = CuminSummary(app, "summary")
        self.add_child(summary)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_script(self, session, *args):
        return None

    def render_icon_resource(self, session, *args):
        return "action-36.png"

    def render_title(self, session, *args):
        return self.frame.render_title(session, *args)

class FormHelp(Widget):
    def __init__(self, app, name):
        super(FormHelp, self).__init__(app, name)

class CuminForm(Form):
    def __init__(self, app, name):
        super(CuminForm, self).__init__(app, name)

        self.__cancel = self.Cancel(app, "cancel")
        self.__cancel.set_tab_index(201)
        self.add_child(self.__cancel)

        self.__submit = self.Submit(app, "submit")
        self.__submit.set_tab_index(200)
        self.add_child(self.__submit)

        self.__help = self.Help(app, "help")
        self.add_child(self.__help)

    def submit(self, session):
        self.__submit.set(session, True)

    def cancel(self, session):
        self.__cancel.set(session, True)

    def get_modal(self, session):
        return True

    def do_process(self, session, *args):
        self.page.set_modal(session, self.get_modal(session))
            
        if self.__cancel.get(session):
            self.__cancel.set(session, False)

            self.process_cancel(session, *args)
        elif self.__submit.get(session):
            self.__submit.set(session, False)

            self.process_submit(session, *args)
        else:
            self.process_display(session, *args)
        
    def process_cancel(self, session, *args):
        self.page.set_redirect_url(session, self.get_origin(session))

    def process_submit(self, session, *args):
        pass

    def process_display(self, session, *args):
        pass

    def render_cancel_content(self, session, *args):
        return "Cancel"

    def render_submit_content(self, session, *args):
        return "Submit"

    def render_form_error(self, session, *args):
        pass

    class Cancel(FormButton):
        def render_class(self, session, *args):
            return "cancel"

        def render_content(self, session, *args):
            cargs = self.parent.get_args(session)
            return self.parent.render_cancel_content(session, *cargs)

    class Submit(FormButton):
        def render_class(self, session, *args):
            return "submit"

        def render_content(self, session, *args):
            cargs = self.parent.get_args(session)
            return self.parent.render_submit_content(session, *cargs)

    class Help(FormHelp):
        def render_help_href(self, session, *args):
            return "resource?name=help.html#%s" % self.path

class CuminFieldForm(CuminForm, FieldForm, Frame):
    def render_body(self, session, *args):
        return self.render_fields(session, *args)
    
class CuminConfirmForm(CuminForm):
    def __init__(self, app, name):
        super(CuminConfirmForm, self).__init__(app, name)

class CuminActionSetForm(ItemSet, CuminForm, Frame):
    def __init__(self, app, name):
        super(CuminActionSetForm, self).__init__(app, name)

class CuminIntIdList(Widget):
    def __init__(self, app, name):
        super(CuminIntIdList, self).__init__(app, name)
    
        param = IntegerParameter(app, "param")
        self.add_parameter(param)

        self.ids = ListParameter(app, "id", param)
        self.add_parameter(self.ids)

    def render_item_content(self, session, id):
        return "Act on object %i" % id
    
class CuminStringIdList(Widget):
    def __init__(self, app, name):
        super(CuminStringIdList, self).__init__(app, name)
    
        param = Parameter(app, "param")
        self.add_parameter(param)

        self.ids = ListParameter(app, "id", param)
        self.add_parameter(self.ids)

    def render_item_content(self, session, id):
        return "Act on object %s" % id
    
class CuminBulk(ItemSet, CuminForm, Frame):
    def get_href(self, session, ids):
        branch = session.branch()
        self.ids.set(branch, ids)
        self.show(branch)
        return branch.marshal()

    def process_submit(self, session, *args):
        items = self.get_items(session, *args)

        for item in items:
            self.process_item(session, item)

        self.process_return(session, *args)

    def process_item(self, session, item):
        pass

    def process_return(self, session, *args):
        self.page.set_redirect_url(session, self.get_origin(session))

    def process_cancel(self, session, *args):
        self.process_return(session, *args)

    def render_form_heading(self, session, *args):
        return "Actions"

class CuminBulkActionForm(CuminBulk, CuminIntIdList):
    def get_items(self, session, *args):
        return self.ids.get(session)

class CuminBulkStringActionForm(CuminBulk, CuminStringIdList):
    def get_items(self, session, *args):
        return self.ids.get(session)

class CuminStatus(Widget):
    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, object):
        return "Status"
    
    def render_color(self, session, object):
        if hasattr(object, "errors"):
            if object.errors:
                return "red"
            elif object.warnings:
                return "yellow"
            else:
                return "green"
        else:
            return "green"

class CuminEditableProperties(PropertySet):
    def __init__(self, app, name):
        self.item_renderer=EditablePropertyRenderer(self, "property_html")
        super(CuminEditableProperties, self).__init__(app, name, item_renderer=self.item_renderer)

class EditablePropertyRenderer(TemplateRenderer, Widget):
    """Display input fields for editing properties

    Parent class needs to override do_get_items() and return a list of items. 
    Each item should be a dictionary. 
            ["name"] is required and should be the display label
            ["value"] is required and should be the value to edit
            ["type"] is required and should be an input type <"number" | "string">
            ["path"] is required and should be the DictParameter path to prepend to the name
            ["error"] is optional and should be the error text to display for that item
            ["property"] is optional and should be a dictionary of item properties
            ["orig"] is optional and should be the original value of the item before user edits
    """
    def __init__(self, widget, template_key):
        super(EditablePropertyRenderer, self).__init__(widget, template_key)

        self.__bool_template = Template(self, "bool_html")
        self.__string_template = Template(self, "string_html")
        self.__bigstring_template = Template(self, "bigstring_html")
        self.__number_template = Template(self, "number_html")
        self.__readonly_template = Template(self, "readonly_html")
        self.__orig_template = Template(self, "orig_html")

    def render_title(self, session, item):
        title = item["name"]
        if "property" in item:
            property = item["property"]
            if property.title:
                title = property.get_title(session) 
        return escape_amp(title)

    def render_value(self, session, item):
        value = item["value"]
        type = item["type"]
        writable = True
        property = "property" in item and item["property"] or None
        if property:
            writable = property.writable
            
        writer = Writer()
        if not writable:
            self.__readonly_template.render(writer, session, item)
        elif type == "float" or type == "integer":
            self.__number_template.render(writer, session, item)
        else:
            uvalue = value.upper()
            if uvalue == "FALSE" or uvalue == "TRUE":
                self.__bool_template.render(writer, session, item)
            else:
                if len(value) > 40:
                    self.__bigstring_template.render(writer, session, item)
                else:
                    self.__string_template.render(writer, session, item)

        return writer.to_string()
            
    def render_pname(self, session, item):
        return DictParameter.sep().join(
                (item["path"], escape_entity(item["name"]), "value"))
    
    def render_ptype_name(self, session, item):
        return DictParameter.sep().join(
                (item["path"], escape_entity(item["name"]), "type"))

    def render_ptype_value(self, session, item):
        return item["type"]
        
    def render_orig_value(self, session, item):
        if "orig" in item:
            writer = Writer()
            self.__orig_template.render(writer, session, item)
            return writer.to_string()
        
    def render_porig_name(self, session, item):
        return DictParameter.sep().join(
            (item["path"], escape_entity(item["name"]), "orig"))
        
    def render_porig_value(self, session, item):
        value = item["orig"]
        return escape_entity(str(value))
        
    def render_val(self, session, item):
        value = self.get_val(session, item)
        return escape_entity(str(value))
    
    def render_display_val(self, session, item):
        value = self.get_val(session, item)
        property = "property" in item and item["property"] or None
        if property:
            renderer = property.renderer
            if renderer:
                value = renderer(session, value)

        return escape_amp(str(value))
        
    def get_val(self, session, item):
        try:
            value = item["value"]
        except KeyError:
            value = ""
        return value
        
    def render_error(self, session, item):
        if "error" in item:
            return "<div class=\"error\">%s</div>" % item["error"]

    def render_inline_help(self, session, item):
        property = "property" in item and item["property"] or None
        if property:
            if property.example:
                example = "<span class=\"prop_example\">%s</span>" % property.example
            else:
                example = ""
            description = property.description or ""
            return " ".join((description, example))
        
    def render_false_selected(self, session, item):
        return item["value"].upper() == "FALSE" and "checked=\"checked\"" or ""
    
    def render_true_selected(self, session, item):
        return item["value"].upper() == "TRUE" and "checked=\"checked\"" or ""
    
    def render_edit_number_class(self, session, item):
        return "error" in item and "numeric_error" or "edit_number"

class CuminProperties(PropertySet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def do_get_items(self, session, object):
        cls = self.app.model.get_class_by_object(object)
        return [(x.get_title(session), x.value(session, object), x.escape)
                for x in cls.properties]

class CuminActions(ActionSet):
    def get_args(self, session):
        return self.frame.get_args(session)

    def do_get_items(self, session, object):
        cls = self.app.model.get_class_by_object(object)
        return [(x.get_href(session, object),
                 x.get_title(session),
                 x.get_enabled(session, object))
                for x in cls.actions if x.navigable and not x.aggregate]

class CuminDetails(Widget):
    def __init__(self, app, name):
        super(CuminDetails, self).__init__(app, name)

        props = CuminProperties(app, "properties")
        self.add_child(props)

        actions = CuminActions(app, "actions")
        self.add_child(actions)

    def render_title(self, session):
        return "Details"

class CuminHeading(Widget):
    def render_title(self, session, *args):
        pass

    def render_icon_href(self, session, *args):
        return "resource?name=action-36.png"

class CuminSummary(Widget):
    def __init__(self, app, name):
        super(CuminSummary, self).__init__(app, name)

        props = self.SummaryProperties(app, "properties")
        self.add_child(props)

        actions = self.SummaryActions(app, "actions")
        self.add_child(actions)

    def get_args(self, session):
        return self.frame.get_args(session)

    def render_title(self, session, object):
        cls = self.app.model.get_class_by_object(object)
        if cls:
            return cls.get_object_title(session, object)

    def render_icon_href(self, session, object):
        cls = self.app.model.get_class_by_object(object)
        if cls:
            return cls.get_icon_href(session)

    class SummaryProperties(CuminProperties):
        def do_get_items(self, session, object):
            cls = self.app.model.get_class_by_object(object)
            if cls:
                return [(x.get_title(session), x.value(session, object), x.escape)
                    for x in cls.properties if x.summary]

    class SummaryActions(CuminActions):
        def do_get_items(self, session, object):
            cls = self.app.model.get_class_by_object(object)
            if cls:
                return [(x.get_href(session, object), x.get_title(session), x.get_enabled(session, object))
                    for x in cls.actions if x.summary and x.navigable]

class StateSwitch(ItemSet):
    def __init__(self, app, name):
        super(StateSwitch, self).__init__(app, name)

        self.param = Parameter(app, "param")
        self.add_parameter(self.param)

        self.__states = list()
        self.__titles = dict()
        self.__hover = dict()
        self.__bookmark = dict()

    def add_state(self, state, title, hover="", bm=None):
        self.__states.append(state)
        self.__titles[state] = title
        self.__hover[state] = hover
        self.__bookmark[state] = bm

        if self.param.default is None:
            self.param.default = state

    def get(self, session):
        return self.param.get(session)

    def set(self, session, value):
        foo = self.param.set(session, value)
        return foo

    def get_items(self, session):
        return self.__states
    
    def get_title(self, state):
        return state in self.__titles and self.__titles[state]

    def get_hover(self, state):
        return state in self.__hover and self.__hover[state]

    def get_bookmark(self, state):
        return state in self.__bookmark and self.__bookmark[state]

    def get_param_id(self):
        """ needed because the SubmitSwitch class uses
            a hidden input to set the param value instead
            of a link name/value pair """
        return self.param.path
        
    def render_item_link(self, session, state):
        branch = session.branch()
        self.set(branch, state)

        title = self.get_title(state)
        hover = self.get_hover(state)
        class_ = self.get(session) == state and "selected"
        bm = self.get_bookmark(state)

        return fmt_link(branch.marshal(), title, class_, link_title=hover, bm=bm)

class GroupSwitch(StateSwitch):
    def __init__(self, app, name):
        super(GroupSwitch, self).__init__(app, name)

        self.add_state("j", "Jobs")
        self.add_state("g", "Groups")

class UnitSwitch(StateSwitch):
    def __init__(self, app, name):
        super(UnitSwitch, self).__init__(app, name)

        self.add_state("m", "Messages")
        self.add_state("b", "Bytes")

        self.brief_sing = {"m": "Msg.", "b": "Byte"}
        self.brief_plur = {"m": "Msgs.", "b": "Bytes"}

    def get_brief_singular(self, session):
        return self.brief_sing[self.get(session)]

    def get_brief_plural(self, session):
        return self.brief_plur[self.get(session)]

class PhaseSwitch(StateSwitch):
    def __init__(self, app, name):
        super(PhaseSwitch, self).__init__(app, name)

        self.add_state("a", "Active")
        self.add_state("i", "Idle")
        self.add_state("d", "Deleted")

    def get_sql_constraint(self, session, *args):
        phase = self.get(session)

        if phase == "a":
            sql = "((c.qmf_update_time is null or " + \
                "c.qmf_update_time > now() - interval '10 minutes')" + \
                " and qmf_delete_time is null)"
        elif phase == "i":
            sql = "((c.qmf_update_time is null or " + \
                "c.qmf_update_time <= now() - interval '10 minutes')" + \
                " and qmf_delete_time is null)"
        else:
            sql = "qmf_delete_time is not null"

        return sql
        
class SubmitSwitch(StateSwitch):
    """ Clicking on the link will submit the form. 
    
        This uses javascript and a hidden input to
        submit the form when the link is clicked. This
        allows other fields on the form to "remember"
        their values. """
    def render_value(self, session, *args):
        return str(self.get(session))

    def render_name(self, session, *args):
        """ this needs to be the path of the param """
        return self.get_param_id()
    
    def render_item_link(self, session, state):
        title = self.get_title(state)
        hover = self.get_hover(state)
        class_ = self.get(session) == state and "selected"
        bm = self.get_bookmark(state)
        name = self.get_param_id()
        click = "submit_state('%s', '%s'); return false;" % (name, state)

        return fmt_link("", title, class_, link_title=hover, bm=bm, click=click)

class TopTableColumn(SqlTableColumn):
    def __init__(self, app, name):
        super(TopTableColumn, self).__init__(app, name)
        
        self.header_class = TopTableColumnHeader

class TopTableColumnHeader(ItemTableColumnHeader):
    pass

class TopTable(SqlTable):
    def __init__(self, app, name):
        super(TopTable, self).__init__(app, name)

        col = self.CountColumn(app, "name")
        col.alignment = "right"
        self.add_column(col)

        self.count = Attribute(app, "count")
        self.count.default = 1
        self.add_attribute(self.count)

    def get_connection(self, session):
        return sqlhub.getConnection().getConnection()

    class CountColumn(TopTableColumn):
        def render_content(self, session, data):
            count = self.parent.count.get(session)
            self.parent.count.set(session, count + 1)
            return count

class CuminTable(SqlTable):
    def __init__(self, app, name):
        super(CuminTable, self).__init__(app, name)

        self.paginator = Paginator(app, "page")
        self.add_child(self.paginator)

    def get_connection(self, session):
        return sqlhub.getConnection().getConnection()

    def do_process(self, session, *args):
        super(CuminTable, self).do_process(session, *args)

        self.paginator.set_count(session, self.get_item_count(session, *args))

    def render_sql_limit(self, session, *args):
        start, end = self.paginator.get_bounds(session)
        return "limit %i offset %i" % (end - start, start)

class CuminTableWithControls(CuminTable):
    def __init__(self, app, name):
        super(CuminTableWithControls, self).__init__(app, name)

        self.switches = WidgetSet(app, "switches")
        self.switches.html_class = "switches"
        self.add_child(self.switches)

        self.filters = WidgetSet(app, "filters")
        self.filters.html_class = "filters"
        self.add_child(self.filters)

    def get_sql_constraint(self, session, *args):
        pass

    def render_switches(self, session, *args):
        if self.switches.children:
            return self.switches.render(session, *args)

    def render_filters(self, session, *args):
        if self.filters.children:
            return self.filters.render(session, *args)

    def render_sql_where(self, session, *args):
        constraints = list()

        for filter in self.filters.children:
            if filter:
                constraints.append(filter.get_sql_constraint(session, *args))

        constraint = self.get_sql_constraint(session, *args)

        if constraint:
            constraints.append(constraint)

        return "where %s" % " and ".join(constraints)

class NullSortColumn(SqlTableColumn):
    def get_orderby_sql(self, session):
        key = self.get_column_key(session)

        if key:
            dir = self.parent.is_reversed(session) and "desc" or "asc"
            return "order by %s_is_null asc, %s %s" % (key, key, dir)

class FreshDataOnlyColumn(SqlTableColumn):
    def __init__(self, app, name):
        super(FreshDataOnlyColumn, self).__init__(app, name)

        self.__ago = self.TimeAgo(app, "ago")
        self.add_attribute(self.__ago)

    def render_content(self, session, data):
        key = self.get_column_key(session)
        value = data["qmf_update_time"]

        if value and value > self.__ago.get(session):
            html = self.render_value(session, data[key])
        else:
            html = fmt_none_brief()

        return html

    class TimeAgo(Attribute):
        def get_default(self, session):
            return datetime.now() - timedelta(minutes=10)

class PaginatedItemSet(ItemSet):
    def __init__(self, app, name):
        super(PaginatedItemSet, self).__init__(app, name)

        self.paginator = Paginator(app, "page")
        self.add_child(self.paginator)

    def do_process(self, session, *args):
        super(PaginatedItemSet, self).do_process(session, *args)

        self.paginator.set_count(session, self.get_item_count(session, *args))

    def get_bounds(self, session):
        return self.paginator.get_bounds(session)

class BrokerGroupInput(OptionInputSet):
    def __init__(self, app, name):
        super(BrokerGroupInput, self).__init__(app, name, None)

        self.param = BrokerGroupParameter(app, "param")
        self.add_parameter(self.param)

    def do_get_items(self, session):
        return list(BrokerGroup.select()) #XXX avoid list()ing this?

    def render_item_value(self, session, group):
        return group.id

    def render_item_content(self, session, group):
        return group.name

class BindingSet(CuminTable):
    def __init__(self, app, name):
        super(BindingSet, self).__init__(app, name)

        self.ids = CheckboxIdColumn(app, "id")
        self.add_column(self.ids)
        
        col = self.QNameColumn(app, "q_id")
        col.visible = False
        self.add_column(col)

        col = self.ENameColumn(app, "e_id")
        col.visible = False
        self.add_column(col)

        col = self.KeyColumn(app, "key")
        self.add_column(col)

        col = self.RateColumn(app, "rate")
        col.alignment = "right"
        self.add_column(col)

        col = self.MatchedColumn(app, "matched")
        col.alignment = "right"
        self.add_column(col)
        
        self.phase = PhaseSwitch(app, "phase")
        self.add_child(self.phase)

    def get_args(self, session):
        return self.frame.get_args(session)
        return (obj,)
    
    def get_sql_values(self, session, obj):
        return {"id": obj.id}

    class QNameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Queue"

        def render_content(self, session, data):
            queue = Queue.get(data["q_id"])
            href = self.page.main.broker.queue.get_href(session, queue)
            return fmt_link(href, fmt_shorten(queue.name))

    class ENameColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Exchange"

        def render_content(self, session, data):
            exchange = Exchange.get(data["e_id"])
            href = self.page.main.broker.exchange.get_href(session, exchange)

            if exchange.name:
                name = fmt_shorten(exchange.name)
            else:
                name = "<em>Default</em>"

            return fmt_link(href, name)

    class KeyColumn(SqlTableColumn):
        def render_title(self, session, data):
            return "Key"

        def render_value(self, session, value):
            return fmt_shorten(value)

    class RateColumn(ItemTableColumn):
        def render_title(self, session, data):
            return "Message Rate"

        def render_content(self, session, data):
            binding = Binding.get(data["id"])
            return self.app.model.binding.msgMatched.rate_html(binding)

    class MatchedColumn(ItemTableColumn):
        def render_title(self, session, data):
            return "Messages Matched"
            
        def render_content(self, session, data):
            binding = Binding.get(data["id"])
            return self.app.model.binding.msgMatched.value(binding)

class CheckboxInputColumn(FormInput, ItemTableColumn):
    def __init__(self, app, name, item_param):
        super(CheckboxInputColumn, self).__init__(app, name, None)

        self.header_class = CheckboxIdColumnHeader

        self.param = ListParameter(app, "param", item_param)
        self.add_parameter(self.param)

    def do_render(self, session, data):
        name = self.param.path
        id = data[self.name]
        attr = id in self.param.get(session) and "checked=\"checked\"" or ""
        html = "<td><input type=\"checkbox\" name=\"%s\" " + \
            "value=\"%i\" %s/></td>"

        return html % (name, id, attr)

class CheckboxIdColumn(FormInput, SqlTableColumn):
    def __init__(self, app, name):
        super(CheckboxIdColumn, self).__init__(app, name, None)

        self.header_class = CheckboxIdColumnHeader

        item = IntegerParameter(app, "param")
        self.add_parameter(item) # XXX lose this?

        self.param = ListParameter(app, "id", item)
        self.add_parameter(self.param)

    def clear(self, session):
        self.param.set(session, list())

    def do_render(self, session, data, disabled=False):
        name = self.param.path
        id = data[self.name]
        attr = id in self.param.get(session) and "checked=\"checked\"" or ""
        disa = disabled and "disabled=\"disabled\"" or ""
        t = "<td><input type=\"checkbox\" name=\"%s\" value=\"%i\" %s %s/></td>"

        return t % (name, id, attr, disa)

class CheckboxStringIdColumn(FormInput, SqlTableColumn):
    def __init__(self, app, name):
        super(CheckboxStringIdColumn, self).__init__(app, name, None)

        self.header_class = CheckboxIdColumnHeader

        param = Parameter(app, "param")
        self.add_parameter(param)

        self.param = ListParameter(app, "id", param)
        self.add_parameter(self.param)

    def clear(self, session):
        self.param.set(session, list())

    def do_render(self, session, data, disabled=False):
        name = self.param.path
        id = data[self.name]
        attr = id in self.param.get(session) and "checked=\"checked\"" or ""
        disa = disabled and "disabled=\"disabled\"" or ""
        t = "<td><input type=\"checkbox\" name=\"%s\" value=\"%s\" %s %s/></td>"

        return t % (name, id, attr, disa)

class CheckboxIdColumnHeader(ItemTableColumnHeader):
    def render_form_id(self, session, *args):
        return self.column.form.path

    def render_elem_name(self, session, *args):
        return self.column.param.path

class FilteredCheckboxIdColumn(CheckboxIdColumn):
    def __init__(self, app, name, form, callback=None):
        super(FilteredCheckboxIdColumn, self).__init__(app, name)
        
        # call back that returns True if the checkbox is to be disabled
        self.__callback = callback

    def do_render(self, session, data):
        disabled = self.__callback and self.__callback(session, data) or False
        return super(FilteredCheckboxIdColumn, self).do_render(session, data, 
            disabled=disabled)

class NameField(StringField):
    def __init__(self, app, name):
        super(NameField, self).__init__(app, name)

        self.required = True
        self.illegal_chars = ""
        self.legal_chars_desc = None

    def render_title(self, session):
        return "Name"

    def set_required(self, required):
        self.required = required
        
    def do_validate(self, session, errors):
        name = self.get(session)

        if name == "" and self.required:
            errors.append(MissingValueError())
        else:
            for char in self.illegal_chars:
                if char in name:
                    msg = "The name contains illegal characters"
                    if self.legal_chars_desc:
                        msg = msg + "; " + self.legal_chars_desc

                    errors.append(FormError(msg))

                    break

class TextField(NameField):
    def __init__(self, app, name):
        super(TextField, self).__init__(app, name)
        
        self.__title = "Title"
        
    def set_title(self, title):
        self.__title = title
        
    def render_title(self, session):
        return self.__title

class ExchangeNameField(NameField):
    def __init__(self, app, name):
        super(ExchangeNameField, self).__init__(app, name)
        
        self.illegal_chars = " (){}[]-<>&%"
        self.legal_chars_desc = """
            The exchange name is invalid; allowed characters are
            letters, digits, ".", and "_"
            """
        
class UniqueNameField(NameField):
    def __init__(self, app, name, cls, fld="name"):
        super(UniqueNameField, self).__init__(app, name)

        self.__class = cls
        self.__field = fld
        self.__object = None

    def set_object_attr(self, attr):
        self.__object = attr

    def do_validate(self, session, errors):
        super(UniqueNameField, self).do_validate(session, errors)

        if not errors:
            name = self.get(session)

            args = {self.__field: name, }
            results = self.__class.selectBy(**args)

            if self.__object:
                object = self.__object.get(session)
                if object:
                    results = results.filter(self.__class.q.id != object.id)

            if results.count() > 0:
                errors.append(DuplicateValueError())

class DuplicateValueError(FormError):
    def __init__(self, fld="name"):
        super(DuplicateValueError, self).__init__()

        self.__field = fld

    def get_message(self, session):
        return "An item with this %s already exists" % self.__field

class TwoOptionRadioField(RadioField):
    def __init__(self, app, name, option1="yes", option2="no"):
        super(TwoOptionRadioField, self).__init__(app, name, None)

        self.param = Parameter(app, "param")
        self.param.default = option2
        self.add_parameter(self.param)

        option = self.Option1(app, option1, self.param)
        self.add_option(option)
        
        option = self.Option2(app, option2, self.param)
        self.add_option(option)

    def render_title(self, session):
        return "Pick an option"

    def render_title_1(self, session):
        pass
    
    def render_title_2(self, session):
        pass
    
    class Option1(RadioFieldOption):
        def render_title(self, session):
            return self.parent.render_title_1(session)

    class Option2(RadioFieldOption):
        def render_title(self, session):
            return self.parent.render_title_2(session)

class MultiplicityField(RadioField):
    def __init__(self, app, name):
        super(MultiplicityField, self).__init__(app, name, None)

        self.param = Parameter(app, "param")
        self.param.default = "all"
        self.add_parameter(self.param)

        self.add_option(self.All(app, "all", self.param))
        self.add_option(self.Top(app, "top", self.param))
        self.top_n = self.TopN(app, "topn", self.param)
        self.add_option(self.top_n)

    def render_title(self, session):
        return "Select all, the top message or the top n messages"

    class All(RadioFieldOption):
        def render_value(self, session):
            return "all"

        def render_title(self, session):
            return "All"

    class Top(RadioFieldOption):
        def render_value(self, session):
            return "top"

        def render_title(self, session):
            return "Top message"

    class TopN(RadioFieldOption):
        def __init__(self, app, name, param):
            super(MultiplicityField.TopN, self).__init__(app, name, param)
        
            self.__n_value = IntegerParameter(app, "arg")
            self.add_parameter(self.__n_value)

        def render_value(self, session):
            return "N"

        def render_arg_name(self, session):
            return self.__n_value.path
            
        def render_arg_value(self, session):
            return 0

        def get_n_value(self, session):
            return self.__n_value.get(session)

        def render_title(self, session):
            return 

class AjaxField(Widget):
    """ Update a single span or div after the page loads """
    
    def render_script(self, session):
        script = """
        <script type="text/javascript">
        function get_%s() { 
            var now = new Date();
            wooly.deferredUpdate('%s'+';ts='+now.getTime(), got_%s, "%s"); 
        }
        addEvent(window, "load", get_%s);
        </script>
        """
        get_fn = self.get_fn(session)
        url = self.get_url(session)
        got_fn = self.got_fn(session)
        elem_id = self.elem_id(session)
        
        return url and script % (get_fn, url, got_fn, elem_id, get_fn) or ""
    
    def get_url(self, session):
        pass

    def get_fn(self, session):
        return self.name
    
    def got_fn(self, session):
        return self.name
    
    def elem_id(self, session):
        return self.name

class StartStopStatus(CuminStatus, AjaxField):
    def render_start_stop_script(self, session, scheduler):
        return self.render_script(session)

    def render_color(self, session, scheduler):
        return "clear"
        
    def got_fn(self, session):
        return "startstop_status"
    
    def elem_id(self, session):
        return self.render_id(session)
    
class MoreFieldSet(FieldForm, FormField):
    """ Displays a button that opens and closes a set of fields
    
        Used in a FieldForm as a FormField. Instead of calling
        add_field on the FieldForm, call add_field on this.
        For example: 
            self.more = MoreFieldSet(app, "more")
            self.add_field(self.more)
            
            self.name = NameField(app, "name")
            self.more.add_field(self.name)

            self.port = StringField(app, "port")
            self.more.add_field(self.port)
        """
    def __init__(self, app, name):
        super(MoreFieldSet, self).__init__(app, name)
        
        self.input = HiddenInput(app, "open")
        self.add_child(self.input)

    def render_inputs(self, session, *args):
        return self.render_fields(session, *args)

    def render_more_text(self, session, *args):
        return "Show Advanced Options..."
    
    def render_less_text(self, session, *args):
        return "Hide Advanced Options..."

    def render_state_text(self, session, *args):
        return self.input.get(session) \
            and self.render_less_text(session, *args) \
            or self.render_more_text(session, *args)
             
    def render_open_state(self, session, *args):
        return self.input.do_marshal(self.input.get(session))
    
    def render_open_path(self, session, *args):
        return self.input.render_name(session, *args)
    
    def render_open_display(self, session, *args):
        return self.input.get(session) and "block" or "none"
