import logging
import qmf.console
import pickle
import struct
import sys
import types
import socket
from threading import Lock, RLock
from sqlobject import *
from traceback import print_exc
from qpid.util import URL
from qpid.datatypes import UUID
from mint.schema import *
from mint import update
from mint.cache import MintCache
from mint.dbexpire import DBExpireThread
from qmf.console import ClassKey
from util import *
from time import sleep
from psycopg2 import OperationalError

log = logging.getLogger("mint")

thisModule = __import__(__name__)

for item in dir(schema):
  cls = getattr(schema, item)

  try:
    if issubclass(cls, SQLObject) and cls is not SQLObject:
      setattr(thisModule, item, cls)
  except TypeError:
    pass

Broker.sqlmeta.addColumn(ForeignKey("BrokerRegistration",
                                    cascade="null", default=None,
                                    name="registration"))

class MintDatabase(object):
  def __init__(self, uri):
    self.uri = uri

  def getConnection(self):
    return connectionForURI(self.uri).getConnection()

  def check(self):
    self.checkConnection()

  def init(self):
    sqlhub.processConnection = connectionForURI(self.uri)

  def checkConnection(self):
    conn = self.getConnection()

    try:
      cursor = conn.cursor()
      cursor.execute("select now()")
    finally:
      conn.close()

  def checkSchema(self):
    pass

  def dropSchema(self):
    conn = self.getConnection()
    try:
      cursor = conn.cursor()

      cursor.execute("drop schema public cascade")

      conn.commit()
    finally:
      conn.close()

  def __splitSQLStatements(self, text):
    result = list()
    unmatchedQuote = False
    tmpStmt = ""

    for stmt in text.split(";"):
      stmt = stmt.rstrip()
      quotePos = stmt.find("'")
      while quotePos > 0:
        quotePos += 1
        if quotePos < len(stmt):
          if stmt[quotePos] != "'":
            unmatchedQuote = not unmatchedQuote
          else:
            # ignore 2 single quotes
            quotePos += 1
        quotePos = stmt.find("'", quotePos)

      if len(stmt.lstrip()) > 0:
          tmpStmt += stmt + ";"
          if not unmatchedQuote:
            # single quote has been matched/closed, generate statement
            result.append(tmpStmt.lstrip())
            tmpStmt = ""

    if unmatchedQuote:
        result.append(tmpStmt.lstrip())
    return result

  def createSchema(self, file_paths):
    conn = self.getConnection()

    scripts = list()

    for path in file_paths:
      file = open(path, "r")
      try:
        scripts.append((path, file.read()))
      finally:
        file.close()

    try:
      cursor = conn.cursor()

      try:
        cursor.execute("create schema public");
      except:
        conn.rollback()
        pass

      for path, text in scripts:
        stmts = self.__splitSQLStatements(text)
        count = 0

        for stmt in stmts:
          stmt = stmt.strip()

          if stmt:
            try:
              cursor.execute(stmt)
            except Exception, e:
              print "Failed executing statement:"
              print stmt

              raise e

            count += 1

        print "Executed %i statements from file '%s'" % (count, path)

      conn.commit()

      info = MintInfo(version="0.1")
      info.sync()

      # Standard roles

      user = Role(name="user")
      user.sync()

      admin = Role(name="admin")
      admin.sync()
    finally:
      conn.close()

  def checkSchema(self):
    conn = self.getConnection()

    try:
      cursor = conn.cursor()

      try:
        cursor.execute("select version from mint_info");
      except Exception, e:
        print "No schema present"
        return

      for rec in cursor:
        print "OK (version %s)" % rec[0]
        return;

      print "No schema present"
    finally:
      conn.close()

class Subject(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  name = StringCol(length=1000, default=None, unique=True, notNone=True)
  password = StringCol(length=1000, default=None)
  lastChallenged = TimestampCol(default=None)
  lastLoggedIn = TimestampCol(default=None)
  lastLoggedOut = TimestampCol(default=None)
  roles = SQLRelatedJoin("Role", intermediateTable="subject_role_mapping",
                         createRelatedTable=False)

  def getByName(cls, name):
    try:
      return Subject.selectBy(name=name)[0]
    except IndexError:
      pass

  getByName = classmethod(getByName)

class Role(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  name = StringCol(length=1000, default=None, unique=True, notNone=True)
  subjects = SQLRelatedJoin("Subject",
                            intermediateTable="subject_role_mapping",
                            createRelatedTable=False)

  def getByName(cls, name):
    try:
      return Role.selectBy(name=name)[0]
    except IndexError:
      pass

  getByName = classmethod(getByName)

class SubjectRoleMapping(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  subject = ForeignKey("Subject", notNull=True, cascade=True)
  role = ForeignKey("Role", notNull=True, cascade=True)
  unique = DatabaseIndex(subject, role, unique=True)

class ObjectNotFound(Exception):
  pass

class MintInfo(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  version = StringCol(length=1000, default="0.1", notNone=True)

class CollectorRegistration(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  name = StringCol(length=1000, default=None)
  collectorId = StringCol(length=1000, default=None)

class BrokerRegistration(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  name = StringCol(length=1000, default=None, unique=True, notNone=True)
  url = StringCol(length=1000, default=None)
  broker = ForeignKey("Broker", cascade="null", default=None)
  groups = SQLRelatedJoin("BrokerGroup",
                          intermediateTable="broker_group_mapping",
                          createRelatedTable=False)
  cluster = ForeignKey("BrokerCluster", cascade="null", default=None)
  profile = ForeignKey("BrokerProfile", cascade="null", default=None)

  url_unique = DatabaseIndex(url, unique=True)

  def getBrokerId(self):
    return self.mintBroker.qmfId

  def getDefaultVhost(self):
    if self.broker:
      try:
        return Vhost.selectBy(broker=self.broker, name="/")[0]
      except IndexError:
        return None

class BrokerGroup(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  name = StringCol(length=1000, default=None, unique=True, notNone=True)
  brokers = SQLRelatedJoin("BrokerRegistration",
                           intermediateTable="broker_group_mapping",
                           createRelatedTable=False)

class BrokerGroupMapping(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  brokerRegistration = ForeignKey("BrokerRegistration", notNull=True,
                                  cascade=True)
  brokerGroup = ForeignKey("BrokerGroup", notNull=True, cascade=True)
  unique = DatabaseIndex(brokerRegistration, brokerGroup, unique=True)

class BrokerCluster(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  name = StringCol(length=1000, default=None)
  brokers = SQLMultipleJoin("BrokerRegistration", joinColumn="cluster_id")

class BrokerProfile(SQLObject):
  class sqlmeta:
    lazyUpdate = True

  name = StringCol(length=1000, default=None)
  brokers = SQLMultipleJoin("BrokerRegistration", joinColumn="profile_id")
  properties = SQLMultipleJoin("ConfigProperty", joinColumn="profile_id")

class MintBroker(object):
  def __init__(self, url, qmfBroker):
    self.url = url
    self.qmfBroker = qmfBroker

    self.qmfId = None
    self.databaseId = None

    self.objectDatabaseIds = MintCache() # database ids by qmf object id
    self.orphans = dict() # updates by qmf object id

    self.connected = False

  def __repr__(self):
    return "%s(%s)" % (self.__class__.__name__, self.url)

  def getAmqpSession(self):
    return self.qmfBroker.getAmqpSession()

  def getAmqpSessionId(self):
    return self.qmfBroker.getSessionId()

  def getFullUrl(self):
    return self.qmfBroker.getFullUrl()

class MintModel(qmf.console.Console):
  staticInstance = None

  def __init__(self, dataUri, dbExpireFrequency, dbExpireThreshold, debug=False):
    self.dataUri = dataUri
    self.debug = debug
    self.updateObjects = True
    self.pollRegistrations = True
    self.dbExpiration = True

    assert MintModel.staticInstance is None
    MintModel.staticInstance = self

    # Lookup tables used for recovering MintBroker objects, which have
    # mint-specific accounting and wrap a qmf broker object

    self.mintBrokersByQmfBroker = dict()
    self.mintBrokersById = dict()
    self.mintBrokersByUrl = dict()

    self.__lock = RLock()

    self.dbConn = None
    self.qmfSession = None

    self.updateThread = update.ModelUpdateThread(self)
    self.dbExpireThread = dbexpire.DBExpireThread \
        (self, frequency=dbExpireFrequency, threshold=dbExpireThreshold)
    self.registrationThread = RegistrationThread(self)

    self.outstandingMethodCalls = dict()

  def lock(self):
    self.__lock.acquire()

  def unlock(self):
    self.__lock.release()

  def check(self):
    try:
      connectionForURI(self.dataUri)
    except Exception, e:
      if hasattr(e, "message") and e.message.find("does not exist"):
        print "Database not found; run cumin-database-init"
      raise e

  def init(self):
    log.info("Object updates are %s",
             self.updateObjects and "enabled" or "disabled")

    log.info("QMF server polling is %s",
             self.pollRegistrations and "enabled" or "disabled")

    log.info("Object expiration is %s",
             self.dbExpiration and "enabled" or "disabled")

    sqlhub.processConnection = self.dbConn = connectionForURI(self.dataUri)

    assert self.qmfSession is None

    self.qmfSession = qmf.console.Session \
        (self, manageConnections=True, rcvObjects=self.updateObjects)

    self.updateThread.init()

  def start(self):
    self.updateThread.start()

    if self.dbExpiration:
      self.dbExpireThread.start()

    if self.pollRegistrations:
      self.registrationThread.start()

  def stop(self):
    self.updateThread.stop()

    if self.dbExpiration:
      self.dbExpireThread.stop()

    if self.pollRegistrations:
      self.registrationThread.stop()

  def callMethod(self, brokerId, objId, classKey, methodName, callback, args):
    self.lock()
    try:
      broker = self.mintBrokersById[brokerId]
    finally:
      self.unlock()

    seq = self.qmfSession._sendMethodRequest \
        (broker.qmfBroker, ClassKey(classKey), objId, methodName, args)

    if seq is not None:
      self.lock()
      try:
        self.outstandingMethodCalls[seq] = callback
      finally:
        self.unlock()
    return seq

  def addBroker(self, url):
    log.info("Adding qmf broker at %s", url)

    self.lock()
    try:
      qbroker = self.qmfSession.addBroker(url)
      mbroker = MintBroker(url, qbroker)

      # Can't add the by-id mapping here, as the id is not set yet;
      # instead we add it when brokerConnected is called

      self.mintBrokersByQmfBroker[qbroker] = mbroker
      self.mintBrokersByUrl[url] = mbroker

      return mbroker
    finally:
      self.unlock()

  def delBroker(self, mbroker):
    assert isinstance(mbroker, MintBroker)

    log.info("Removing qmf broker at %s", mbroker.url)

    self.lock()
    try:
      self.qmfSession.delBroker(mbroker.qmfBroker)

      del self.mintBrokersByQmfBroker[mbroker.qmfBroker]
      del self.mintBrokersById[mbroker.qmfId]
      del self.mintBrokersByUrl[mbroker.url]
    finally:
      self.unlock()

  def brokerConnected(self, qbroker):
    """ Invoked when a connection is established to a broker """
    self.lock()
    try:
      mbroker = self.mintBrokersByQmfBroker[qbroker]

      assert mbroker.connected is False

      mbroker.connected = True
    finally:
      self.unlock()

  def brokerInfo(self, qbroker):
    self.lock()
    try:
      id = str(qbroker.getBrokerId())
      mbroker = self.mintBrokersByQmfBroker[qbroker]

      mbroker.qmfId = id
      self.mintBrokersById[id] = mbroker
    finally:
      self.unlock()

  def brokerDisconnected(self, qbroker):
    """ Invoked when the connection to a broker is lost """
    self.lock()
    try:
      mbroker = self.mintBrokersByQmfBroker[qbroker]

      assert mbroker.connected is True

      mbroker.connected = False
    finally:
      self.unlock()

  def getMintBrokerByQmfBroker(self, qbroker):
    self.lock()
    try:
      return self.mintBrokersByQmfBroker[qbroker]
    finally:
      self.unlock()

  def newPackage(self, name):
    """ Invoked when a QMF package is discovered. """
    pass

  def newClass(self, kind, classKey):
    """ Invoked when a new class is discovered.  Session.getSchema can be
    used to obtain details about the class."""
    pass

  def newAgent(self, agent):
    """ Invoked when a QMF agent is discovered. """
    pass

  def delAgent(self, agent):
    """ Invoked when a QMF agent disconects. """
    pass

  def objectProps(self, broker, record):
    """ Invoked when an object is updated. """

    if not self.updateObjects:
      return

    mbroker = self.getMintBrokerByQmfBroker(broker)

    up = update.PropertyUpdate(self, mbroker, record)

    if record.getClassKey().getClassName() == "job":
      up.priority = 1

    self.updateThread.enqueue(up)

  def objectStats(self, broker, record):
    """ Invoked when an object is updated. """

    if not self.updateObjects:
      return

    if record.getClassKey().getClassName() == "job":
      return

    mbroker = self.getMintBrokerByQmfBroker(broker)
    up = update.StatisticUpdate(self, mbroker, record)

    self.updateThread.enqueue(up)

  def event(self, broker, event):
    """ Invoked when an event is raised. """
    pass

  def heartbeat(self, agent, timestamp):
    pass

  def methodResponse(self, broker, seq, response):
    mbroker = self.getMintBrokerByQmfBroker(broker)
    up = update.MethodUpdate(self, mbroker, seq, response)

    self.updateThread.enqueue(up)

class RegistrationThread(MintDaemonThread):
  def run(self):
    log.info("Polling for registration changes every 5 seconds")

    while True:
      try:
        if self.stopRequested:
          break

        self.do_run()
      except OperationalError, e:
        log.exception(e)
        if str(e).find("server closed the connection unexpectedly") >= 0:
          sys.exit()
      except Exception, e:
        log.exception(e)

  def do_run(self):
    regUrls = set()

    for reg in BrokerRegistration.select():
      if reg.url not in self.model.mintBrokersByUrl:
        try:
          self.model.addBroker(reg.url)
        except socket.error, e:
          log.info("Can't connect to broker at %s: %s", reg.url, e)
          pass

      regUrls.add(reg.url)

    for mbroker in self.model.mintBrokersByQmfBroker.values():
      if mbroker.url not in regUrls:
        self.model.delBroker(mbroker)

    sleep(5)
