import mllib
from sqlobject import *

class SchemaParser:
  """parses broker XML schema"""

  def __init__(self, pythonFilePath, sqlTriggersFilePath, xmlFilePaths):
    self.pythonFilePath = pythonFilePath
    self.sqlTriggersFilePath = sqlTriggersFilePath
    self.xmlFilePaths = xmlFilePaths
    self.style = MixedCaseUnderscoreStyle()
    self.additionalPythonOutput = ""
    self.currentClass = ""
    self.pythonOutput = ""
    self.finalPythonOutput = ""
    self.sqlTriggersOutput = ""
    self.entityClasses = []
    self.statsClasses = []
    self.groups = dict()
    # mapping between xml schema types and database column types
    # see xml/MintTypes.xml
    self.dataTypesMap = dict()
    self.dataTypesMap["objId"] = "ForeignKey"
    self.dataTypesMap["uuid"] = "BLOBCol"
    self.dataTypesMap["int32"] = "IntCol"
    self.dataTypesMap["uint8"] = self.dataTypesMap["hilo8"] = self.dataTypesMap["count8"] = self.dataTypesMap["mma8"] = "SmallIntCol"
    self.dataTypesMap["hilo16"] = self.dataTypesMap["count16"] = self.dataTypesMap["mma16"] = "SmallIntCol"
    self.dataTypesMap["uint16"] = "IntCol"
    self.dataTypesMap["hilo32"] = self.dataTypesMap["count32"] = self.dataTypesMap["mma32"] = "BigIntCol"
    self.dataTypesMap["uint32"] = "BigIntCol"
    self.dataTypesMap["uint64"] = self.dataTypesMap["hilo64"] = self.dataTypesMap["count64"] = self.dataTypesMap["mma64"] = self.dataTypesMap["mmaTime"] = "BigIntCol"
    self.dataTypesMap["float"] = self.dataTypesMap["double"] = "FloatCol"
    self.dataTypesMap["absTime"] = "TimestampCol"
    self.dataTypesMap["deltaTime"] = "BigIntCol"
    self.dataTypesMap["bool"] = "BoolCol"
    self.dataTypesMap["sstr"] = self.dataTypesMap["lstr"] = "StringCol"
    self.dataTypesMap["map"] = "PickleCol"
    # mapping for identifiers in the XML schema that are reserved words in either SQL or Python
    self.reservedWords = {"in": "inRsv", "In": "InRsv", 
                          "connection": "clientConnection", "Connection": "ClientConnection", 
                          "connectionRef": "clientConnectionRef", 
                          "user": "gridUser", "User": "GridUser",
                          "registeredTo": "broker",
                          "sync": "syncRsv"}

  def renameReservedWord(self, name):
    if (name in self.reservedWords.keys()):
      print "Notice: %s is a reserved word, automatically translating to %s" % (name, self.reservedWords[name])
      return self.reservedWords[name]
    else:
      return name
 
  def attrNameFromDbColumn(self, name, removeSuffix=""):
    return self.style.dbColumnToPythonAttr(name.replace(removeSuffix, ""))
  
  def generateAttrib(self, attribName, attribType, params=""):
    if (params.find("default") < 0):
      if (params == ""):
        params = "default=None"
      else:
        params += ", default=None"
    if attribName == "id":
      attribName = "id_"
    # special case for "port" attrib, needs to be a 2-byte unsigned 
    # but uint16 converts to a signed int (SmallIntCol), so forcing the next size up (IntCol)
    if (attribName == "port" and attribType == self.dataTypesMap["uint16"]):
      attribType = self.dataTypesMap["uint32"]
    self.pythonOutput += "  %s = %s(%s)\n" % (attribName, attribType, params)

  def generateTimestampAttrib(self, col):
    self.generateAttrib("qmf" + col + "Time", "TimestampCol") #, "default=datetime.min")

  def generateForeignKeyAttrib(self, name, reference):
    params = "'%s', cascade='null'" % (reference)
    name = self.renameReservedWord(name)
    self.generateAttrib(name, "ForeignKey", params)

  def generateForeignKeyAttribWithIndex(self, name, reference):
    self.generateForeignKeyAttrib(name, reference)
    name = self.renameReservedWord(name)
    self.pythonOutput += "  %sIndex = DatabaseIndex(%s)\n" % (name, name)
    
  def generateHiLoAttrib(self, name, type):
    self.generateAttrib(name, type)
    self.generateAttrib(name + "Low", type)
    self.generateAttrib(name + "High", type)

  def generateMinMaxAvgAttrib(self, name, type):
    self.generateAttrib(name + "Min", type)
    self.generateAttrib(name + "Max", type)
    self.generateAttrib(name + "Average", type)
    self.generateAttrib(name + "Samples", type)

  def generateMultipleJoin(self, tableFrom, tableTo, attrib=""):
    if (attrib == ""):
      attrib = tableTo[0].lower() + tableTo[1:] + "s"
    self.additionalPythonOutput += "\n%s.sqlmeta.addJoin(SQLMultipleJoin('%s', joinMethodName='%s'))\n" % (tableFrom, tableTo, attrib)

  def generateLazyUpdate(self, lazyUpdate=True):
    self.pythonOutput += "  class sqlmeta:\n"
    self.pythonOutput += "    lazyUpdate = %s\n\n" % (lazyUpdate)

  def generateQmfIdsIndex(self, className):
    self.pythonOutput += "  qmfIdsUnique = DatabaseIndex(qmfBrokerId, qmfScopeId, qmfObjectId, unique=True)\n"

  def generateClassAttribs(self, schemaName, elements):
    if (schemaName == "JournalStats"):
      print schemaName
    for elem in elements:
      elemName = self.renameReservedWord(elem["@name"])
      if (elem["@type"] == "objId"):
        reference = elem["@references"]
        #XXX: TO-DO: properly handle namespaces
        # handle cases where the referenced class is in a different namespace (ie, contains a "." or a ":");
        # for now, discard namespace
        namespaceIndex = max(reference.rfind("."), reference.rfind(":"))
        if (namespaceIndex > 0):
          reference = reference[namespaceIndex + 1:]
        reference = self.style.dbTableToPythonClass(reference)
        reference = self.renameReservedWord(reference)
        attrib = reference[0].lower() + reference[1:]
        self.generateForeignKeyAttrib(attrib, reference)
        self.generateMultipleJoin(reference, self.currentClass)
      elif (elem["@type"].startswith("hilo")):
        self.generateHiLoAttrib(self.attrNameFromDbColumn(elemName), self.dataTypesMap[elem["@type"]])
      elif (elem["@type"].startswith("mma")):
        self.generateMinMaxAvgAttrib(self.attrNameFromDbColumn(elemName), self.dataTypesMap[elem["@type"]])
      else:
        args = ""
        if (elem["@type"] == "sstr"):
          args += "length=1000"
        elif (elem["@type"] == "lstr" or elem["@type"] == "ftable"):
          args += "length=4000"
        self.generateAttrib(self.attrNameFromDbColumn(elemName), self.dataTypesMap[elem["@type"]], args)
    self.pythonOutput += "\n"

  def startClass(self, schemaName, stats=False):
    schemaName = self.renameReservedWord(schemaName)
    if (stats):
      origPythonName = self.style.dbTableToPythonClass(schemaName)
      pythonName = self.style.dbTableToPythonClass(schemaName + "_stats")
      colPythonName = self.style.dbColumnToPythonAttr(schemaName)
      keyPythonName = self.style.dbTableToPythonClass(schemaName)
      sqlTable = self.style.pythonClassToDBTable(pythonName)
      sqlParentTable = self.style.pythonClassToDBTable(self.style.dbTableToPythonClass(schemaName))
      self.sqlTriggersOutput += self.sqlTriggerFunction % (sqlParentTable, sqlParentTable, sqlParentTable)
      self.sqlTriggersOutput += "\n"
      self.sqlTriggersOutput += "CREATE TRIGGER update_%s_stats AFTER INSERT ON %s \n" % (sqlParentTable, sqlTable)
      self.sqlTriggersOutput += "  FOR EACH ROW EXECUTE PROCEDURE update_%s_stats(); \n\n" % (sqlParentTable)
      self.sqlTriggersOutput += "CREATE INDEX %s_update_time ON %s (qmf_update_time);\n\n" % (sqlTable, sqlTable)
    else:
      pythonName = self.style.dbTableToPythonClass(schemaName)
      statsPythonName = self.style.dbTableToPythonClass(schemaName + "_stats")
    self.currentClass = pythonName
    self.pythonOutput += "\nclass %s(SQLObject):\n" % (pythonName)
    self.generateLazyUpdate()
    if (stats):
      self.statsClasses.append(str(pythonName))
      self.generateTimestampAttrib("Update")
      self.generateForeignKeyAttrib(colPythonName[0].lower() + colPythonName[1:], keyPythonName)
      self.generateMultipleJoin(origPythonName, pythonName, "stats")
    else:
      self.entityClasses.append(str(pythonName))
      self.generateAttrib("qmfBrokerId", "StringCol", "length=1000")
      self.generateAttrib("qmfScopeId", "BigIntCol")
      self.generateAttrib("qmfObjectId", "BigIntCol")
      self.generateQmfIdsIndex(pythonName)
      self.generateAttrib("qmfClassKey", "StringCol", "length=1000")
      self.generateTimestampAttrib("Update")
      self.generateTimestampAttrib("Create")
      self.generateTimestampAttrib("Delete")
      self.generateForeignKeyAttribWithIndex("statsCurr", statsPythonName)
      self.generateForeignKeyAttribWithIndex("statsPrev", statsPythonName)
      self.finalPythonOutput += "classToSchemaNameMap['%s'] = '%s'\n" % (pythonName, schemaName)
      self.finalPythonOutput += "schemaNameToClassMap['%s'] = %s\n" % (schemaName, pythonName)
    # XXX get rid of this
    self.pythonOutput += "  classInfos = dict() # brokerId => classInfo\n"
      
  def generateMethod(self, elem):
    if (elem["@desc"] != None):
      comment = '    """' + elem["@desc"] + '"""\n'
    else:
      comment = ""
    formalArgs = ", "
    actualArgs = "    actualArgs = list()\n"
    for arg in elem.query["arg"]:
      formalArgs += "%s, " % (arg["@name"])
      actualArgs += "    if %s is not None:\n" % (arg["@name"])
      actualArgs += "        actualArgs.append(%s)\n" % (arg["@name"])

    if (formalArgs != ", "):
      formalArgs = formalArgs[:-2]
    else:
      formalArgs = ""
    self.pythonOutput += "\n  def %s(self, model, callback%s):\n" % (elem["@name"], formalArgs)
    self.pythonOutput += comment
    self.pythonOutput += actualArgs
    self.pythonOutput += "    originalId = ObjectId(None, self.qmfScopeId, self.qmfObjectId)\n"
    self.pythonOutput += "    model.callMethod(self.qmfBrokerId, originalId, self.qmfClassKey, \"%s\",\n" % elem["@name"]
    self.pythonOutput += "                     callback, args=actualArgs)\n"

  def endClass(self):
    if (self.additionalPythonOutput != ""):
      self.finalPythonOutput += self.additionalPythonOutput + "\n"
      self.additionalPythonOutput = ""
    if (self.pythonOutput.endswith("(SQLObject):\n")):
        self.pythonOutput += "  pass\n"
    self.currentClass = ""
    
  def generateCode(self):
    self.pythonOutput += "import mint\n"
    self.pythonOutput += "from sqlobject import *\n"
    self.pythonOutput += "from datetime import datetime\n"
    self.pythonOutput += "from qmf.console import ObjectId\n\n"
    self.pythonOutput += "class Pool(SQLObject):\n"
    self.pythonOutput += "  class sqlmeta:\n"
    self.pythonOutput += "    lazyUpdate = True\n"
    self.pythonOutput += "  sourceId = StringCol(length=1000, default=None, unique=True)\n\n"
    self.finalPythonOutput += "\nclassToSchemaNameMap = dict()\n"
    self.finalPythonOutput += "schemaNameToClassMap = dict()\n"
    self.finalPythonOutput += 'schemaReservedWordsMap = {"in": "inRsv", "In": "InRsv", \n'
    self.finalPythonOutput += '  "connection": "clientConnection", "Connection": "ClientConnection", \n'
    self.finalPythonOutput += '  "connectionRef": "clientConnectionRef", \n'
    self.finalPythonOutput += '  "user": "gridUser", "User": "GridUser", \n'
    self.finalPythonOutput += '  "registeredTo": "broker",\n'
    self.finalPythonOutput += '  "sync": "syncRsv"} \n\n'

    # TODO: optimize getting the id to the parent table from new.parent_table_id
    self.sqlTriggersOutput += """
CREATE OR REPLACE FUNCTION create_plpgsql() RETURNS TEXT AS '
  CREATE LANGUAGE plpgsql;
  SELECT ''plpgsql language created''::TEXT;
' LANGUAGE sql;

SELECT CASE WHEN
  (SELECT true
     FROM pg_language
    WHERE lanname='plpgsql')
  THEN
    (SELECT 'plpgsql language already installed'::TEXT)
  ELSE
    (SELECT create_plpgsql())
END;

DROP FUNCTION create_plpgsql();

"""
    self.sqlTriggerFunction = """
CREATE OR REPLACE FUNCTION update_%s_stats() RETURNS trigger AS '
BEGIN
  UPDATE %s SET stats_prev_id = stats_curr_id, stats_curr_id = new.id WHERE id = new.%s_id;
  RETURN new;
END
' LANGUAGE plpgsql;
"""

    outputFile = open(self.pythonFilePath, "w")
    sqlTriggersFile = open(self.sqlTriggersFilePath, "w")
    for xmlFile in self.xmlFilePaths:
      schema = mllib.xml_parse(xmlFile)
      # parse groups and store their structure as is
      groups = schema.query["schema/group"]
      for grp in groups:
        self.groups[grp["@name"]] = grp.query["property"], grp.query["statistic"]

      # parse class definitions
      classes = schema.query["schema/class"]
      for cls in classes:
        self.startClass(cls["@name"])
        self.generateClassAttribs(cls["@name"], cls.query["property"])
        # generate properties attribs from any groups included in this class
        for clsGroup in cls.query["group"]:
          self.generateClassAttribs(cls["@name"], self.groups[clsGroup["@name"]][0])
        for elem in cls.query["method"]:
          self.generateMethod(elem)
        self.endClass()

        self.startClass(cls["@name"], stats=True)
        self.generateClassAttribs(cls["@name"], cls.query["statistic"])
        # generate statistics attribs from any groups included in this class
        for clsGroup in cls.query["group"]:
          self.generateClassAttribs(cls["@name"], self.groups[clsGroup["@name"]][1])
        self.endClass()
        self.pythonOutput += "\n\n"
    self.finalPythonOutput += "\nentityClasses = %s\n" % (self.entityClasses)
    self.finalPythonOutput += "\nstatsClasses = %s\n" % (self.statsClasses)
    outputFile.write(self.pythonOutput + self.finalPythonOutput)
    outputFile.close()
    sqlTriggersFile.write(self.sqlTriggersOutput)
    sqlTriggersFile.close()

if __name__ == "__main__":
  import sys

  if len(sys.argv) < 3:
    print "Usage: schemaparser.py OUTPUT-PYTHON-FILE OUTPUT-SQL-TRIGGERS-FILE INPUT-XML-SCHEMA [INPUT-XML-SCHEMA]*"
    sys.exit(1)
  else:
    parser = SchemaParser(sys.argv[1], sys.argv[2], sys.argv[3:])
    parser.generateCode()
