import sys, os

from time import sleep, clock
from parsley.config import *
from parsley.command import *
from parsley.loggingex import *

from mint import *

class BaseMintTool(Command):
    def __init__(self, name):
        super(BaseMintTool, self).__init__(None, name)

        opt = CommandOption(self, "data")
        opt.argument = "URI"
        opt.description = "Connect to database at URI"

        opt = CommandOption(self, "log-file")
        opt.argument = "PATH"
        opt.description = "Log to file at PATH"

        opt = CommandOption(self, "log-level")
        opt.argument = "LEVEL"
        opt.description = "Log messages at or above LEVEL " + \
            "('debug', 'info', 'warn', 'error')"

        opt = CommandOption(self, "debug")
        opt.description = "Enable debugging; print logging to console"

        opt = CommandOption(self, "expire-frequency")
        opt.argument = "FREQUENCY"
        opt.description = "Frequency (in seconds) of the expiration thread"

        opt = CommandOption(self, "expire-threshold")
        opt.argument = "THRESHOLD"
        opt.description = "Threshold (in seconds) of the age of data before it is expired"

    def check(self):
        if os.getuid() not in (os.stat(__file__).st_uid, 0):
            print "Error: You have insufficient privileges"
            sys.exit(1)

    def init(self):
        super(BaseMintTool, self).init()

        #self.config.init()

        try:
            import psyco
            psyco.full()
        except ImportError:
            pass

    def run(self):
        try:
            opts, args = self.parse(sys.argv)
        except CommandException, e:
            print "Error: %s" % e.message
            e.command.print_help()
            sys.exit(1)

        if "help" in opts:
            self.print_help()
            sys.exit(0)

        level = opts.get("log-level", "warn")

        if "debug" in opts:
            level = "debug"
            file = sys.stderr
        else:
            file = opts.get("log-file", sys.stderr)

        enable_logging("mint", level, file)

        data = opts.get("data", "postgresql://cumin@localhost/cumin")
        freq = int(opts.get("expire-frequency", 600))
        threshold = int(opts.get("expire-threshold", 24 * 3600))

        model = MintModel(data, freq, threshold, debug="debug" in opts)

        self.do_run(opts, args, model)

    def do_run(self, opts, args, model):
        raise Exception("Not implemented")

    def main(self):
        self.check()
        self.init()
        self.run()

class MintServerTool(BaseMintTool):
    def __init__(self, name):
        super(MintServerTool, self).__init__(name)

    def init(self):
        super(MintServerTool, self).init()

        # get better thread switching performance
        sys.setcheckinterval(200)

    def do_run(self, opts, args, model):
        model.check()
        model.init()
        model.start()

        try:
            for arg in args[1:]:
                model.addBroker(arg)

            while True:
                sleep(2)
        finally:
            model.stop()

class MintTestTool(BaseMintTool):
    def __init__(self, name):
        super(MintTestTool, self).__init__(name)

    def do_run(self, opts, args, model):
        model.pollRegistrations = False

        model.check()
        model.init()
        model.start()

        added = list()

        try:
            for arg in args[1:]:
                added.append(model.addBroker(arg))

            while True:
                sleep(2)
        finally:
            for broker in added:
                model.delBroker(broker)

            model.stop()

class MintBenchTool(BaseMintTool):
    def __init__(self, name):
        super(MintBenchTool, self).__init__(name)

    def init(self):
        super(MintBenchTool, self).init()

        # get better thread switching performance
        sys.setcheckinterval(200)

    def do_run(self, opts, args, model):
        model.pollRegistrations = False

        model.check()
        model.init()
        model.start()

        added = list()

        try:
            for arg in args[1:]:
                added.append(model.addBroker(arg))

            enq_last = 0
            deq_last = 0
            drp_last = 0
            dfr_last = 0
            prop_last = 0
            stat_last = 0
            meth_last = 0
            exp_last = 0

            head = "%6s %6s %6s %6s %6s %6s %6s %6s %6s" % \
                ("enqs", "deqs", "depth", "drop", "defer",
                 "prop", "stat", "meth", "exp")
            row = "%6i %6i %6i %6i %6i %6i %6i %6i %6i"

            samples = 0

            while True:
                if samples % 24 == 0:
                    print head

                samples += 1

                sleep(1)

                ut = model.updateThread

                enq = ut.enqueueCount
                deq = ut.dequeueCount
                drp = ut.dropCount
                dfr = ut.deferCount

                prop = ut.propUpdateCount
                stat = ut.statUpdateCount
                meth = ut.methUpdateCount
                exp = ut.expireUpdateCount

                print row % (enq - enq_last,
                             deq - deq_last,
                             enq - deq,
                             drp - drp_last,
                             dfr - dfr_last,
                             prop - prop_last,
                             stat - stat_last,
                             meth - meth_last,
                             exp - exp_last)

                enq_last = enq
                deq_last = deq
                drp_last = drp
                dfr_last = dfr

                prop_last = prop
                stat_last = stat
                meth_last = meth
                exp_last = exp
        finally:

            print "Totals:"

            print row % (enq,
                         deq,
                         enq - deq,
                         drp,
                         dfr,
                         prop,
                         stat,
                         meth,
                         exp)

            #from threading import enumerate
            #for item in enumerate():
            #    print item

            for broker in added:
                model.delBroker(broker)

            model.stop()
