import sys
import logging
import types
import pickle
import psycopg2
import mint
from Queue import Queue as ConcurrentQueue, Full, Empty
from threading import Thread
from traceback import print_exc
from qpid.datatypes import UUID
from mint.schema import *
from time import clock, sleep
from sql import *
from collections import deque
from util import *
from datetime import datetime, timedelta

log = logging.getLogger("mint.update")

class ModelUpdateThread(MintDaemonThread):
  def __init__(self, model):
    super(ModelUpdateThread, self).__init__(model)

    self.updates = UpdateQueue(slotCount=2)

    self.enqueueCount = 0
    self.dequeueCount = 0
    self.statUpdateCount = 0
    self.propUpdateCount = 0
    self.methUpdateCount = 0
    self.expireUpdateCount = 0
    self.dropCount = 0
    self.deferCount = 0
    self.commitThreshold = 100

    self.conn = None

  def init(self):
    self.conn = self.model.dbConn.getConnection()

  def enqueue(self, update):
    try:
      self.updates.put(update)

      self.enqueueCount += 1
    except Full:
      log.exception("Queue is full")

    if self.updates.qsize() > 1000:
      # This is an attempt to yield from the enqueueing thread (this
      # method's caller) to the update thread

      sleep(0.1)

  def run(self):
    while True:
      try:
        update = self.updates.get(True, 1)

        log.debug("Processing %s", update)

        self.dequeueCount += 1

        if self.stopRequested:
          break
      except Empty:
        if self.stopRequested:
          break
        else:
          #log.debug("Queue is empty")
          continue

      self.process_update(update)
        
  def process_update(self, update):
    try:
      update.process(self)

      if self.dequeueCount % self.commitThreshold == 0 \
              or self.updates.qsize() == 0:
        # commit only every "commitThreshold" updates, or whenever
        # the update queue is empty

        self.commit()
    except:
      log.exception("Update failed")

      self.rollback()

  def cursor(self):
    return self.conn.cursor()

  def commit(self):
    self.conn.commit()

    for broker in self.model.mintBrokersByQmfBroker.values():
      broker.objectDatabaseIds.commit()

  def rollback(self):
    self.conn.rollback()

    for broker in self.model.mintBrokersByQmfBroker.values():
      broker.objectDatabaseIds.rollback()

class ReferenceException(Exception):
    def __init__(self, sought):
        self.sought = sought

    def __str__(self):
        return repr(self.sought)

class ModelUpdate(object):
  def __init__(self, model, broker, object):
    self.model = model
    self.broker = broker
    self.object = object
    self.priority = 0

  def __repr__(self):
    return "%s(%s, %s, %i)" % (self.__class__.__name__,
                               str(self.broker),
                               str(self.object),
                               self.priority)

  def getClass(self):
    # XXX this is unfortunate
    name = self.object.getClassKey().getClassName()
    name = mint.schema.schemaReservedWordsMap.get(name, name)
    name = name[0].upper() + name[1:]

    try:
      return schemaNameToClassMap[name]
    except KeyError:
      raise ReferenceException(name)

  def process(self, thread):
    pass

  def processAttributes(self, attrs, cls):
    results = dict()

    for key, value in attrs:
      name = key.__repr__()
      name = mint.schema.schemaReservedWordsMap.get(name, name)

      if key.type == 10:
        self.processReference(name, value, results)
      elif key.type == 8:
        self.processTimestamp(name, value, results)
      elif key.type == 14:
        # Convert UUIDs into their string representation, to be
        # handled by sqlobject
        results[name] = str(value)
      elif key.type == 15:
        #if value:
        results[name] = pickle.dumps(value)
      elif not hasattr(cls, name):
        # Discard attrs that we don't have in our schema
        log.debug("%s has no field '%s'" % (cls, name))
      else:
        results[name] = value

    return results

  # XXX this needs to be a much more straightforward procedure
  def processReference(self, name, oid, results):
    if name.endswith("Ref"):
      name = name[:-3]

    className = name[0].upper() + name[1:]
    otherClass = getattr(mint, className)

    foreignKey = name + "_id"

    id = self.broker.objectDatabaseIds.get(oid)

    if id is None:
      raise ReferenceException(oid)

    results[foreignKey] = id

  def processTimestamp(self, name, tstamp, results):
    if tstamp:
      t = datetime.fromtimestamp(tstamp / 1000000000)
      results[name] = t

class PropertyUpdate(ModelUpdate):
  def process(self, thread):
    try:
      cls = self.getClass()
    except ReferenceException, e:
      log.info("Referenced class %r not found", e.sought)

      thread.dropCount += 1
      return

    oid = self.object.getObjectId()

    try:
      attrs = self.processAttributes(self.object.getProperties(), cls)
    except ReferenceException, e:
      log.info("Referenced object %r not found", e.sought)

      try:
        orphans = self.broker.orphans[oid]
        orphans.append(self)
      except KeyError:
        self.broker.orphans[oid] = list((self,))

      thread.deferCount += 1
      return

    timestamps = self.object.getTimestamps()

    self.processTimestamp("qmfUpdateTime", timestamps[0], attrs)

    if cls is Job and attrs["qmfUpdateTime"] < datetime.now() - timedelta(seconds=thread.model.dbExpireThread.threshold):
      # drop updates for Jobs that are older then the expiration threshold, 
      # since they would be deleted anyway in the next run of the db expiration thread
      log.debug("Property update is older than expiration threshold; skipping it")
      thread.dropCount += 1
      return

    self.processTimestamp("qmfCreateTime", timestamps[1], attrs)

    if timestamps[2] != 0:
      self.processTimestamp("qmfDeleteTime", timestamps[2], attrs)

      log.debug("%s(%s) marked deleted", cls.__name__, oid)

    attrs["qmfScopeId"] = oid.first
    attrs["qmfObjectId"] = oid.second
    attrs["qmfClassKey"] = str(self.object.getClassKey())
    attrs["qmfBrokerId"] = str(self.broker.qmfBroker.getBrokerId())

    cursor = thread.cursor()

    # Cases:
    #
    # 1. Object is utterly new to mint
    # 2. Object is in mint's db, but id is not yet cached
    # 3. Object is in mint's db, and id is cached

    id = self.broker.objectDatabaseIds.get(oid)

    if id is None:
      # Case 1 or 2

      op = SqlGetId(cls)
      op.execute(cursor, attrs)

      rec = cursor.fetchone()

      if rec:
        id = rec[0]

      if id is None:
        # Case 1

        op = SqlInsert(cls, attrs)
        op.execute(cursor, attrs)

        id = cursor.fetchone()[0]

        log.debug("%s(%i) created", cls.__name__, id)

      else:
        # Case 2

        attrs["id"] = id

        op = SqlUpdate(cls, attrs)
        op.execute(cursor, attrs)

        assert cursor.rowcount == 1

      self.broker.objectDatabaseIds.set(oid, id)

      if cls is Broker:
        self.processBroker(cursor, id)
    else:
      # Case 3

      attrs["id"] = id

      op = SqlUpdate(cls, attrs)
      op.execute(cursor, attrs)

      #assert cursor.rowcount == 1

    try:
      orphans = self.broker.orphans.pop(oid)

      if orphans:
        log.info("Re-enqueueing %i orphans whose creation had been deferred",
                 len(orphans))

        for orphan in orphans:
          thread.enqueue(orphan)
    except KeyError:
      pass

    thread.propUpdateCount += 1

  def processBroker(self, cursor, id):
    if self.broker.databaseId is None:
      op = SqlGetBrokerRegistration()
      op.execute(cursor, {"url": self.broker.url})

      rec = cursor.fetchone()

      #print op.text, {"url": self.broker.url}

      if rec:
        rid = rec[0]

        op = SqlAttachBroker()
        op.execute(cursor, {"id": id, "registrationId": rid})

        #print op.text, {"id": id, "registrationId": rid}

      self.broker.databaseId = id

class StatisticUpdate(ModelUpdate):
  def process(self, thread):
    try:
      cls = self.getClass()
    except ReferenceException, e:
      log.info("Referenced class %r not found", e.sought)
      return

    statsCls = getattr(mint, "%sStats" % cls.__name__)

    oid = self.object.getObjectId()
    id = self.broker.objectDatabaseIds.get(oid)

    if id is None:
      thread.dropCount += 1
      return

    timestamps = self.object.getTimestamps()

    tnow = datetime.now()
    t = datetime.fromtimestamp(timestamps[0] / 1000000000)

    if t < tnow - timedelta(seconds=30):
      log.debug("Update is %i seconds old; skipping it", (tnow - t).seconds)

      thread.dropCount += 1
      return

    try:
      attrs = self.processAttributes(self.object.getStatistics(), statsCls)
    except ReferenceException:
      thread.dropCount += 1
      return

    attrs["qmfUpdateTime"] = t > tnow and tnow or t
    attrs["%s_id" % cls.sqlmeta.table] = id

    cursor = thread.cursor()

    op = SqlInsert(statsCls, attrs)
    op.execute(cursor, attrs)

    log.debug("%s(%s) created", statsCls.__name__, id)

    thread.statUpdateCount += 1

class MethodUpdate(ModelUpdate):
  def __init__(self, model, broker, seq, response):
    super(MethodUpdate, self).__init__(model, broker, response)

    self.seq = seq

  def process(self, thread):
    self.model.lock()

    try:
      methodCallback = self.model.outstandingMethodCalls.pop(self.seq)
      methodCallback(self.object.text, self.object.outArgs)
    finally:
      self.model.unlock()

    thread.methUpdateCount += 1

class DBExpireUpdate(ModelUpdate):
  def __init__(self, model):
    super(DBExpireUpdate, self).__init__(model, None, None)

  def process(self, thread):
    cursor = thread.cursor()
    attrs = self.model.dbExpireThread.attrs
    total = 0

    thread.commit()

    for op in self.model.dbExpireThread.ops:
      log.debug("Running expire op %s", op)

      count = op.execute(cursor, attrs)

      thread.commit()

      log.debug("%i records expired", count)

      total += count

    log.debug("%i total records expired", total)

    thread.expireUpdateCount += 1

class UpdateQueue(ConcurrentQueue):
  def __init__(self, maxsize=0, slotCount=1):
    self.slotCount = slotCount
    ConcurrentQueue.__init__(self, maxsize)

  def _init(self, maxsize):
    self.maxsize = maxsize
    self.slots = []

    for i in range(self.slotCount):
      self.slots.append(deque())

  def _qsize(self):
    size = 0

    for i in range(self.slotCount):
      size += len(self.slots[i])

    return size

  def _empty(self):
    return self._qsize() == 0

  def _full(self):
    return self.maxsize > 0 and self._qsize() == self.maxsize

  def _put(self, update):
    slot = update.priority

    if slot in range(self.slotCount):
      self.slots[slot].append(update)
    else:
      raise ValueError("Invalid priority slot")

  def _get(self):
    for slot in range(self.slotCount):
      if len(self.slots[slot]) > 0:
        return self.slots[slot].popleft()
    return None
