import sys, os, logging
from cStringIO import StringIO
from urllib import quote, unquote_plus, unquote
from copy import copy
from time import clock
from datetime import datetime
from urlparse import urlsplit
from time import gmtime

from resources import ResourceFinder, StringCatalog

strings = StringCatalog(__file__)

class Attribute(object):
    def __init__(self, app, name):
        self.app = app
        self.name = name
        self.widget = None
        self.default = None
        self.required = True

        self.path = None

    def init(self):
        if self.widget.path:
            self.path = ".".join((self.widget.path, self.name))
        else:
            self.path = self.name

    def validate(self, session):
        value = self.get(session)

        if value is None and self.required:
            raise Exception("%s not set" % self)

    def get(self, session):
        value = session.get(self.path)

        # Use strict test because empty collections are False
        if value is None:
            default = self.get_default(session)

            if default is not None:
                value = self.set(session, default)

        return value

    def add(self, session, value):
        self.set(session, value)

    def set(self, session, value):
        return session.set(self.path, value)

    def get_default(self, session):
        return self.default

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self.path)

class Parameter(Attribute):
    def __init__(self, app, name):
        super(Parameter, self).__init__(app, name)

        self.is_collection = False
        self.is_dictionary = False

        app.add_parameter(self)

    def marshal(self, object):
        if object == None:
            string = ""
        else:
            string = self.do_marshal(object)

        return string

    def do_marshal(self, object):
        return str(object)

    def unmarshal(self, string):
        return self.do_unmarshal(string)

    def do_unmarshal(self, string):
        return string

class Widget(object):
    def __init__(self, app, name):
        self.app = app
        self.name = name
        self.parent = None
        self.children = list()
        self.children_by_name = dict()
        self.attributes = list()
        self.parameters = list()

        # These are set in the init() pass
        self.ancestors = None
        self.path = None
        self.page = None
        self.frame = None

        self.__main_tmpl = Template(self, "html")

        app.add_widget(self)

        for cls in self.__class__.__mro__:
            app.add_widget_class(cls)
            if cls is Widget:
                break

        self.html_class = "_"

        self.sealed = False

    def init(self):
        #print "Initializing %s" % str(self)

        assert not self.sealed

        ancestors = list()
        widget = self.parent

        while widget is not None:
            ancestors.append(widget)
            widget = widget.parent

        for widget in ancestors:
            if isinstance(widget, Frame):
                self.frame = widget
                break

        self.ancestors = tuple(ancestors)

        pelems = [x.name for x in reversed(ancestors)]
        pelems.append(self.name)
        self.path = ".".join(pelems[1:])

        assert isinstance(self.ancestors[-1], Page)
        self.page = self.ancestors[-1]

        self.sealed = True

        self.init_children()

    def init_children(self):
        for attr in self.attributes:
            attr.init()

        for param in self.parameters:
            param.init()

        for child in self.children:
            child.init()

    def add_child(self, child):
        assert not self.sealed
        assert child is not None
        assert child is not self

        self.children.append(child)
        self.children_by_name[child.name] = child
        child.parent = self

    def add_attribute(self, attribute):
        assert not self.sealed
        self.attributes.append(attribute)
        attribute.widget = self

    def add_parameter(self, parameter):
        assert not self.sealed
        self.parameters.append(parameter)
        parameter.widget = self

    def get_ancestor(self, name):
        for anc in self.ancestors:
            if anc.name == name:
                return anc

    @classmethod
    def get_module_strings(cls):
        module = sys.modules[cls.__module__]
        return module.__dict__.get("strings")

    def get_string(self, key):
        for cls in self.__class__.__mro__:
            if cls is object:
                return

            str = None
            strs = cls.get_module_strings()

            if strs:
                str = strs.get(cls.__name__ + "." + key)

                if str:
                    return str

            str = cls.__dict__.get(key)

            if str:
                return str

    def show(self, session):
        assert self.parent

        self.parent.show_child(session, self)
        self.parent.show(session)

        return self

    def show_child(self, session, child):
        pass

    def save_parameters(self, session, params):
        params.extend(self.parameters)

        for child in self.children:
            child.save_parameters(session, params)

    def get_args(self, session):
        return ()

    def process(self, session):
        if session.debug:
            call = WidgetCall(session.debug.process_stack,
                              self, session, None)
            call.open()

        args = self.get_args(session)
        self.do_process(session, *args)

        if session.debug:
            call.close()

    def do_process(self, session, *args):
        for child in self.children:
            child.process(session)

    def render(self, session):
        if session.debug:
            call = WidgetCall(session.debug.render_stack,
                              self, session, None)
            call.open()

        args = self.get_args(session)
        string = self.do_render(session, *args)

        if session.debug:
            call.close()

        return string

    def do_render(self, session, *args):
        writer = Writer()
        self.__main_tmpl.render(writer, session, *args)
        return writer.to_string()

    def render_id(self, session, *args):
        return self.path

    def render_class(self, session, *args):
        return self.html_class

    def render_href(self, session, *args):
        return session.marshal()

    def render_title(self, session, *args):
        return None

    def render_content(self, session, *args):
        writer = Writer()

        for child in self.children:
           writer.write(child.render(session))

        return writer.to_string()

    def prt(self):
        print self

        for key, value in sorted(self.__dict__.items()):
            print "    %-30s %s" % (key + ":", value)

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self.path)

class Frame(Widget):
    def show(self, session):
        super(Frame, self).show(session)

        return self.page.set_current_frame(session, self)

    def save_parameters(self, session, params):
        frame = self.page.get_current_frame(session)

        if self is frame or self in frame.ancestors:
            super(Frame, self).save_parameters(session, params)

class Page(Frame):
    xml_content_type = "text/xml"
    html_content_type = "text/html"
    xhtml_content_type = "application/xhtml+xml"
    xml_1_0_declaration = """<?xml version="1.0"?>"""
    xhtml_1_1_doctype = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">"""
    xhtml_namespace = "http://www.w3.org/1999/xhtml"

    def __init__(self, app, name):
        super(Page, self).__init__(app, name)

        self.redirect = Attribute(app, "redirect")
        self.add_attribute(self.redirect)

        self.current_frame = self.FrameParameter(app, "frame")
        self.add_parameter(self.current_frame)
        self.set_default_frame(self)

        self.__saved_params = dict()

        self.agent = Attribute(app, "agent")
        self.add_attribute(self.agent)
        
    def init(self):
        assert not self.sealed
        assert self.parent is None

        self.ancestors = ()
        self.path = ""
        self.page = self
        self.frame = self

        self.sealed = True

        self.init_children()

    def get_last_modified(self, session):
        return datetime.utcnow()

    def get_content_type(self, session):
        agent = self.agent.get(session)
        if self.name.endswith(".html") and (agent.find("MSIE 6") == -1):
            return self.xhtml_content_type
        else:
            return self.html_content_type

    def get_cache_control(self, session):
        return None

    def show(self, session):
        pass

    def show_child(self, session, child):
        pass

    def save_session(self, session):
        pass

    def set_redirect_url(self, session, url):
        self.redirect.set(session, url)

        return url

    def get_redirect_url(self, session):
        return self.redirect.get(session)

    # XXX take "current" out of these
    def set_current_frame(self, session, frame):
        self.current_frame.set(session, frame)
        return frame

    def get_current_frame(self, session):
        return self.current_frame.get(session)

    def pop_current_frame(self, session):
        frame = self.get_current_frame(session)
        #print "Popping current frame", frame
        self.set_current_frame(session, frame.frame)

    def set_agent(self, session, agent):
        self.agent.set(session, agent)
        
    def set_default_frame(self, frame):
        self.current_frame.default = frame

    def get_saved_parameters(self, session):
        """
        Gets a list of parameters saved for this page and its current
        frame.

        This, combined with [gs]et_current_frame, serves to discard
        state that is out of scope.  The current pattern is to
        preserve the state of parameters in the current frame and all
        of its ancestor frames.
        """

        frame = self.get_current_frame(session)

        try:
            params = self.__saved_params[frame]
        except KeyError:
            params = list()
            self.save_parameters(session, params)
            self.__saved_params[frame] = params

        return params

    class FrameParameter(Parameter):
        def do_marshal(self, frame):
            return frame.path

        def do_unmarshal(self, path):
            return self.app.get_widget(path)

from parameters import DictParameter

class Application(object):
    def __init__(self):
        self.pages = dict()
        self.default_page = None

        self.widgets = list()
        self.widget_index = None

        self.widget_classes = set()

        self.parameters = list()
        self.parameter_index = None

        self.finder = ResourceFinder()

        self.debug = None

    def enable_debug(self):
        self.debug = self.Debug(self)

    class Debug(object):
        def __init__(self, app):
            self.app = app
            self.sessions = list()
            self.urls = set()

    def add_page(self, page):
        if page.parent:
            raise Exception("Page '%s' is not a root widget" % page.name)

        self.pages[page.name] = page

        page.init()

    def set_default_page(self, page):
        self.pages[""] = page

    def add_widget(self, widget):
        if self.widget_index:
            raise Exception()

        self.widgets.append(widget)

    def add_widget_class(self, cls):
        self.widget_classes.add(cls)

    def get_widget(self, key):
        if not self.widget_index:
            index = dict()

            for widget in self.widgets:
                index[widget.path] = widget

            self.widget_index = index

        return self.widget_index.get(key)

    def add_parameter(self, param):
        if self.parameter_index:
            raise Exception()

        self.parameters.append(param)

    def get_parameter(self, page, key):
        if not self.parameter_index:
            index = dict()

            for param in self.parameters:
                if param.widget:
                    index[(param.widget.page, param.path)] = param

            self.parameter_index = index

        param = self.parameter_index.get((page, key))

        # check for partial match for dictparameters
        if not param:
            if DictParameter.sep() in key:
                dict_key = key.split(DictParameter.sep(), 1)[0]
                param = self.parameter_index.get((page, dict_key))

        return param

    def add_resource_dir(self, dir):
        self.finder.add_dir(dir)

    def get_resource(self, name):
        return self.finder.find(name)

    def __repr__(self):
        return "%s(default_page=%s)" % (self.__class__.__name__, self.default_page)

class Session(object):
    http_date_gmt = "%a, %d %b %Y %H:%M:%S GMT"

    def __init__(self, page, response_headers=None):
        assert isinstance(page, Page)

        self.app = page.app
        self.page = page
        self.trunk = None
        self.values = dict()
        self.origin = None
        self.cookies = dict()

        if response_headers is None:
            self.response_headers = list()
        else:
            self.response_headers = response_headers

        if self.app.debug:
            self.debug = self.Debug(self)
        else:
            self.debug = None

    class Debug(object):
        def __init__(self, session):
            self.session = session
            self.process_stack = list()
            self.render_stack = list()

        def print_process_calls(self, out=sys.stdout):
            if self.process_stack:
                self.process_stack[0].write(out)

        def print_render_calls(self, out=sys.stdout):
            if self.render_stack:
                self.render_stack[0].write(out)

        def print_last_call(self, out=sys.stdout):
            if self.render_stack:
                self.render_stack[-1].write(out)
            elif self.process_stack:
                self.process_stack[-1].write(out)

    def branch(self):
        session = Session(self.page)
        session.trunk = self
        session.response_headers = list(self.response_headers)
        return session

    def add_header(self, name, value):
        self.response_headers.append((name, value))

    def set_cookie(self, name, value, expires=None):
        crumbs = list()
        crumbs.append("%s=%s" % (name, value))

        if expires:
            when = expires.strftime(self.http_date_gmt)
            crumbs.append("expires=%s" % when)

        self.add_header("Set-Cookie", "; ".join(crumbs))

    def expire_cookie(self, name):
        self.set_cookie(name, "", datetime(*gmtime(1)[0:6]))

    # This is the biggest hotspot in cumin-bench profiling
    def get(self, key):
        if key in self.values:
            value = self.values[key]
        elif self.trunk:
            value = self.trunk.get(key)
        else:
            value = None

        return value

    def set(self, key, value):
        #if key.startswith("server.vhost.queue.edit.durable"):
        #    raise Exception()

        self.values[key] = value

        return value

    def unset(self, key):
        if key in self.values:
            del self.values[key]

    def marshal(self):
        page = self.page.name
        vars = self.marshal_url_vars()

        if vars:
            url = page + "?" + vars
        else:
            url = page

        if self.debug:
            self.app.debug.urls.add(url)

        return url

    def marshal_url_vars(self, separator=";"):
        params = self.page.get_saved_parameters(self)
        vars = list()

        for param in params:
            key = param.path

            if param.is_collection:
                collection = self.get(key)

                if collection:
                    for value in collection:
                        svalue = quote(param.marshal(value))
                        vars.append("%s=%s" % (key, svalue))
            else:
                #value = self.get(key)

                # Inlined below saving about a second in a
                # 1000-hit profile

                if key in self.values:
                    value = self.values[key]
                elif self.trunk:
                    value = self.trunk.get(key)
                else:
                    value = None

                default = param.get_default(self)

                if value not in (default, None):
                    svalue = quote(param.marshal(value))
                    vars.append("%s=%s" % (key, svalue))

        return separator.join(vars)

    def unmarshal(cls, app, string):
        elems = string.split("?")

        try:
            name = elems[0]

            if name.startswith("/"):
                name = name[1:]

            page = app.pages[name]
        except KeyError:
            raise Exception("Page '%s' not found" % name)

        session = Session(page)

        try:
            session.unmarshal_url_vars(elems[1])
        except IndexError:
            pass

        return session

    unmarshal = classmethod(unmarshal)

    def unmarshal_url_vars(self, string, separator=";"):
        vars = string.split(separator)

        for var in vars:
            try:
                skey, svalue = var.split("=")
                key = unquote(skey)
                value = unquote_plus(svalue)

                param = self.app.get_parameter(self.page, key)

                if param:
                    if param.is_dictionary:
                        param.add(self, param.unmarshal(value), key)
                    else:
                        param.add(self, param.unmarshal(value))
            except ValueError:
                pass

    def unmarshal_cookies(self, string):
        if string is not None:
            cookies = dict()
            crumbs = string.split(";")

            for crumb in crumbs:
                name, value = crumb.split("=", 1)
                cookies[name.strip()] = value.strip()

        self.cookies = cookies

    def set_origin(self, origin):
        self.origin = self.fix_origin(origin)

    def fix_origin(self, origin):
        if origin:
            try:
                # clean up refering url
                url = urlsplit(origin)
                query = url.query
                if query:
                    namevals = query.split(";")
                    newquery = dict()
                    for nameval in namevals:
                        name, value = nameval.split("=", 1)
                        # remove the parent's origin
                        if not name.endswith(".origin"):
                            newquery[name] = value

                    origin = ";".join(["%s=%s" % item for item in newquery.iteritems()])
                    return "?".join([url.path, origin])
            except:
                pass

    def get_origin(self):
        return self.origin

    def __repr__(self):
        return "%s(trunk=%s,app=%s)" % \
            (self.__class__.__name__, self.trunk, self.app)

class StringIOWriter(object):
    def __init__(self):
        self.writer = StringIO()

    def write(self, string):
        self.writer.write(string)

    def to_string(self):
        string = self.writer.getvalue()

        self.writer.close()

        return string

class ListWriter(object):
    def __init__(self):
        self.strings = list()

    def write(self, string):
        self.strings.append(string)

    def to_string(self):
        return "".join(self.strings)

class Writer(StringIOWriter):
    pass

class ObjectBoundTemplate(object):
    def __init__(self, obj, text):
        self.__object = obj
        self.__text = text
        self.__fragments = None

    def compile(self):
        return self.resolve(self.parse(self.__text))

    def parse(self, text):
        strings = list()

        start = 0
        end = text.find("{")

        while True:
            if (end == -1):
                strings.append(text[start:])
                break

            strings.append(text[start:end])

            ccurly = text.find("}", end + 1)

            if ccurly == -1:
                start = end
                end = -1
            else:
                ocurly = text.find("{", end + 1)

                if ocurly == -1:
                    start = end
                    end = ccurly + 1
                elif ocurly < ccurly:
                    start = end
                    end = ocurly
                else:
                    strings.append("{" + text[end + 1:ccurly] + "}")

                    start = ccurly + 1
                    end = ocurly

        return strings

    def resolve(self, strings):
        fragments = list()

        for string in strings:
            if string.startswith("{") and string.endswith("}"):
                name = string[1:-1]
                method = self.find_method("render_" + name)

                if method:
                    fragments.append(method)
                else:
                    child = self.find_child(name)

                    if child:
                        fragments.append(child)
                    else:
                        fragments.append(string)
            else:
                fragments.append(string)

        return fragments

    def find_method(self, name):
        for cls in self.__object.__class__.__mro__:
            meth = getattr(cls, name, None)

            if meth and callable(meth):
                return meth

    def find_child(self, name):
        return None

    def render(self, writer, session, *args):
        if not self.__fragments:
            self.__fragments = self.compile()

        for frag in self.__fragments:
            if type(frag) is str:
                writer.write(frag)
            elif callable(frag):
                #print "tc", frag, args

                result = frag(self.__object, session, *args)

                if result is not None:
                    writer.write(str(result))
            else:
                result = frag.render(session)

                if result is not None:
                    writer.write(str(result))

    def __repr__(self):
        name = self.__object.__class__.__name__ + "." + self.key
        return "%s('%s')" % (self.__class__.__name__, name)

class Template(ObjectBoundTemplate):
    def __init__(self, widget, key):
        text = widget.get_string(key)

        if not text:
            raise Exception("Template '%s.%s' not found" \
                            % (widget.__class__.__name__, key))

        super(Template, self).__init__(widget, text)

        self.widget = widget

    def find_child(self, name):
        return self.widget.children_by_name.get(name)

class WidgetCall(object):
    def __init__(self, stack, widget, session, object):
        self.stack = stack
        self.widget = widget
        self.session = session
        self.session_values = copy(session.values)
        self.object = object

        self.caller = None
        self.callees = list()

        self.start = None
        self.end = None

    def open(self):
        if self.stack:
            self.caller = self.stack[-1]
            self.caller.callees.append(self)

        self.stack.append(self)

        self.start = clock()

    def close(self):
        self.end = clock()

        if len(self.stack) > 1:
            self.stack.pop()

    def write(self, writer):
        writer.write(str(self.widget))
        writer.write(" (call %i, caller %i)" % (id(self), id(self.caller)))
        writer.write(os.linesep)

        writer.write("  session: " + str(self.session))
        writer.write(os.linesep)

        for item in sorted(self.session_values.items()):
            writer.write("    value: %s = %s" % item)
            writer.write(os.linesep)

        writer.write("   object: " + str(self.object))
        writer.write(os.linesep)

        writer.write("    times: %f, %f" % (self.start, self.end or -1))
        writer.write(os.linesep)

        for call in self.callees:
            call.write(writer)
