from wooly import *
from parameters import *
from resources import *
from widgets import ItemSet

strings = StringCatalog(__file__)

class Form(Widget):
    def __init__(self, app, name):
        super(Form, self).__init__(app, name)

        self.form_params = set()

        self.origin = Parameter(app, "origin")
        self.add_parameter(self.origin)

    def get_origin(self, session):
        origin = self.origin.get(session)

        # this will happen if we load the page
        # directly from a saved url and not through
        # a link on another page
        if not origin:
            try:
                branch = session.branch()
                self.frame.show_view(branch)
                origin = branch.marshal()
            except: # we don't have a show_view
                pass
            
        return origin
    
    def render_hidden_inputs(self, session, *args):
        writer = Writer()
        # remember our roots
        if not self.origin.get(session):
            self.origin.set(session, session.get_origin())

        params = set(session.page.get_saved_parameters(session))
        params.difference_update(self.form_params)

        for param in params:
            key = param.path
            
            if param.is_collection:
                collection = session.get(key)

                if collection:
                    for value in collection:
                        svalue = param.marshal(value)
                        self.write_hidden_input(key, svalue, writer)
            elif param.is_dictionary:
                pass
            else:
                value = session.get(key)
                default = param.get_default(session)

                if value not in (default, None):
                    svalue = param.marshal(value)
                    self.write_hidden_input(key, svalue, writer)

        return writer.to_string()

    def write_hidden_input(self, name, value, writer):
        writer.write("<input type='hidden' name='%s' value='%s'/>" \
                     % (name, value))
        
class FormError(object):
    def __init__(self, message=None):
        self.message = message

    def get_message(self, session):
        return self.message

class FormErrorSet(ItemSet):
    def __init__(self, app, name):
        super(FormErrorSet, self).__init__(app, name)

        self.__errors = self.Errors(app, "errors")
        self.add_attribute(self.__errors)

    def add(self, session, error):
        assert isinstance(error, FormError)

        self.__errors.get(session).append(error)

    def get(self, session):
        return self.__errors.get(session)

    def get_items(self, session, *args):
        return self.get(session)

    def do_render(self, session, *args):
        if self.get(session):
            return super(FormErrorSet, self).do_render(session, *args)

    def render_item_content(self, session, item):
        return item.get_message(session)

    class Errors(Attribute):
        def get_default(self, session):
            return list()

class FormInput(Widget):
    def __init__(self, app, name, param):
        super(FormInput, self).__init__(app, name)

        self.param = param
        self.tab_index = 100
        self.disabled = False

        self.form = None

    def init(self):
        super(FormInput, self).init()

        if self.param is None:
            raise Exception("Parameter not set for %s" % self)

        assert isinstance(self.param, Parameter)

        for anc in self.ancestors:
            if isinstance(anc, Form):
                self.form = anc

        if self.form:
            self.form.form_params.add(self.param)
        else:
            print "Warning: FormInput '%s' not inside a form" % self

    def get(self, session):
        return self.param.get(session)

    def set(self, session, value):
        return self.param.set(session, value)

    def get_default(self, session):
        return self.param.get_default(session)

    def set_tab_index(self, tab_index):
        self.tab_index = tab_index

    def set_disabled(self, disabled):
        self.disabled = disabled

    def render_name(self, session, *args):
        return self.param.path

    def render_value(self, session, *args):
        return self.param.marshal(self.param.get(session))

    def render_tab_index(self, session, *args):
        return self.tab_index

    def render_disabled_attr(self, session, *args):
        return self.disabled and "disabled=\"disabled\"" or None

class MissingValueError(FormError):
    def get_message(self, session):
        return "This value is required"

class EmptyInputError(FormError):
    def __init__(self):
        super(EmptyInputError, self).__init__("This value is required")

class ScalarInput(FormInput):
    def __init__(self, app, name, param):
        super(ScalarInput, self).__init__(app, name, param)

        self.size = 32

    def render_size(self, session, *args):
        return self.size

class StringInput(ScalarInput):
    def __init__(self, app, name):
        super(StringInput, self).__init__(app, name, None)

        self.param = Parameter(app, "param")
        self.add_parameter(self.param)

        self.size = 30

class PasswordInput(StringInput):
    pass

# XXX Why does this have a boolean param?  Shouldn't the name suggest
# that somehow?  Would some folks want a hidden input with a different
# param?  I think this needs to take a param
class HiddenInput(ScalarInput):
    def __init__(self, app, name):
        super(HiddenInput, self).__init__(app, name, None)
        
        self.param = BooleanParameter(app, "param")
        self.add_parameter(self.param)
        
class IntegerInput(ScalarInput):
    def __init__(self, app, name):
        super(IntegerInput, self).__init__(app, name, None)

        self.param = IntegerParameter(app, "param")
        self.add_parameter(self.param)

        self.size = 15

class CheckboxInput(FormInput):
    def __init__(self, app, name):
        super(CheckboxInput, self).__init__(app, name, None)

        self.param = VoidBooleanParameter(app, "param")
        self.add_parameter(self.param)

    def render_checked_attr(self, session, *args):
        if self.get(session):
            return "checked=\"checked\""

class RadioInput(FormInput):
    def __init__(self, app, name, param):
        super(RadioInput, self).__init__(app, name, param)

        self.value = name

    def render_value(self, session, *args):
        return self.value

    def render_checked_attr(self, session, *args):
        value = self.get(session)

        if value and value == self.value:
            return "checked=\"checked\""

class FormButton(FormInput):
    def __init__(self, app, name):
        super(FormButton, self).__init__(app, name, None)

        self.param = BooleanParameter(app, "param")
        self.add_parameter(self.param)

    def do_process(self, session, *args):
        if self.get(session):
            self.set(session, False)

            self.process_submit(session, *args)

    def process_submit(self, session, *args):
        pass

    def render_value(self, session, *args):
        branch = session.branch()

        self.set(branch, True)

        return super(FormButton, self).render_value(branch, *args)
    
    def render_type(self, session, *args):
        return "submit"

    def render_onclick(self, session, *args):
        return "click_button"
    
class CheckboxInputSet(FormInput, ItemSet):
    def render_item_value(self, session, *args):
        return None

    def render_item_checked_attr(self, session, *args):
        return None

class RadioInputSet(FormInput, ItemSet):
    def render_item_value(self, session, *args):
        return None

    def render_item_checked_attr(self, session, *args):
        return None

class OptionInputSet(FormInput, ItemSet):
    def render_item_value(self, session, item):
        return None

    def render_item_selected_attr(self, session, item):
        return None

class FormField(Widget):
    def __init__(self, app, name):
        super(FormField, self).__init__(app, name)

        self.__errors = FormErrorSet(app, "errors")
        self.add_child(self.__errors)

        self.css_class = "field"
        
    def validate(self, session):
        errors = self.__errors.get(session)

        self.do_validate(session, errors)

        return errors

    def do_validate(self, session, errors):
        pass

    def render_field_help(self, session, *args):
        pass
    
    def render_form_field_class(self, session, *args):
        return self.css_class
    
class ScalarField(FormField):
    def __init__(self, app, name, input):
        super(ScalarField, self).__init__(app, name)

        self.input = input

    def init(self):
        super(ScalarField, self).init()

        assert self.input is not None
        assert isinstance(self.input, ScalarInput)

    def get(self, session):
        return self.input.get(session)

    def set(self, session, value):
        return self.input.set(session, value)

    def render_inputs(self, session, *args):
        return self.input.render(session)

class StringField(ScalarField):
    def __init__(self, app, name):
        super(StringField, self).__init__(app, name, None)

        self.input = StringInput(app, "input")
        self.add_child(self.input)

class PasswordField(ScalarField):
    def __init__(self, app, name):
        super(PasswordField, self).__init__(app, name, None)

        self.input = PasswordInput(app, "input")
        self.add_child(self.input)
        
class IntegerField(ScalarField):
    def __init__(self, app, name):
        super(IntegerField, self).__init__(app, name, None)

        self.input = IntegerInput(app, "input")
        self.add_child(self.input)

    def do_validate(self, session, errors):
        val = self.get(session)
        try:
            val = int(val)
        except:
            if val:
                error = FormError("A Numeric Value is Expected")
                errors.append(error)
            else:
                self.set(session, self.input.param.get_default(session))
            
# XXX make this use a RadioInputSet instead?
class RadioField(FormField):
    def __init__(self, app, name, param):
        super(RadioField, self).__init__(app, name)

        self.param = param
        self.options = list()

    def init(self):
        super(RadioField, self).init()

        assert self.param is not None
        assert isinstance(self.param, Parameter)

    def get(self, session):
        return self.param.get(session)

    def add_option(self, option):
        assert isinstance(option, RadioFieldOption)

        self.options.append(option)
        self.add_child(option)

        option.param = self.param

    def get(self, session):
        return self.param.get(session)

    def set(self, session, value):
        self.param.set(session, value)

    def render_inputs(self, session, *args):
        writer = Writer()

        for option in self.options:
            writer.write(option.render(session))

        return writer.to_string()

class RadioFieldOption(RadioInput):
    pass

class ButtonForm(Form):
    def __init__(self, app, name):
        super(ButtonForm. self).__init__(app, name)

        self.buttons = list()

    def add_button(self, button):
        assert type(button) is FormButton

        self.buttons.append(button)
        self.add_child(button)

    def render_buttons(self, session, *args):
        writer = Writer()

        for button in self.buttons:
            writer.write(button.render(session))

        return writer.to_string()

class FieldForm(Form):
    def __init__(self, app, name):
        super(FieldForm, self).__init__(app, name)

        self.fields = list()

    def add_field(self, field):
        assert isinstance(field, FormField)

        self.fields.append(field)
        self.add_child(field)

    def validate(self, session):
        errors = list()

        for field in self.fields:
            errors.extend(field.validate(session))

        return errors

    def render_fields(self, session, *args):
        writer = Writer()

        for field in self.fields:
            writer.write(field.render(session))

        return writer.to_string()
