from traceback import print_exc
from datetime import datetime
from time import strptime
from base64 import decodestring

from wooly import *
from devel import DevelPage
from wsgiserver import CherryPyWSGIServer

class WebServer(object):
    http_date = "%a, %d %b %Y %H:%M:%S %Z"
    http_date_gmt = "%a, %d %b %Y %H:%M:%S GMT"

    def __init__(self, app, addr, port):
        self.app = app
        self.addr = addr
        self.port = port

        fqaddr = (self.addr, self.port)
        apps = [("", self.service)]
        self.__server = CherryPyWSGIServer(fqaddr, apps)

    def set_ssl_cert_path(self, path):
        self.__server.ssl_certificate = path

    def set_ssl_key_path(self, path):
        self.__server.ssl_private_key = path

    def start(self):
        self.__server.start()

    def stop(self):
        self.__server.stop()

    def authorized(self, session):
        return False

    def service(self, env, respond):
        headers = list()

        name = env["PATH_INFO"][1:]

        try:
            page = self.app.pages[name]
        except KeyError:
            return self.not_found(respond, headers)

        session = Session(page, headers)
        session.unmarshal_url_vars(env["QUERY_STRING"])

        if "HTTP_USER_AGENT" in env:
            page.set_agent(session, env["HTTP_USER_AGENT"])
            
        if env["REQUEST_METHOD"] == "POST":
            if env["CONTENT_TYPE"] == "application/x-www-form-urlencoded":
                length = int(env["CONTENT_LENGTH"])
                vars = env["wsgi.input"].read(length)
            else:
                raise Exception("Content type '%s' is not supported" \
                                % env["CONTENT_TYPE"])

            if vars:
                session.unmarshal_url_vars(vars, "&")

        try:
            session.unmarshal_cookies(env["HTTP_COOKIE"])
        except KeyError:
            pass

        if "HTTP_REFERER" in env:
            # XXX set this on page instead
            session.set_origin(env["HTTP_REFERER"])

        if not self.authorized(session):
            url = page.get_redirect_url(session)

            if url:
                return self.redirect(respond, headers, url)
            else:
                return self.unauthorized(respond, headers)
        
        try:
            page.process(session)
        except:
            return self.error(respond, session)

        url = page.get_redirect_url(session)

        if url:
            return self.redirect(respond, headers, url)

        ims = env.get("HTTP_IF_MODIFIED_SINCE")
        modified = page.get_last_modified(session).replace(microsecond=0)

        if ims:
            since = datetime(*strptime(str(ims), self.http_date)[0:6])

            if modified <= since:
                return self.not_modified(respond, headers)

        if modified:
            ts = modified.strftime(self.http_date_gmt)
            headers.append(("Last-Modified", ts))

        type = page.get_content_type(session)

        if type:
            headers.append(("Content-Type", type))

        cache = page.get_cache_control(session)

        if cache:
            headers.append(("Cache-Control", cache))

        try:
            response = page.render(session)
        except:
            return self.error(respond, session)

        headers.append(("Content-Length", str(len(response))))

        respond("200 OK", headers)

        page.save_session(session)

        return (response,)

    def message(self, respond, headers, message):
        headers.append(("Content-Length", str(len(message))))
        headers.append(("Content-Type", "text/plain"))

        respond(message, headers)

        return (message,)

    def not_found(self, respond, headers):
        return self.message(respond, headers, "404 Not Found")

    def unauthorized(self, respond, headers):
        return self.message(respond, headers, "401 Unauthorized")

    def redirect(self, respond, headers, url):
        headers.append(("Location", url))
        headers.append(("Content-Length", "0"))

        respond("303 See Other", headers)

        return ()

    def not_modified(self, respond, headers):
        headers.append(("Content-Length", "0"))

        respond("304 Not Modified", headers)

        return ()

    def error(self, respond, session):
        session.add_header("Content-Type", "text/plain")

        respond("500 Error", session.response_headers)

        writer = Writer()
        writer.write("APPLICATION ERROR\n")
        writer.write("\n----- python trace -----\n")
        print_exc(None, writer)

        if session.debug:
            writer.write("\n----- process trace -----\n")
            session.debug.print_process_calls(writer)
            writer.write("\n----- render trace -----\n")
            session.debug.print_render_calls(writer)

        return writer.to_string()
