from wooly import *
from widgets import *
from parameters import *

strings = StringCatalog(__file__)

class ItemTable(ItemSet):
    def __init__(self, app, name):
        super(ItemTable, self).__init__(app, name)

        self.columns = list()
        self.headers_by_column = dict()

        # which column are we sorting on
        self.scolumn = Parameter(app, "col")
        self.add_parameter(self.scolumn)

        # is sort asc
        self.reversed = BooleanParameter(app, "rev")
        self.reversed.default = False
        self.add_parameter(self.reversed)

    def get_visible_columns(self, session):
        # default impl
        return [col for col in self.columns if col.visible]
    
    def get_request_visible_columns(self, session, vlist):
        return [col for col in self.columns if col.visible or col.name in vlist]

    def render_headers(self, session, *args):
        writer = Writer()

        vlist = self.get_visible_columns(session)
        for column in vlist:
            header = self.headers_by_column[column]
            writer.write(header.render(session))

        return writer.to_string()

    def render_column_count(self, session, *args):
        vlist = self.get_visible_columns(session)
        return len(vlist)

    def render_cells(self, session, item):
        writer = Writer()

        vlist = self.get_visible_columns(session)
        for col in vlist:
            col.set_item(session, item)
            writer.write(col.render(session))

        return writer.to_string()

    def add_column(self, column):
        self.columns.append(column)
        self.add_child(column)

        header = column.header_class(self.app, "head", column)
        self.set_header(column, header)
        self.add_child(header)

        if self.scolumn.default is None:
            self.scolumn.default = column.name

    def set_header(self, column, header):
        self.headers_by_column[column] = header

    def get_selected_column(self, session):
        name = self.scolumn.get(session)
        for column in self.columns:
            if column.name == name:
                return column

    def set_default_column(self, column):
        self.scolumn.default = column.name

    def set_default_column_name(self, name):
        self.scolumn.default = name
        
    def is_reversed(self, session):
        return self.reversed.get(session)

    def do_get_items(self, session, *args):
        """Gets the rows"""

        return None

    def render_count(self, session, *args):
        count = self.get_item_count(session, *args)
        return "%i %s" % (count, count == 1 and "item" or "items")

    def render_none(self, session, *args):
        """For producing a message when the table is empty"""

        return None

class ItemTableColumn(Widget):
    def __init__(self, app, name):
        super(ItemTableColumn, self).__init__(app, name)

        self.header_class = ItemTableColumnHeader
        self.alignment = None
        self.visible = True

        self.__item = Attribute(app, "item")
        self.add_attribute(self.__item)

    def set_item(self, session, item):
        return self.__item.set(session, item)

    def get_args(self, session):
        return (self.__item.get(session),)

    def get_column_key(self, session):
        return self.name

    def get_class_attr(self, session):
        return self.alignment == "right" and "class=\"ralign\"" or ""

    def do_render(self, session, data):
        content = self.render_content(session, data)
        return "<td %s>%s</td>" % (self.get_class_attr(session) or "", content)

    def render_content(self, session, data):
        key = self.get_column_key(session)
        return self.render_value(session, data[key])

    def render_value(self, session, value):
        return str(value)

    # in case a non-sql column is included in an sqltable # XXX ugh
    def get_orderby_sql(self, session):
        pass
    
class ItemTableColumnHeader(Widget):
    def __init__(self, app, name, column):
        super(ItemTableColumnHeader, self).__init__(app, name)

        self.column = column

    def render_href(self, session, *args):
        branch = session.branch()

        sel = self.parent.get_selected_column(session)

        if sel is self.column:
            self.parent.reversed.set \
                (branch, not self.parent.reversed.get(session))

        self.parent.scolumn.set(branch, self.column.name)

        return branch.marshal()

    def render_content(self, session, *args):
        cargs = self.column.get_args(session)
        return self.column.render_title(session, *cargs)

    def render_class_attr(self, session, *args):
        return self.column.get_class_attr(session)

    def render_sorted_dir(self, session, *args):
        sel = self.parent.get_selected_column(session)

        if sel is self.column:
            if self.parent.reversed.get(session):
                return "up"
            else:
                return "down"
        elif self.parent.reversed.get(session):
            return "unsorted_up"
        else:
            return "unsorted_down"
        
class SqlTable(ItemTable):
    def __init__(self, app, name):
        super(SqlTable, self).__init__(app, name)

        self.__sql_tmpl = Template(self, "sql")
        self.__count_sql_tmpl = Template(self, "count_sql")
        self.__find_sql_tmpl = Template(self, "find_sql")

    def render_sql(self, session, *args):
        writer = Writer()
        self.__sql_tmpl.render(writer, session, *args)
        return writer.to_string()

    def render_sql_where(self, session, *args):
        pass

    def render_find_sql_where(self, session, *args):
        pass

    def render_sql_orderby(self, session, *args):
        scol = self.get_selected_column(session)
        if scol:
            return scol.get_orderby_sql(session)

    def render_sql_limit(self, session, *args):
        return None

    def render_count_sql(self, session, *args):
        writer = Writer()
        self.__count_sql_tmpl.render(writer, session, *args)
        return writer.to_string()

    def render_find_sql(self, session, *args):
        writer = Writer()
        self.__find_sql_tmpl.render(writer, session, *args)
        return writer.to_string()

    def build_sql(self, elems):
        writer = Writer()
        
        for elem in elems:
            if elem:
                writer.write(elem)
                writer.write("\n")

        return writer.to_string()

    def get_sql_values(self, session, *args):
        return None

    def get_find_sql_values(self, session, *args):
        return None

    def get_connection(self, session):
        pass

    def get_item_count(self, session, *args):
        conn = self.get_connection(session)

        if conn:
            cursor = conn.cursor()
            sql = self.render_count_sql(session, *args)
            sql_values = self.get_sql_values(session, *args)

            cursor.execute(sql, sql_values)
            data = cursor.fetchone()

            return data[0]

    def do_get_items(self, session, *args):
        conn = self.get_connection(session)

        if conn:
            cursor = conn.cursor()
            sql = self.render_sql(session, *args)
            sql_values = self.get_sql_values(session, *args)

            #print "SQL TEXT", sql
            #print "SQL VALS", sql_values

            cursor.execute(sql, sql_values)

            return cursor

    def render_items(self, session, *args):
        cursor = self.get_items(session, *args)
        cols = [spec[0] for spec in cursor.description]
        data = dict()

        writer = Writer()

        for tuple in cursor:
            for col, datum in zip(cols, tuple):
                data[col] = datum

            self.item_tmpl.render(writer, session, data)

        return writer.to_string()

    def find_item(self, session, *args):
        """ Find items in the current ItemSet
        
            To use this an SqlTable derived object needs to have a
            [<class>.find_sql] section and override render_find_sql_where
            and get_find_sql_values.
            Returns a list of dictionaries. Each dictionary is a matched row. """
            
        conn = self.get_connection(session)
        if conn:
            cursor = conn.cursor()
            select = self.render_find_sql(session, *args)
            where = self.render_sql_where(session, *args)
            sql_values = self.get_sql_values(session, *args)
            
            find_where = self.render_find_sql_where(session, *args)
            find_values = self.get_find_sql_values(session, *args)
            sql = " and ".join(["%s %s" % (select, where), find_where])
            if sql_values:
                for sql_val in sql_values:
                    find_values[sql_val] = sql_values[sql_val]
            
            try:
                cursor.execute(sql, find_values)
                return self.cursor_to_rows(cursor)
            except Exception, e:
                pass

    def cursor_to_rows(self, cursor):
        cols = [spec[0] for spec in cursor.description]
        rows = list()
        
        for tuple in cursor:
            row = dict()
            for col, datum in zip(cols, tuple):
                row[col] = datum

            rows.append(row)
        
        return rows
        
class SqlTableColumn(ItemTableColumn):
    # XXX to be consistent with similar methods, rename to
    # get_sql_order_by

    def get_orderby_sql(self, session):
        key = self.get_column_key(session)

        if key:
            dir = self.parent.is_reversed(session) and "desc" or "asc"
            return "order by %s %s" % (key, dir)

class SqlTableFilter(Widget):
    def get_sql_constraint(self, session, *args):
        pass
